=== modified file 'bzrlib/annotate.py'
--- bzrlib/annotate.py	2008-04-24 07:22:53 +0000
+++ bzrlib/annotate.py	2008-06-11 04:20:16 +0000
@@ -112,8 +112,8 @@
 
 def _annotations(repo, file_id, rev_id):
     """Return the list of (origin,text) for a revision of a file in a repository."""
-    w = repo.weave_store.get_weave(file_id, repo.get_transaction())
-    return w.annotate(rev_id)
+    annotations = repo.texts.annotate((file_id, rev_id))
+    return [(key[-1], line) for key, line in annotations]
 
 
 def _annotate_file(branch, rev_id, file_id):

=== modified file 'bzrlib/branch.py'
--- bzrlib/branch.py	2008-05-08 04:33:38 +0000
+++ bzrlib/branch.py	2008-06-11 04:20:16 +0000
@@ -1413,6 +1413,9 @@
         """See Branch.set_revision_history."""
         if 'evil' in debug.debug_flags:
             mutter_callsite(3, "set_revision_history scales with history.")
+        check_not_reserved_id = _mod_revision.check_not_reserved_id
+        for rev_id in rev_history:
+            check_not_reserved_id(rev_id)
         self._write_revision_history(rev_history)
         self._clear_cached_state()
         self._cache_revision_history(rev_history)
@@ -1476,9 +1479,11 @@
             raise errors.NoSuchRevision(self, revision_id)
         current_rev_id = revision_id
         new_history = []
+        check_not_reserved_id = _mod_revision.check_not_reserved_id
         # Do not include ghosts or graph origin in revision_history
         while (current_rev_id in parents_map and
                len(parents_map[current_rev_id]) > 0):
+            check_not_reserved_id(current_rev_id)
             new_history.append(current_rev_id)
             current_rev_id = parents_map[current_rev_id][0]
             parents_map = graph.get_parent_map([current_rev_id])

=== modified file 'bzrlib/builtins.py'
--- bzrlib/builtins.py	2008-05-08 04:33:38 +0000
+++ bzrlib/builtins.py	2008-06-11 04:20:16 +0000
@@ -4554,5 +4554,5 @@
     cmd_bundle_info,
     )
 from bzrlib.sign_my_commits import cmd_sign_my_commits
-from bzrlib.weave_commands import cmd_versionedfile_list, cmd_weave_join, \
+from bzrlib.weave_commands import cmd_versionedfile_list, \
         cmd_weave_plan_merge, cmd_weave_merge_text

=== modified file 'bzrlib/bundle/serializer/v4.py'
--- bzrlib/bundle/serializer/v4.py	2008-05-08 04:33:38 +0000
+++ bzrlib/bundle/serializer/v4.py	2008-06-11 04:20:16 +0000
@@ -278,14 +278,19 @@
             revision_ids = set(repository.get_ancestry(target,
                                                        topo_sorted=False))
             self.revision_ids = revision_ids.difference(self.base_ancestry)
+        self.revision_keys = set([(revid,) for revid in self.revision_ids])
 
     def do_write(self):
         """Write all data to the bundle"""
-        self.bundle.begin()
-        self.write_info()
-        self.write_files()
-        self.write_revisions()
-        self.bundle.end()
+        self.repository.lock_read()
+        try:
+            self.bundle.begin()
+            self.write_info()
+            self.write_files()
+            self.write_revisions()
+            self.bundle.end()
+        finally:
+            self.repository.unlock()
         return self.revision_ids
 
     def write_info(self):
@@ -296,56 +301,27 @@
         self.bundle.add_info_record(serializer=serializer_format,
                                     supports_rich_root=supports_rich_root)
 
-    def iter_file_revisions(self):
-        """Iterate through all relevant revisions of all files.
-
-        This is the correct implementation, but is not compatible with bzr.dev,
-        because certain old revisions were not converted correctly, and have
-        the wrong "revision" marker in inventories.
-        """
-        transaction = self.repository.get_transaction()
-        altered = self.repository.fileids_altered_by_revision_ids(
-            self.revision_ids)
-        for file_id, file_revision_ids in altered.iteritems():
-            vf = self.repository.weave_store.get_weave(file_id, transaction)
-            yield vf, file_id, file_revision_ids
-
-    def iter_file_revisions_aggressive(self):
-        """Iterate through all relevant revisions of all files.
-
-        This uses the standard iter_file_revisions to determine what revisions
-        are referred to by inventories, but then uses the versionedfile to
-        determine what the build-dependencies of each required revision.
-
-        All build dependencies which are not ancestors of the base revision
-        are emitted.
-        """
-        for vf, file_id, file_revision_ids in self.iter_file_revisions():
-            new_revision_ids = set()
-            pending = list(file_revision_ids)
-            while len(pending) > 0:
-                revision_id = pending.pop()
-                if revision_id in new_revision_ids:
-                    continue
-                if revision_id in self.base_ancestry:
-                    continue
-                new_revision_ids.add(revision_id)
-                pending.extend(vf.get_parent_map([revision_id])[revision_id])
-            yield vf, file_id, new_revision_ids
-
     def write_files(self):
         """Write bundle records for all revisions of all files"""
-        for vf, file_id, revision_ids in self.iter_file_revisions():
-            self.add_mp_records('file', file_id, vf, revision_ids)
+        texts = self.repository.texts
+        text_keys = []
+        for file_id, revision_ids in \
+            self.repository.fileids_altered_by_revision_ids(
+                self.revision_ids).iteritems():
+            revision_ids = list(revision_ids)
+            for revision_id in revision_ids:
+                text_keys.append((file_id, revision_id))
+        self.add_mp_records_keys('file', texts, text_keys)
 
     def write_revisions(self):
         """Write bundle records for all revisions and signatures"""
-        inv_vf = self.repository.get_inventory_weave()
-        revision_order = list(multiparent.topo_iter(inv_vf, self.revision_ids))
+        inv_vf = self.repository.inventories
+        revision_order = [key[-1] for key in multiparent.topo_iter_keys(inv_vf,
+            self.revision_keys)]
         if self.target is not None and self.target in self.revision_ids:
             revision_order.remove(self.target)
             revision_order.append(self.target)
-        self.add_mp_records('inventory', None, inv_vf, revision_order)
+        self.add_mp_records_keys('inventory', inv_vf, [(revid,) for revid in revision_order])
         parent_map = self.repository.get_parent_map(revision_order)
         for revision_id in revision_order:
             parents = parent_map.get(revision_id, None)
@@ -374,17 +350,22 @@
                 base = parents[0]
         return base, target
 
-    def add_mp_records(self, repo_kind, file_id, vf, revision_ids):
+    def add_mp_records_keys(self, repo_kind, vf, keys):
         """Add multi-parent diff records to a bundle"""
-        revision_ids = list(multiparent.topo_iter(vf, revision_ids))
-        mpdiffs = vf.make_mpdiffs(revision_ids)
-        sha1s = vf.get_sha1s(revision_ids)
-        parent_map = vf.get_parent_map(revision_ids)
-        for mpdiff, revision_id, sha1, in zip(mpdiffs, revision_ids, sha1s):
-            parents = parent_map[revision_id]
+        ordered_keys = list(multiparent.topo_iter_keys(vf, keys))
+        mpdiffs = vf.make_mpdiffs(ordered_keys)
+        sha1s = vf.get_sha1s(ordered_keys)
+        parent_map = vf.get_parent_map(ordered_keys)
+        for mpdiff, item_key, sha1, in zip(mpdiffs, ordered_keys, sha1s):
+            parents = [key[-1] for key in parent_map[item_key]]
             text = ''.join(mpdiff.to_patch())
+            # Infer file id records as appropriate.
+            if len(item_key) == 2:
+                file_id = item_key[0]
+            else:
+                file_id = None
             self.bundle.add_multiparent_record(text, sha1, parents, repo_kind,
-                                               revision_id, file_id)
+                                               item_key[-1], file_id)
 
 
 class BundleInfoV4(object):
@@ -500,23 +481,19 @@
                 if self._info is not None:
                     raise AssertionError()
                 self._handle_info(metadata)
-            if (repo_kind, file_id) != ('file', current_file):
-                if len(pending_file_records) > 0:
-                    self._install_mp_records(current_versionedfile,
-                                             pending_file_records)
+            if (pending_file_records and
+                (repo_kind, file_id) != ('file', current_file)):
+                # Flush the data for a single file - prevents memory
+                # spiking due to buffering all files in memory.
+                self._install_mp_records_keys(self._repository.texts,
+                    pending_file_records)
                 current_file = None
-                current_versionedfile = None
-                pending_file_records = []
+                del pending_file_records[:]
             if len(pending_inventory_records) > 0 and repo_kind != 'inventory':
-                self._install_inventory_records(inventory_vf,
-                                                pending_inventory_records)
+                self._install_inventory_records(pending_inventory_records)
                 pending_inventory_records = []
             if repo_kind == 'inventory':
-                if inventory_vf is None:
-                    inventory_vf = self._repository.get_inventory_weave()
-                if revision_id not in inventory_vf:
-                    pending_inventory_records.append((revision_id, metadata,
-                                                      bytes))
+                pending_inventory_records.append(((revision_id,), metadata, bytes))
             if repo_kind == 'revision':
                 target_revision = revision_id
                 self._install_revision(revision_id, metadata, bytes)
@@ -524,15 +501,8 @@
                 self._install_signature(revision_id, metadata, bytes)
             if repo_kind == 'file':
                 current_file = file_id
-                if current_versionedfile is None:
-                    current_versionedfile = \
-                        self._repository.weave_store.get_weave_or_empty(
-                        file_id, self._repository.get_transaction())
-                    pending_file_records = []
-                if revision_id in current_versionedfile:
-                    continue
-                pending_file_records.append((revision_id, metadata, bytes))
-        self._install_mp_records(current_versionedfile, pending_file_records)
+                pending_file_records.append(((file_id, revision_id), metadata, bytes))
+        self._install_mp_records_keys(self._repository.texts, pending_file_records)
         return target_revision
 
     def _handle_info(self, info):
@@ -553,10 +523,25 @@
                       records if r not in versionedfile]
         versionedfile.add_mpdiffs(vf_records)
 
-    def _install_inventory_records(self, vf, records):
+    def _install_mp_records_keys(self, versionedfile, records):
+        d_func = multiparent.MultiParent.from_patch
+        vf_records = []
+        for key, meta, text in records:
+            # Adapt to tuple interface:
+            if len(key) == 2:
+                prefix = key[:1]
+            else:
+                prefix = ()
+            parents = [prefix + (parent,) for parent in meta['parents']]
+            vf_records.append((key, parents, meta['sha1'], d_func(text)))
+        versionedfile.add_mpdiffs(vf_records)
+
+    def _install_inventory_records(self, records):
         if self._info['serializer'] == self._repository._serializer.format_num:
-            return self._install_mp_records(vf, records)
-        for revision_id, metadata, bytes in records:
+            return self._install_mp_records_keys(self._repository.inventories,
+                records)
+        for key, metadata, bytes in records:
+            revision_id = key[-1]
             parent_ids = metadata['parents']
             parents = [self._repository.get_inventory(p)
                        for p in parent_ids]
@@ -579,29 +564,22 @@
     def _handle_root(self, target_inv, parent_ids):
         revision_id = target_inv.revision_id
         if self.update_root:
-            target_inv.root.revision = revision_id
-            store = self._repository.weave_store
-            transaction = self._repository.get_transaction()
-            vf = store.get_weave_or_empty(target_inv.root.file_id, transaction)
-            vf.add_lines(revision_id, parent_ids, [])
+            text_key = (target_inv.root.file_id, revision_id)
+            parent_keys = [(target_inv.root.file_id, parent) for
+                parent in parent_ids]
+            self._repository.texts.add_lines(text_key, parent_keys, [])
         elif not self._repository.supports_rich_root():
             if target_inv.root.revision != revision_id:
                 raise errors.IncompatibleRevision(repr(self._repository))
 
-
     def _install_revision(self, revision_id, metadata, text):
         if self._repository.has_revision(revision_id):
             return
-        if self._info['serializer'] == self._repository._serializer.format_num:
-            self._repository._add_revision_text(revision_id, text)
-        else:
-            revision = self._source_serializer.read_revision_from_string(text)
-            self._repository.add_revision(revision.revision_id, revision)
+        revision = self._source_serializer.read_revision_from_string(text)
+        self._repository.add_revision(revision.revision_id, revision)
 
     def _install_signature(self, revision_id, metadata, text):
         transaction = self._repository.get_transaction()
-        if self._repository._revision_store.has_signature(revision_id,
-                                                          transaction):
+        if self._repository.has_signature_for_revision_id(revision_id):
             return
-        self._repository._revision_store.add_revision_signature_text(
-            revision_id, text, transaction)
+        self._repository.add_signature_text(revision_id, text)

=== modified file 'bzrlib/bzrdir.py'
--- bzrlib/bzrdir.py	2008-05-12 04:20:02 +0000
+++ bzrlib/bzrdir.py	2008-06-11 04:20:16 +0000
@@ -50,6 +50,7 @@
     symbol_versioning,
     ui,
     urlutils,
+    versionedfile,
     win32utils,
     workingtree,
     workingtree_4,
@@ -62,8 +63,7 @@
     )
 from bzrlib.smart.client import _SmartClient
 from bzrlib.smart import protocol
-from bzrlib.store.revision.text import TextRevisionStore
-from bzrlib.store.text import TextStore
+from bzrlib.repofmt.weaverepo import TextRevisionStore, RevisionTextStore
 from bzrlib.store.versioned import WeaveStore
 from bzrlib.transactions import WriteTransaction
 from bzrlib.transport import (
@@ -2012,14 +2012,17 @@
         self.bzrdir.transport.mkdir('revision-store')
         revision_transport = self.bzrdir.transport.clone('revision-store')
         # TODO permissions
-        _revision_store = TextRevisionStore(TextStore(revision_transport,
-                                                      prefixed=False,
-                                                      compressed=True))
+        from bzrlib.xml5 import serializer_v5
+        revision_store = RevisionTextStore(revision_transport,
+            serializer_v5, False, versionedfile.PrefixMapper(),
+            lambda:True, lambda:True)
         try:
-            transaction = WriteTransaction()
             for i, rev_id in enumerate(self.converted_revs):
                 self.pb.update('write revision', i, len(self.converted_revs))
-                _revision_store.add_revision(self.revisions[rev_id], transaction)
+                text = serializer_v5.write_revision_to_string(
+                    self.revisions[rev_id])
+                key = (rev_id,)
+                revision_store.add_lines(key, None, osutils.split_lines(text))
         finally:
             self.pb.clear()
             
@@ -2038,8 +2041,7 @@
                          rev_id)
             self.absent_revisions.add(rev_id)
         else:
-            rev = self.branch.repository._revision_store.get_revision(rev_id,
-                self.branch.repository.get_transaction())
+            rev = self.branch.repository.get_revision(rev_id)
             for parent_id in rev.parent_ids:
                 self.known_revisions.add(parent_id)
                 self.to_read.append(parent_id)
@@ -2135,7 +2137,7 @@
                 ie.revision = previous_ie.revision
                 return
         if ie.has_text():
-            text = self.branch.repository.weave_store.get(ie.text_id)
+            text = self.branch.repository._text_store.get(ie.text_id)
             file_lines = text.readlines()
             w.add_lines(rev_id, previous_revisions, file_lines)
             self.text_count += 1

=== modified file 'bzrlib/check.py'
--- bzrlib/check.py	2007-11-27 23:39:22 +0000
+++ bzrlib/check.py	2008-06-11 04:20:16 +0000
@@ -55,7 +55,7 @@
         self.missing_revision_cnt = 0
         # maps (file-id, version) -> sha1; used by InventoryFile._check
         self.checked_texts = {}
-        self.checked_weaves = {}
+        self.checked_weaves = set()
         self.unreferenced_versions = set()
         self.inconsistent_parents = []
 
@@ -66,7 +66,7 @@
             self.progress.update('retrieving inventory', 0, 2)
             # do not put in init, as it should be done with progess,
             # and inside the lock.
-            self.inventory_weave = self.repository.get_inventory_weave()
+            self.inventory_weave = self.repository.inventories
             self.progress.update('checking revision graph', 1)
             self.check_revision_graph()
             self.plan_revisions()
@@ -96,7 +96,7 @@
         repository = self.repository
         self.planned_revisions = repository.all_revision_ids()
         self.progress.clear()
-        inventoried = set(self.inventory_weave.versions())
+        inventoried = set(key[-1] for key in self.inventory_weave.keys())
         awol = set(self.planned_revisions) - inventoried
         if len(awol) > 0:
             raise BzrCheckError('Stored revisions missing from inventory'
@@ -193,31 +193,26 @@
     def check_weaves(self):
         """Check all the weaves we can get our hands on.
         """
-        n_weaves = 1
         weave_ids = []
-        if self.repository.weave_store.listable():
-            weave_ids = list(self.repository.weave_store)
-            n_weaves = len(weave_ids) + 1
-        self.progress.update('checking versionedfile', 0, n_weaves)
+        self.progress.update('checking inventory', 0, 2)
         self.inventory_weave.check(progress_bar=self.progress)
-        files_in_revisions = {}
-        revisions_of_files = {}
+        self.progress.update('checking text storage', 1, 2)
+        self.repository.texts.check(progress_bar=self.progress)
         weave_checker = self.repository._get_versioned_file_checker()
-        for i, weave_id in enumerate(weave_ids):
-            self.progress.update('checking versionedfile', i, n_weaves)
-            w = self.repository.weave_store.get_weave(weave_id,
-                    self.repository.get_transaction())
-            # No progress here, because it looks ugly.
-            w.check()
-            result = weave_checker.check_file_version_parents(w, weave_id)
-            bad_parents, unused_versions = result
-            bad_parents = bad_parents.items()
-            for revision_id, (weave_parents, correct_parents) in bad_parents:
-                self.inconsistent_parents.append(
-                    (revision_id, weave_id, weave_parents, correct_parents))
-            for revision_id in unused_versions:
-                self.unreferenced_versions.add((weave_id, revision_id))
-            self.checked_weaves[weave_id] = True
+        result = weave_checker.check_file_version_parents(
+            self.repository.texts, progress_bar=self.progress)
+        self.checked_weaves = weave_checker.file_ids
+        bad_parents, unused_versions = result
+        bad_parents = bad_parents.items()
+        for text_key, (stored_parents, correct_parents) in bad_parents:
+            # XXX not ready for id join/split operations.
+            weave_id = text_key[0]
+            revision_id = text_key[-1]
+            weave_parents = tuple([parent[-1] for parent in stored_parents])
+            correct_parents = tuple([parent[-1] for parent in correct_parents])
+            self.inconsistent_parents.append(
+                (revision_id, weave_id, weave_parents, correct_parents))
+        self.unreferenced_versions.update(unused_versions)
 
     def _check_revision_tree(self, rev_id):
         tree = self.repository.revision_tree(rev_id)

=== modified file 'bzrlib/fetch.py'
--- bzrlib/fetch.py	2008-05-12 04:20:02 +0000
+++ bzrlib/fetch.py	2008-06-11 04:20:16 +0000
@@ -30,6 +30,8 @@
 the ids we need, and then pull those ids and then return to the inventories.
 """
 
+import operator
+
 import bzrlib
 import bzrlib.errors as errors
 from bzrlib.errors import InstallFailed
@@ -38,9 +40,10 @@
 from bzrlib.symbol_versioning import (deprecated_function,
         deprecated_method,
         )
+from bzrlib.tsort import topo_sort
 from bzrlib.trace import mutter
 import bzrlib.ui
-from bzrlib.versionedfile import filter_absent
+from bzrlib.versionedfile import filter_absent, FulltextContentFactory
 
 # TODO: Avoid repeatedly opening weaves so many times.
 
@@ -125,8 +128,6 @@
         This initialises all the needed variables, and then fetches the 
         requested revisions, finally clearing the progress bar.
         """
-        self.to_weaves = self.to_repository.weave_store
-        self.from_weaves = self.from_repository.weave_store
         self.count_total = 0
         self.file_ids_names = {}
         pp = ProgressPhase('Transferring', 4, self.pb)
@@ -161,6 +162,7 @@
             revs = list(graph.iter_topo_order(revs))
             data_to_fetch = self.from_repository.item_keys_introduced_by(revs,
                                                                          pb)
+            text_keys = []
             for knit_kind, file_id, revisions in data_to_fetch:
                 if knit_kind != phase:
                     phase = knit_kind
@@ -169,14 +171,24 @@
                     pp.next_phase()
                     pb = bzrlib.ui.ui_factory.nested_progress_bar()
                 if knit_kind == "file":
-                    self._fetch_weave_text(file_id, revisions)
+                    # Accumulate file texts
+                    text_keys.extend([(file_id, revision) for revision in
+                        revisions])
                 elif knit_kind == "inventory":
+                    # Now copy the file texts.
+                    to_texts = self.to_repository.texts
+                    from_texts = self.from_repository.texts
+                    to_texts.insert_record_stream(from_texts.get_record_stream(
+                        text_keys, 'topological', False))
+                    # Cause an error if a text occurs after we have done the
+                    # copy.
+                    text_keys = None
                     # Before we process the inventory we generate the root
                     # texts (if necessary) so that the inventories references
                     # will be valid.
                     self._generate_root_texts(revs)
                     # NB: This currently reopens the inventory weave in source;
-                    # using a full get_data_stream instead would avoid this.
+                    # using a single stream interface instead would avoid this.
                     self._fetch_inventory_weave(revs, pb)
                 elif knit_kind == "signatures":
                     # Nothing to do here; this will be taken care of when
@@ -209,32 +221,25 @@
                 self.from_repository, self._last_revision,
                 find_ghosts=self.find_ghosts)
         except errors.NoSuchRevision, e:
+            import pdb;pdb.set_trace()
             raise InstallFailed([self._last_revision])
 
-    def _fetch_weave_text(self, file_id, required_versions):
-        to_weave = self.to_weaves.get_weave_or_empty(file_id,
-            self.to_repository.get_transaction())
-        from_weave = self.from_weaves.get_weave(file_id,
-            self.from_repository.get_transaction())
-        # Fetch all the texts.
-        to_weave.insert_record_stream(from_weave.get_record_stream(
-            required_versions, 'topological', False))
-
     def _fetch_inventory_weave(self, revs, pb):
         pb.update("fetch inventory", 0, 2)
-        to_weave = self.to_repository.get_inventory_weave()
+        to_weave = self.to_repository.inventories
         child_pb = bzrlib.ui.ui_factory.nested_progress_bar()
         try:
             # just merge, this is optimisable and its means we don't
             # copy unreferenced data such as not-needed inventories.
             pb.update("fetch inventory", 1, 3)
-            from_weave = self.from_repository.get_inventory_weave()
+            from_weave = self.from_repository.inventories
             pb.update("fetch inventory", 2, 3)
             # we fetch only the referenced inventories because we do not
             # know for unselected inventories whether all their required
             # texts are present in the other repository - it could be
             # corrupt.
-            to_weave.insert_record_stream(from_weave.get_record_stream(revs,
+            to_weave.insert_record_stream(from_weave.get_record_stream(
+                [(rev_id,) for rev_id in revs],
                 'topological', False))
         finally:
             child_pb.finished()
@@ -258,19 +263,17 @@
 
     def _fetch_revision_texts(self, revs, pb):
         """Fetch revision object texts"""
-        to_txn = self.to_transaction = self.to_repository.get_transaction()
         count = 0
         total = len(revs)
-        to_store = self.to_repository._revision_store
         for rev in revs:
             pb.update('copying revisions', count, total)
             try:
                 sig_text = self.from_repository.get_signature_text(rev)
-                to_store.add_revision_signature_text(rev, sig_text, to_txn)
+                self.to_repository.add_signature_text(rev, sig_text)
             except errors.NoSuchRevision:
                 # not signed.
                 pass
-            self._copy_revision(rev, to_txn)
+            self._copy_revision(rev)
             count += 1
         # fixup inventory if needed: 
         # this is expensive because we have no inverse index to current ghosts.
@@ -279,9 +282,9 @@
         # FIXME: repository should inform if this is needed.
         self.to_repository.reconcile()
 
-    def _copy_revision(self, rev, to_txn):
-        to_store = self.to_repository._revision_store
-        to_store.add_revision(self.from_repository.get_revision(rev), to_txn)
+    def _copy_revision(self, rev_id):
+        rev = self.from_repository.get_revision(rev_id)
+        self.to_repository.add_revision(rev_id, rev)
 
 
 class KnitRepoFetcher(RepoFetcher):
@@ -294,24 +297,19 @@
 
     def _fetch_revision_texts(self, revs, pb):
         # may need to be a InterRevisionStore call here.
-        from_transaction = self.from_repository.get_transaction()
-        to_transaction = self.to_repository.get_transaction()
-        to_sf = self.to_repository._revision_store.get_signature_file(
-            to_transaction)
-        from_sf = self.from_repository._revision_store.get_signature_file(
-            from_transaction)
+        to_sf = self.to_repository.signatures
+        from_sf = self.from_repository.signatures
         # A missing signature is just skipped.
-        to_sf.insert_record_stream(filter_absent(from_sf.get_record_stream(revs,
+        to_sf.insert_record_stream(filter_absent(from_sf.get_record_stream(
+            [(rev_id,) for rev_id in revs],
             'unordered', False)))
-        self._fetch_just_revision_texts(revs, from_transaction, to_transaction)
+        self._fetch_just_revision_texts(revs)
 
-    def _fetch_just_revision_texts(self, version_ids, from_transaction,
-                                   to_transaction):
-        to_rf = self.to_repository._revision_store.get_revision_file(
-            to_transaction)
-        from_rf = self.from_repository._revision_store.get_revision_file(
-            from_transaction)
-        to_rf.insert_record_stream(from_rf.get_record_stream(version_ids,
+    def _fetch_just_revision_texts(self, version_ids):
+        to_rf = self.to_repository.revisions
+        from_rf = self.from_repository.revisions
+        to_rf.insert_record_stream(from_rf.get_record_stream(
+            [(rev_id,) for rev_id in version_ids],
             'topological', False))
 
 
@@ -374,29 +372,37 @@
 
         :param revs: the revisions to include
         """
-        to_store = self.target.weave_store
+        to_texts = self.target.texts
         graph = self.source.get_graph()
         parent_map = graph.get_parent_map(revs)
-        revision_root, planned_versions = self._find_root_ids(
+        rev_order = topo_sort(parent_map)
+        rev_id_to_root_id, root_id_to_rev_ids = self._find_root_ids(
             revs, parent_map, graph)
-        for root_id, versions in planned_versions.iteritems():
-            versionedfile = to_store.get_weave_or_empty(root_id,
-                self.target.get_transaction())
-            parent_texts = {}
-            for revision_id in versions:
-                if revision_id in versionedfile:
-                    continue
-                parents = parent_map[revision_id]
+        root_id_order = [(rev_id_to_root_id[rev_id], rev_id) for rev_id in
+            rev_order]
+        # Guaranteed stable, this groups all the file id operations together
+        # retaining topological order within the revisions of a file id.
+        # File id splits and joins would invalidate this, but they don't exist
+        # yet, and are unlikely to in non-rich-root environments anyway.
+        root_id_order.sort(key=operator.itemgetter(0))
+        # Create a record stream containing the roots to create.
+        def yield_roots():
+            for root_id, rev_id in root_id_order:
+                key = (root_id, rev_id)
+                rev_parents = parent_map[rev_id]
                 # We drop revision parents with different file-ids, because
-                # a version cannot have a version with another file-id as its
-                # parent.
+                # that represents a rename of the root to a different location
+                # - its not actually a parent for us. (We could look for that
+                # file id in the revision tree at considerably more expense,
+                # but for now this is sufficient (and reconcile will catch and
+                # correct this anyway).
                 # When a parent revision is a ghost, we guess that its root id
-                # was unchanged.
-                parents = tuple(p for p in parents if p != NULL_REVISION
-                    and revision_root.get(p, root_id) == root_id)
-                result = versionedfile.add_lines_with_ghosts(
-                    revision_id, parents, [], parent_texts)
-                parent_texts[revision_id] = result[2]
+                # was unchanged (rather than trimming it from the parent list).
+                parent_keys = tuple((root_id, parent) for parent in rev_parents
+                    if parent != NULL_REVISION and
+                        rev_id_to_root_id.get(parent, root_id) == root_id)
+                yield FulltextContentFactory(key, parent_keys, None, '')
+        to_texts.insert_record_stream(yield_roots())
 
     def regenerate_inventory(self, revs):
         """Generate a new inventory versionedfile in target, convertin data.
@@ -430,7 +436,7 @@
     def _fetch_inventory_weave(self, revs, pb):
         self.helper.regenerate_inventory(revs)
 
-    def _copy_revision(self, rev, to_txn):
+    def _copy_revision(self, rev):
         self.helper.fetch_revisions([rev])
 
 
@@ -449,15 +455,5 @@
     def _fetch_inventory_weave(self, revs, pb):
         self.helper.regenerate_inventory(revs)
 
-    def _fetch_just_revision_texts(self, version_ids, from_transaction,
-                                   to_transaction):
+    def _fetch_just_revision_texts(self, version_ids):
         self.helper.fetch_revisions(version_ids)
-
-
-class RemoteToOtherFetcher(GenericRepoFetcher):
-
-    def _fetch_everything_for_search(self, search, pp):
-        data_stream = self.from_repository.get_data_stream_for_search(search)
-        self.to_repository.insert_data_stream(data_stream)
-
-

=== modified file 'bzrlib/inventory.py'
--- bzrlib/inventory.py	2008-05-08 04:33:38 +0000
+++ bzrlib/inventory.py	2008-06-11 04:20:16 +0000
@@ -484,23 +484,13 @@
                 checker.repeated_text_cnt += 1
                 return
 
-        if self.file_id not in checker.checked_weaves:
-            mutter('check weave {%s}', self.file_id)
-            w = tree._get_weave(self.file_id)
-            # Not passing a progress bar, because it creates a new
-            # progress, which overwrites the current progress,
-            # and doesn't look nice
-            w.check()
-            checker.checked_weaves[self.file_id] = True
-        else:
-            w = tree._get_weave(self.file_id)
-
         mutter('check version {%s} of {%s}', tree_revision_id, self.file_id)
         checker.checked_text_cnt += 1
         # We can't check the length, because Weave doesn't store that
         # information, and the whole point of looking at the weave's
         # sha1sum is that we don't have to extract the text.
-        if self.text_sha1 != w.get_sha1s([self.revision])[0]:
+        if (self.text_sha1 !=
+            tree._repository.texts.get_sha1s([(self.file_id, self.revision)])[0]):
             raise BzrCheckError('text {%s} version {%s} wrong sha1' 
                                 % (self.file_id, self.revision))
         checker.checked_texts[t] = self.text_sha1

=== modified file 'bzrlib/knit.py'
--- bzrlib/knit.py	2008-05-12 02:40:40 +0000
+++ bzrlib/knit.py	2008-06-11 04:20:16 +0000
@@ -65,6 +65,7 @@
 from itertools import izip, chain
 import operator
 import os
+import urllib
 import sys
 import warnings
 from zlib import Z_DEFAULT_COMPRESSION
@@ -75,6 +76,7 @@
 from bzrlib import (
     annotate,
     graph as _mod_graph,
+    index as _mod_index,
     lru_cache,
     pack,
     trace,
@@ -115,9 +117,11 @@
 from bzrlib.versionedfile import (
     AbsentContentFactory,
     adapter_registry,
+    ConstantMapper,
     ContentFactory,
-    InterVersionedFile,
+    FulltextContentFactory,
     VersionedFile,
+    VersionedFiles,
     )
 import bzrlib.weave
 
@@ -148,7 +152,7 @@
         :param basis_vf: A versioned file to access basis texts of deltas from.
             May be None for adapters that do not need to access basis texts.
         """
-        self._data = _KnitData(None)
+        self._data = KnitVersionedFiles(None, None)
         self._annotate_factory = KnitAnnotateFactory()
         self._plain_factory = KnitPlainFactory()
         self._basis_vf = basis_vf
@@ -161,7 +165,7 @@
         rec, contents = \
             self._data._parse_record_unchecked(annotated_compressed_bytes)
         content = self._annotate_factory.parse_fulltext(contents, rec[1])
-        size, bytes = self._data._record_to_data(rec[1], rec[3], content.text())
+        size, bytes = self._data._record_to_data((rec[1],), rec[3], content.text())
         return bytes
 
 
@@ -174,7 +178,7 @@
         delta = self._annotate_factory.parse_line_delta(contents, rec[1],
             plain=True)
         contents = self._plain_factory.lower_line_delta(delta)
-        size, bytes = self._data._record_to_data(rec[1], rec[3], contents)
+        size, bytes = self._data._record_to_data((rec[1],), rec[3], contents)
         return bytes
 
 
@@ -184,7 +188,7 @@
     def get_bytes(self, factory, annotated_compressed_bytes):
         rec, contents = \
             self._data._parse_record_unchecked(annotated_compressed_bytes)
-        content, delta = self._annotate_factory.parse_record(factory.key[0],
+        content, delta = self._annotate_factory.parse_record(factory.key[-1],
             contents, factory._build_details, None)
         return ''.join(content.text())
 
@@ -197,8 +201,12 @@
             self._data._parse_record_unchecked(annotated_compressed_bytes)
         delta = self._annotate_factory.parse_line_delta(contents, rec[1],
             plain=True)
-        compression_parent = factory.parents[0][0]
-        basis_lines = self._basis_vf.get_lines(compression_parent)
+        compression_parent = factory.parents[0]
+        basis_entry = self._basis_vf.get_record_stream(
+            [compression_parent], 'unordered', True).next()
+        if basis_entry.storage_kind == 'absent':
+            raise errors.RevisionNotPresent(compression_parent, self._basis_vf)
+        basis_lines = split_lines(basis_entry.get_bytes_as('fulltext'))
         # Manually apply the delta because we have one annotated content and
         # one plain.
         basis_content = PlainKnitContent(basis_lines, compression_parent)
@@ -213,7 +221,7 @@
     def get_bytes(self, factory, compressed_bytes):
         rec, contents = \
             self._data._parse_record_unchecked(compressed_bytes)
-        content, delta = self._plain_factory.parse_record(factory.key[0],
+        content, delta = self._plain_factory.parse_record(factory.key[-1],
             contents, factory._build_details, None)
         return ''.join(content.text())
 
@@ -225,8 +233,13 @@
         rec, contents = \
             self._data._parse_record_unchecked(compressed_bytes)
         delta = self._plain_factory.parse_line_delta(contents, rec[1])
-        compression_parent = factory.parents[0][0]
-        basis_lines = self._basis_vf.get_lines(compression_parent)
+        compression_parent = factory.parents[0]
+        # XXX: string splitting overhead.
+        basis_entry = self._basis_vf.get_record_stream(
+            [compression_parent], 'unordered', True).next()
+        if basis_entry.storage_kind == 'absent':
+            raise errors.RevisionNotPresent(compression_parent, self._basis_vf)
+        basis_lines = split_lines(basis_entry.get_bytes_as('fulltext'))
         basis_content = PlainKnitContent(basis_lines, compression_parent)
         # Manually apply the delta because we have one annotated content and
         # one plain.
@@ -241,11 +254,11 @@
     :seealso ContentFactory:
     """
 
-    def __init__(self, version, parents, build_details, sha1, raw_record,
+    def __init__(self, key, parents, build_details, sha1, raw_record,
         annotated, knit=None):
-        """Create a KnitContentFactory for version.
+        """Create a KnitContentFactory for key.
         
-        :param version: The version.
+        :param key: The key.
         :param parents: The parents.
         :param build_details: The build details as returned from
             get_build_details.
@@ -255,8 +268,8 @@
         """
         ContentFactory.__init__(self)
         self.sha1 = sha1
-        self.key = (version,)
-        self.parents = tuple((parent,) for parent in parents)
+        self.key = key
+        self.parents = parents
         if build_details[0] == 'line-delta':
             kind = 'delta'
         else:
@@ -562,9 +575,19 @@
                        for origin, text in lines)
         return out
 
-    def annotate(self, knit, version_id):
-        content = knit._get_content(version_id)
-        return content.annotate()
+    def annotate(self, knit, key):
+        content = knit._get_content(key)
+        # adjust for the fact that serialised annotations are only key suffixes
+        # for this factory.
+        if type(key) == tuple:
+            prefix = key[:-1]
+            origins = content.annotate()
+            result = []
+            for origin, line in origins:
+                result.append((prefix + (origin,), line))
+            return result
+        else:
+            return content.annotate()
 
 
 class KnitPlainFactory(_KnitFactory):
@@ -624,91 +647,240 @@
             out.extend(lines)
         return out
 
-    def annotate(self, knit, version_id):
+    def annotate(self, knit, key):
         annotator = _KnitAnnotator(knit)
-        return annotator.annotate(version_id)
-
-
-def make_empty_knit(transport, relpath):
-    """Construct a empty knit at the specified location."""
-    k = make_file_knit(transport, relpath, 'w', KnitPlainFactory)
-
-
-def make_file_knit(name, transport, file_mode=None, access_mode='w',
-    factory=None, delta=True, create=False, create_parent_dir=False,
-    delay_create=False, dir_mode=None, get_scope=None):
-    """Factory to create a KnitVersionedFile for a .knit/.kndx file pair."""
-    if factory is None:
-        factory = KnitAnnotateFactory()
-    if get_scope is None:
-        get_scope = lambda:None
-    index = _KnitIndex(transport, name + INDEX_SUFFIX,
-        access_mode, create=create, file_mode=file_mode,
-        create_parent_dir=create_parent_dir, delay_create=delay_create,
-        dir_mode=dir_mode, get_scope=get_scope)
-    access = _KnitAccess(transport, name + DATA_SUFFIX, file_mode,
-        dir_mode, ((create and not len(index)) and delay_create),
-        create_parent_dir)
-    return KnitVersionedFile(name, transport, factory=factory,
-        create=create, delay_create=delay_create, index=index,
-        access_method=access)
-
-
-def get_suffixes():
-    """Return the suffixes used by file based knits."""
-    return [DATA_SUFFIX, INDEX_SUFFIX]
-make_file_knit.get_suffixes = get_suffixes
-
-
-class KnitVersionedFile(VersionedFile):
-    """Weave-like structure with faster random access.
-
-    A knit stores a number of texts and a summary of the relationships
-    between them.  Texts are identified by a string version-id.  Texts
-    are normally stored and retrieved as a series of lines, but can
-    also be passed as single strings.
-
-    Lines are stored with the trailing newline (if any) included, to
-    avoid special cases for files with no final newline.  Lines are
-    composed of 8-bit characters, not unicode.  The combination of
-    these approaches should mean any 'binary' file can be safely
-    stored and retrieved.
-    """
-
-    def __init__(self, relpath, transport, file_mode=None,
-        factory=None, delta=True, create=False, create_parent_dir=False,
-        delay_create=False, dir_mode=None, index=None, access_method=None):
-        """Construct a knit at location specified by relpath.
-        
-        :param create: If not True, only open an existing knit.
-        :param create_parent_dir: If True, create the parent directory if 
-            creating the file fails. (This is used for stores with 
-            hash-prefixes that may not exist yet)
-        :param delay_create: The calling code is aware that the knit won't 
-            actually be created until the first data is stored.
-        :param index: An index to use for the knit.
+        return annotator.annotate(key)
+
+
+
+def make_file_factory(annotated, mapper):
+    """Create a factory for creating a file based KnitVersionedFiles.
+    
+    :param annotated: knit annotations are wanted.
+    :param mapper: The mapper from keys to paths.
+    """
+    def factory(transport):
+        index = _KndxIndex(transport, mapper, lambda:None, lambda:True, lambda:True)
+        access = _KnitKeyAccess(transport, mapper)
+        return KnitVersionedFiles(index, access, annotated=annotated)
+    return factory
+
+
+def make_pack_factory(graph, delta, keylength):
+    """Create a factory for creating a pack based VersionedFiles.
+
+    This is only functional enough to run interface tests, it doesn't try to
+    provide a full pack environment.
+    
+    :param graph: Store a graph.
+    :param delta: Delta compress contents.
+    :param keylength: How long should keys be.
+    """
+    def factory(transport):
+        parents = graph or delta
+        ref_length = 0
+        if graph:
+            ref_length += 1
+        if delta:
+            ref_length += 1
+            max_delta_chain = 200
+        else:
+            max_delta_chain = 0
+        graph_index = _mod_index.InMemoryGraphIndex(reference_lists=ref_length,
+            key_elements=keylength)
+        stream = transport.open_write_stream('newpack')
+        writer = pack.ContainerWriter(stream.write)
+        writer.begin()
+        index = _KnitGraphIndex(graph_index, lambda:True, parents=parents,
+            deltas=delta, add_callback=graph_index.add_nodes)
+        access = _DirectPackAccess({})
+        access.set_writer(writer, graph_index, (transport, 'newpack'))
+        result = KnitVersionedFiles(index, access,
+            max_delta_chain=max_delta_chain)
+        result.stream = stream
+        result.writer = writer
+        return result
+    return factory
+
+
+def cleanup_pack_knit(versioned_files):
+    versioned_files.stream.close()
+    versioned_files.writer.end()
+
+
+class KnitVersionedFiles(VersionedFiles):
+    """Storage for many versioned files using knit compression.
+
+    Backend storage is managed by indices and data objects.
+    """
+
+    def __init__(self, index, data_access, max_delta_chain=200,
+        annotated=False):
+        """Create a KnitVersionedFiles with index and data_access.
+
+        :param index: The index for the knit data.
+        :param data_access: The access object to store and retrieve knit
+            records.
+        :param max_delta_chain: The maximum number of deltas to permit during
+            insertion. Set to 0 to prohibit the use of deltas.
+        :param annotated: Set to True to cause annotations to be calculated and
+            stored during insertion.
         """
-        super(KnitVersionedFile, self).__init__()
-        self.transport = transport
-        self.filename = relpath
-        self.factory = factory or KnitAnnotateFactory()
-        self.delta = delta
-
-        self._max_delta_chain = 200
-
-        if None in (access_method, index):
-            raise ValueError("No default access_method or index any more")
         self._index = index
-        _access = access_method
-        if create and not len(self) and not delay_create:
-            _access.create()
-        self._data = _KnitData(_access)
-
-    def __repr__(self):
-        return '%s(%s)' % (self.__class__.__name__,
-                           self.transport.abspath(self.filename))
-    
-    def _check_should_delta(self, first_parents):
+        self._access = data_access
+        self._max_delta_chain = max_delta_chain
+        if annotated:
+            self._factory = KnitAnnotateFactory()
+        else:
+            self._factory = KnitPlainFactory()
+
+    def add_lines(self, key, parents, lines, parent_texts=None,
+        left_matching_blocks=None, nostore_sha=None, random_id=False,
+        check_content=True):
+        """See VersionedFiles.add_lines()."""
+        self._index._check_write_ok()
+        self._check_add(key, lines, random_id, check_content)
+        if parents is None:
+            # For no-graph knits, have the public interface use None for
+            # parents.
+            parents = ()
+        return self._add(key, lines, parents,
+            parent_texts, left_matching_blocks, nostore_sha, random_id)
+
+    def _add(self, key, lines, parents, parent_texts,
+        left_matching_blocks, nostore_sha, random_id):
+        """Add a set of lines on top of version specified by parents.
+
+        Any versions not present will be converted into ghosts.
+        """
+        # first thing, if the content is something we don't need to store, find
+        # that out.
+        line_bytes = ''.join(lines)
+        digest = sha_string(line_bytes)
+        if nostore_sha == digest:
+            raise errors.ExistingContent
+
+        present_parents = []
+        if parent_texts is None:
+            parent_texts = {}
+        # Do a single query to ascertain parent presence.
+        present_parent_map = self.get_parent_map(parents)
+        for parent in parents:
+            if parent in present_parent_map:
+                present_parents.append(parent)
+
+        # Currently we can only compress against the left most present parent.
+        if (len(present_parents) == 0 or
+            present_parents[0] != parents[0]):
+            delta = False
+        else:
+            # To speed the extract of texts the delta chain is limited
+            # to a fixed number of deltas.  This should minimize both
+            # I/O and the time spend applying deltas.
+            delta = self._check_should_delta(present_parents[0])
+
+        text_length = len(line_bytes)
+        options = []
+        if lines:
+            if lines[-1][-1] != '\n':
+                # copy the contents of lines.
+                lines = lines[:]
+                options.append('no-eol')
+                lines[-1] = lines[-1] + '\n'
+                line_bytes += '\n'
+
+        for element in key:
+            if type(element) != str:
+                raise TypeError("key contains non-strings: %r" % (key,))
+        # Knit hunks are still last-element only
+        version_id = key[-1]
+        content = self._factory.make(lines, version_id)
+        if 'no-eol' in options:
+            # Hint to the content object that its text() call should strip the
+            # EOL.
+            content._should_strip_eol = True
+        if delta or (self._factory.annotated and len(present_parents) > 0):
+            # Merge annotations from parent texts if needed.
+            delta_hunks = self._merge_annotations(content, present_parents,
+                parent_texts, delta, self._factory.annotated,
+                left_matching_blocks)
+
+        if delta:
+            options.append('line-delta')
+            store_lines = self._factory.lower_line_delta(delta_hunks)
+            size, bytes = self._record_to_data(key, digest,
+                store_lines)
+        else:
+            options.append('fulltext')
+            # isinstance is slower and we have no hierarchy.
+            if self._factory.__class__ == KnitPlainFactory:
+                # Use the already joined bytes saving iteration time in
+                # _record_to_data.
+                size, bytes = self._record_to_data(key, digest,
+                    lines, [line_bytes])
+            else:
+                # get mixed annotation + content and feed it into the
+                # serialiser.
+                store_lines = self._factory.lower_fulltext(content)
+                size, bytes = self._record_to_data(key, digest,
+                    store_lines)
+
+        access_memo = self._access.add_raw_records([(key, size)], bytes)[0]
+        self._index.add_records(
+            ((key, options, access_memo, parents),),
+            random_id=random_id)
+        return digest, text_length, content
+
+    def annotate(self, key):
+        """See VersionedFiles.annotate."""
+        return self._factory.annotate(self, key)
+
+    def check(self, progress_bar=None):
+        """See VersionedFiles.check()."""
+        # This doesn't actually test extraction of everything, but that will
+        # impact 'bzr check' substantially, and needs to be integrated with
+        # care. However, it does check for the obvious problem of a delta with
+        # no basis.
+        keys = self.keys()
+        parent_map = self.get_parent_map(keys)
+        for key in keys:
+            if self._index.get_method(key) != 'fulltext':
+                compression_parent = parent_map[key][0]
+                if compression_parent not in parent_map:
+                    raise errors.KnitCorrupt(self,
+                        "Missing basis parent %s for %s" % (
+                        compression_parent, key))
+
+    def _check_add(self, key, lines, random_id, check_content):
+        """check that version_id and lines are safe to add."""
+        if contains_whitespace(key[-1]):
+            raise InvalidRevisionId(key[-1], self.filename)
+        self.check_not_reserved_id(key[-1])
+        # Technically this could be avoided if we are happy to allow duplicate
+        # id insertion when other things than bzr core insert texts, but it
+        # seems useful for folk using the knit api directly to have some safety
+        # blanket that we can disable.
+        ##if not random_id and self.has_version(key):
+        ##    raise RevisionAlreadyPresent(key, self)
+        if check_content:
+            self._check_lines_not_unicode(lines)
+            self._check_lines_are_lines(lines)
+
+    def _check_header(self, key, line):
+        rec = self._split_header(line)
+        self._check_header_version(rec, key[-1])
+        return rec
+
+    def _check_header_version(self, rec, version_id):
+        """Checks the header version on original format knit records.
+        
+        These have the last component of the key embedded in the record.
+        """
+        if rec[1] != version_id:
+            raise KnitCorrupt(self,
+                'unexpected version, wanted %r, got %r' % (version_id, rec[1]))
+
+    def _check_should_delta(self, parent):
         """Iterate back through the parent listing, looking for a fulltext.
 
         This is used when we want to decide whether to add a delta or a new
@@ -721,143 +893,170 @@
         """
         delta_size = 0
         fulltext_size = None
-        delta_parents = first_parents
         for count in xrange(self._max_delta_chain):
-            parent = delta_parents[0]
+            # XXX: Collapse these two queries:
             method = self._index.get_method(parent)
             index, pos, size = self._index.get_position(parent)
             if method == 'fulltext':
                 fulltext_size = size
                 break
             delta_size += size
-            delta_parents = self._index.get_parent_map([parent])[parent]
+            # No exception here because we stop at first fulltext anyway, an
+            # absent parent indicates a corrupt knit anyway.
+            # TODO: This should be asking for compression parent, not graph
+            # parent.
+            parent = self._index.get_parent_map([parent])[parent][0]
         else:
             # We couldn't find a fulltext, so we must create a new one
             return False
-
+        # Simple heuristic - if the total I/O wold be greater as a delta than
+        # the originally installed fulltext, we create a new fulltext.
         return fulltext_size > delta_size
 
-    def _check_write_ok(self):
-        return self._index._check_write_ok()
-
-    def _add_raw_records(self, records, data):
-        """Add all the records 'records' with data pre-joined in 'data'.
-
-        :param records: A list of tuples(version_id, options, parents, size).
-        :param data: The data for the records. When it is written, the records
-                     are adjusted to have pos pointing into data by the sum of
-                     the preceding records sizes.
-        """
-        # write all the data
-        raw_record_sizes = [record[3] for record in records]
-        positions = self._data.add_raw_records(raw_record_sizes, data)
-        index_entries = []
-        for (version_id, options, parents, _), access_memo in zip(
-            records, positions):
-            index_entries.append((version_id, options, access_memo, parents))
-        self._index.add_versions(index_entries)
-
-    def copy_to(self, name, transport):
-        """See VersionedFile.copy_to()."""
-        # copy the current index to a temp index to avoid racing with local
-        # writes
-        transport.put_file_non_atomic(name + INDEX_SUFFIX + '.tmp',
-                self.transport.get(self._index._filename))
-        # copy the data file
-        f = self._data._open_file()
-        try:
-            transport.put_file(name + DATA_SUFFIX, f)
-        finally:
-            f.close()
-        # move the copied index into place
-        transport.move(name + INDEX_SUFFIX + '.tmp', name + INDEX_SUFFIX)
-
-    def get_data_stream(self, required_versions):
-        """Get a data stream for the specified versions.
-
-        Versions may be returned in any order, not necessarily the order
-        specified.  They are returned in a partial order by compression
-        parent, so that the deltas can be applied as the data stream is
-        inserted; however note that compression parents will not be sent
-        unless they were specifically requested, as the client may already
-        have them.
-
-        :param required_versions: The exact set of versions to be extracted.
-            Unlike some other knit methods, this is not used to generate a
-            transitive closure, rather it is used precisely as given.
-        
-        :returns: format_signature, list of (version, options, length, parents),
-            reader_callable.
-        """
-        required_version_set = frozenset(required_versions)
-        version_index = {}
-        # list of revisions that can just be sent without waiting for their
-        # compression parent
-        ready_to_send = []
-        # map from revision to the children based on it
-        deferred = {}
-        # first, read all relevant index data, enough to sort into the right
-        # order to return
-        for version_id in required_versions:
-            options = self._index.get_options(version_id)
-            parents = self._index.get_parents_with_ghosts(version_id)
-            index_memo = self._index.get_position(version_id)
-            version_index[version_id] = (index_memo, options, parents)
-            if ('line-delta' in options
-                and parents[0] in required_version_set):
-                # must wait until the parent has been sent
-                deferred.setdefault(parents[0], []). \
-                    append(version_id)
-            else:
-                # either a fulltext, or a delta whose parent the client did
-                # not ask for and presumably already has
-                ready_to_send.append(version_id)
-        # build a list of results to return, plus instructions for data to
-        # read from the file
-        copy_queue_records = []
-        temp_version_list = []
-        while ready_to_send:
-            # XXX: pushing and popping lists may be a bit inefficient
-            version_id = ready_to_send.pop(0)
-            (index_memo, options, parents) = version_index[version_id]
-            copy_queue_records.append((version_id, index_memo))
-            none, data_pos, data_size = index_memo
-            temp_version_list.append((version_id, options, data_size,
-                parents))
-            if version_id in deferred:
-                # now we can send all the children of this revision - we could
-                # put them in anywhere, but we hope that sending them soon
-                # after the fulltext will give good locality in the receiver
-                ready_to_send[:0] = deferred.pop(version_id)
-        if not (len(deferred) == 0):
-            raise AssertionError("Still have compressed child versions waiting to be sent")
-        # XXX: The stream format is such that we cannot stream it - we have to
-        # know the length of all the data a-priori.
-        raw_datum = []
-        result_version_list = []
-        for (version_id, raw_data, _), \
-            (version_id2, options, _, parents) in \
-            izip(self._data.read_records_iter_raw(copy_queue_records),
-                 temp_version_list):
-            if not (version_id == version_id2):
-                raise AssertionError('logic error, inconsistent results')
-            raw_datum.append(raw_data)
-            result_version_list.append(
-                (version_id, options, len(raw_data), parents))
-        # provide a callback to get data incrementally.
-        pseudo_file = StringIO(''.join(raw_datum))
-        def read(length):
-            if length is None:
-                return pseudo_file.read()
-            else:
-                return pseudo_file.read(length)
-        return (self.get_format_signature(), result_version_list, read)
-
-    def get_record_stream(self, versions, ordering, include_delta_closure):
-        """Get a stream of records for versions.
-
-        :param versions: The versions to include. Each version is a tuple
-            (version,).
+    def _build_details_to_components(self, build_details):
+        """Convert a build_details tuple to a position tuple."""
+        # record_details, access_memo, compression_parent
+        return build_details[3], build_details[0], build_details[1]
+
+    def _get_components_positions(self, keys, noraise=False):
+        """Produce a map of position data for the components of keys.
+
+        This data is intended to be used for retrieving the knit records.
+
+        A dict of key to (record_details, index_memo, next, parents) is
+        returned.
+        method is the way referenced data should be applied.
+        index_memo is the handle to pass to the data access to actually get the
+            data
+        next is the build-parent of the version, or None for fulltexts.
+        parents is the version_ids of the parents of this version
+
+        :param noraise: If True do not raise an error on a missing component,
+            just ignore it.
+        """
+        component_data = {}
+        pending_components = keys
+        while pending_components:
+            build_details = self._index.get_build_details(pending_components)
+            current_components = set(pending_components)
+            pending_components = set()
+            for key, details in build_details.iteritems():
+                (index_memo, compression_parent, parents,
+                 record_details) = details
+                method = record_details[0]
+                if compression_parent is not None:
+                    pending_components.add(compression_parent)
+                component_data[key] = self._build_details_to_components(details)
+            missing = current_components.difference(build_details)
+            if missing and not noraise:
+                raise errors.RevisionNotPresent(missing.pop(), self)
+        return component_data
+       
+    def _get_content(self, key, parent_texts={}):
+        """Returns a content object that makes up the specified
+        version."""
+        cached_version = parent_texts.get(key, None)
+        if cached_version is not None:
+            # Ensure the cache dict is valid.
+            if not self.get_parent_map([key]):
+                raise RevisionNotPresent(key, self)
+            return cached_version
+        text_map, contents_map = self._get_content_maps([key])
+        return contents_map[key]
+
+    def _get_content_maps(self, keys):
+        """Produce maps of text and KnitContents
+        
+        :return: (text_map, content_map) where text_map contains the texts for
+        the requested versions and content_map contains the KnitContents.
+        """
+        # FUTURE: This function could be improved for the 'extract many' case
+        # by tracking each component and only doing the copy when the number of
+        # children than need to apply delta's to it is > 1 or it is part of the
+        # final output.
+        keys = list(keys)
+        multiple_versions = len(keys) != 1
+        record_map = self._get_record_map(keys)
+
+        text_map = {}
+        content_map = {}
+        final_content = {}
+        for key in keys:
+            components = []
+            cursor = key
+            while cursor is not None:
+                record, record_details, digest, next = record_map[cursor]
+                components.append((cursor, record, record_details, digest))
+                if cursor in content_map:
+                    break
+                cursor = next
+
+            content = None
+            for (component_id, record, record_details,
+                 digest) in reversed(components):
+                if component_id in content_map:
+                    content = content_map[component_id]
+                else:
+                    content, delta = self._factory.parse_record(key[-1],
+                        record, record_details, content,
+                        copy_base_content=multiple_versions)
+                    if multiple_versions:
+                        content_map[component_id] = content
+
+            content.cleanup_eol(copy_on_mutate=multiple_versions)
+            final_content[key] = content
+
+            # digest here is the digest from the last applied component.
+            text = content.text()
+            actual_sha = sha_strings(text)
+            if actual_sha != digest:
+                raise KnitCorrupt(self,
+                    '\n  sha-1 %s'
+                    '\n  of reconstructed text does not match'
+                    '\n  expected %s'
+                    '\n  for version %s' %
+                    (actual_sha, digest, key))
+            text_map[key] = text
+        return text_map, final_content
+
+    def get_parent_map(self, keys):
+        """Get a map of the parents of keys.
+
+        :param keys: The keys to look up parents for.
+        :return: A mapping from keys to parents. Absent keys are absent from
+            the mapping.
+        """
+        return self._index.get_parent_map(keys)
+
+    def _get_record_map(self, keys):
+        """Produce a dictionary of knit records.
+        
+        :return: {key:(record, record_details, digest, next)}
+            record
+                data returned from read_records
+            record_details
+                opaque information to pass to parse_record
+            digest
+                SHA1 digest of the full text after all steps are done
+            next
+                build-parent of the version, i.e. the leftmost ancestor.
+                Will be None if the record is not a delta.
+        """
+        position_map = self._get_components_positions(keys)
+        # key = component_id, r = record_details, i_m = index_memo, n = next
+        records = [(key, i_m) for key, (r, i_m, n)
+                             in position_map.iteritems()]
+        record_map = {}
+        for key, record, digest in \
+                self._read_records_iter(records):
+            (record_details, index_memo, next) = position_map[key]
+            record_map[key] = record, record_details, digest, next
+        return record_map
+
+    def get_record_stream(self, keys, ordering, include_delta_closure):
+        """Get a stream of records for keys.
+
+        :param keys: The keys to include.
         :param ordering: Either 'unordered' or 'topological'. A topologically
             sorted stream has compression parents strictly before their
             children.
@@ -866,198 +1065,85 @@
         :return: An iterator of ContentFactory objects, each of which is only
             valid until the iterator is advanced.
         """
+        # keys might be a generator
+        keys = set(keys)
+        if not self._index.has_graph:
+            # Cannot topological order when no graph has been stored.
+            ordering = 'unordered'
         if include_delta_closure:
-            # Nb: what we should do is plan the data to stream to allow
-            # reconstruction of all the texts without excessive buffering,
-            # including re-sending common bases as needed. This makes the most
-            # sense when we start serialising these streams though, so for now
-            # we just fallback to individual text construction behind the
-            # abstraction barrier.
-            knit = self
+            positions = self._get_components_positions(keys, noraise=True)
         else:
-            knit = None
-        # We end up doing multiple index lookups here for parents details and
-        # disk layout details - we need a unified api ?
-        parent_map = self.get_parent_map(versions)
-        absent_versions = set(versions) - set(parent_map)
+            build_details = self._index.get_build_details(keys)
+            positions = dict((key, self._build_details_to_components(details))
+                for key, details in build_details.iteritems())
+        absent_keys = keys.difference(set(positions))
+        # There may be more absent keys : if we're missing the basis component
+        # and are trying to include the delta closure.
+        if include_delta_closure:
+            # key:True means key can be reconstructed
+            checked_keys = {}
+            for key in keys:
+                # the delta chain
+                try:
+                    chain = [key, positions[key][2]]
+                except KeyError:
+                    absent_keys.add(key)
+                    continue
+                result = True
+                while chain[-1] is not None:
+                    if chain[-1] in checked_keys:
+                        result = checked_keys[chain[-1]]
+                        break
+                    else:
+                        try:
+                            chain.append(positions[chain[-1]][2])
+                        except KeyError:
+                            # missing basis component
+                            result = False
+                            break
+                for chain_key in chain[:-1]:
+                    checked_keys[chain_key] = result
+                if not result:
+                    absent_keys.add(key)
+        for key in absent_keys:
+            yield AbsentContentFactory(key)
+        # restrict our view to the keys we can answer.
+        keys = keys - absent_keys
+        # Double index lookups here : need a unified api ?
+        parent_map = self.get_parent_map(keys)
         if ordering == 'topological':
-            present_versions = topo_sort(parent_map)
-        else:
-            # List comprehension to keep the requested order (as that seems
-            # marginally useful, at least until we start doing IO optimising
-            # here.
-            present_versions = [version for version in versions if version in
-                parent_map]
-        position_map = self._get_components_positions(present_versions)
-        records = [(version, position_map[version][1]) for version in
-            present_versions]
-        record_map = {}
-        for version in absent_versions:
-            yield AbsentContentFactory((version,))
-        for version, raw_data, sha1 in \
-                self._data.read_records_iter_raw(records):
-            (record_details, index_memo, _) = position_map[version]
-            yield KnitContentFactory(version, parent_map[version],
-                record_details, sha1, raw_data, self.factory.annotated, knit)
-
-    def _extract_blocks(self, version_id, source, target):
-        if self._index.get_method(version_id) != 'line-delta':
-            return None
-        parent, sha1, noeol, delta = self.get_delta(version_id)
-        return KnitContent.get_line_delta_blocks(delta, source, target)
-
-    def get_delta(self, version_id):
-        """Get a delta for constructing version from some other version."""
-        self.check_not_reserved_id(version_id)
-        parents = self.get_parent_map([version_id])[version_id]
-        if len(parents):
-            parent = parents[0]
-        else:
-            parent = None
-        index_memo = self._index.get_position(version_id)
-        data, sha1 = self._data.read_records(((version_id, index_memo),))[version_id]
-        noeol = 'no-eol' in self._index.get_options(version_id)
-        if 'fulltext' == self._index.get_method(version_id):
-            new_content = self.factory.parse_fulltext(data, version_id)
-            if parent is not None:
-                reference_content = self._get_content(parent)
-                old_texts = reference_content.text()
-            else:
-                old_texts = []
-            new_texts = new_content.text()
-            delta_seq = patiencediff.PatienceSequenceMatcher(None, old_texts,
-                                                             new_texts)
-            return parent, sha1, noeol, self._make_line_delta(delta_seq, new_content)
-        else:
-            delta = self.factory.parse_line_delta(data, version_id)
-            return parent, sha1, noeol, delta
-
-    def get_format_signature(self):
-        """See VersionedFile.get_format_signature()."""
-        if self.factory.annotated:
-            annotated_part = "annotated"
-        else:
-            annotated_part = "plain"
-        return "knit-%s" % (annotated_part,)
-        
-    def get_sha1s(self, version_ids):
-        """See VersionedFile.get_sha1s()."""
-        record_map = self._get_record_map(version_ids)
+            present_keys = topo_sort(parent_map)
+        else:
+            present_keys = keys
+        # XXX: Memory: TODO: batch data here to cap buffered data at (say) 1MB.
+        # XXX: At that point we need to consider double reads by utilising
+        # components multiple times.
+        if include_delta_closure:
+            # XXX: get_content_maps performs its own index queries; allow state
+            # to be passed in.
+            text_map, _ = self._get_content_maps(present_keys)
+            for key in present_keys:
+                yield FulltextContentFactory(key, parent_map[key], None,
+                    ''.join(text_map[key]))
+        else:
+            records = [(key, positions[key][1]) for key in present_keys]
+            for key, raw_data, sha1 in self._read_records_iter_raw(records):
+                (record_details, index_memo, _) = positions[key]
+                yield KnitContentFactory(key, parent_map[key],
+                    record_details, sha1, raw_data, self._factory.annotated, None)
+
+    def get_sha1s(self, keys):
+        """See VersionedFiles.get_sha1s()."""
+        record_map = self._get_record_map(keys)
         # record entry 2 is the 'digest'.
-        return [record_map[v][2] for v in version_ids]
-
-    def insert_data_stream(self, (format, data_list, reader_callable)):
-        """Insert knit records from a data stream into this knit.
-
-        If a version in the stream is already present in this knit, it will not
-        be inserted a second time.  It will be checked for consistency with the
-        stored version however, and may cause a KnitCorrupt error to be raised
-        if the data in the stream disagrees with the already stored data.
-        
-        :seealso: get_data_stream
-        """
-        if format != self.get_format_signature():
-            if 'knit' in debug.debug_flags:
-                trace.mutter(
-                    'incompatible format signature inserting to %r', self)
-            source = self._knit_from_datastream(
-                (format, data_list, reader_callable))
-            stream = source.get_record_stream(source.versions(), 'unordered', False)
-            self.insert_record_stream(stream)
-            return
-
-        for version_id, options, length, parents in data_list:
-            if self.has_version(version_id):
-                # First check: the list of parents.
-                my_parents = self.get_parents_with_ghosts(version_id)
-                if tuple(my_parents) != tuple(parents):
-                    # XXX: KnitCorrupt is not quite the right exception here.
-                    raise KnitCorrupt(
-                        self.filename,
-                        'parents list %r from data stream does not match '
-                        'already recorded parents %r for %s'
-                        % (parents, my_parents, version_id))
-
-                # Also check the SHA-1 of the fulltext this content will
-                # produce.
-                raw_data = reader_callable(length)
-                my_fulltext_sha1 = self.get_sha1s([version_id])[0]
-                df, rec = self._data._parse_record_header(version_id, raw_data)
-                stream_fulltext_sha1 = rec[3]
-                if my_fulltext_sha1 != stream_fulltext_sha1:
-                    # Actually, we don't know if it's this knit that's corrupt,
-                    # or the data stream we're trying to insert.
-                    raise KnitCorrupt(
-                        self.filename, 'sha-1 does not match %s' % version_id)
-            else:
-                if 'line-delta' in options:
-                    # Make sure that this knit record is actually useful: a
-                    # line-delta is no use unless we have its parent.
-                    # Fetching from a broken repository with this problem
-                    # shouldn't break the target repository.
-                    #
-                    # See https://bugs.launchpad.net/bzr/+bug/164443
-                    if not self._index.has_version(parents[0]):
-                        raise KnitCorrupt(
-                            self.filename,
-                            'line-delta from stream '
-                            'for version %s '
-                            'references '
-                            'missing parent %s\n'
-                            'Try running "bzr check" '
-                            'on the source repository, and "bzr reconcile" '
-                            'if necessary.' %
-                            (version_id, parents[0]))
-                    if not self.delta:
-                        # We received a line-delta record for a non-delta knit.
-                        # Convert it to a fulltext.
-                        gzip_bytes = reader_callable(length)
-                        self._convert_line_delta_to_fulltext(
-                            gzip_bytes, version_id, parents)
-                        continue
-
-                self._add_raw_records(
-                    [(version_id, options, parents, length)],
-                    reader_callable(length))
-
-    def _convert_line_delta_to_fulltext(self, gzip_bytes, version_id, parents):
-        lines, sha1 = self._data._parse_record(version_id, gzip_bytes)
-        delta = self.factory.parse_line_delta(lines, version_id)
-        content = self.factory.make(self.get_lines(parents[0]), parents[0])
-        content.apply_delta(delta, version_id)
-        digest, len, content = self.add_lines(
-            version_id, parents, content.text())
-        if digest != sha1:
-            raise errors.VersionedFileInvalidChecksum(version_id)
-
-    def _knit_from_datastream(self, (format, data_list, reader_callable)):
-        """Create a knit object from a data stream.
-
-        This method exists to allow conversion of data streams that do not
-        match the signature of this knit. Generally it will be slower and use
-        more memory to use this method to insert data, but it will work.
-
-        :seealso: get_data_stream for details on datastreams.
-        :return: A knit versioned file which can be used to join the datastream
-            into self.
-        """
-        if format == "knit-plain":
-            factory = KnitPlainFactory()
-        elif format == "knit-annotated":
-            factory = KnitAnnotateFactory()
-        else:
-            raise errors.KnitDataStreamUnknown(format)
-        index = _StreamIndex(data_list, self._index)
-        access = _StreamAccess(reader_callable, index, self, factory)
-        return KnitVersionedFile(self.filename, self.transport,
-            factory=factory, index=index, access_method=access)
+        return [record_map[key][2] for key in keys]
 
     def insert_record_stream(self, stream):
-        """Insert a record stream into this versioned file.
+        """Insert a record stream into this container.
 
         :param stream: A stream of records to insert. 
         :return: None
-        :seealso VersionedFile.get_record_stream:
+        :seealso VersionedFiles.get_record_stream:
         """
         def get_adapter(adapter_key):
             try:
@@ -1067,18 +1153,20 @@
                 adapter = adapter_factory(self)
                 adapters[adapter_key] = adapter
                 return adapter
-        if self.factory.annotated:
+        if self._factory.annotated:
             # self is annotated, we need annotated knits to use directly.
             annotated = "annotated-"
             convertibles = []
         else:
             # self is not annotated, but we can strip annotations cheaply.
             annotated = ""
-            convertibles = set(["knit-annotated-delta-gz",
-                "knit-annotated-ft-gz"])
+            convertibles = set(["knit-annotated-ft-gz"])
+            if self._max_delta_chain:
+                convertibles.add("knit-annotated-delta-gz")
         # The set of types we can cheaply adapt without needing basis texts.
         native_types = set()
-        native_types.add("knit-%sdelta-gz" % annotated)
+        if self._max_delta_chain:
+            native_types.add("knit-%sdelta-gz" % annotated)
         native_types.add("knit-%sft-gz" % annotated)
         knit_types = native_types.union(convertibles)
         adapters = {}
@@ -1091,11 +1179,10 @@
         # key = basis_parent, value = index entry to add
         buffered_index_entries = {}
         for record in stream:
+            parents = record.parents
             # Raise an error when a record is missing.
             if record.storage_kind == 'absent':
-                raise RevisionNotPresent([record.key[0]], self)
-            # adapt to non-tuple interface
-            parents = [parent[0] for parent in record.parents]
+                raise RevisionNotPresent([record.key], self)
             if record.storage_kind in knit_types:
                 if record.storage_kind not in native_types:
                     try:
@@ -1119,20 +1206,26 @@
                 # deprecated format this is tolerable. It can be fixed if
                 # needed by in the kndx index support raising on a duplicate
                 # add with identical parents and options.
-                access_memo = self._data.add_raw_records([len(bytes)], bytes)[0]
-                index_entry = (record.key[0], options, access_memo, parents)
+                access_memo = self._access.add_raw_records(
+                    [(record.key, len(bytes))], bytes)[0]
+                index_entry = (record.key, options, access_memo, parents)
                 buffered = False
                 if 'fulltext' not in options:
                     basis_parent = parents[0]
-                    if not self.has_version(basis_parent):
+                    # Note that pack backed knits don't need to buffer here
+                    # because they buffer all writes to the transaction level,
+                    # but we don't expose that differnet at the index level. If
+                    # the query here has sufficient cost to show up in
+                    # profiling we should do that.
+                    if basis_parent not in self.get_parent_map([basis_parent]):
                         pending = buffered_index_entries.setdefault(
                             basis_parent, [])
                         pending.append(index_entry)
                         buffered = True
                 if not buffered:
-                    self._index.add_versions([index_entry])
+                    self._index.add_records([index_entry])
             elif record.storage_kind == 'fulltext':
-                self.add_lines(record.key[0], parents,
+                self.add_lines(record.key, parents,
                     split_lines(record.get_bytes_as('fulltext')))
             else:
                 adapter_key = record.storage_kind, 'fulltext'
@@ -1140,16 +1233,16 @@
                 lines = split_lines(adapter.get_bytes(
                     record, record.get_bytes_as(record.storage_kind)))
                 try:
-                    self.add_lines(record.key[0], parents, lines)
+                    self.add_lines(record.key, parents, lines)
                 except errors.RevisionAlreadyPresent:
                     pass
             # Add any records whose basis parent is now available.
-            added_keys = [record.key[0]]
+            added_keys = [record.key]
             while added_keys:
                 key = added_keys.pop(0)
                 if key in buffered_index_entries:
                     index_entries = buffered_index_entries[key]
-                    self._index.add_versions(index_entries)
+                    self._index.add_records(index_entries)
                     added_keys.extend(
                         [index_entry[0] for index_entry in index_entries])
                     del buffered_index_entries[key]
@@ -1158,34 +1251,86 @@
             raise errors.RevisionNotPresent(buffered_index_entries.keys()[0],
                 self)
 
-    def versions(self):
-        """See VersionedFile.versions."""
-        if 'evil' in debug.debug_flags:
-            trace.mutter_callsite(2, "versions scales with size of history")
-        return self._index.get_versions()
-
-    def has_version(self, version_id):
-        """See VersionedFile.has_version."""
-        if 'evil' in debug.debug_flags:
-            trace.mutter_callsite(2, "has_version is a LBYL scenario")
-        return self._index.has_version(version_id)
-
-    __contains__ = has_version
+    def iter_lines_added_or_present_in_keys(self, keys, pb=None):
+        """Iterate over the lines in the versioned files from keys.
+
+        This may return lines from other keys. Each item the returned
+        iterator yields is a tuple of a line and a text version that that line
+        is present in (not introduced in).
+
+        Ordering of results is in whatever order is most suitable for the
+        underlying storage format.
+
+        If a progress bar is supplied, it may be used to indicate progress.
+        The caller is responsible for cleaning up progress bars (because this
+        is an iterator).
+
+        NOTES:
+         * Lines are normalised by the underlying store: they will all have \n
+           terminators.
+         * Lines are returned in arbitrary order.
+
+        :return: An iterator over (line, key).
+        """
+        if pb is None:
+            pb = progress.DummyProgress()
+        keys = set(keys)
+        # filter for available keys
+        parent_map = self.get_parent_map(keys)
+        if len(parent_map) != len(keys):
+            missing = set(parent_map) - requested_keys
+            raise RevisionNotPresent(key, self.filename)
+        # we don't care about inclusions, the caller cares.
+        # but we need to setup a list of records to visit.
+        # we need key, position, length
+        key_records = []
+        build_details = self._index.get_build_details(keys)
+        for key in keys:
+            key_records.append((key, build_details[key][0]))
+        total = len(key_records)
+        for key_idx, (key, data, sha_value) in \
+            enumerate(self._read_records_iter(key_records)):
+            pb.update('Walking content.', key_idx, total)
+            compression_parent = build_details[key][1]
+            if compression_parent is None:
+                # fulltext
+                line_iterator = self._factory.get_fulltext_content(data)
+            else:
+                # Delta 
+                line_iterator = self._factory.get_linedelta_content(data)
+            # XXX: It might be more efficient to yield (key,
+            # line_iterator) in the future. However for now, this is a simpler
+            # change to integrate into the rest of the codebase. RBC 20071110
+            for line in line_iterator:
+                yield line, key
+        pb.update('Walking content.', total, total)
+
+    def _make_line_delta(self, delta_seq, new_content):
+        """Generate a line delta from delta_seq and new_content."""
+        diff_hunks = []
+        for op in delta_seq.get_opcodes():
+            if op[0] == 'equal':
+                continue
+            diff_hunks.append((op[1], op[2], op[4]-op[3], new_content._lines[op[3]:op[4]]))
+        return diff_hunks
 
     def _merge_annotations(self, content, parents, parent_texts={},
                            delta=None, annotated=None,
                            left_matching_blocks=None):
-        """Merge annotations for content.  This is done by comparing
-        the annotations based on changed to the text.
+        """Merge annotations for content and generate deltas.
+        
+        This is done by comparing the annotations based on changes to the text
+        and generating a delta on the resulting full texts. If annotations are
+        not being created then a simple delta is created.
         """
         if left_matching_blocks is not None:
             delta_seq = diff._PrematchedMatcher(left_matching_blocks)
         else:
             delta_seq = None
         if annotated:
-            for parent_id in parents:
-                merge_content = self._get_content(parent_id, parent_texts)
-                if (parent_id == parents[0] and delta_seq is not None):
+            for parent_key in parents:
+                merge_content = self._get_content(parent_key, parent_texts)
+                if (parent_key == parents[0] and delta_seq is not None):
                     seq = delta_seq
                 else:
                     seq = patiencediff.PatienceSequenceMatcher(
@@ -1197,6 +1342,12 @@
                     # new content for any line that matches the last-checked
                     # parent.
                     content._lines[j:j+n] = merge_content._lines[i:i+n]
+            if content._lines and content._lines[-1][1][-1] != '\n':
+                # The copied annotation was from a line without a trailing EOL,
+                # reinstate one for the content object, to ensure correct
+                # serialization.
+                line = content._lines[-1][1] + '\n'
+                content._lines[-1] = (content._lines[-1][0], line)
         if delta:
             if delta_seq is None:
                 reference_content = self._get_content(parents[0], parent_texts)
@@ -1206,405 +1357,158 @@
                                                  None, old_texts, new_texts)
             return self._make_line_delta(delta_seq, content)
 
-    def _make_line_delta(self, delta_seq, new_content):
-        """Generate a line delta from delta_seq and new_content."""
-        diff_hunks = []
-        for op in delta_seq.get_opcodes():
-            if op[0] == 'equal':
-                continue
-            diff_hunks.append((op[1], op[2], op[4]-op[3], new_content._lines[op[3]:op[4]]))
-        return diff_hunks
-
-    def _get_components_positions(self, version_ids):
-        """Produce a map of position data for the components of versions.
-
-        This data is intended to be used for retrieving the knit records.
-
-        A dict of version_id to (record_details, index_memo, next, parents) is
-        returned.
-        method is the way referenced data should be applied.
-        index_memo is the handle to pass to the data access to actually get the
-            data
-        next is the build-parent of the version, or None for fulltexts.
-        parents is the version_ids of the parents of this version
-        """
-        component_data = {}
-        pending_components = version_ids
-        while pending_components:
-            build_details = self._index.get_build_details(pending_components)
-            current_components = set(pending_components)
-            pending_components = set()
-            for version_id, details in build_details.iteritems():
-                (index_memo, compression_parent, parents,
-                 record_details) = details
-                method = record_details[0]
-                if compression_parent is not None:
-                    pending_components.add(compression_parent)
-                component_data[version_id] = (record_details, index_memo,
-                                              compression_parent)
-            missing = current_components.difference(build_details)
-            if missing:
-                raise errors.RevisionNotPresent(missing.pop(), self.filename)
-        return component_data
-       
-    def _get_content(self, version_id, parent_texts={}):
-        """Returns a content object that makes up the specified
-        version."""
-        cached_version = parent_texts.get(version_id, None)
-        if cached_version is not None:
-            if not self.has_version(version_id):
-                raise RevisionNotPresent(version_id, self.filename)
-            return cached_version
-
-        text_map, contents_map = self._get_content_maps([version_id])
-        return contents_map[version_id]
-
-    def _check_versions_present(self, version_ids):
-        """Check that all specified versions are present."""
-        self._index.check_versions_present(version_ids)
-
-    def _add_lines_with_ghosts(self, version_id, parents, lines, parent_texts,
-        nostore_sha, random_id, check_content, left_matching_blocks):
-        """See VersionedFile.add_lines_with_ghosts()."""
-        self._check_add(version_id, lines, random_id, check_content)
-        return self._add(version_id, lines, parents, self.delta,
-            parent_texts, left_matching_blocks, nostore_sha, random_id)
-
-    def _add_lines(self, version_id, parents, lines, parent_texts,
-        left_matching_blocks, nostore_sha, random_id, check_content):
-        """See VersionedFile.add_lines."""
-        self._check_add(version_id, lines, random_id, check_content)
-        self._check_versions_present(parents)
-        return self._add(version_id, lines[:], parents, self.delta,
-            parent_texts, left_matching_blocks, nostore_sha, random_id)
-
-    def _check_add(self, version_id, lines, random_id, check_content):
-        """check that version_id and lines are safe to add."""
-        if contains_whitespace(version_id):
-            raise InvalidRevisionId(version_id, self.filename)
-        self.check_not_reserved_id(version_id)
-        # Technically this could be avoided if we are happy to allow duplicate
-        # id insertion when other things than bzr core insert texts, but it
-        # seems useful for folk using the knit api directly to have some safety
-        # blanket that we can disable.
-        if not random_id and self.has_version(version_id):
-            raise RevisionAlreadyPresent(version_id, self.filename)
-        if check_content:
-            self._check_lines_not_unicode(lines)
-            self._check_lines_are_lines(lines)
-
-    def _add(self, version_id, lines, parents, delta, parent_texts,
-        left_matching_blocks, nostore_sha, random_id):
-        """Add a set of lines on top of version specified by parents.
-
-        If delta is true, compress the text as a line-delta against
-        the first parent.
-
-        Any versions not present will be converted into ghosts.
-        """
-        # first thing, if the content is something we don't need to store, find
-        # that out.
-        line_bytes = ''.join(lines)
-        digest = sha_string(line_bytes)
-        if nostore_sha == digest:
-            raise errors.ExistingContent
-
-        present_parents = []
-        if parent_texts is None:
-            parent_texts = {}
-        for parent in parents:
-            if self.has_version(parent):
-                present_parents.append(parent)
-
-        # can only compress against the left most present parent.
-        if (delta and
-            (len(present_parents) == 0 or
-             present_parents[0] != parents[0])):
-            delta = False
-
-        text_length = len(line_bytes)
-        options = []
-        if lines:
-            if lines[-1][-1] != '\n':
-                # copy the contents of lines.
-                lines = lines[:]
-                options.append('no-eol')
-                lines[-1] = lines[-1] + '\n'
-                line_bytes += '\n'
-
-        if delta:
-            # To speed the extract of texts the delta chain is limited
-            # to a fixed number of deltas.  This should minimize both
-            # I/O and the time spend applying deltas.
-            delta = self._check_should_delta(present_parents)
-
-        content = self.factory.make(lines, version_id)
-        if delta or (self.factory.annotated and len(present_parents) > 0):
-            # Merge annotations from parent texts if needed.
-            delta_hunks = self._merge_annotations(content, present_parents,
-                parent_texts, delta, self.factory.annotated,
-                left_matching_blocks)
-
-        if delta:
-            options.append('line-delta')
-            store_lines = self.factory.lower_line_delta(delta_hunks)
-            size, bytes = self._data._record_to_data(version_id, digest,
-                store_lines)
-        else:
-            options.append('fulltext')
-            # isinstance is slower and we have no hierarchy.
-            if self.factory.__class__ == KnitPlainFactory:
-                # Use the already joined bytes saving iteration time in
-                # _record_to_data.
-                size, bytes = self._data._record_to_data(version_id, digest,
-                    lines, [line_bytes])
-            else:
-                # get mixed annotation + content and feed it into the
-                # serialiser.
-                store_lines = self.factory.lower_fulltext(content)
-                size, bytes = self._data._record_to_data(version_id, digest,
-                    store_lines)
-
-        access_memo = self._data.add_raw_records([size], bytes)[0]
-        self._index.add_versions(
-            ((version_id, options, access_memo, parents),),
-            random_id=random_id)
-        return digest, text_length, content
-
-    def check(self, progress_bar=None):
-        """See VersionedFile.check()."""
-        # This doesn't actually test extraction of everything, but that will
-        # impact 'bzr check' substantially, and needs to be integrated with
-        # care. However, it does check for the obvious problem of a delta with
-        # no basis.
-        versions = self.versions()
-        parent_map = self.get_parent_map(versions)
-        for version in versions:
-            if self._index.get_method(version) != 'fulltext':
-                compression_parent = parent_map[version][0]
-                if compression_parent not in parent_map:
-                    raise errors.KnitCorrupt(self,
-                        "Missing basis parent %s for %s" % (
-                        compression_parent, version))
-
-    def get_lines(self, version_id):
-        """See VersionedFile.get_lines()."""
-        return self.get_line_list([version_id])[0]
-
-    def _get_record_map(self, version_ids):
-        """Produce a dictionary of knit records.
-        
-        :return: {version_id:(record, record_details, digest, next)}
-            record
-                data returned from read_records
-            record_details
-                opaque information to pass to parse_record
-            digest
-                SHA1 digest of the full text after all steps are done
-            next
-                build-parent of the version, i.e. the leftmost ancestor.
-                Will be None if the record is not a delta.
-        """
-        position_map = self._get_components_positions(version_ids)
-        # c = component_id, r = record_details, i_m = index_memo, n = next
-        records = [(c, i_m) for c, (r, i_m, n)
-                             in position_map.iteritems()]
-        record_map = {}
-        for component_id, record, digest in \
-                self._data.read_records_iter(records):
-            (record_details, index_memo, next) = position_map[component_id]
-            record_map[component_id] = record, record_details, digest, next
-
-        return record_map
-
-    def get_text(self, version_id):
-        """See VersionedFile.get_text"""
-        return self.get_texts([version_id])[0]
-
-    def get_texts(self, version_ids):
-        return [''.join(l) for l in self.get_line_list(version_ids)]
-
-    def get_line_list(self, version_ids):
-        """Return the texts of listed versions as a list of strings."""
-        for version_id in version_ids:
-            self.check_not_reserved_id(version_id)
-        text_map, content_map = self._get_content_maps(version_ids)
-        return [text_map[v] for v in version_ids]
-
-    _get_lf_split_line_list = get_line_list
-
-    def _get_content_maps(self, version_ids):
-        """Produce maps of text and KnitContents
-        
-        :return: (text_map, content_map) where text_map contains the texts for
-        the requested versions and content_map contains the KnitContents.
-        Both dicts take version_ids as their keys.
-        """
-        # FUTURE: This function could be improved for the 'extract many' case
-        # by tracking each component and only doing the copy when the number of
-        # children than need to apply delta's to it is > 1 or it is part of the
-        # final output.
-        version_ids = list(version_ids)
-        multiple_versions = len(version_ids) != 1
-        record_map = self._get_record_map(version_ids)
-
-        text_map = {}
-        content_map = {}
-        final_content = {}
-        for version_id in version_ids:
-            components = []
-            cursor = version_id
-            while cursor is not None:
-                record, record_details, digest, next = record_map[cursor]
-                components.append((cursor, record, record_details, digest))
-                if cursor in content_map:
-                    break
-                cursor = next
-
-            content = None
-            for (component_id, record, record_details,
-                 digest) in reversed(components):
-                if component_id in content_map:
-                    content = content_map[component_id]
-                else:
-                    content, delta = self.factory.parse_record(version_id,
-                        record, record_details, content,
-                        copy_base_content=multiple_versions)
-                    if multiple_versions:
-                        content_map[component_id] = content
-
-            content.cleanup_eol(copy_on_mutate=multiple_versions)
-            final_content[version_id] = content
-
-            # digest here is the digest from the last applied component.
-            text = content.text()
-            actual_sha = sha_strings(text)
-            if actual_sha != digest:
-                raise KnitCorrupt(self.filename,
-                    '\n  sha-1 %s'
-                    '\n  of reconstructed text does not match'
-                    '\n  expected %s'
-                    '\n  for version %s' %
-                    (actual_sha, digest, version_id))
-            text_map[version_id] = text
-        return text_map, final_content
-
-    def iter_lines_added_or_present_in_versions(self, version_ids=None, 
-                                                pb=None):
-        """See VersionedFile.iter_lines_added_or_present_in_versions()."""
-        if version_ids is None:
-            version_ids = self.versions()
-        if pb is None:
-            pb = progress.DummyProgress()
-        # we don't care about inclusions, the caller cares.
-        # but we need to setup a list of records to visit.
-        # we need version_id, position, length
-        version_id_records = []
-        requested_versions = set(version_ids)
-        # filter for available versions
-        for version_id in requested_versions:
-            if not self.has_version(version_id):
-                raise RevisionNotPresent(version_id, self.filename)
-        # get a in-component-order queue:
-        for version_id in self.versions():
-            if version_id in requested_versions:
-                index_memo = self._index.get_position(version_id)
-                version_id_records.append((version_id, index_memo))
-
-        total = len(version_id_records)
-        for version_idx, (version_id, data, sha_value) in \
-            enumerate(self._data.read_records_iter(version_id_records)):
-            pb.update('Walking content.', version_idx, total)
-            method = self._index.get_method(version_id)
-            if method == 'fulltext':
-                line_iterator = self.factory.get_fulltext_content(data)
-            elif method == 'line-delta':
-                line_iterator = self.factory.get_linedelta_content(data)
-            else:
-                raise ValueError('invalid method %r' % (method,))
-            # XXX: It might be more efficient to yield (version_id,
-            # line_iterator) in the future. However for now, this is a simpler
-            # change to integrate into the rest of the codebase. RBC 20071110
-            for line in line_iterator:
-                yield line, version_id
-
-        pb.update('Walking content.', total, total)
-        
-    def num_versions(self):
-        """See VersionedFile.num_versions()."""
-        return self._index.num_versions()
-
-    __len__ = num_versions
-
-    def annotate(self, version_id):
-        """See VersionedFile.annotate."""
-        return self.factory.annotate(self, version_id)
-
-    def get_parent_map(self, version_ids):
-        """See VersionedFile.get_parent_map."""
-        return self._index.get_parent_map(version_ids)
-
-    def get_ancestry(self, versions, topo_sorted=True):
-        """See VersionedFile.get_ancestry."""
-        if isinstance(versions, basestring):
-            versions = [versions]
-        if not versions:
-            return []
-        return self._index.get_ancestry(versions, topo_sorted)
-
-    def get_ancestry_with_ghosts(self, versions):
-        """See VersionedFile.get_ancestry_with_ghosts."""
-        if isinstance(versions, basestring):
-            versions = [versions]
-        if not versions:
-            return []
-        return self._index.get_ancestry_with_ghosts(versions)
-
-    def plan_merge(self, ver_a, ver_b):
-        """See VersionedFile.plan_merge."""
-        ancestors_b = set(self.get_ancestry(ver_b, topo_sorted=False))
-        ancestors_a = set(self.get_ancestry(ver_a, topo_sorted=False))
-        annotated_a = self.annotate(ver_a)
-        annotated_b = self.annotate(ver_b)
-        return merge._plan_annotate_merge(annotated_a, annotated_b,
-                                          ancestors_a, ancestors_b)
-
-
-class _KnitComponentFile(object):
-    """One of the files used to implement a knit database"""
-
-    def __init__(self, transport, filename, mode, file_mode=None,
-                 create_parent_dir=False, dir_mode=None):
-        self._transport = transport
-        self._filename = filename
-        self._mode = mode
-        self._file_mode = file_mode
-        self._dir_mode = dir_mode
-        self._create_parent_dir = create_parent_dir
-        self._need_to_create = False
-
-    def _full_path(self):
-        """Return the full path to this file."""
-        return self._transport.base + self._filename
-
-    def check_header(self, fp):
-        line = fp.readline()
-        if line == '':
-            # An empty file can actually be treated as though the file doesn't
-            # exist yet.
-            raise errors.NoSuchFile(self._full_path())
-        if line != self.HEADER:
-            raise KnitHeaderError(badline=line,
-                              filename=self._transport.abspath(self._filename))
-
-    def __repr__(self):
-        return '%s(%s)' % (self.__class__.__name__, self._filename)
-
-
-class _KnitIndex(_KnitComponentFile):
-    """Manages knit index file.
-
-    The index is already kept in memory and read on startup, to enable
+    def _parse_record(self, version_id, data):
+        """Parse an original format knit record.
+
+        These have the last element of the key only present in the stored data.
+        """
+        rec, record_contents = self._parse_record_unchecked(data)
+        self._check_header_version(rec, version_id)
+        return record_contents, rec[3]
+
+    def _parse_record_header(self, key, raw_data):
+        """Parse a record header for consistency.
+
+        :return: the header and the decompressor stream.
+                 as (stream, header_record)
+        """
+        df = GzipFile(mode='rb', fileobj=StringIO(raw_data))
+        try:
+            # Current serialise
+            rec = self._check_header(key, df.readline())
+        except Exception, e:
+            raise KnitCorrupt(self,
+                              "While reading {%s} got %s(%s)"
+                              % (key, e.__class__.__name__, str(e)))
+        return df, rec
+
+    def _parse_record_unchecked(self, data):
+        # profiling notes:
+        # 4168 calls in 2880 217 internal
+        # 4168 calls to _parse_record_header in 2121
+        # 4168 calls to readlines in 330
+        df = GzipFile(mode='rb', fileobj=StringIO(data))
+        try:
+            record_contents = df.readlines()
+        except Exception, e:
+            raise KnitCorrupt(self, "Corrupt compressed record %r, got %s(%s)" %
+                (data, e.__class__.__name__, str(e)))
+        header = record_contents.pop(0)
+        rec = self._split_header(header)
+        last_line = record_contents.pop()
+        if len(record_contents) != int(rec[2]):
+            raise KnitCorrupt(self,
+                              'incorrect number of lines %s != %s'
+                              ' for version {%s} %s'
+                              % (len(record_contents), int(rec[2]),
+                                 rec[1], record_contents))
+        if last_line != 'end %s\n' % rec[1]:
+            raise KnitCorrupt(self,
+                              'unexpected version end line %r, wanted %r' 
+                              % (last_line, rec[1]))
+        df.close()
+        return rec, record_contents
+
+    def _read_records_iter(self, records):
+        """Read text records from data file and yield result.
+
+        The result will be returned in whatever is the fastest to read.
+        Not by the order requested. Also, multiple requests for the same
+        record will only yield 1 response.
+        :param records: A list of (key, access_memo) entries
+        :return: Yields (key, contents, digest) in the order
+                 read, not the order requested
+        """
+        if not records:
+            return
+
+        # XXX: This smells wrong, IO may not be getting ordered right.
+        needed_records = sorted(set(records), key=operator.itemgetter(1))
+        if not needed_records:
+            return
+
+        # The transport optimizes the fetching as well 
+        # (ie, reads continuous ranges.)
+        raw_data = self._access.get_raw_records(
+            [index_memo for key, index_memo in needed_records])
+
+        for (key, index_memo), data in \
+                izip(iter(needed_records), raw_data):
+            content, digest = self._parse_record(key[-1], data)
+            yield key, content, digest
+
+    def _read_records_iter_raw(self, records):
+        """Read text records from data file and yield raw data.
+
+        This unpacks enough of the text record to validate the id is
+        as expected but thats all.
+
+        Each item the iterator yields is (key, bytes, sha1_of_full_text).
+        """
+        # setup an iterator of the external records:
+        # uses readv so nice and fast we hope.
+        if len(records):
+            # grab the disk data needed.
+            needed_offsets = [index_memo for key, index_memo
+                                           in records]
+            raw_records = self._access.get_raw_records(needed_offsets)
+
+        for key, index_memo in records:
+            data = raw_records.next()
+            # validate the header (note that we can only use the suffix in
+            # current knit records).
+            df, rec = self._parse_record_header(key, data)
+            df.close()
+            yield key, data, rec[3]
+
+    def _record_to_data(self, key, digest, lines, dense_lines=None):
+        """Convert key, digest, lines into a raw data block.
+        
+        :param key: The key of the record. Currently keys are always serialised
+            using just the trailing component.
+        :param dense_lines: The bytes of lines but in a denser form. For
+            instance, if lines is a list of 1000 bytestrings each ending in \n,
+            dense_lines may be a list with one line in it, containing all the
+            1000's lines and their \n's. Using dense_lines if it is already
+            known is a win because the string join to create bytes in this
+            function spends less time resizing the final string.
+        :return: (len, a StringIO instance with the raw data ready to read.)
+        """
+        # Note: using a string copy here increases memory pressure with e.g.
+        # ISO's, but it is about 3 seconds faster on a 1.2Ghz intel machine
+        # when doing the initial commit of a mozilla tree. RBC 20070921
+        bytes = ''.join(chain(
+            ["version %s %d %s\n" % (key[-1],
+                                     len(lines),
+                                     digest)],
+            dense_lines or lines,
+            ["end %s\n" % key[-1]]))
+        if type(bytes) != str:
+            raise AssertionError(
+                'data must be plain bytes was %s' % type(bytes))
+        if lines and lines[-1][-1] != '\n':
+            raise ValueError('corrupt lines value %r' % lines)
+        compressed_bytes = bytes_to_gzip(bytes)
+        return len(compressed_bytes), compressed_bytes
+
+    def _split_header(self, line):
+        rec = line.split()
+        if len(rec) != 4:
+            raise KnitCorrupt(self,
+                              'unexpected number of elements in record header')
+        return rec
+
+    def keys(self):
+        """See VersionedFiles.keys."""
+        if 'evil' in debug.debug_flags:
+            trace.mutter_callsite(2, "keys scales with size of history")
+        return self._index.keys()
+
+
+class _KndxIndex(object):
+    """Manages knit index files
+
+    The index is kept in memorya  already kept in memory and read on startup, to enable
     fast lookups of revision information.  The cursor of the index
     file is always pointing to the end, making it easy to append
     entries.
@@ -1656,116 +1560,122 @@
 
     HEADER = "# bzr knit index 8\n"
 
-    # speed of knit parsing went from 280 ms to 280 ms with slots addition.
-    # __slots__ = ['_cache', '_history', '_transport', '_filename']
-
-    def _cache_version(self, version_id, options, pos, size, parents):
+    def __init__(self, transport, mapper, get_scope, allow_writes, is_locked):
+        """Create a _KndxIndex on transport using mapper."""
+        self._transport = transport
+        self._mapper = mapper
+        self._get_scope = get_scope
+        self._allow_writes = allow_writes
+        self._is_locked = is_locked
+        self._reset_cache()
+        self.has_graph = True
+
+    def add_records(self, records, random_id=False):
+        """Add multiple records to the index.
+        
+        :param records: a list of tuples:
+                         (key, options, access_memo, parents).
+        :param random_id: If True the ids being added were randomly generated
+            and no check for existence will be performed.
+        """
+        paths = {}
+        for record in records:
+            key = record[0]
+            prefix = key[:-1]
+            path = self._mapper.map(key) + '.kndx'
+            path_keys = paths.setdefault(path, (prefix, []))
+            path_keys[1].append(record)
+        for path in sorted(paths):
+            prefix, path_keys = paths[path]
+            self._load_prefixes([prefix])
+            lines = []
+            orig_history = self._kndx_cache[prefix][1][:]
+            orig_cache = self._kndx_cache[prefix][0].copy()
+
+            try:
+                for key, options, (_, pos, size), parents in path_keys:
+                    if parents is None:
+                        # kndx indices cannot be parentless.
+                        parents = ()
+                    line = "\n%s %s %s %s %s :" % (
+                        key[-1], ','.join(options), pos, size,
+                        self._dictionary_compress(parents))
+                    if type(line) != str:
+                        raise AssertionError(
+                            'data must be utf8 was %s' % type(line))
+                    lines.append(line)
+                    self._cache_key(key, options, pos, size, parents)
+                if len(orig_history):
+                    self._transport.append_bytes(path, ''.join(lines))
+                else:
+                    self._init_index(path, lines)
+            except:
+                # If any problems happen, restore the original values and re-raise
+                self._kndx_cache[prefix] = (orig_cache, orig_history)
+                raise
+
+    def _cache_key(self, key, options, pos, size, parent_keys):
         """Cache a version record in the history array and index cache.
 
         This is inlined into _load_data for performance. KEEP IN SYNC.
         (It saves 60ms, 25% of the __init__ overhead on local 4000 record
          indexes).
         """
+        prefix = key[:-1]
+        version_id = key[-1]
+        # last-element only for compatibilty with the C load_data.
+        parents = tuple(parent[-1] for parent in parent_keys)
+        for parent in parent_keys:
+            if parent[:-1] != prefix:
+                raise ValueError("mismatched prefixes for %r, %r" % (
+                    key, parent_keys))
+        cache, history = self._kndx_cache[prefix]
         # only want the _history index to reference the 1st index entry
         # for version_id
-        if version_id not in self._cache:
-            index = len(self._history)
-            self._history.append(version_id)
+        if version_id not in cache:
+            index = len(history)
+            history.append(version_id)
         else:
-            index = self._cache[version_id][5]
-        self._cache[version_id] = (version_id,
+            index = cache[version_id][5]
+        cache[version_id] = (version_id,
                                    options,
                                    pos,
                                    size,
                                    parents,
                                    index)
 
+    def check_header(self, fp):
+        line = fp.readline()
+        if line == '':
+            # An empty file can actually be treated as though the file doesn't
+            # exist yet.
+            raise errors.NoSuchFile(self)
+        if line != self.HEADER:
+            raise KnitHeaderError(badline=line, filename=self)
+
+    def _check_read(self):
+        if not self._is_locked():
+            raise errors.ObjectNotLocked(self)
+        if self._get_scope() != self._scope:
+            self._reset_cache()
+
     def _check_write_ok(self):
+        """Assert if not writes are permitted."""
+        if not self._is_locked():
+            raise errors.ObjectNotLocked(self)
         if self._get_scope() != self._scope:
-            raise errors.OutSideTransaction()
+            self._reset_cache()
         if self._mode != 'w':
             raise errors.ReadOnlyObjectDirtiedError(self)
 
-    def __init__(self, transport, filename, mode, create=False, file_mode=None,
-        create_parent_dir=False, delay_create=False, dir_mode=None,
-        get_scope=None):
-        _KnitComponentFile.__init__(self, transport, filename, mode,
-                                    file_mode=file_mode,
-                                    create_parent_dir=create_parent_dir,
-                                    dir_mode=dir_mode)
-        self._cache = {}
-        # position in _history is the 'official' index for a revision
-        # but the values may have come from a newer entry.
-        # so - wc -l of a knit index is != the number of unique names
-        # in the knit.
-        self._history = []
-        try:
-            fp = self._transport.get(self._filename)
-            try:
-                # _load_data may raise NoSuchFile if the target knit is
-                # completely empty.
-                _load_data(self, fp)
-            finally:
-                fp.close()
-        except NoSuchFile:
-            if mode != 'w' or not create:
-                raise
-            elif delay_create:
-                self._need_to_create = True
-            else:
-                self._transport.put_bytes_non_atomic(
-                    self._filename, self.HEADER, mode=self._file_mode)
-        self._scope = get_scope()
-        self._get_scope = get_scope
-
-    def get_ancestry(self, versions, topo_sorted=True):
-        """See VersionedFile.get_ancestry."""
-        # get a graph of all the mentioned versions:
-        graph = {}
-        pending = set(versions)
-        cache = self._cache
-        while pending:
-            version = pending.pop()
-            # trim ghosts
-            try:
-                parents = [p for p in cache[version][4] if p in cache]
-            except KeyError:
-                raise RevisionNotPresent(version, self._filename)
-            # if not completed and not a ghost
-            pending.update([p for p in parents if p not in graph])
-            graph[version] = parents
-        if not topo_sorted:
-            return graph.keys()
-        return topo_sort(graph.items())
-
-    def get_ancestry_with_ghosts(self, versions):
-        """See VersionedFile.get_ancestry_with_ghosts."""
-        # get a graph of all the mentioned versions:
-        self.check_versions_present(versions)
-        cache = self._cache
-        graph = {}
-        pending = set(versions)
-        while pending:
-            version = pending.pop()
-            try:
-                parents = cache[version][4]
-            except KeyError:
-                # ghost, fake it
-                graph[version] = []
-            else:
-                # if not completed
-                pending.update([p for p in parents if p not in graph])
-                graph[version] = parents
-        return topo_sort(graph.items())
-
-    def get_build_details(self, version_ids):
-        """Get the method, index_memo and compression parent for version_ids.
+    def get_build_details(self, keys):
+        """Get the method, index_memo and compression parent for keys.
 
         Ghosts are omitted from the result.
 
-        :param version_ids: An iterable of version_ids.
-        :return: A dict of version_id:(index_memo, compression_parent,
-                                       parents, record_details).
+        :param keys: An iterable of keys.
+        :return: A dict of key:(access_memo, compression_parent, parents,
+            record_details).
             index_memo
                 opaque structure to pass to read_records to extract the raw
                 data
@@ -1777,434 +1687,234 @@
                 extra information about the content which needs to be passed to
                 Factory.parse_record
         """
+        prefixes = self._partition_keys(keys)
+        parent_map = self.get_parent_map(keys)
         result = {}
-        for version_id in version_ids:
-            if version_id not in self._cache:
-                # ghosts are omitted
-                continue
-            method = self.get_method(version_id)
-            parents = self.get_parents_with_ghosts(version_id)
+        for key in keys:
+            if key not in parent_map:
+                continue # Ghost
+            method = self.get_method(key)
+            parents = parent_map[key]
             if method == 'fulltext':
                 compression_parent = None
             else:
                 compression_parent = parents[0]
-            noeol = 'no-eol' in self.get_options(version_id)
-            index_memo = self.get_position(version_id)
-            result[version_id] = (index_memo, compression_parent,
+            noeol = 'no-eol' in self.get_options(key)
+            index_memo = self.get_position(key)
+            result[key] = (index_memo, compression_parent,
                                   parents, (method, noeol))
         return result
 
-    def num_versions(self):
-        return len(self._history)
-
-    __len__ = num_versions
-
-    def get_versions(self):
-        """Get all the versions in the file. not topologically sorted."""
-        return self._history
-
-    def _version_list_to_index(self, versions):
+    def get_method(self, key):
+        """Return compression method of specified key."""
+        options = self.get_options(key)
+        if 'fulltext' in options:
+            return 'fulltext'
+        elif 'line-delta' in options:
+            return 'line-delta'
+        else:
+            raise errors.KnitIndexUnknownMethod(self, options)
+
+    def get_options(self, key):
+        """Return a list representing options.
+
+        e.g. ['foo', 'bar']
+        """
+        prefix, suffix = self._split_key(key)
+        self._load_prefixes([prefix])
+        return self._kndx_cache[prefix][0][suffix][1]
+
+    def get_parent_map(self, keys):
+        """Get a map of the parents of keys.
+
+        :param keys: The keys to look up parents for.
+        :return: A mapping from keys to parents. Absent keys are absent from
+            the mapping.
+        """
+        # Parse what we need to up front, this potentially trades off I/O
+        # locality (.kndx and .knit in the same block group for the same file
+        # id) for less checking in inner loops.
+        prefixes = set()
+        prefixes.update(key[:-1] for key in keys)
+        self._load_prefixes(prefixes)
+        result = {}
+        for key in keys:
+            prefix = key[:-1]
+            try:
+                suffix_parents = self._kndx_cache[prefix][0][key[-1]][4]
+            except KeyError:
+                pass
+            else:
+                result[key] = tuple(prefix + (suffix,) for
+                    suffix in suffix_parents)
+        return result
+
+    def get_position(self, key):
+        """Return details needed to access the version.
+        
+        :return: a tuple (key, data position, size) to hand to the access
+            logic to get the record.
+        """
+        prefix, suffix = self._split_key(key)
+        self._load_prefixes([prefix])
+        entry = self._kndx_cache[prefix][0][suffix]
+        return key, entry[2], entry[3]
+
+    def _init_index(self, path, extra_lines=[]):
+        """Initialize an index."""
+        sio = StringIO()
+        sio.write(self.HEADER)
+        sio.writelines(extra_lines)
+        sio.seek(0)
+        self._transport.put_file_non_atomic(path, sio,
+                            create_parent_dir=True)
+                           # self._create_parent_dir)
+                           # mode=self._file_mode,
+                           # dir_mode=self._dir_mode)
+
+    def keys(self):
+        """Get all the keys in the collection.
+        
+        The keys are not ordered.
+        """
+        result = set()
+        # Identify all key prefixes.
+        # XXX: A bit hacky, needs polish.
+        if type(self._mapper) == ConstantMapper:
+            prefixes = [()]
+        else:
+            relpaths = set()
+            for quoted_relpath in self._transport.iter_files_recursive():
+                path, ext = os.path.splitext(quoted_relpath)
+                relpaths.add(path)
+            prefixes = [self._mapper.unmap(path) for path in relpaths]
+        self._load_prefixes(prefixes)
+        for prefix in prefixes:
+            for suffix in self._kndx_cache[prefix][1]:
+                result.add(prefix + (suffix,))
+        return result
+    
+    def _load_prefixes(self, prefixes):
+        """Load the indices for prefixes."""
+        self._check_read()
+        for prefix in prefixes:
+            if prefix not in self._kndx_cache:
+                # the load_data interface writes to these variables.
+                self._cache = {}
+                self._history = []
+                self._filename = prefix
+                try:
+                    path = self._mapper.map(prefix) + '.kndx'
+                    fp = self._transport.get(path)
+                    try:
+                        # _load_data may raise NoSuchFile if the target knit is
+                        # completely empty.
+                        _load_data(self, fp)
+                    finally:
+                        fp.close()
+                    self._kndx_cache[prefix] = (self._cache, self._history)
+                    del self._cache
+                    del self._filename
+                    del self._history
+                except NoSuchFile:
+                    self._kndx_cache[prefix] = ({}, [])
+                    if type(self._mapper) == ConstantMapper:
+                        # preserve behaviour for revisions.kndx etc.
+                        self._init_index(path)
+                    del self._cache
+                    del self._filename
+                    del self._history
+
+    def _partition_keys(self, keys):
+        """Turn keys into a dict of prefix:suffix_list."""
+        result = {}
+        for key in keys:
+            prefix_keys = result.setdefault(key[:-1], [])
+            prefix_keys.append(key[-1])
+        return result
+
+    def _dictionary_compress(self, keys):
+        """Dictionary compress keys.
+        
+        :param keys: The keys to generate references to.
+        :return: A string representation of keys. keys which are present are
+            dictionary compressed, and others are emitted as fulltext with a
+            '.' prefix.
+        """
+        if not keys:
+            return ''
         result_list = []
-        cache = self._cache
-        for version in versions:
-            if version in cache:
+        prefix = keys[0][:-1]
+        cache = self._kndx_cache[prefix][0]
+        for key in keys:
+            if key[:-1] != prefix:
+                # kndx indices cannot refer across partitioned storage.
+                raise ValueError("mismatched prefixes for %r" % keys)
+            if key[-1] in cache:
                 # -- inlined lookup() --
-                result_list.append(str(cache[version][5]))
+                result_list.append(str(cache[key[-1]][5]))
                 # -- end lookup () --
             else:
-                result_list.append('.' + version)
+                result_list.append('.' + key[-1])
         return ' '.join(result_list)
 
-    def add_version(self, version_id, options, index_memo, parents):
-        """Add a version record to the index."""
-        self.add_versions(((version_id, options, index_memo, parents),))
-
-    def add_versions(self, versions, random_id=False):
-        """Add multiple versions to the index.
-        
-        :param versions: a list of tuples:
-                         (version_id, options, pos, size, parents).
-        :param random_id: If True the ids being added were randomly generated
-            and no check for existence will be performed.
-        """
-        lines = []
-        orig_history = self._history[:]
-        orig_cache = self._cache.copy()
-
-        try:
-            for version_id, options, (index, pos, size), parents in versions:
-                line = "\n%s %s %s %s %s :" % (version_id,
-                                               ','.join(options),
-                                               pos,
-                                               size,
-                                               self._version_list_to_index(parents))
-                lines.append(line)
-                self._cache_version(version_id, options, pos, size, tuple(parents))
-            if not self._need_to_create:
-                self._transport.append_bytes(self._filename, ''.join(lines))
-            else:
-                sio = StringIO()
-                sio.write(self.HEADER)
-                sio.writelines(lines)
-                sio.seek(0)
-                self._transport.put_file_non_atomic(self._filename, sio,
-                                    create_parent_dir=self._create_parent_dir,
-                                    mode=self._file_mode,
-                                    dir_mode=self._dir_mode)
-                self._need_to_create = False
-        except:
-            # If any problems happen, restore the original values and re-raise
-            self._history = orig_history
-            self._cache = orig_cache
-            raise
-
-    def has_version(self, version_id):
-        """True if the version is in the index."""
-        return version_id in self._cache
-
-    def get_position(self, version_id):
-        """Return details needed to access the version.
-        
-        .kndx indices do not support split-out data, so return None for the 
-        index field.
-
-        :return: a tuple (None, data position, size) to hand to the access
-            logic to get the record.
-        """
-        entry = self._cache[version_id]
-        return None, entry[2], entry[3]
-
-    def get_method(self, version_id):
-        """Return compression method of specified version."""
-        try:
-            options = self._cache[version_id][1]
-        except KeyError:
-            raise RevisionNotPresent(version_id, self._filename)
-        if 'fulltext' in options:
-            return 'fulltext'
+    def _reset_cache(self):
+        # Possibly this should be a LRU cache. A dictionary from key_prefix to
+        # (cache_dict, history_vector) for parsed kndx files.
+        self._kndx_cache = {}
+        self._scope = self._get_scope()
+        allow_writes = self._allow_writes()
+        if allow_writes:
+            self._mode = 'w'
         else:
-            if 'line-delta' not in options:
-                raise errors.KnitIndexUnknownMethod(self._full_path(), options)
-            return 'line-delta'
-
-    def get_options(self, version_id):
-        """Return a list representing options.
-
-        e.g. ['foo', 'bar']
-        """
-        return self._cache[version_id][1]
-
-    def get_parent_map(self, version_ids):
-        """Passed through to by KnitVersionedFile.get_parent_map."""
-        result = {}
-        for version_id in version_ids:
-            try:
-                result[version_id] = tuple(self._cache[version_id][4])
-            except KeyError:
-                pass
-        return result
-
-    def get_parents_with_ghosts(self, version_id):
-        """Return parents of specified version with ghosts."""
-        try:
-            return self.get_parent_map([version_id])[version_id]
-        except KeyError:
-            raise RevisionNotPresent(version_id, self)
-
-    def check_versions_present(self, version_ids):
-        """Check that all specified versions are present."""
-        cache = self._cache
-        for version_id in version_ids:
-            if version_id not in cache:
-                raise RevisionNotPresent(version_id, self._filename)
-
-
-class KnitGraphIndex(object):
-    """A knit index that builds on GraphIndex."""
-
-    def __init__(self, graph_index, deltas=False, parents=True, add_callback=None):
+            self._mode = 'r'
+
+    def _split_key(self, key):
+        """Split key into a prefix and suffix."""
+        return key[:-1], key[-1]
+
+
+class _KnitGraphIndex(object):
+    """A KnitVersionedFiles index layered on GraphIndex."""
+
+    def __init__(self, graph_index, is_locked, deltas=False, parents=True,
+        add_callback=None):
         """Construct a KnitGraphIndex on a graph_index.
 
         :param graph_index: An implementation of bzrlib.index.GraphIndex.
+        :param is_locked: A callback to check whether the object should answer
+            queries.
         :param deltas: Allow delta-compressed records.
+        :param parents: If True, record knits parents, if not do not record 
+            parents.
         :param add_callback: If not None, allow additions to the index and call
             this callback with a list of added GraphIndex nodes:
             [(node, value, node_refs), ...]
-        :param parents: If True, record knits parents, if not do not record 
-            parents.
+        :param is_locked: A callback, returns True if the index is locked and
+            thus usable.
         """
+        self._add_callback = add_callback
         self._graph_index = graph_index
         self._deltas = deltas
-        self._add_callback = add_callback
         self._parents = parents
         if deltas and not parents:
+            # XXX: TODO: Delta tree and parent graph should be conceptually
+            # separate.
             raise KnitCorrupt(self, "Cannot do delta compression without "
                 "parent tracking.")
-
-    def _check_write_ok(self):
-        pass
-
-    def _get_entries(self, keys, check_present=False):
-        """Get the entries for keys.
-        
-        :param keys: An iterable of index keys, - 1-tuples.
-        """
-        keys = set(keys)
-        found_keys = set()
-        if self._parents:
-            for node in self._graph_index.iter_entries(keys):
-                yield node
-                found_keys.add(node[1])
-        else:
-            # adapt parentless index to the rest of the code.
-            for node in self._graph_index.iter_entries(keys):
-                yield node[0], node[1], node[2], ()
-                found_keys.add(node[1])
-        if check_present:
-            missing_keys = keys.difference(found_keys)
-            if missing_keys:
-                raise RevisionNotPresent(missing_keys.pop(), self)
-
-    def _present_keys(self, version_ids):
-        return set([
-            node[1] for node in self._get_entries(version_ids)])
-
-    def _parentless_ancestry(self, versions):
-        """Honour the get_ancestry API for parentless knit indices."""
-        wanted_keys = self._version_ids_to_keys(versions)
-        present_keys = self._present_keys(wanted_keys)
-        missing = set(wanted_keys).difference(present_keys)
-        if missing:
-            raise RevisionNotPresent(missing.pop(), self)
-        return list(self._keys_to_version_ids(present_keys))
-
-    def get_ancestry(self, versions, topo_sorted=True):
-        """See VersionedFile.get_ancestry."""
-        if not self._parents:
-            return self._parentless_ancestry(versions)
-        # XXX: This will do len(history) index calls - perhaps
-        # it should be altered to be a index core feature?
-        # get a graph of all the mentioned versions:
-        graph = {}
-        ghosts = set()
-        versions = self._version_ids_to_keys(versions)
-        pending = set(versions)
-        while pending:
-            # get all pending nodes
-            this_iteration = pending
-            new_nodes = self._get_entries(this_iteration)
-            found = set()
-            pending = set()
-            for (index, key, value, node_refs) in new_nodes:
-                # dont ask for ghosties - otherwise
-                # we we can end up looping with pending
-                # being entirely ghosted.
-                graph[key] = [parent for parent in node_refs[0]
-                    if parent not in ghosts]
-                # queue parents
-                for parent in graph[key]:
-                    # dont examine known nodes again
-                    if parent in graph:
-                        continue
-                    pending.add(parent)
-                found.add(key)
-            ghosts.update(this_iteration.difference(found))
-        if versions.difference(graph):
-            raise RevisionNotPresent(versions.difference(graph).pop(), self)
-        if topo_sorted:
-            result_keys = topo_sort(graph.items())
-        else:
-            result_keys = graph.iterkeys()
-        return [key[0] for key in result_keys]
-
-    def get_ancestry_with_ghosts(self, versions):
-        """See VersionedFile.get_ancestry."""
-        if not self._parents:
-            return self._parentless_ancestry(versions)
-        # XXX: This will do len(history) index calls - perhaps
-        # it should be altered to be a index core feature?
-        # get a graph of all the mentioned versions:
-        graph = {}
-        versions = self._version_ids_to_keys(versions)
-        pending = set(versions)
-        while pending:
-            # get all pending nodes
-            this_iteration = pending
-            new_nodes = self._get_entries(this_iteration)
-            pending = set()
-            for (index, key, value, node_refs) in new_nodes:
-                graph[key] = node_refs[0]
-                # queue parents 
-                for parent in graph[key]:
-                    # dont examine known nodes again
-                    if parent in graph:
-                        continue
-                    pending.add(parent)
-            missing_versions = this_iteration.difference(graph)
-            missing_needed = versions.intersection(missing_versions)
-            if missing_needed:
-                raise RevisionNotPresent(missing_needed.pop(), self)
-            for missing_version in missing_versions:
-                # add a key, no parents
-                graph[missing_version] = []
-                pending.discard(missing_version) # don't look for it
-        result_keys = topo_sort(graph.items())
-        return [key[0] for key in result_keys]
-
-    def get_build_details(self, version_ids):
-        """Get the method, index_memo and compression parent for version_ids.
-
-        Ghosts are omitted from the result.
-
-        :param version_ids: An iterable of version_ids.
-        :return: A dict of version_id:(index_memo, compression_parent,
-                                       parents, record_details).
-            index_memo
-                opaque structure to pass to read_records to extract the raw
-                data
-            compression_parent
-                Content that this record is built upon, may be None
-            parents
-                Logical parents of this node
-            record_details
-                extra information about the content which needs to be passed to
-                Factory.parse_record
-        """
-        result = {}
-        entries = self._get_entries(self._version_ids_to_keys(version_ids), True)
-        for entry in entries:
-            version_id = self._keys_to_version_ids((entry[1],))[0]
-            if not self._parents:
-                parents = ()
-            else:
-                parents = self._keys_to_version_ids(entry[3][0])
-            if not self._deltas:
-                compression_parent = None
-            else:
-                compression_parent_key = self._compression_parent(entry)
-                if compression_parent_key:
-                    compression_parent = self._keys_to_version_ids(
-                    (compression_parent_key,))[0]
-                else:
-                    compression_parent = None
-            noeol = (entry[2][0] == 'N')
-            if compression_parent:
-                method = 'line-delta'
-            else:
-                method = 'fulltext'
-            result[version_id] = (self._node_to_position(entry),
-                                  compression_parent, parents,
-                                  (method, noeol))
-        return result
-
-    def _compression_parent(self, an_entry):
-        # return the key that an_entry is compressed against, or None
-        # Grab the second parent list (as deltas implies parents currently)
-        compression_parents = an_entry[3][1]
-        if not compression_parents:
-            return None
-        return compression_parents[0]
-
-    def _get_method(self, node):
-        if not self._deltas:
-            return 'fulltext'
-        if self._compression_parent(node):
-            return 'line-delta'
-        else:
-            return 'fulltext'
-
-    def num_versions(self):
-        return len(list(self._graph_index.iter_all_entries()))
-
-    __len__ = num_versions
-
-    def get_versions(self):
-        """Get all the versions in the file. not topologically sorted."""
-        return [node[1][0] for node in self._graph_index.iter_all_entries()]
-    
-    def has_version(self, version_id):
-        """True if the version is in the index."""
-        return len(self._present_keys(self._version_ids_to_keys([version_id]))) == 1
-
-    def _keys_to_version_ids(self, keys):
-        return tuple(key[0] for key in keys)
-
-    def get_position(self, version_id):
-        """Return details needed to access the version.
-        
-        :return: a tuple (index, data position, size) to hand to the access
-            logic to get the record.
-        """
-        node = self._get_node(version_id)
-        return self._node_to_position(node)
-
-    def _node_to_position(self, node):
-        """Convert an index value to position details."""
-        bits = node[2][1:].split(' ')
-        return node[0], int(bits[0]), int(bits[1])
-
-    def get_method(self, version_id):
-        """Return compression method of specified version."""
-        return self._get_method(self._get_node(version_id))
-
-    def _get_node(self, version_id):
-        try:
-            return list(self._get_entries(self._version_ids_to_keys([version_id])))[0]
-        except IndexError:
-            raise RevisionNotPresent(version_id, self)
-
-    def get_options(self, version_id):
-        """Return a list representing options.
-
-        e.g. ['foo', 'bar']
-        """
-        node = self._get_node(version_id)
-        options = [self._get_method(node)]
-        if node[2][0] == 'N':
-            options.append('no-eol')
-        return options
-
-    def get_parent_map(self, version_ids):
-        """Passed through to by KnitVersionedFile.get_parent_map."""
-        nodes = self._get_entries(self._version_ids_to_keys(version_ids))
-        result = {}
-        if self._parents:
-            for node in nodes:
-                result[node[1][0]] = self._keys_to_version_ids(node[3][0])
-        else:
-            for node in nodes:
-                result[node[1][0]] = ()
-        return result
-
-    def get_parents_with_ghosts(self, version_id):
-        """Return parents of specified version with ghosts."""
-        try:
-            return self.get_parent_map([version_id])[version_id]
-        except KeyError:
-            raise RevisionNotPresent(version_id, self)
-
-    def check_versions_present(self, version_ids):
-        """Check that all specified versions are present."""
-        keys = self._version_ids_to_keys(version_ids)
-        present = self._present_keys(keys)
-        missing = keys.difference(present)
-        if missing:
-            raise RevisionNotPresent(missing.pop(), self)
-
-    def add_version(self, version_id, options, access_memo, parents):
-        """Add a version record to the index."""
-        return self.add_versions(((version_id, options, access_memo, parents),))
-
-    def add_versions(self, versions, random_id=False):
-        """Add multiple versions to the index.
+        self.has_graph = parents
+        self._is_locked = is_locked
+
+    def add_records(self, records, random_id=False):
+        """Add multiple records to the index.
         
         This function does not insert data into the Immutable GraphIndex
         backing the KnitGraphIndex, instead it prepares data for insertion by
         the caller and checks that it is safe to insert then calls
         self._add_callback with the prepared GraphIndex nodes.
 
-        :param versions: a list of tuples:
-                         (version_id, options, pos, size, parents).
+        :param records: a list of tuples:
+                         (key, options, access_memo, parents).
         :param random_id: If True the ids being added were randomly generated
             and no check for existence will be performed.
         """
@@ -2212,13 +1922,12 @@
             raise errors.ReadOnlyError(self)
         # we hope there are no repositories with inconsistent parentage
         # anymore.
-        # check for dups
 
         keys = {}
-        for (version_id, options, access_memo, parents) in versions:
+        for (key, options, access_memo, parents) in records:
+            if self._parents:
+                parents = tuple(parents)
             index, pos, size = access_memo
-            key = (version_id, )
-            parents = tuple((parent, ) for parent in parents)
             if 'no-eol' in options:
                 value = 'N'
             else:
@@ -2241,11 +1950,13 @@
                         "in parentless index.")
                 node_refs = ()
             keys[key] = (value, node_refs)
+        # check for dups
         if not random_id:
             present_nodes = self._get_entries(keys)
             for (index, key, value, node_refs) in present_nodes:
-                if (value, node_refs) != keys[key]:
-                    raise KnitCorrupt(self, "inconsistent details in add_versions"
+                if (value[0] != keys[key][0][0] or
+                    node_refs != keys[key][1]):
+                    raise KnitCorrupt(self, "inconsistent details in add_records"
                         ": %s %s" % ((value, node_refs), keys[key]))
                 del keys[key]
         result = []
@@ -2257,132 +1968,274 @@
                 result.append((key, value))
         self._add_callback(result)
         
-    def _version_ids_to_keys(self, version_ids):
-        return set((version_id, ) for version_id in version_ids)
-
-
-class _KnitAccess(object):
-    """Access to knit records in a .knit file."""
-
-    def __init__(self, transport, filename, _file_mode, _dir_mode,
-        _need_to_create, _create_parent_dir):
-        """Create a _KnitAccess for accessing and inserting data.
-
-        :param transport: The transport the .knit is located on.
-        :param filename: The filename of the .knit.
+    def _check_read(self):
+        """raise if reads are not permitted."""
+        if not self._is_locked():
+            raise errors.ObjectNotLocked(self)
+
+    def _check_write_ok(self):
+        """Assert if writes are not permitted."""
+        if not self._is_locked():
+            raise errors.ObjectNotLocked(self)
+
+    def _compression_parent(self, an_entry):
+        # return the key that an_entry is compressed against, or None
+        # Grab the second parent list (as deltas implies parents currently)
+        compression_parents = an_entry[3][1]
+        if not compression_parents:
+            return None
+        if len(compression_parents) != 1:
+            raise AssertionError(
+                "Too many compression parents: %r" % compression_parents)
+        return compression_parents[0]
+
+    def get_build_details(self, keys):
+        """Get the method, index_memo and compression parent for version_ids.
+
+        Ghosts are omitted from the result.
+
+        :param keys: An iterable of keys.
+        :return: A dict of key:
+            (index_memo, compression_parent, parents, record_details).
+            index_memo
+                opaque structure to pass to read_records to extract the raw
+                data
+            compression_parent
+                Content that this record is built upon, may be None
+            parents
+                Logical parents of this node
+            record_details
+                extra information about the content which needs to be passed to
+                Factory.parse_record
+        """
+        self._check_read()
+        result = {}
+        entries = self._get_entries(keys, False)
+        for entry in entries:
+            key = entry[1]
+            if not self._parents:
+                parents = ()
+            else:
+                parents = entry[3][0]
+            if not self._deltas:
+                compression_parent_key = None
+            else:
+                compression_parent_key = self._compression_parent(entry)
+            noeol = (entry[2][0] == 'N')
+            if compression_parent_key:
+                method = 'line-delta'
+            else:
+                method = 'fulltext'
+            result[key] = (self._node_to_position(entry),
+                                  compression_parent_key, parents,
+                                  (method, noeol))
+        return result
+
+    def _get_entries(self, keys, check_present=False):
+        """Get the entries for keys.
+        
+        :param keys: An iterable of index key tuples.
+        """
+        keys = set(keys)
+        found_keys = set()
+        if self._parents:
+            for node in self._graph_index.iter_entries(keys):
+                yield node
+                found_keys.add(node[1])
+        else:
+            # adapt parentless index to the rest of the code.
+            for node in self._graph_index.iter_entries(keys):
+                yield node[0], node[1], node[2], ()
+                found_keys.add(node[1])
+        if check_present:
+            missing_keys = keys.difference(found_keys)
+            if missing_keys:
+                raise RevisionNotPresent(missing_keys.pop(), self)
+
+    def get_method(self, key):
+        """Return compression method of specified key."""
+        return self._get_method(self._get_node(key))
+
+    def _get_method(self, node):
+        if not self._deltas:
+            return 'fulltext'
+        if self._compression_parent(node):
+            return 'line-delta'
+        else:
+            return 'fulltext'
+
+    def _get_node(self, key):
+        try:
+            return list(self._get_entries([key]))[0]
+        except IndexError:
+            raise RevisionNotPresent(key, self)
+
+    def get_options(self, key):
+        """Return a list representing options.
+
+        e.g. ['foo', 'bar']
+        """
+        node = self._get_node(key)
+        options = [self._get_method(node)]
+        if node[2][0] == 'N':
+            options.append('no-eol')
+        return options
+
+    def get_parent_map(self, keys):
+        """Get a map of the parents of keys.
+
+        :param keys: The keys to look up parents for.
+        :return: A mapping from keys to parents. Absent keys are absent from
+            the mapping.
+        """
+        self._check_read()
+        nodes = self._get_entries(keys)
+        result = {}
+        if self._parents:
+            for node in nodes:
+                result[node[1]] = node[3][0]
+        else:
+            for node in nodes:
+                result[node[1]] = None
+        return result
+
+    def get_position(self, key):
+        """Return details needed to access the version.
+        
+        :return: a tuple (index, data position, size) to hand to the access
+            logic to get the record.
+        """
+        node = self._get_node(key)
+        return self._node_to_position(node)
+
+    def keys(self):
+        """Get all the keys in the collection.
+        
+        The keys are not ordered.
+        """
+        self._check_read()
+        return [node[1] for node in self._graph_index.iter_all_entries()]
+    
+    def _node_to_position(self, node):
+        """Convert an index value to position details."""
+        bits = node[2][1:].split(' ')
+        return node[0], int(bits[0]), int(bits[1])
+
+
+class _KnitKeyAccess(object):
+    """Access to records in .knit files."""
+
+    def __init__(self, transport, mapper):
+        """Create a _KnitKeyAccess with transport and mapper.
+
+        :param transport: The transport the access object is rooted at.
+        :param mapper: The mapper used to map keys to .knit files.
         """
         self._transport = transport
-        self._filename = filename
-        self._file_mode = _file_mode
-        self._dir_mode = _dir_mode
-        self._need_to_create = _need_to_create
-        self._create_parent_dir = _create_parent_dir
+        self._mapper = mapper
 
-    def add_raw_records(self, sizes, raw_data):
+    def add_raw_records(self, key_sizes, raw_data):
         """Add raw knit bytes to a storage area.
 
-        The data is spooled to whereever the access method is storing data.
+        The data is spooled to the container writer in one bytes-record per
+        raw data item.
 
-        :param sizes: An iterable containing the size of each raw data segment.
+        :param sizes: An iterable of tuples containing the key and size of each
+            raw data segment.
         :param raw_data: A bytestring containing the data.
-        :return: A list of memos to retrieve the record later. Each memo is a
-            tuple - (index, pos, length), where the index field is always None
-            for the .knit access method.
+        :return: A list of memos to retrieve the record later. Each memo is an
+            opaque index memo. For _KnitKeyAccess the memo is (key, pos,
+            length), where the key is the record key.
         """
-        if not self._need_to_create:
-            base = self._transport.append_bytes(self._filename, raw_data)
-        else:
-            self._transport.put_bytes_non_atomic(self._filename, raw_data,
-                                   create_parent_dir=self._create_parent_dir,
-                                   mode=self._file_mode,
-                                   dir_mode=self._dir_mode)
-            self._need_to_create = False
-            base = 0
+        if type(raw_data) != str:
+            raise AssertionError(
+                'data must be plain bytes was %s' % type(raw_data))
         result = []
-        for size in sizes:
-            result.append((None, base, size))
-            base += size
+        offset = 0
+        # TODO: This can be tuned for writing to sftp and other servers where
+        # append() is relatively expensive by grouping the writes to each key
+        # prefix.
+        for key, size in key_sizes:
+            path = self._mapper.map(key)
+            try:
+                base = self._transport.append_bytes(path + '.knit',
+                    raw_data[offset:offset+size])
+            except errors.NoSuchFile:
+                self._transport.mkdir(osutils.dirname(path))
+                base = self._transport.append_bytes(path + '.knit',
+                    raw_data[offset:offset+size])
+            # if base == 0:
+            # chmod.
+            offset += size
+            result.append((key, base, size))
         return result
 
-    def create(self):
-        """IFF this data access has its own storage area, initialise it.
-
-        :return: None.
-        """
-        self._transport.put_bytes_non_atomic(self._filename, '',
-                                             mode=self._file_mode)
-
-    def open_file(self):
-        """IFF this data access can be represented as a single file, open it.
-
-        For knits that are not mapped to a single file on disk this will
-        always return None.
-
-        :return: None or a file handle.
-        """
-        try:
-            return self._transport.get(self._filename)
-        except NoSuchFile:
-            pass
-        return None
-
     def get_raw_records(self, memos_for_retrieval):
         """Get the raw bytes for a records.
 
-        :param memos_for_retrieval: An iterable containing the (index, pos, 
-            length) memo for retrieving the bytes. The .knit method ignores
-            the index as there is always only a single file.
+        :param memos_for_retrieval: An iterable containing the access memo for
+            retrieving the bytes.
         :return: An iterator over the bytes of the records.
         """
-        read_vector = [(pos, size) for (index, pos, size) in memos_for_retrieval]
-        for pos, data in self._transport.readv(self._filename, read_vector):
-            yield data
-
-
-class _PackAccess(object):
-    """Access to knit records via a collection of packs."""
-
-    def __init__(self, index_to_packs, writer=None):
-        """Create a _PackAccess object.
+        # first pass, group into same-index request to minimise readv's issued.
+        request_lists = []
+        current_prefix = None
+        for (key, offset, length) in memos_for_retrieval:
+            if current_prefix == key[:-1]:
+                current_list.append((offset, length))
+            else:
+                if current_prefix is not None:
+                    request_lists.append((current_prefix, current_list))
+                current_prefix = key[:-1]
+                current_list = [(offset, length)]
+        # handle the last entry
+        if current_prefix is not None:
+            request_lists.append((current_prefix, current_list))
+        for prefix, read_vector in request_lists:
+            path = self._mapper.map(prefix) + '.knit'
+            for pos, data in self._transport.readv(path, read_vector):
+                yield data
+
+
+class _DirectPackAccess(object):
+    """Access to data in one or more packs with less translation."""
+
+    def __init__(self, index_to_packs):
+        """Create a _DirectPackAccess object.
 
         :param index_to_packs: A dict mapping index objects to the transport
             and file names for obtaining data.
-        :param writer: A tuple (pack.ContainerWriter, write_index) which
-            contains the pack to write, and the index that reads from it will
-            be associated with.
         """
-        if writer:
-            self.container_writer = writer[0]
-            self.write_index = writer[1]
-        else:
-            self.container_writer = None
-            self.write_index = None
-        self.indices = index_to_packs
+        self._container_writer = None
+        self._write_index = None
+        self._indices = index_to_packs
 
-    def add_raw_records(self, sizes, raw_data):
+    def add_raw_records(self, key_sizes, raw_data):
         """Add raw knit bytes to a storage area.
 
         The data is spooled to the container writer in one bytes-record per
         raw data item.
 
-        :param sizes: An iterable containing the size of each raw data segment.
+        :param sizes: An iterable of tuples containing the key and size of each
+            raw data segment.
         :param raw_data: A bytestring containing the data.
-        :return: A list of memos to retrieve the record later. Each memo is a
-            tuple - (index, pos, length), where the index field is the 
-            write_index object supplied to the PackAccess object.
+        :return: A list of memos to retrieve the record later. Each memo is an
+            opaque index memo. For _DirectPackAccess the memo is (index, pos,
+            length), where the index field is the write_index object supplied
+            to the PackAccess object.
         """
+        if type(raw_data) != str:
+            raise AssertionError(
+                'data must be plain bytes was %s' % type(raw_data))
         result = []
         offset = 0
-        for size in sizes:
-            p_offset, p_length = self.container_writer.add_bytes_record(
+        for key, size in key_sizes:
+            p_offset, p_length = self._container_writer.add_bytes_record(
                 raw_data[offset:offset+size], [])
             offset += size
-            result.append((self.write_index, p_offset, p_length))
+            result.append((self._write_index, p_offset, p_length))
         return result
 
-    def create(self):
-        """Pack based knits do not get individually created."""
-
     def get_raw_records(self, memos_for_retrieval):
         """Get the raw bytes for a records.
 
@@ -2407,659 +2260,17 @@
         if current_index is not None:
             request_lists.append((current_index, current_list))
         for index, offsets in request_lists:
-            transport, path = self.indices[index]
+            transport, path = self._indices[index]
             reader = pack.make_readv_reader(transport, path, offsets)
             for names, read_func in reader.iter_records():
                 yield read_func(None)
 
-    def open_file(self):
-        """Pack based knits have no single file."""
-        return None
-
-    def set_writer(self, writer, index, (transport, packname)):
+    def set_writer(self, writer, index, transport_packname):
         """Set a writer to use for adding data."""
         if index is not None:
-            self.indices[index] = (transport, packname)
-        self.container_writer = writer
-        self.write_index = index
-
-
-class _StreamAccess(object):
-    """A Knit Access object that provides data from a datastream.
-    
-    It also provides a fallback to present as unannotated data, annotated data
-    from a *backing* access object.
-
-    This is triggered by a index_memo which is pointing to a different index
-    than this was constructed with, and is used to allow extracting full
-    unannotated texts for insertion into annotated knits.
-    """
-
-    def __init__(self, reader_callable, stream_index, backing_knit,
-        orig_factory):
-        """Create a _StreamAccess object.
-
-        :param reader_callable: The reader_callable from the datastream.
-            This is called to buffer all the data immediately, for 
-            random access.
-        :param stream_index: The index the data stream this provides access to
-            which will be present in native index_memo's.
-        :param backing_knit: The knit object that will provide access to 
-            annotated texts which are not available in the stream, so as to
-            create unannotated texts.
-        :param orig_factory: The original content factory used to generate the
-            stream. This is used for checking whether the thunk code for
-            supporting _copy_texts will generate the correct form of data.
-        """
-        self.data = reader_callable(None)
-        self.stream_index = stream_index
-        self.backing_knit = backing_knit
-        self.orig_factory = orig_factory
-
-    def get_raw_records(self, memos_for_retrieval):
-        """Get the raw bytes for a records.
-
-        :param memos_for_retrieval: An iterable of memos from the
-            _StreamIndex object identifying bytes to read; for these classes
-            they are (from_backing_knit, index, start, end) and can point to
-            either the backing knit or streamed data.
-        :return: An iterator yielding a byte string for each record in 
-            memos_for_retrieval.
-        """
-        # use a generator for memory friendliness
-        for from_backing_knit, version_id, start, end in memos_for_retrieval:
-            if not from_backing_knit:
-                if version_id is not self.stream_index:
-                    raise AssertionError()
-                yield self.data[start:end]
-                continue
-            # we have been asked to thunk. This thunking only occurs when
-            # we are obtaining plain texts from an annotated backing knit
-            # so that _copy_texts will work.
-            # We could improve performance here by scanning for where we need
-            # to do this and using get_line_list, then interleaving the output
-            # as desired. However, for now, this is sufficient.
-            if self.orig_factory.__class__ != KnitPlainFactory:
-                raise errors.KnitCorrupt(
-                    self, 'Bad thunk request %r cannot be backed by %r' %
-                        (version_id, self.orig_factory))
-            lines = self.backing_knit.get_lines(version_id)
-            line_bytes = ''.join(lines)
-            digest = sha_string(line_bytes)
-            # the packed form of the fulltext always has a trailing newline,
-            # even if the actual text does not, unless the file is empty.  the
-            # record options including the noeol flag are passed through by
-            # _StreamIndex, so this is safe.
-            if lines:
-                if lines[-1][-1] != '\n':
-                    lines[-1] = lines[-1] + '\n'
-                    line_bytes += '\n'
-            # We want plain data, because we expect to thunk only to allow text
-            # extraction.
-            size, bytes = self.backing_knit._data._record_to_data(version_id,
-                digest, lines, line_bytes)
-            yield bytes
-
-
-class _StreamIndex(object):
-    """A Knit Index object that uses the data map from a datastream."""
-
-    def __init__(self, data_list, backing_index):
-        """Create a _StreamIndex object.
-
-        :param data_list: The data_list from the datastream.
-        :param backing_index: The index which will supply values for nodes
-            referenced outside of this stream.
-        """
-        self.data_list = data_list
-        self.backing_index = backing_index
-        self._by_version = {}
-        pos = 0
-        for key, options, length, parents in data_list:
-            self._by_version[key] = options, (pos, pos + length), parents
-            pos += length
-
-    def get_ancestry(self, versions, topo_sorted):
-        """Get an ancestry list for versions."""
-        if topo_sorted:
-            # Not needed for basic joins
-            raise NotImplementedError(self.get_ancestry)
-        # get a graph of all the mentioned versions:
-        # Little ugly - basically copied from KnitIndex, but don't want to
-        # accidentally incorporate too much of that index's code.
-        ancestry = set()
-        pending = set(versions)
-        cache = self._by_version
-        while pending:
-            version = pending.pop()
-            # trim ghosts
-            try:
-                parents = [p for p in cache[version][2] if p in cache]
-            except KeyError:
-                raise RevisionNotPresent(version, self)
-            # if not completed and not a ghost
-            pending.update([p for p in parents if p not in ancestry])
-            ancestry.add(version)
-        return list(ancestry)
-
-    def get_build_details(self, version_ids):
-        """Get the method, index_memo and compression parent for version_ids.
-
-        Ghosts are omitted from the result.
-
-        :param version_ids: An iterable of version_ids.
-        :return: A dict of version_id:(index_memo, compression_parent,
-                                       parents, record_details).
-            index_memo
-                opaque memo that can be passed to _StreamAccess.read_records
-                to extract the raw data; for these classes it is
-                (from_backing_knit, index, start, end) 
-            compression_parent
-                Content that this record is built upon, may be None
-            parents
-                Logical parents of this node
-            record_details
-                extra information about the content which needs to be passed to
-                Factory.parse_record
-        """
-        result = {}
-        for version_id in version_ids:
-            try:
-                method = self.get_method(version_id)
-            except errors.RevisionNotPresent:
-                # ghosts are omitted
-                continue
-            parent_ids = self.get_parents_with_ghosts(version_id)
-            noeol = ('no-eol' in self.get_options(version_id))
-            index_memo = self.get_position(version_id)
-            from_backing_knit = index_memo[0]
-            if from_backing_knit:
-                # texts retrieved from the backing knit are always full texts
-                method = 'fulltext'
-            if method == 'fulltext':
-                compression_parent = None
-            else:
-                compression_parent = parent_ids[0]
-            result[version_id] = (index_memo, compression_parent,
-                                  parent_ids, (method, noeol))
-        return result
-
-    def get_method(self, version_id):
-        """Return compression method of specified version."""
-        options = self.get_options(version_id)
-        if 'fulltext' in options:
-            return 'fulltext'
-        elif 'line-delta' in options:
-            return 'line-delta'
-        else:
-            raise errors.KnitIndexUnknownMethod(self, options)
-
-    def get_options(self, version_id):
-        """Return a list representing options.
-
-        e.g. ['foo', 'bar']
-        """
-        try:
-            return self._by_version[version_id][0]
-        except KeyError:
-            options = list(self.backing_index.get_options(version_id))
-            if 'fulltext' in options:
-                pass
-            elif 'line-delta' in options:
-                # Texts from the backing knit are always returned from the stream
-                # as full texts
-                options.remove('line-delta')
-                options.append('fulltext')
-            else:
-                raise errors.KnitIndexUnknownMethod(self, options)
-            return tuple(options)
-
-    def get_parent_map(self, version_ids):
-        """Passed through to by KnitVersionedFile.get_parent_map."""
-        result = {}
-        pending_ids = set()
-        for version_id in version_ids:
-            try:
-                result[version_id] = self._by_version[version_id][2]
-            except KeyError:
-                pending_ids.add(version_id)
-        result.update(self.backing_index.get_parent_map(pending_ids))
-        return result
-
-    def get_parents_with_ghosts(self, version_id):
-        """Return parents of specified version with ghosts."""
-        try:
-            return self.get_parent_map([version_id])[version_id]
-        except KeyError:
-            raise RevisionNotPresent(version_id, self)
-
-    def get_position(self, version_id):
-        """Return details needed to access the version.
-        
-        _StreamAccess has the data as a big array, so we return slice
-        coordinates into that (as index_memo's are opaque outside the
-        index and matching access class).
-
-        :return: a tuple (from_backing_knit, index, start, end) that can 
-            be passed e.g. to get_raw_records.  
-            If from_backing_knit is False, index will be self, otherwise it
-            will be a version id.
-        """
-        try:
-            start, end = self._by_version[version_id][1]
-            return False, self, start, end
-        except KeyError:
-            # Signal to the access object to handle this from the backing knit.
-            return (True, version_id, None, None)
-
-    def get_versions(self):
-        """Get all the versions in the stream."""
-        return self._by_version.keys()
-
-
-class _KnitData(object):
-    """Manage extraction of data from a KnitAccess, caching and decompressing.
-    
-    The KnitData class provides the logic for parsing and using knit records,
-    making use of an access method for the low level read and write operations.
-    """
-
-    def __init__(self, access):
-        """Create a KnitData object.
-
-        :param access: The access method to use. Access methods such as
-            _KnitAccess manage the insertion of raw records and the subsequent
-            retrieval of the same.
-        """
-        self._access = access
-        self._checked = False
-
-    def _open_file(self):
-        return self._access.open_file()
-
-    def _record_to_data(self, version_id, digest, lines, dense_lines=None):
-        """Convert version_id, digest, lines into a raw data block.
-        
-        :param dense_lines: The bytes of lines but in a denser form. For
-            instance, if lines is a list of 1000 bytestrings each ending in \n,
-            dense_lines may be a list with one line in it, containing all the
-            1000's lines and their \n's. Using dense_lines if it is already
-            known is a win because the string join to create bytes in this
-            function spends less time resizing the final string.
-        :return: (len, a StringIO instance with the raw data ready to read.)
-        """
-        # Note: using a string copy here increases memory pressure with e.g.
-        # ISO's, but it is about 3 seconds faster on a 1.2Ghz intel machine
-        # when doing the initial commit of a mozilla tree. RBC 20070921
-        bytes = ''.join(chain(
-            ["version %s %d %s\n" % (version_id,
-                                     len(lines),
-                                     digest)],
-            dense_lines or lines,
-            ["end %s\n" % version_id]))
-        compressed_bytes = bytes_to_gzip(bytes)
-        return len(compressed_bytes), compressed_bytes
-
-    def add_raw_records(self, sizes, raw_data):
-        """Append a prepared record to the data file.
-        
-        :param sizes: An iterable containing the size of each raw data segment.
-        :param raw_data: A bytestring containing the data.
-        :return: a list of index data for the way the data was stored.
-            See the access method add_raw_records documentation for more
-            details.
-        """
-        return self._access.add_raw_records(sizes, raw_data)
-        
-    def _parse_record_header(self, version_id, raw_data):
-        """Parse a record header for consistency.
-
-        :return: the header and the decompressor stream.
-                 as (stream, header_record)
-        """
-        df = GzipFile(mode='rb', fileobj=StringIO(raw_data))
-        try:
-            rec = self._check_header(version_id, df.readline())
-        except Exception, e:
-            raise KnitCorrupt(self._access,
-                              "While reading {%s} got %s(%s)"
-                              % (version_id, e.__class__.__name__, str(e)))
-        return df, rec
-
-    def _split_header(self, line):
-        rec = line.split()
-        if len(rec) != 4:
-            raise KnitCorrupt(self._access,
-                              'unexpected number of elements in record header')
-        return rec
-
-    def _check_header_version(self, rec, version_id):
-        if rec[1] != version_id:
-            raise KnitCorrupt(self._access,
-                              'unexpected version, wanted %r, got %r'
-                              % (version_id, rec[1]))
-
-    def _check_header(self, version_id, line):
-        rec = self._split_header(line)
-        self._check_header_version(rec, version_id)
-        return rec
-
-    def _parse_record_unchecked(self, data):
-        # profiling notes:
-        # 4168 calls in 2880 217 internal
-        # 4168 calls to _parse_record_header in 2121
-        # 4168 calls to readlines in 330
-        df = GzipFile(mode='rb', fileobj=StringIO(data))
-        try:
-            record_contents = df.readlines()
-        except Exception, e:
-            raise KnitCorrupt(self._access, "Corrupt compressed record %r, got %s(%s)" %
-                (data, e.__class__.__name__, str(e)))
-        header = record_contents.pop(0)
-        rec = self._split_header(header)
-        last_line = record_contents.pop()
-        if len(record_contents) != int(rec[2]):
-            raise KnitCorrupt(self._access,
-                              'incorrect number of lines %s != %s'
-                              ' for version {%s}'
-                              % (len(record_contents), int(rec[2]),
-                                 rec[1]))
-        if last_line != 'end %s\n' % rec[1]:
-            raise KnitCorrupt(self._access,
-                              'unexpected version end line %r, wanted %r' 
-                              % (last_line, rec[1]))
-        df.close()
-        return rec, record_contents
-
-    def _parse_record(self, version_id, data):
-        rec, record_contents = self._parse_record_unchecked(data)
-        self._check_header_version(rec, version_id)
-        return record_contents, rec[3]
-
-    def read_records_iter_raw(self, records):
-        """Read text records from data file and yield raw data.
-
-        This unpacks enough of the text record to validate the id is
-        as expected but thats all.
-
-        Each item the iterator yields is (version_id, bytes,
-        sha1_of_full_text).
-        """
-        # setup an iterator of the external records:
-        # uses readv so nice and fast we hope.
-        if len(records):
-            # grab the disk data needed.
-            needed_offsets = [index_memo for version_id, index_memo
-                                           in records]
-            raw_records = self._access.get_raw_records(needed_offsets)
-
-        for version_id, index_memo in records:
-            data = raw_records.next()
-            # validate the header
-            df, rec = self._parse_record_header(version_id, data)
-            df.close()
-            yield version_id, data, rec[3]
-
-    def read_records_iter(self, records):
-        """Read text records from data file and yield result.
-
-        The result will be returned in whatever is the fastest to read.
-        Not by the order requested. Also, multiple requests for the same
-        record will only yield 1 response.
-        :param records: A list of (version_id, pos, len) entries
-        :return: Yields (version_id, contents, digest) in the order
-                 read, not the order requested
-        """
-        if not records:
-            return
-
-        needed_records = sorted(set(records), key=operator.itemgetter(1))
-        if not needed_records:
-            return
-
-        # The transport optimizes the fetching as well 
-        # (ie, reads continuous ranges.)
-        raw_data = self._access.get_raw_records(
-            [index_memo for version_id, index_memo in needed_records])
-
-        for (version_id, index_memo), data in \
-                izip(iter(needed_records), raw_data):
-            content, digest = self._parse_record(version_id, data)
-            yield version_id, content, digest
-
-    def read_records(self, records):
-        """Read records into a dictionary."""
-        components = {}
-        for record_id, content, digest in \
-                self.read_records_iter(records):
-            components[record_id] = (content, digest)
-        return components
-
-
-class InterKnit(InterVersionedFile):
-    """Optimised code paths for knit to knit operations."""
-    
-    _matching_file_from_factory = staticmethod(make_file_knit)
-    _matching_file_to_factory = staticmethod(make_file_knit)
-    
-    @staticmethod
-    def is_compatible(source, target):
-        """Be compatible with knits.  """
-        try:
-            return (isinstance(source, KnitVersionedFile) and
-                    isinstance(target, KnitVersionedFile))
-        except AttributeError:
-            return False
-
-    def _copy_texts(self, pb, msg, version_ids, ignore_missing=False):
-        """Copy texts to the target by extracting and adding them one by one.
-
-        see join() for the parameter definitions.
-        """
-        version_ids = self._get_source_version_ids(version_ids, ignore_missing)
-        # --- the below is factorable out with VersionedFile.join, but wait for
-        # VersionedFiles, it may all be simpler then.
-        graph = Graph(self.source)
-        search = graph._make_breadth_first_searcher(version_ids)
-        transitive_ids = set()
-        map(transitive_ids.update, list(search))
-        parent_map = self.source.get_parent_map(transitive_ids)
-        order = topo_sort(parent_map.items())
-
-        def size_of_content(content):
-            return sum(len(line) for line in content.text())
-        # Cache at most 10MB of parent texts
-        parent_cache = lru_cache.LRUSizeCache(max_size=10*1024*1024,
-                                              compute_size=size_of_content)
-        # TODO: jam 20071116 It would be nice to have a streaming interface to
-        #       get multiple texts from a source. The source could be smarter
-        #       about how it handled intermediate stages.
-        #       get_line_list() or make_mpdiffs() seem like a possibility, but
-        #       at the moment they extract all full texts into memory, which
-        #       causes us to store more than our 3x fulltext goal.
-        #       Repository.iter_files_bytes() may be another possibility
-        to_process = [version for version in order
-                               if version not in self.target]
-        total = len(to_process)
-        pb = ui.ui_factory.nested_progress_bar()
-        try:
-            for index, version in enumerate(to_process):
-                pb.update('Converting versioned data', index, total)
-                sha1, num_bytes, parent_text = self.target.add_lines(version,
-                    self.source.get_parents_with_ghosts(version),
-                    self.source.get_lines(version),
-                    parent_texts=parent_cache)
-                parent_cache[version] = parent_text
-        finally:
-            pb.finished()
-        return total
-
-    def join(self, pb=None, msg=None, version_ids=None, ignore_missing=False):
-        """See InterVersionedFile.join."""
-        # If the source and target are mismatched w.r.t. annotations vs
-        # plain, the data needs to be converted accordingly
-        if self.source.factory.annotated == self.target.factory.annotated:
-            converter = None
-        elif self.source.factory.annotated:
-            converter = self._anno_to_plain_converter
-        else:
-            # We're converting from a plain to an annotated knit. Copy them
-            # across by full texts.
-            return self._copy_texts(pb, msg, version_ids, ignore_missing)
-
-        version_ids = self._get_source_version_ids(version_ids, ignore_missing)
-        if not version_ids:
-            return 0
-
-        pb = ui.ui_factory.nested_progress_bar()
-        try:
-            version_ids = list(version_ids)
-            if None in version_ids:
-                version_ids.remove(None)
-    
-            self.source_ancestry = set(self.source.get_ancestry(version_ids,
-                topo_sorted=False))
-            this_versions = set(self.target._index.get_versions())
-            # XXX: For efficiency we should not look at the whole index,
-            #      we only need to consider the referenced revisions - they
-            #      must all be present, or the method must be full-text.
-            #      TODO, RBC 20070919
-            needed_versions = self.source_ancestry - this_versions
-    
-            if not needed_versions:
-                return 0
-            full_list = topo_sort(
-                self.source.get_parent_map(self.source.versions()))
-    
-            version_list = [i for i in full_list if (not self.target.has_version(i)
-                            and i in needed_versions)]
-    
-            # plan the join:
-            copy_queue = []
-            copy_queue_records = []
-            copy_set = set()
-            for version_id in version_list:
-                options = self.source._index.get_options(version_id)
-                parents = self.source._index.get_parents_with_ghosts(version_id)
-                # check that its will be a consistent copy:
-                for parent in parents:
-                    # if source has the parent, we must :
-                    # * already have it or
-                    # * have it scheduled already
-                    # otherwise we don't care
-                    if not (self.target.has_version(parent) or
-                            parent in copy_set or
-                            not self.source.has_version(parent)):
-                        raise AssertionError("problem joining parent %r "
-                            "from %r to %r"
-                            % (parent, self.source, self.target))
-                index_memo = self.source._index.get_position(version_id)
-                copy_queue_records.append((version_id, index_memo))
-                copy_queue.append((version_id, options, parents))
-                copy_set.add(version_id)
-
-            # data suck the join:
-            count = 0
-            total = len(version_list)
-            raw_datum = []
-            raw_records = []
-            for (version_id, raw_data, _), \
-                (version_id2, options, parents) in \
-                izip(self.source._data.read_records_iter_raw(copy_queue_records),
-                     copy_queue):
-                if not (version_id == version_id2):
-                    raise AssertionError('logic error, inconsistent results')
-                count = count + 1
-                pb.update("Joining knit", count, total)
-                if converter:
-                    size, raw_data = converter(raw_data, version_id, options,
-                        parents)
-                else:
-                    size = len(raw_data)
-                raw_records.append((version_id, options, parents, size))
-                raw_datum.append(raw_data)
-            self.target._add_raw_records(raw_records, ''.join(raw_datum))
-            return count
-        finally:
-            pb.finished()
-
-    def _anno_to_plain_converter(self, raw_data, version_id, options,
-                                 parents):
-        """Convert annotated content to plain content."""
-        data, digest = self.source._data._parse_record(version_id, raw_data)
-        if 'fulltext' in options:
-            content = self.source.factory.parse_fulltext(data, version_id)
-            lines = self.target.factory.lower_fulltext(content)
-        else:
-            delta = self.source.factory.parse_line_delta(data, version_id,
-                plain=True)
-            lines = self.target.factory.lower_line_delta(delta)
-        return self.target._data._record_to_data(version_id, digest, lines)
-
-
-InterVersionedFile.register_optimiser(InterKnit)
-
-
-class WeaveToKnit(InterVersionedFile):
-    """Optimised code paths for weave to knit operations."""
-    
-    _matching_file_from_factory = bzrlib.weave.WeaveFile
-    _matching_file_to_factory = staticmethod(make_file_knit)
-    
-    @staticmethod
-    def is_compatible(source, target):
-        """Be compatible with weaves to knits."""
-        try:
-            return (isinstance(source, bzrlib.weave.Weave) and
-                    isinstance(target, KnitVersionedFile))
-        except AttributeError:
-            return False
-
-    def join(self, pb=None, msg=None, version_ids=None, ignore_missing=False):
-        """See InterVersionedFile.join."""
-        version_ids = self._get_source_version_ids(version_ids, ignore_missing)
-
-        if not version_ids:
-            return 0
-
-        pb = ui.ui_factory.nested_progress_bar()
-        try:
-            version_ids = list(version_ids)
-    
-            self.source_ancestry = set(self.source.get_ancestry(version_ids))
-            this_versions = set(self.target._index.get_versions())
-            needed_versions = self.source_ancestry - this_versions
-    
-            if not needed_versions:
-                return 0
-            full_list = topo_sort(
-                self.source.get_parent_map(self.source.versions()))
-    
-            version_list = [i for i in full_list if (not self.target.has_version(i)
-                            and i in needed_versions)]
-    
-            # do the join:
-            count = 0
-            total = len(version_list)
-            parent_map = self.source.get_parent_map(version_list)
-            for version_id in version_list:
-                pb.update("Converting to knit", count, total)
-                parents = parent_map[version_id]
-                # check that its will be a consistent copy:
-                for parent in parents:
-                    # if source has the parent, we must already have it
-                    if not self.target.has_version(parent):
-                        raise AssertionError("%r does not have parent %r"
-                            % (self.target, parent))
-                self.target.add_lines(
-                    version_id, parents, self.source.get_lines(version_id))
-                count = count + 1
-            return count
-        finally:
-            pb.finished()
-
-
-InterVersionedFile.register_optimiser(WeaveToKnit)
+            self._indices[index] = transport_packname
+        self._container_writer = writer
+        self._write_index = index
 
 
 # Deprecated, use PatienceSequenceMatcher instead
@@ -3172,7 +2383,7 @@
             self._check_parents(child, nodes_to_annotate)
         return nodes_to_annotate
 
-    def _get_build_graph(self, revision_id):
+    def _get_build_graph(self, key):
         """Get the graphs for building texts and annotations.
 
         The data you need for creating a full text may be different than the
@@ -3180,14 +2391,14 @@
         parents to create an annotation, but only need 1 parent to generate the
         fulltext.)
 
-        :return: A list of (revision_id, index_memo) records, suitable for
+        :return: A list of (key, index_memo) records, suitable for
             passing to read_records_iter to start reading in the raw data fro/
             the pack file.
         """
-        if revision_id in self._annotated_lines:
+        if key in self._annotated_lines:
             # Nothing to do
             return []
-        pending = set([revision_id])
+        pending = set([key])
         records = []
         generation = 0
         kept_generation = 0
@@ -3199,26 +2410,26 @@
             self._all_build_details.update(build_details)
             # new_nodes = self._knit._index._get_entries(this_iteration)
             pending = set()
-            for rev_id, details in build_details.iteritems():
+            for key, details in build_details.iteritems():
                 (index_memo, compression_parent, parents,
                  record_details) = details
-                self._revision_id_graph[rev_id] = parents
-                records.append((rev_id, index_memo))
+                self._revision_id_graph[key] = parents
+                records.append((key, index_memo))
                 # Do we actually need to check _annotated_lines?
                 pending.update(p for p in parents
                                  if p not in self._all_build_details)
                 if compression_parent:
                     self._compression_children.setdefault(compression_parent,
-                        []).append(rev_id)
+                        []).append(key)
                 if parents:
                     for parent in parents:
                         self._annotate_children.setdefault(parent,
-                            []).append(rev_id)
+                            []).append(key)
                     num_gens = generation - kept_generation
                     if ((num_gens >= self._generations_until_keep)
                         and len(parents) > 1):
                         kept_generation = generation
-                        self._nodes_to_keep_annotations.add(rev_id)
+                        self._nodes_to_keep_annotations.add(key)
 
             missing_versions = this_iteration.difference(build_details.keys())
             self._ghosts.update(missing_versions)
@@ -3250,7 +2461,7 @@
         # still need parents, cleaning them up when those parents are
         # processed.
         for (rev_id, record,
-             digest) in self._knit._data.read_records_iter(records):
+             digest) in self._knit._read_records_iter(records):
             if rev_id in self._annotated_lines:
                 continue
             parent_ids = self._revision_id_graph[rev_id]
@@ -3265,7 +2476,7 @@
             if len(parent_ids) == 0:
                 # There are no parents for this node, so just add it
                 # TODO: This probably needs to be decoupled
-                fulltext_content, delta = self._knit.factory.parse_record(
+                fulltext_content, delta = self._knit._factory.parse_record(
                     rev_id, record, record_details, None)
                 fulltext = self._add_fulltext_content(rev_id, fulltext_content)
                 nodes_to_annotate.extend(self._add_annotation(rev_id, fulltext,
@@ -3299,7 +2510,7 @@
                         parent_fulltext_content = self._fulltext_contents[compression_parent]
                         parent_fulltext = parent_fulltext_content.text()
                     comp_children.remove(rev_id)
-                    fulltext_content, delta = self._knit.factory.parse_record(
+                    fulltext_content, delta = self._knit._factory.parse_record(
                         rev_id, record, record_details,
                         parent_fulltext_content,
                         copy_base_content=(not reuse_content))
@@ -3308,7 +2519,7 @@
                     blocks = KnitContent.get_line_delta_blocks(delta,
                             parent_fulltext, fulltext)
                 else:
-                    fulltext_content = self._knit.factory.parse_fulltext(
+                    fulltext_content = self._knit._factory.parse_fulltext(
                         record, rev_id)
                     fulltext = self._add_fulltext_content(rev_id,
                         fulltext_content)
@@ -3328,16 +2539,16 @@
         self._heads_provider = head_cache
         return head_cache
 
-    def annotate(self, revision_id):
-        """Return the annotated fulltext at the given revision.
+    def annotate(self, key):
+        """Return the annotated fulltext at the given key.
 
-        :param revision_id: The revision id for this file
+        :param key: The key to annotate.
         """
-        records = self._get_build_graph(revision_id)
-        if revision_id in self._ghosts:
-            raise errors.RevisionNotPresent(revision_id, self._knit)
+        records = self._get_build_graph(key)
+        if key in self._ghosts:
+            raise errors.RevisionNotPresent(key, self._knit)
         self._annotate_records(records)
-        return self._annotated_lines[revision_id]
+        return self._annotated_lines[key]
 
 
 try:

=== modified file 'bzrlib/log.py'
--- bzrlib/log.py	2008-05-08 04:33:38 +0000
+++ bzrlib/log.py	2008-06-11 04:20:16 +0000
@@ -455,9 +455,6 @@
     :return: A list of (revision_id, dotted_revno, merge_depth) tuples.
     """
     # find all the revisions that change the specific file
-    file_weave = branch.repository.weave_store.get_weave(file_id,
-                branch.repository.get_transaction())
-    weave_modifed_revisions = set(file_weave.versions())
     # build the ancestry of each revision in the graph
     # - only listing the ancestors that change the specific file.
     graph = branch.repository.get_graph()
@@ -469,16 +466,19 @@
     parent_map = dict(((key, value) for key, value in
         graph.iter_ancestry(mainline_revisions[1:]) if value is not None))
     sorted_rev_list = topo_sort(parent_map.items())
+    text_keys = [(file_id, rev_id) for rev_id in sorted_rev_list]
+    modified_text_versions = branch.repository.texts.get_parent_map(text_keys)
     ancestry = {}
     for rev in sorted_rev_list:
+        text_key = (file_id, rev)
         parents = parent_map[rev]
-        if rev not in weave_modifed_revisions and len(parents) == 1:
+        if text_key not in modified_text_versions and len(parents) == 1:
             # We will not be adding anything new, so just use a reference to
             # the parent ancestry.
             rev_ancestry = ancestry[parents[0]]
         else:
             rev_ancestry = set()
-            if rev in weave_modifed_revisions:
+            if text_key in modified_text_versions:
                 rev_ancestry.add(rev)
             for parent in parents:
                 if parent not in ancestry:
@@ -502,7 +502,7 @@
     # filter from the view the revisions that did not change or merge 
     # the specific file
     return [(r, n, d) for r, n, d in view_revs_iter
-            if r in weave_modifed_revisions or is_merging_rev(r)]
+            if (file_id, r) in modified_text_versions or is_merging_rev(r)]
 
 
 def get_view_revisions(mainline_revs, rev_nos, branch, direction,

=== modified file 'bzrlib/merge.py'
--- bzrlib/merge.py	2008-05-08 04:12:06 +0000
+++ bzrlib/merge.py	2008-06-11 04:20:16 +0000
@@ -15,8 +15,9 @@
 # Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 
 
+import errno
+from itertools import chain
 import os
-import errno
 import warnings
 
 from bzrlib import (
@@ -42,6 +43,7 @@
                            WorkingTreeNotRevision,
                            BinaryFile,
                            )
+from bzrlib.graph import Graph
 from bzrlib.merge3 import Merge3
 from bzrlib.osutils import rename, pathjoin
 from progress import DummyProgress, ProgressPhase
@@ -256,7 +258,7 @@
     def compare_basis(self):
         try:
             basis_tree = self.revision_tree(self.this_tree.last_revision())
-        except errors.RevisionNotPresent:
+        except errors.NoSuchRevision:
             basis_tree = self.this_tree.basis_tree()
         changes = self.this_tree.changes_from(basis_tree)
         if not changes.has_changed():
@@ -276,7 +278,7 @@
         for revision_id in new_parents:
             try:
                 tree = self.revision_tree(revision_id)
-            except errors.RevisionNotPresent:
+            except errors.NoSuchRevision:
                 tree = None
             else:
                 tree.lock_read()
@@ -1240,21 +1242,38 @@
 
 class _PlanMergeBase(object):
 
-    def __init__(self, a_rev, b_rev, vf):
+    def __init__(self, a_rev, b_rev, vf, prefix):
         """Contructor.
 
         :param a_rev: Revision-id of one revision to merge
         :param b_rev: Revision-id of the other revision to merge
-        :param vf: A versionedfile containing both revisions
+        :param vf: A VersionedFiles containing both revisions
+        :param prefix: A prefix for accessing keys in vf.
         """
         self.a_rev = a_rev
         self.b_rev = b_rev
-        self.lines_a = vf.get_lines(a_rev)
-        self.lines_b = vf.get_lines(b_rev)
         self.vf = vf
         self._last_lines = None
         self._last_lines_revision_id = None
         self._cached_matching_blocks = {}
+        self._prefix = prefix
+        lines = self.get_lines([a_rev, b_rev])
+        self.lines_a = lines[a_rev]
+        self.lines_b = lines[b_rev]
+
+    def get_lines(self, revisions):
+        """Get lines for revisions from the backing VersionedFiles.
+        
+        :raises RevisionNotPresent on absent texts.
+        """
+        keys = dict((self._prefix + (rev,), rev) for rev in revisions)
+        result = {}
+        for record in self.vf.get_record_stream(keys, 'unordered', True):
+            if record.storage_kind == 'absent':
+                raise errors.RevisionNotPresent(record.key, self.vf)
+            result[keys[record.key]] = osutils.split_lines(
+                record.get_bytes_as('fulltext'))
+        return result
 
     def plan_merge(self):
         """Generate a 'plan' for merging the two revisions.
@@ -1308,9 +1327,11 @@
             return cached
         if self._last_lines_revision_id == left_revision:
             left_lines = self._last_lines
+            right_lines = self.get_lines([right_revision])[right_revision]
         else:
-            left_lines = self.vf.get_lines(left_revision)
-        right_lines = self.vf.get_lines(right_revision)
+            lines = self.get_lines([left_revision, right_revision])
+            left_lines = lines[left_revision]
+            right_lines = lines[right_revision]
         self._last_lines = right_lines
         self._last_lines_revision_id = right_revision
         matcher = patiencediff.PatienceSequenceMatcher(None, left_lines,
@@ -1369,11 +1390,16 @@
 class _PlanMerge(_PlanMergeBase):
     """Plan an annotate merge using on-the-fly annotation"""
 
-    def __init__(self, a_rev, b_rev, vf):
-       _PlanMergeBase.__init__(self, a_rev, b_rev, vf)
-       a_ancestry = set(vf.get_ancestry(a_rev, topo_sorted=False))
-       b_ancestry = set(vf.get_ancestry(b_rev, topo_sorted=False))
-       self.uncommon = a_ancestry.symmetric_difference(b_ancestry)
+    def __init__(self, a_rev, b_rev, vf, prefix):
+        _PlanMergeBase.__init__(self, a_rev, b_rev, vf, prefix)
+        graph = Graph(vf)
+        # XXX: There is probably a better API to use to examine less history.
+        a_ancestry = set(chain(*graph._make_breadth_first_searcher(
+            [prefix + (a_rev,)])))
+        b_ancestry = set(chain(*graph._make_breadth_first_searcher(
+            [prefix + (b_rev,)])))
+        self.uncommon = set(key[-1] for key in
+            a_ancestry.symmetric_difference(b_ancestry))
 
     def _determine_status(self, revision_id, unique_line_numbers):
         """Determines the status unique lines versus all lcas.
@@ -1399,9 +1425,11 @@
         """
         if version_id not in self.uncommon:
             return set()
-        parents = self.vf.get_parent_map([version_id])[version_id]
+        key = self._prefix + (version_id,)
+        parent_map = self.vf.get_parent_map([key])
+        parents = tuple(parent[-1] for parent in parent_map[key])
         if len(parents) == 0:
-            return set(range(len(self.vf.get_lines(version_id))))
+            return set(range(len(self.get_lines([version_id])[version_id])))
         new = None
         for parent in parents:
             blocks = self._get_matching_blocks(version_id, parent)
@@ -1429,11 +1457,12 @@
     This is faster, and hopefully produces more useful output.
     """
 
-    def __init__(self, a_rev, b_rev, vf, graph):
-        _PlanMergeBase.__init__(self, a_rev, b_rev, vf)
-        self.lcas = graph.find_lca(a_rev, b_rev)
+    def __init__(self, a_rev, b_rev, vf, prefix, graph):
+        _PlanMergeBase.__init__(self, a_rev, b_rev, vf, prefix)
+        lcas = graph.find_lca(prefix + (a_rev,), prefix + (b_rev,))
+        self.lcas = set(lca[-1] for lca in lcas)
         for lca in self.lcas:
-            lca_lines = self.vf.get_lines(lca)
+            lca_lines = self.get_lines([lca])[lca]
             matcher = patiencediff.PatienceSequenceMatcher(None, self.lines_a,
                                                            lca_lines)
             blocks = list(matcher.get_matching_blocks())

=== modified file 'bzrlib/multiparent.py'
--- bzrlib/multiparent.py	2008-04-24 07:22:53 +0000
+++ bzrlib/multiparent.py	2008-06-11 04:20:16 +0000
@@ -33,16 +33,30 @@
 from bzrlib.tuned_gzip import GzipFile
 
 
+def topo_iter_keys(vf, keys=None):
+    if keys is None:
+        keys = vf.keys()
+    parents = vf.get_parent_map(keys)
+    return _topo_iter(parents, keys)
+
 def topo_iter(vf, versions=None):
-    seen = set()
-    descendants = {}
     if versions is None:
         versions = vf.versions()
     parents = vf.get_parent_map(versions)
+    return _topo_iter(parents, versions)
+
+def _topo_iter(parents, versions):
+    seen = set()
+    descendants = {}
     def pending_parents(version):
+        if parents[version] is None:
+            return []
         return [v for v in parents[version] if v in versions and
                 v not in seen]
     for version_id in versions:
+        if parents[version_id] is None:
+            # parentless
+            continue
         for parent_id in parents[version_id]:
             descendants.setdefault(parent_id, []).append(version_id)
     cur = [v for v in versions if len(pending_parents(v)) == 0]

=== modified file 'bzrlib/reconcile.py'
--- bzrlib/reconcile.py	2008-05-08 04:33:38 +0000
+++ bzrlib/reconcile.py	2008-06-11 04:20:16 +0000
@@ -34,6 +34,7 @@
     )
 from bzrlib.trace import mutter, note
 from bzrlib.tsort import TopoSorter
+from bzrlib.versionedfile import AdapterFactory, FulltextContentFactory
 
 
 def reconcile(dir, other=None):
@@ -213,9 +214,10 @@
         from bzrlib.weave import WeaveFile, Weave
         transaction = self.repo.get_transaction()
         self.pb.update('Reading inventory data.')
-        self.inventory = self.repo.get_inventory_weave()
+        self.inventory = self.repo.inventories
+        self.revisions = self.repo.revisions
         # the total set of revisions to process
-        self.pending = set([rev_id for rev_id in self.repo._revision_store.all_revision_ids(transaction)])
+        self.pending = set([key[-1] for key in self.revisions.keys()])
 
         # mapping from revision_id to parents
         self._rev_graph = {}
@@ -234,47 +236,49 @@
             self.pb.note('Inventory ok.')
             return
         self.pb.update('Backing up inventory...', 0, 0)
-        self.repo.control_weaves.copy(self.inventory, 'inventory.backup', self.repo.get_transaction())
+        self.repo._backup_inventory()
         self.pb.note('Backup Inventory created.')
-        # asking for '' should never return a non-empty weave
-        new_inventory_vf = self.repo.control_weaves.get_empty('inventory.new',
-            self.repo.get_transaction())
+        new_inventories = self.repo._temp_inventories()
 
         # we have topological order of revisions and non ghost parents ready.
         self._setup_steps(len(self._rev_graph))
-        for rev_id in TopoSorter(self._rev_graph.items()).iter_topo_order():
-            parents = self._rev_graph[rev_id]
-            # double check this really is in topological order.
-            unavailable = [p for p in parents if p not in new_inventory_vf]
-            if unavailable:
-                raise AssertionError('unavailable parents: %r'
-                    % unavailable)
-            # this entry has all the non ghost parents in the inventory
-            # file already.
-            self._reweave_step('adding inventories')
-            if isinstance(new_inventory_vf, WeaveFile):
-                # It's really a WeaveFile, but we call straight into the
-                # Weave's add method to disable the auto-write-out behaviour.
-                # This is done to avoid a revision_count * time-to-write additional overhead on 
-                # reconcile.
-                new_inventory_vf._check_write_ok()
-                Weave._add_lines(new_inventory_vf, rev_id, parents,
-                    self.inventory.get_lines(rev_id), None, None, None, False, True)
-            else:
-                new_inventory_vf.add_lines(rev_id, parents, self.inventory.get_lines(rev_id))
-
-        if isinstance(new_inventory_vf, WeaveFile):
-            new_inventory_vf._save()
-        # if this worked, the set of new_inventory_vf.names should equal
+        revision_keys = [(rev_id,) for rev_id in
+            TopoSorter(self._rev_graph.items()).iter_topo_order()]
+        stream = self._change_inv_parents(
+            self.inventory.get_record_stream(revision_keys, 'unsorted', True),
+            self._new_inv_parents,
+            set(revision_keys))
+        new_inventories.insert_record_stream(stream)
+        # if this worked, the set of new_inventories.keys should equal
         # self.pending
-        if not (set(new_inventory_vf.versions()) == self.pending):
+        if not (set(new_inventories.keys()) ==
+            set([(revid,) for revid in self.pending])):
             raise AssertionError()
         self.pb.update('Writing weave')
-        self.repo.control_weaves.copy(new_inventory_vf, 'inventory', self.repo.get_transaction())
-        self.repo.control_weaves.delete('inventory.new', self.repo.get_transaction())
+        self.repo._activate_new_inventory()
         self.inventory = None
         self.pb.note('Inventory regenerated.')
 
+    def _new_inv_parents(self, revision_key):
+        """Lookup ghost-filtered parents for revision_key."""
+        # Use the filtered ghostless parents list:
+        return tuple([(revid,) for revid in self._rev_graph[revision_key[-1]]])
+
+    def _change_inv_parents(self, stream, get_parents, all_revision_keys):
+        """Adapt a record stream to reconcile the parents."""
+        for record in stream:
+            wanted_parents = get_parents(record.key)
+            if wanted_parents and wanted_parents[0] not in all_revision_keys:
+                # The check for the left most parent only handles knit
+                # compressors, but this code only applies to knit and weave
+                # repositories anyway.
+                bytes = record.get_bytes_as('fulltext')
+                yield FulltextContentFactory(record.key, wanted_parents, record.sha1, bytes)
+            else:
+                adapted_record = AdapterFactory(record.key, wanted_parents, record)
+                yield adapted_record
+            self._reweave_step('adding inventories')
+
     def _setup_steps(self, new_total):
         """Setup the markers we need to control the progress bar."""
         self.total = new_total
@@ -293,34 +297,6 @@
             else:
                 mutter('found ghost %s', parent)
         self._rev_graph[rev_id] = parents
-        if self._parents_are_inconsistent(rev_id, parents):
-            self.inconsistent_parents += 1
-            mutter('Inconsistent inventory parents: id {%s} '
-                   'inventory claims %r, '
-                   'available parents are %r, '
-                   'unavailable parents are %r',
-                   rev_id,
-                   set(self.inventory.get_parent_map([rev_id])[rev_id]),
-                   set(parents),
-                   set(rev.parent_ids).difference(set(parents)))
-
-    def _parents_are_inconsistent(self, rev_id, parents):
-        """Return True if the parents list of rev_id does not match the weave.
-
-        This detects inconsistencies based on the self.thorough value:
-        if thorough is on, the first parent value is checked as well as ghost
-        differences.
-        Otherwise only the ghost differences are evaluated.
-        """
-        weave_parents = self.inventory.get_parent_map([rev_id])[rev_id]
-        weave_missing_old_ghosts = set(weave_parents) != set(parents)
-        first_parent_is_wrong = (
-            len(weave_parents) and len(parents) and
-            parents[0] != weave_parents[0])
-        if self.thorough:
-            return weave_missing_old_ghosts or first_parent_is_wrong
-        else:
-            return weave_missing_old_ghosts
 
     def _check_garbage_inventories(self):
         """Check for garbage inventories which we cannot trust
@@ -330,12 +306,12 @@
         """
         if not self.thorough:
             return
-        inventories = set(self.inventory.versions())
-        revisions = set(self._rev_graph.keys())
+        inventories = set(self.inventory.keys())
+        revisions = set(self.revisions.keys())
         garbage = inventories.difference(revisions)
         self.garbage_inventories = len(garbage)
-        for revision_id in garbage:
-            mutter('Garbage inventory {%s} found.', revision_id)
+        for revision_key in garbage:
+            mutter('Garbage inventory {%s} found.', revision_key[-1])
 
     def _parent_is_available(self, parent):
         """True if parent is a fully available revision
@@ -343,8 +319,10 @@
         A fully available revision has a inventory and a revision object in the
         repository.
         """
-        return (parent in self._rev_graph or 
-                (parent in self.inventory and self.repo.has_revision(parent)))
+        if parent in self._rev_graph:
+            return True
+        inv_present = (1 == len(self.inventory.get_parent_map([(parent,)])))
+        return (inv_present and self.repo.has_revision(parent))
 
     def _reweave_step(self, message):
         """Mark a single step of regeneration complete."""
@@ -374,10 +352,10 @@
         """Load indexes for the reconciliation."""
         self.transaction = self.repo.get_transaction()
         self.pb.update('Reading indexes.', 0, 2)
-        self.inventory = self.repo.get_inventory_weave()
+        self.inventory = self.repo.inventories
         self.pb.update('Reading indexes.', 1, 2)
         self.repo._check_for_inconsistent_revision_parents()
-        self.revisions = self.repo._revision_store.get_revision_file(self.transaction)
+        self.revisions = self.repo.revisions
         self.pb.update('Reading indexes.', 2, 2)
 
     def _gc_inventory(self):
@@ -389,54 +367,29 @@
             self.pb.note('Inventory ok.')
             return
         self.pb.update('Backing up inventory...', 0, 0)
-        self.repo.control_weaves.copy(self.inventory, 'inventory.backup', self.transaction)
+        self.repo._backup_inventory()
         self.pb.note('Backup Inventory created.')
         # asking for '' should never return a non-empty weave
-        new_inventory_vf = self.repo.control_weaves.get_empty('inventory.new',
-            self.transaction)
-
+        new_inventories = self.repo._temp_inventories()
         # we have topological order of revisions and non ghost parents ready.
-        self._setup_steps(len(self.revisions))
-        revision_ids = self.revisions.versions()
-        graph = self.revisions.get_parent_map(revision_ids)
-        for rev_id in TopoSorter(graph.items()).iter_topo_order():
-            parents = graph[rev_id]
-            # double check this really is in topological order, ignoring existing ghosts.
-            unavailable = [p for p in parents if p not in new_inventory_vf and
-                p in self.revisions]
-            if unavailable:
-                raise AssertionError(
-                    'unavailable parents: %r' % (unavailable,))
-            # this entry has all the non ghost parents in the inventory
-            # file already.
-            self._reweave_step('adding inventories')
-            # ugly but needed, weaves are just way tooooo slow else.
-            new_inventory_vf.add_lines_with_ghosts(rev_id, parents,
-                self.inventory.get_lines(rev_id))
-
+        graph = self.revisions.get_parent_map(self.revisions.keys())
+        revision_keys = list(TopoSorter(graph).iter_topo_order())
+        revision_ids = [key[-1] for key in revision_keys]
+        self._setup_steps(len(revision_keys))
+        stream = self._change_inv_parents(
+            self.inventory.get_record_stream(revision_keys, 'unsorted', True),
+            graph.__getitem__,
+            set(revision_keys))
+        new_inventories.insert_record_stream(stream)
         # if this worked, the set of new_inventory_vf.names should equal
-        # self.pending
-        if not(set(new_inventory_vf.versions()) == set(self.revisions.versions())):
+        # the revisionds list
+        if not(set(new_inventories.keys()) == set(revision_keys)):
             raise AssertionError()
         self.pb.update('Writing weave')
-        self.repo.control_weaves.copy(new_inventory_vf, 'inventory', self.transaction)
-        self.repo.control_weaves.delete('inventory.new', self.transaction)
+        self.repo._activate_new_inventory()
         self.inventory = None
         self.pb.note('Inventory regenerated.')
 
-    def _check_garbage_inventories(self):
-        """Check for garbage inventories which we cannot trust
-
-        We cant trust them because their pre-requisite file data may not
-        be present - all we know is that their revision was not installed.
-        """
-        inventories = set(self.inventory.versions())
-        revisions = set(self.revisions.versions())
-        garbage = inventories.difference(revisions)
-        self.garbage_inventories = len(garbage)
-        for revision_id in garbage:
-            mutter('Garbage inventory {%s} found.', revision_id)
-
     def _fix_text_parents(self):
         """Fix bad versionedfile parent entries.
 
@@ -447,59 +400,78 @@
         parent lists, and replaces the versionedfile with a corrected version.
         """
         transaction = self.repo.get_transaction()
-        versions = self.revisions.versions()
+        versions = [key[-1] for key in self.revisions.keys()]
         mutter('Prepopulating revision text cache with %d revisions',
                 len(versions))
         vf_checker = self.repo._get_versioned_file_checker()
-        # List all weaves before altering, to avoid race conditions when we
-        # delete unused weaves.
-        weaves = list(enumerate(self.repo.weave_store))
-        for num, file_id in weaves:
+        bad_parents, unused_versions = vf_checker.check_file_version_parents(
+            self.repo.texts, self.pb)
+        text_index = vf_checker.text_index
+        per_id_bad_parents = {}
+        for key in unused_versions:
+            # Ensure that every file with unused versions gets rewritten.
+            # NB: This is really not needed, reconcile != pack.
+            per_id_bad_parents[key[0]] = {}
+        # Generate per-knit/weave data.
+        for key, details in bad_parents.iteritems():
+            file_id = key[0]
+            rev_id = key[1]
+            knit_parents = tuple([parent[-1] for parent in details[0]])
+            correct_parents = tuple([parent[-1] for parent in details[1]])
+            file_details = per_id_bad_parents.setdefault(file_id, {})
+            file_details[rev_id] = (knit_parents, correct_parents)
+        file_id_versions = {}
+        for text_key in text_index:
+            versions_list = file_id_versions.setdefault(text_key[0], [])
+            versions_list.append(text_key[1])
+        # Do the reconcile of individual weaves.
+        for num, file_id in enumerate(per_id_bad_parents):
             self.pb.update('Fixing text parents', num,
-                           len(self.repo.weave_store))
-            vf = self.repo.weave_store.get_weave(file_id, transaction)
-            versions_with_bad_parents, unused_versions = \
-                vf_checker.check_file_version_parents(vf, file_id)
-            if (len(versions_with_bad_parents) == 0 and
-                len(unused_versions) == 0):
-                continue
-            full_text_versions = set()
-            self._fix_text_parent(file_id, vf, versions_with_bad_parents,
-                full_text_versions, unused_versions)
+                           len(per_id_bad_parents))
+            versions_with_bad_parents = per_id_bad_parents[file_id]
+            id_unused_versions = set(key[-1] for key in unused_versions
+                if key[0] == file_id)
+            if file_id in file_id_versions:
+                file_versions = file_id_versions[file_id]
+            else:
+                # This id was present in the disk store but is not referenced
+                # by any revision at all.
+                file_versions = []
+            self._fix_text_parent(file_id, versions_with_bad_parents,
+                 id_unused_versions, file_versions)
 
-    def _fix_text_parent(self, file_id, vf, versions_with_bad_parents,
-            full_text_versions, unused_versions):
+    def _fix_text_parent(self, file_id, versions_with_bad_parents,
+            unused_versions, all_versions):
         """Fix bad versionedfile entries in a single versioned file."""
         mutter('fixing text parent: %r (%d versions)', file_id,
                 len(versions_with_bad_parents))
-        mutter('(%d need to be full texts, %d are unused)',
-                len(full_text_versions), len(unused_versions))
-        new_vf = self.repo.weave_store.get_empty('temp:%s' % file_id,
-            self.transaction)
+        mutter('(%d are unused)', len(unused_versions))
+        new_file_id = 'temp:%s' % file_id
         new_parents = {}
-        for version in vf.versions():
+        needed_keys = set()
+        for version in all_versions:
             if version in unused_versions:
                 continue
             elif version in versions_with_bad_parents:
                 parents = versions_with_bad_parents[version][1]
             else:
-                parents = vf.get_parent_map([version])[version]
-            new_parents[version] = parents
-        if not len(new_parents):
-            # No used versions, remove the VF.
-            self.repo.weave_store.delete(file_id, self.transaction)
-            return
-        for version in TopoSorter(new_parents.items()).iter_topo_order():
-            lines = vf.get_lines(version)
-            parents = new_parents[version]
-            if parents and (parents[0] in full_text_versions):
-                # Force this record to be a fulltext, not a delta.
-                new_vf._add(version, lines, parents, False,
-                    None, None, None, False)
-            else:
-                new_vf.add_lines(version, parents, lines)
-        self.repo.weave_store.copy(new_vf, file_id, self.transaction)
-        self.repo.weave_store.delete('temp:%s' % file_id, self.transaction)
+                pmap = self.repo.texts.get_parent_map([(file_id, version)])
+                parents = [key[-1] for key in pmap[(file_id, version)]]
+            new_parents[(new_file_id, version)] = [
+                (new_file_id, parent) for parent in parents]
+            needed_keys.add((file_id, version))
+        def fix_parents(stream):
+            for record in stream:
+                bytes = record.get_bytes_as('fulltext')
+                new_key = (new_file_id, record.key[-1])
+                parents = new_parents[new_key]
+                yield FulltextContentFactory(new_key, parents, record.sha1, bytes)
+        stream = self.repo.texts.get_record_stream(needed_keys, 'topological', True)
+        self.repo._remove_file_id(new_file_id)
+        self.repo.texts.insert_record_stream(fix_parents(stream))
+        self.repo._remove_file_id(file_id)
+        if len(new_parents):
+            self.repo._move_file_id(new_file_id, file_id)
 
 
 class PackReconciler(RepoReconciler):

=== modified file 'bzrlib/remote.py'
--- bzrlib/remote.py	2008-05-08 04:33:38 +0000
+++ bzrlib/remote.py	2008-06-11 04:20:16 +0000
@@ -709,11 +709,6 @@
         self._ensure_real()
         return self._real_repository.get_revision(revision_id)
 
-    @property
-    def weave_store(self):
-        self._ensure_real()
-        return self._real_repository.weave_store
-
     def get_transaction(self):
         self._ensure_real()
         return self._real_repository.get_transaction()
@@ -768,20 +763,15 @@
         self._ensure_real()
         self._real_repository.create_bundle(target, base, fileobj, format)
 
-    @property
-    def control_weaves(self):
-        self._ensure_real()
-        return self._real_repository.control_weaves
-
     @needs_read_lock
     def get_ancestry(self, revision_id, topo_sorted=True):
         self._ensure_real()
         return self._real_repository.get_ancestry(revision_id, topo_sorted)
 
-    @needs_read_lock
-    def get_inventory_weave(self):
+    @property
+    def inventories(self):
         self._ensure_real()
-        return self._real_repository.get_inventory_weave()
+        return self._real_repository.inventories
 
     def fileids_altered_by_revision_ids(self, revision_ids):
         self._ensure_real()
@@ -1011,6 +1001,16 @@
         # TODO: Suggestion from john: using external tar is much faster than
         # python's tarfile library, but it may not work on windows.
 
+    @property
+    def inventories(self):
+        """Decorate the real repository for now.
+
+        In the long term a full blown network facility is needed to
+        avoid creating a real repository object locally.
+        """
+        self._ensure_real()
+        return self._real_repository.inventories
+
     @needs_write_lock
     def pack(self):
         """Compress the data within the repository.
@@ -1020,15 +1020,47 @@
         self._ensure_real()
         return self._real_repository.pack()
 
+    @property
+    def revisions(self):
+        """Decorate the real repository for now.
+
+        In the short term this should become a real object to intercept graph
+        lookups.
+
+        In the long term a full blown network facility is needed.
+        """
+        self._ensure_real()
+        return self._real_repository.revisions
+
     def set_make_working_trees(self, new_value):
         self._ensure_real()
         self._real_repository.set_make_working_trees(new_value)
 
+    @property
+    def signatures(self):
+        """Decorate the real repository for now.
+
+        In the long term a full blown network facility is needed to avoid
+        creating a real repository object locally.
+        """
+        self._ensure_real()
+        return self._real_repository.signatures
+
     @needs_write_lock
     def sign_revision(self, revision_id, gpg_strategy):
         self._ensure_real()
         return self._real_repository.sign_revision(revision_id, gpg_strategy)
 
+    @property
+    def texts(self):
+        """Decorate the real repository for now.
+
+        In the long term a full blown network facility is needed to avoid
+        creating a real repository object locally.
+        """
+        self._ensure_real()
+        return self._real_repository.texts
+
     @needs_read_lock
     def get_revisions(self, revision_ids):
         self._ensure_real()
@@ -1060,62 +1092,6 @@
         self._ensure_real()
         return self._real_repository.has_signature_for_revision_id(revision_id)
 
-    def get_data_stream_for_search(self, search):
-        medium = self._client._medium
-        if not medium._remote_is_at_least_1_2:
-            self._ensure_real()
-            return self._real_repository.get_data_stream_for_search(search)
-        REQUEST_NAME = 'Repository.stream_revisions_chunked'
-        path = self.bzrdir._path_for_remote_call(self._client)
-        body = self._serialise_search_recipe(search.get_recipe())
-        try:
-            result = self._client.call_with_body_bytes_expecting_body(
-                REQUEST_NAME, (path,), body)
-            response, protocol = result
-        except errors.UnknownSmartMethod:
-            # Server does not support this method, so fall back to VFS.
-            # Worse, we have to force a disconnection, because the server now
-            # doesn't realise it has a body on the wire to consume, so the
-            # only way to recover is to abandon the connection.
-            warning(
-                'Server is too old for streaming pull, reconnecting.  '
-                '(Upgrade the server to Bazaar 1.2 to avoid this)')
-            medium.disconnect()
-            # To avoid having to disconnect repeatedly, we keep track of the
-            # fact the server doesn't understand this remote method.
-            medium._remote_is_at_least_1_2 = False
-            self._ensure_real()
-            return self._real_repository.get_data_stream_for_search(search)
-
-        if response == ('ok',):
-            return self._deserialise_stream(protocol)
-        if response == ('NoSuchRevision', ):
-            # We cannot easily identify the revision that is missing in this
-            # situation without doing much more network IO. For now, bail.
-            raise NoSuchRevision(self, "unknown")
-        else:
-            raise errors.UnexpectedSmartServerResponse(response)
-
-    def _deserialise_stream(self, protocol):
-        stream = protocol.read_streamed_body()
-        container_parser = ContainerPushParser()
-        for bytes in stream:
-            container_parser.accept_bytes(bytes)
-            records = container_parser.read_pending_records()
-            for record_names, record_bytes in records:
-                if len(record_names) != 1:
-                    # These records should have only one name, and that name
-                    # should be a one-element tuple.
-                    raise errors.SmartProtocolError(
-                        'Repository data stream had invalid record name %r'
-                        % (record_names,))
-                name_tuple = record_names[0]
-                yield name_tuple, record_bytes
-
-    def insert_data_stream(self, stream):
-        self._ensure_real()
-        self._real_repository.insert_data_stream(stream)
-
     def item_keys_introduced_by(self, revision_ids, _files_pb=None):
         self._ensure_real()
         return self._real_repository.item_keys_introduced_by(revision_ids,

=== modified file 'bzrlib/repofmt/knitrepo.py'
--- bzrlib/repofmt/knitrepo.py	2008-05-12 04:20:02 +0000
+++ bzrlib/repofmt/knitrepo.py	2008-06-11 04:20:16 +0000
@@ -35,8 +35,8 @@
     xml6,
     xml7,
     )
-
 from bzrlib.decorators import needs_read_lock, needs_write_lock
+from bzrlib.knit import KnitVersionedFiles, _KndxIndex, _KnitKeyAccess
 from bzrlib.repository import (
     CommitBuilder,
     MetaDirRepository,
@@ -48,6 +48,7 @@
 from bzrlib.store.versioned import VersionedFileStore
 from bzrlib.trace import mutter, mutter_callsite
 from bzrlib.util import bencode
+from bzrlib.versionedfile import ConstantMapper, HashEscapedPrefixMapper
 
 
 class _KnitParentsProvider(object):
@@ -83,6 +84,34 @@
         return parent_map
 
 
+class _KnitsParentsProvider(object):
+
+    def __init__(self, knit, prefix=()):
+        """Create a parent provider for string keys mapped to tuple keys."""
+        self._knit = knit
+        self._prefix = prefix
+
+    def __repr__(self):
+        return 'KnitsParentsProvider(%r)' % self._knit
+
+    def get_parent_map(self, keys):
+        """See graph._StackedParentsProvider.get_parent_map"""
+        parent_map = self._knit.get_parent_map(
+            [self._prefix + (key,) for key in keys])
+        result = {}
+        for key, parents in parent_map.items():
+            revid = key[-1]
+            if len(parents) == 0:
+                parents = (_mod_revision.NULL_REVISION,)
+            else:
+                parents = tuple(parent[-1] for parent in parents)
+            result[revid] = parents
+        for revision_id in keys:
+            if revision_id == _mod_revision.NULL_REVISION:
+                result[revision_id] = ()
+        return result
+
+
 class KnitRepository(MetaDirRepository):
     """Knit format repository."""
 
@@ -93,31 +122,73 @@
     _commit_builder_class = None
     _serializer = None
 
-    def __init__(self, _format, a_bzrdir, control_files, _revision_store,
-        control_store, text_store, _commit_builder_class, _serializer):
-        MetaDirRepository.__init__(self, _format, a_bzrdir, control_files,
-            _revision_store, control_store, text_store)
+    def __init__(self, _format, a_bzrdir, control_files, _commit_builder_class,
+        _serializer):
+        MetaDirRepository.__init__(self, _format, a_bzrdir, control_files)
         self._commit_builder_class = _commit_builder_class
         self._serializer = _serializer
         self._reconcile_fixes_text_parents = True
-        control_store.get_scope = self.get_transaction
-        text_store.get_scope = self.get_transaction
-        _revision_store.get_scope = self.get_transaction
 
     def _warn_if_deprecated(self):
         # This class isn't deprecated
         pass
 
-    def _inventory_add_lines(self, inv_vf, revid, parents, lines, check_content):
-        return inv_vf.add_lines_with_ghosts(revid, parents, lines,
-            check_content=check_content)[0]
-
     @needs_read_lock
     def _all_revision_ids(self):
         """See Repository.all_revision_ids()."""
-        # Knits get the revision graph from the index of the revision knit, so
-        # it's always possible even if they're on an unlistable transport.
-        return self._revision_store.all_revision_ids(self.get_transaction())
+        return [key[0] for key in self.revisions.keys()]
+
+    def _activate_new_inventory(self):
+        """Put a replacement inventory.new into use as inventories."""
+        # Copy the content across
+        t = self._transport
+        t.copy('inventory.new.kndx', 'inventory.kndx')
+        try:
+            t.copy('inventory.new.knit', 'inventory.knit')
+        except errors.NoSuchFile:
+            # empty inventories knit
+            t.delete('inventory.knit')
+        # delete the temp inventory
+        t.delete('inventory.new.kndx')
+        try:
+            t.delete('inventory.new.knit')
+        except errors.NoSuchFile:
+            # empty inventories knit
+            pass
+        # Force index reload (sanity check)
+        self.inventories._index._reset_cache()
+        self.inventories.keys()
+
+    def _backup_inventory(self):
+        t = self._transport
+        t.copy('inventory.kndx', 'inventory.backup.kndx')
+        t.copy('inventory.knit', 'inventory.backup.knit')
+
+    def _move_file_id(self, from_id, to_id):
+        t = self._transport.clone('knits')
+        from_rel_url = self.texts._index._mapper.map((from_id, None))
+        to_rel_url = self.texts._index._mapper.map((to_id, None))
+        # We expect both files to always exist in this case.
+        for suffix in ('.knit', '.kndx'):
+            t.rename(from_rel_url + suffix, to_rel_url + suffix)
+
+    def _remove_file_id(self, file_id):
+        t = self._transport.clone('knits')
+        rel_url = self.texts._index._mapper.map((file_id, None))
+        for suffix in ('.kndx', '.knit'):
+            try:
+                t.delete(rel_url + suffix)
+            except errors.NoSuchFile:
+                pass
+
+    def _temp_inventories(self):
+        result = self._format._get_inventories(self._transport, self,
+            'inventory.new')
+        # Reconciling when the output has no revisions would result in no
+        # writes - but we want to ensure there is an inventory for
+        # compatibility with older clients that don't lazy-load.
+        result.get_parent_map([('A',)])
+        return result
 
     def fileid_involved_between_revs(self, from_revid, to_revid):
         """Find file_id(s) which are involved in the changes between revisions.
@@ -145,70 +216,11 @@
         return self._fileid_involved_by_set(changed)
 
     @needs_read_lock
-    def get_ancestry(self, revision_id, topo_sorted=True):
-        """Return a list of revision-ids integrated by a revision.
-        
-        This is topologically sorted, unless 'topo_sorted' is specified as
-        False.
-        """
-        if _mod_revision.is_null(revision_id):
-            return [None]
-        vf = self._get_revision_vf()
-        try:
-            return [None] + vf.get_ancestry(revision_id, topo_sorted)
-        except errors.RevisionNotPresent:
-            raise errors.NoSuchRevision(self, revision_id)
-
-    @symbol_versioning.deprecated_method(symbol_versioning.one_two)
-    def get_data_stream(self, revision_ids):
-        """See Repository.get_data_stream.
-        
-        Deprecated in 1.2 for get_data_stream_for_search.
-        """
-        search_result = self.revision_ids_to_search_result(set(revision_ids))
-        return self.get_data_stream_for_search(search_result)
-
-    def get_data_stream_for_search(self, search):
-        """See Repository.get_data_stream_for_search."""
-        item_keys = self.item_keys_introduced_by(search.get_keys())
-        for knit_kind, file_id, versions in item_keys:
-            name = (knit_kind,)
-            if knit_kind == 'file':
-                name = ('file', file_id)
-                knit = self.weave_store.get_weave_or_empty(
-                    file_id, self.get_transaction())
-            elif knit_kind == 'inventory':
-                knit = self.get_inventory_weave()
-            elif knit_kind == 'revisions':
-                knit = self._revision_store.get_revision_file(
-                    self.get_transaction())
-            elif knit_kind == 'signatures':
-                knit = self._revision_store.get_signature_file(
-                    self.get_transaction())
-            else:
-                raise AssertionError('Unknown knit kind %r' % (knit_kind,))
-            yield name, _get_stream_as_bytes(knit, versions)
-
-    @needs_read_lock
     def get_revision(self, revision_id):
         """Return the Revision object for a named revision"""
         revision_id = osutils.safe_revision_id(revision_id)
         return self.get_revision_reconcile(revision_id)
 
-    def _get_revision_vf(self):
-        """:return: a versioned file containing the revisions."""
-        vf = self._revision_store.get_revision_file(self.get_transaction())
-        return vf
-
-    def has_revisions(self, revision_ids):
-        """See Repository.has_revisions()."""
-        result = set()
-        transaction = self.get_transaction()
-        for revision_id in revision_ids:
-            if self._revision_store.has_revision_id(revision_id, transaction):
-                result.add(revision_id)
-        return result
-
     @needs_write_lock
     def reconcile(self, other=None, thorough=False):
         """Reconcile this repository."""
@@ -222,7 +234,7 @@
         return self._get_revision_vf().get_parents(revision_id)
 
     def _make_parents_provider(self):
-        return _KnitParentsProvider(self._get_revision_vf())
+        return _KnitsParentsProvider(self.revisions)
 
     def _find_inconsistent_revision_parents(self):
         """Find revisions with different parent lists in the revision object
@@ -233,14 +245,15 @@
         """
         if not self.is_locked():
             raise AssertionError()
-        vf = self._get_revision_vf()
-        for index_version in vf.versions():
-            parents_according_to_index = tuple(vf.get_parents_with_ghosts(
-                index_version))
-            revision = self.get_revision(index_version)
+        vf = self.revisions
+        for index_version in vf.keys():
+            parent_map = vf.get_parent_map([index_version])
+            parents_according_to_index = tuple(parent[-1] for parent in
+                parent_map[index_version])
+            revision = self.get_revision(index_version[-1])
             parents_according_to_revision = tuple(revision.parent_ids)
             if parents_according_to_index != parents_according_to_revision:
-                yield (index_version, parents_according_to_index,
+                yield (index_version[-1], parents_according_to_index,
                     parents_according_to_revision)
 
     def _check_for_inconsistent_revision_parents(self):
@@ -284,43 +297,37 @@
     # External lookups are not supported in this format.
     supports_external_lookups = False
 
-    def _get_control_store(self, repo_transport, control_files):
-        """Return the control store for this repository."""
-        return VersionedFileStore(
-            repo_transport,
-            prefixed=False,
-            file_mode=control_files._file_mode,
-            versionedfile_class=knit.make_file_knit,
-            versionedfile_kwargs={'factory':knit.KnitPlainFactory()},
-            )
-
-    def _get_revision_store(self, repo_transport, control_files):
-        """See RepositoryFormat._get_revision_store()."""
-        versioned_file_store = VersionedFileStore(
-            repo_transport,
-            file_mode=control_files._file_mode,
-            prefixed=False,
-            precious=True,
-            versionedfile_class=knit.make_file_knit,
-            versionedfile_kwargs={'delta':False,
-                                  'factory':knit.KnitPlainFactory(),
-                                 },
-            escaped=False,
-            )
-        return KnitRevisionStore(versioned_file_store)
-
-    def _get_text_store(self, transport, control_files):
-        """See RepositoryFormat._get_text_store()."""
-        return self._get_versioned_file_store('knits',
-                                  transport,
-                                  control_files,
-                                  versionedfile_class=knit.make_file_knit,
-                                  versionedfile_kwargs={
-                                      'create_parent_dir':True,
-                                      'delay_create':True,
-                                      'dir_mode':control_files._dir_mode,
-                                  },
-                                  escaped=True)
+    def _get_inventories(self, repo_transport, repo, name='inventory'):
+        mapper = ConstantMapper(name)
+        index = _KndxIndex(repo_transport, mapper, repo.get_transaction,
+            repo.is_write_locked, repo.is_locked)
+        access = _KnitKeyAccess(repo_transport, mapper)
+        return KnitVersionedFiles(index, access, annotated=False)
+
+    def _get_revisions(self, repo_transport, repo):
+        mapper = ConstantMapper('revisions')
+        index = _KndxIndex(repo_transport, mapper, repo.get_transaction,
+            repo.is_write_locked, repo.is_locked)
+        access = _KnitKeyAccess(repo_transport, mapper)
+        return KnitVersionedFiles(index, access, max_delta_chain=0,
+            annotated=False)
+
+    def _get_signatures(self, repo_transport, repo):
+        mapper = ConstantMapper('signatures')
+        index = _KndxIndex(repo_transport, mapper, repo.get_transaction,
+            repo.is_write_locked, repo.is_locked)
+        access = _KnitKeyAccess(repo_transport, mapper)
+        return KnitVersionedFiles(index, access, max_delta_chain=0,
+            annotated=False)
+
+    def _get_texts(self, repo_transport, repo):
+        mapper = HashEscapedPrefixMapper()
+        base_transport = repo_transport.clone('knits')
+        index = _KndxIndex(base_transport, mapper, repo.get_transaction,
+            repo.is_write_locked, repo.is_locked)
+        access = _KnitKeyAccess(base_transport, mapper)
+        return KnitVersionedFiles(index, access, max_delta_chain=200,
+            annotated=True)
 
     def initialize(self, a_bzrdir, shared=False):
         """Create a knit format 1 repository.
@@ -339,16 +346,16 @@
         repo_transport = a_bzrdir.get_repository_transport(None)
         control_files = lockable_files.LockableFiles(repo_transport,
                                 'lock', lockdir.LockDir)
-        control_store = self._get_control_store(repo_transport, control_files)
         transaction = transactions.WriteTransaction()
-        # trigger a write of the inventory store.
-        control_store.get_weave_or_empty('inventory', transaction)
-        _revision_store = self._get_revision_store(repo_transport, control_files)
+        result = self.open(a_bzrdir=a_bzrdir, _found=True)
+        result.lock_write()
         # the revision id here is irrelevant: it will not be stored, and cannot
-        # already exist.
-        _revision_store.has_revision_id('A', transaction)
-        _revision_store.get_signature_file(transaction)
-        return self.open(a_bzrdir=a_bzrdir, _found=True)
+        # already exist, we do this to create files on disk for older clients.
+        result.inventories.get_parent_map([('A',)])
+        result.revisions.get_parent_map([('A',)])
+        result.signatures.get_parent_map([('A',)])
+        result.unlock()
+        return result
 
     def open(self, a_bzrdir, _found=False, _override_transport=None):
         """See RepositoryFormat.open().
@@ -365,17 +372,17 @@
             repo_transport = a_bzrdir.get_repository_transport(None)
         control_files = lockable_files.LockableFiles(repo_transport,
                                 'lock', lockdir.LockDir)
-        text_store = self._get_text_store(repo_transport, control_files)
-        control_store = self._get_control_store(repo_transport, control_files)
-        _revision_store = self._get_revision_store(repo_transport, control_files)
-        return self.repository_class(_format=self,
+        repo = self.repository_class(_format=self,
                               a_bzrdir=a_bzrdir,
                               control_files=control_files,
-                              _revision_store=_revision_store,
-                              control_store=control_store,
-                              text_store=text_store,
                               _commit_builder_class=self._commit_builder_class,
                               _serializer=self._serializer)
+        repo.revisions = self._get_revisions(repo_transport, repo)
+        repo.signatures = self._get_signatures(repo_transport, repo)
+        repo.inventories = self._get_inventories(repo_transport, repo)
+        repo.texts = self._get_texts(repo_transport, repo)
+        repo._transport = repo_transport
+        return repo
 
 
 class RepositoryFormatKnit1(RepositoryFormatKnit):
@@ -502,26 +509,3 @@
     def get_format_description(self):
         """See RepositoryFormat.get_format_description()."""
         return "Knit repository format 4"
-
-
-def _get_stream_as_bytes(knit, required_versions):
-    """Generate a serialised data stream.
-
-    The format is a bencoding of a list.  The first element of the list is a
-    string of the format signature, then each subsequent element is a list
-    corresponding to a record.  Those lists contain:
-
-      * a version id
-      * a list of options
-      * a list of parents
-      * the bytes
-
-    :returns: a bencoded list.
-    """
-    knit_stream = knit.get_data_stream(required_versions)
-    format_signature, data_list, callable = knit_stream
-    data = []
-    data.append(format_signature)
-    for version, options, length, parents in data_list:
-        data.append([version, options, parents, callable(length)])
-    return bencode.bencode(data)

=== modified file 'bzrlib/repofmt/pack_repo.py'
--- bzrlib/repofmt/pack_repo.py	2008-05-12 04:20:02 +0000
+++ bzrlib/repofmt/pack_repo.py	2008-06-11 04:20:16 +0000
@@ -34,8 +34,13 @@
     CombinedGraphIndex,
     GraphIndexPrefixAdapter,
     )
-from bzrlib.knit import KnitGraphIndex, _PackAccess, _KnitData
-from bzrlib.osutils import rand_chars
+from bzrlib.knit import (
+    KnitPlainFactory,
+    KnitVersionedFiles,
+    _KnitGraphIndex,
+    _DirectPackAccess,
+    )
+from bzrlib.osutils import rand_chars, split_lines
 from bzrlib.pack import ContainerWriter
 from bzrlib.store import revision
 from bzrlib import tsort
@@ -43,7 +48,6 @@
 from bzrlib import (
     bzrdir,
     errors,
-    knit,
     lockable_files,
     lockdir,
     osutils,
@@ -64,7 +68,6 @@
     RootCommitBuilder,
     )
 import bzrlib.revision as _mod_revision
-from bzrlib.store.revision.knit import KnitRevisionStore
 from bzrlib.store.versioned import VersionedFileStore
 from bzrlib.trace import (
     mutter,
@@ -90,11 +93,6 @@
         self._file_graph = graph.Graph(
             repository._pack_collection.text_index.combined_index)
 
-    def _add_text_to_weave(self, file_id, new_lines, parents, nostore_sha):
-        return self.repository._pack_collection._add_text_to_weave(file_id,
-            self._new_revision_id, new_lines, parents, nostore_sha,
-            self.random_revid)
-
     def _heads(self, file_id, revision_ids):
         keys = [(file_id, revision_id) for revision_id in revision_ids]
         return set([key[1] for key in self._file_graph.heads(keys)])
@@ -116,11 +114,6 @@
         self._file_graph = graph.Graph(
             repository._pack_collection.text_index.combined_index)
 
-    def _add_text_to_weave(self, file_id, new_lines, parents, nostore_sha):
-        return self.repository._pack_collection._add_text_to_weave(file_id,
-            self._new_revision_id, new_lines, parents, nostore_sha,
-            self.random_revid)
-
     def _heads(self, file_id, revision_ids):
         keys = [(file_id, revision_id) for revision_id in revision_ids]
         return set([key[1] for key in self._file_graph.heads(keys)])
@@ -457,7 +450,8 @@
         """Create an AggregateIndex."""
         self.index_to_pack = {}
         self.combined_index = CombinedGraphIndex([])
-        self.knit_access = _PackAccess(self.index_to_pack)
+        self.data_access = _DirectPackAccess(self.index_to_pack)
+        self.add_callback = None
 
     def replace_indices(self, index_to_pack, indices):
         """Replace the current mappings with fresh ones.
@@ -509,12 +503,12 @@
         # allow writing: queue writes to a new index
         self.add_index(index, pack)
         # Updates the index to packs mapping as a side effect,
-        self.knit_access.set_writer(pack._writer, index, pack.access_tuple())
+        self.data_access.set_writer(pack._writer, index, pack.access_tuple())
         self.add_callback = index.add_nodes
 
     def clear(self):
         """Reset all the aggregate data to nothing."""
-        self.knit_access.set_writer(None, None, (None, None))
+        self.data_access.set_writer(None, None, (None, None))
         self.index_to_pack.clear()
         del self.combined_index._indices[:]
         self.add_callback = None
@@ -530,7 +524,7 @@
         if (self.add_callback is not None and
             getattr(index, 'add_nodes', None) == self.add_callback):
             self.add_callback = None
-            self.knit_access.set_writer(None, None, (None, None))
+            self.data_access.set_writer(None, None, (None, None))
 
 
 class Packer(object):
@@ -589,6 +583,8 @@
                 return None
             else:
                 self.revision_ids = frozenset(self.revision_ids)
+                self.revision_keys = frozenset((revid,) for revid in
+                    self.revision_ids)
         if pb is None:
             self.pb = ui.ui_factory.nested_progress_bar()
         else:
@@ -763,7 +759,7 @@
 
     def _do_copy_nodes(self, nodes, index_map, writer, write_index, pb):
         # for record verification
-        knit_data = _KnitData(None)
+        knit = KnitVersionedFiles(None, None)
         # plan a readv on each source pack:
         # group by pack
         nodes = sorted(nodes)
@@ -795,7 +791,7 @@
                 izip(reader.iter_records(), pack_readv_requests):
                 raw_data = read_func(None)
                 # check the header only
-                df, _ = knit_data._parse_record_header(key[-1], raw_data)
+                df, _ = knit._parse_record_header(key, raw_data)
                 df.close()
                 pos, size = writer.add_bytes_record(raw_data, names)
                 write_index.add_node(key, eol_flag + "%d %d" % (pos, size))
@@ -824,10 +820,10 @@
     def _do_copy_nodes_graph(self, index_map, writer, write_index,
         output_lines, pb, readv_group_iter, total_items):
         # for record verification
-        knit_data = _KnitData(None)
+        knit = KnitVersionedFiles(None, None)
         # for line extraction when requested (inventories only)
         if output_lines:
-            factory = knit.KnitPlainFactory()
+            factory = KnitPlainFactory()
         record_index = 0
         pb.update("Copied record", record_index, total_items)
         for index, readv_vector, node_vector in readv_group_iter:
@@ -837,19 +833,18 @@
             for (names, read_func), (key, eol_flag, references) in \
                 izip(reader.iter_records(), node_vector):
                 raw_data = read_func(None)
-                version_id = key[-1]
                 if output_lines:
                     # read the entire thing
-                    content, _ = knit_data._parse_record(version_id, raw_data)
+                    content, _ = knit._parse_record(key[-1], raw_data)
                     if len(references[-1]) == 0:
                         line_iterator = factory.get_fulltext_content(content)
                     else:
                         line_iterator = factory.get_linedelta_content(content)
                     for line in line_iterator:
-                        yield line, version_id
+                        yield line, key
                 else:
                     # check the header only
-                    df, _ = knit_data._parse_record_header(version_id, raw_data)
+                    df, _ = knit._parse_record_header(key, raw_data)
                     df.close()
                 pos, size = writer.add_bytes_record(raw_data, names)
                 write_index.add_node(key, eol_flag + "%d %d" % (pos, size), references)
@@ -911,7 +906,7 @@
         """Use up the inv_lines generator and setup a text key filter."""
         repo = self._pack_collection.repo
         fileid_revisions = repo._find_file_ids_from_xml_inventory_lines(
-            inv_lines, self.revision_ids)
+            inv_lines, self.revision_keys)
         text_filter = []
         for fileid, file_revids in fileid_revisions.iteritems():
             text_filter.extend([(fileid, file_revid) for file_revid in file_revids])
@@ -1064,16 +1059,15 @@
             self.new_pack.text_index,
             ('blank', ), 1,
             add_nodes_callback=self.new_pack.text_index.add_nodes)
-        knit_index = KnitGraphIndex(file_id_index,
-            add_callback=file_id_index.add_nodes,
-            deltas=True, parents=True)
-        output_knit = knit.KnitVersionedFile('reconcile-texts',
-            self._pack_collection.transport,
-            index=knit_index,
-            access_method=_PackAccess(
-                {self.new_pack.text_index:self.new_pack.access_tuple()},
-                (self.new_pack._writer, self.new_pack.text_index)),
-            factory=knit.KnitPlainFactory())
+        data_access = _DirectPackAccess(
+                {self.new_pack.text_index:self.new_pack.access_tuple()})
+        data_access.set_writer(self.new_pack._writer, self.new_pack.text_index,
+            self.new_pack.access_tuple())
+        output_texts = KnitVersionedFiles(
+            _KnitGraphIndex(self.new_pack.text_index,
+                add_callback=self.new_pack.text_index.add_nodes,
+                deltas=True, parents=True, is_locked=repo.is_locked),
+            data_access=data_access, max_delta_chain=200)
         for key, parent_keys in bad_texts:
             # We refer to the new pack to delta data being output.
             # A possible improvement would be to catch errors on short reads
@@ -1086,17 +1080,10 @@
                     raise errors.BzrError('Mismatched key parent %r:%r' %
                         (key, parent_keys))
                 parents.append(parent_key[1])
-            source_weave = repo.weave_store.get_weave(key[0], transaction)
-            text_lines = source_weave.get_lines(key[1])
-            # adapt the 'knit' to the current file_id.
-            file_id_index = GraphIndexPrefixAdapter(
-                self.new_pack.text_index,
-                (key[0], ), 1,
-                add_nodes_callback=self.new_pack.text_index.add_nodes)
-            knit_index._graph_index = file_id_index
-            knit_index._add_callback = file_id_index.add_nodes
-            output_knit.add_lines_with_ghosts(
-                key[1], parents, text_lines, random_id=True, check_content=False)
+            text_lines = split_lines(repo.texts.get_record_stream(
+                [key], 'unordered', True).next().get_bytes_as('fulltext'))
+            output_texts.add_lines(key, parent_keys, text_lines,
+                random_id=True, check_content=False)
         # 5) check that nothing inserted has a reference outside the keyspace.
         missing_text_keys = self.new_pack._external_compression_parents_of_texts()
         if missing_text_keys:
@@ -1166,18 +1153,6 @@
         self.text_index.add_index(pack.text_index, pack)
         self.signature_index.add_index(pack.signature_index, pack)
         
-    def _add_text_to_weave(self, file_id, revision_id, new_lines, parents,
-        nostore_sha, random_revid):
-        file_id_index = GraphIndexPrefixAdapter(
-            self.text_index.combined_index,
-            (file_id, ), 1,
-            add_nodes_callback=self.text_index.add_callback)
-        self.repo._text_knit._index._graph_index = file_id_index
-        self.repo._text_knit._index._add_callback = file_id_index.add_nodes
-        return self.repo._text_knit.add_lines_with_ghosts(
-            revision_id, parents, new_lines, nostore_sha=nostore_sha,
-            random_id=random_revid, check_content=False)[0:2]
-
     def all_packs(self):
         """Return a list of all the Pack objects this repository has.
 
@@ -1641,19 +1616,10 @@
         self.signature_index.add_writable_index(self._new_pack.signature_index,
             self._new_pack)
 
-        # reused revision and signature knits may need updating
-        #
-        # "Hysterical raisins. client code in bzrlib grabs those knits outside
-        # of write groups and then mutates it inside the write group."
-        if self.repo._revision_knit is not None:
-            self.repo._revision_knit._index._add_callback = \
-                self.revision_index.add_callback
-        if self.repo._signature_knit is not None:
-            self.repo._signature_knit._index._add_callback = \
-                self.signature_index.add_callback
-        # create a reused knit object for text addition in commit.
-        self.repo._text_knit = self.repo.weave_store.get_weave_or_empty(
-            'all-texts', None)
+        self.repo.inventories._index._add_callback = self.inventory_index.add_callback
+        self.repo.revisions._index._add_callback = self.revision_index.add_callback
+        self.repo.signatures._index._add_callback = self.signature_index.add_callback
+        self.repo.texts._index._add_callback = self.text_index.add_callback
 
     def _abort_write_group(self):
         # FIXME: just drop the transient index.
@@ -1681,167 +1647,42 @@
         self.repo._text_knit = None
 
 
-class KnitPackRevisionStore(KnitRevisionStore):
-    """An object to adapt access from RevisionStore's to use KnitPacks.
-
-    This class works by replacing the original RevisionStore.
-    We need to do this because the KnitPackRevisionStore is less
-    isolated in its layering - it uses services from the repo.
-    """
-
-    def __init__(self, repo, transport, revisionstore):
-        """Create a KnitPackRevisionStore on repo with revisionstore.
-
-        This will store its state in the Repository, use the
-        indices to provide a KnitGraphIndex,
-        and at the end of transactions write new indices.
-        """
-        KnitRevisionStore.__init__(self, revisionstore.versioned_file_store)
-        self.repo = repo
-        self._serializer = revisionstore._serializer
-        self.transport = transport
-
-    def get_revision_file(self, transaction):
-        """Get the revision versioned file object."""
-        if getattr(self.repo, '_revision_knit', None) is not None:
-            return self.repo._revision_knit
-        self.repo._pack_collection.ensure_loaded()
-        add_callback = self.repo._pack_collection.revision_index.add_callback
-        # setup knit specific objects
-        knit_index = KnitGraphIndex(
-            self.repo._pack_collection.revision_index.combined_index,
-            add_callback=add_callback)
-        self.repo._revision_knit = knit.KnitVersionedFile(
-            'revisions', self.transport.clone('..'),
-            self.repo.control_files._file_mode,
-            create=False,
-            index=knit_index, delta=False, factory=knit.KnitPlainFactory(),
-            access_method=self.repo._pack_collection.revision_index.knit_access)
-        return self.repo._revision_knit
-
-    def get_signature_file(self, transaction):
-        """Get the signature versioned file object."""
-        if getattr(self.repo, '_signature_knit', None) is not None:
-            return self.repo._signature_knit
-        self.repo._pack_collection.ensure_loaded()
-        add_callback = self.repo._pack_collection.signature_index.add_callback
-        # setup knit specific objects
-        knit_index = KnitGraphIndex(
-            self.repo._pack_collection.signature_index.combined_index,
-            add_callback=add_callback, parents=False)
-        self.repo._signature_knit = knit.KnitVersionedFile(
-            'signatures', self.transport.clone('..'),
-            self.repo.control_files._file_mode,
-            create=False,
-            index=knit_index, delta=False, factory=knit.KnitPlainFactory(),
-            access_method=self.repo._pack_collection.signature_index.knit_access)
-        return self.repo._signature_knit
-
-
-class KnitPackTextStore(VersionedFileStore):
-    """Presents a TextStore abstraction on top of packs.
-
-    This class works by replacing the original VersionedFileStore.
-    We need to do this because the KnitPackRevisionStore is less
-    isolated in its layering - it uses services from the repo and shares them
-    with all the data written in a single write group.
-    """
-
-    def __init__(self, repo, transport, weavestore):
-        """Create a KnitPackTextStore on repo with weavestore.
-
-        This will store its state in the Repository, use the
-        indices FileNames to provide a KnitGraphIndex,
-        and at the end of transactions write new indices.
-        """
-        # don't call base class constructor - it's not suitable.
-        # no transient data stored in the transaction
-        # cache.
-        self._precious = False
-        self.repo = repo
-        self.transport = transport
-        self.weavestore = weavestore
-        # XXX for check() which isn't updated yet
-        self._transport = weavestore._transport
-
-    def get_weave_or_empty(self, file_id, transaction):
-        """Get a 'Knit' backed by the .tix indices.
-
-        The transaction parameter is ignored.
-        """
-        self.repo._pack_collection.ensure_loaded()
-        add_callback = self.repo._pack_collection.text_index.add_callback
-        # setup knit specific objects
-        file_id_index = GraphIndexPrefixAdapter(
-            self.repo._pack_collection.text_index.combined_index,
-            (file_id, ), 1, add_nodes_callback=add_callback)
-        knit_index = KnitGraphIndex(file_id_index,
-            add_callback=file_id_index.add_nodes,
-            deltas=True, parents=True)
-        return knit.KnitVersionedFile('text:' + file_id,
-            self.transport.clone('..'),
-            None,
-            index=knit_index,
-            access_method=self.repo._pack_collection.text_index.knit_access,
-            factory=knit.KnitPlainFactory())
-
-    get_weave = get_weave_or_empty
-
-    def __iter__(self):
-        """Generate a list of the fileids inserted, for use by check."""
-        self.repo._pack_collection.ensure_loaded()
-        ids = set()
-        for index, key, value, refs in \
-            self.repo._pack_collection.text_index.combined_index.iter_all_entries():
-            ids.add(key[0])
-        return iter(ids)
-
-
-class InventoryKnitThunk(object):
-    """An object to manage thunking get_inventory_weave to pack based knits."""
-
-    def __init__(self, repo, transport):
-        """Create an InventoryKnitThunk for repo at transport.
-
-        This will store its state in the Repository, use the
-        indices FileNames to provide a KnitGraphIndex,
-        and at the end of transactions write a new index..
-        """
-        self.repo = repo
-        self.transport = transport
-
-    def get_weave(self):
-        """Get a 'Knit' that contains inventory data."""
-        self.repo._pack_collection.ensure_loaded()
-        add_callback = self.repo._pack_collection.inventory_index.add_callback
-        # setup knit specific objects
-        knit_index = KnitGraphIndex(
-            self.repo._pack_collection.inventory_index.combined_index,
-            add_callback=add_callback, deltas=True, parents=True)
-        return knit.KnitVersionedFile(
-            'inventory', self.transport.clone('..'),
-            self.repo.control_files._file_mode,
-            create=False,
-            index=knit_index, delta=True, factory=knit.KnitPlainFactory(),
-            access_method=self.repo._pack_collection.inventory_index.knit_access)
-
-
 class KnitPackRepository(KnitRepository):
     """Experimental graph-knit using repository."""
 
-    def __init__(self, _format, a_bzrdir, control_files, _revision_store,
-        control_store, text_store, _commit_builder_class, _serializer):
-        KnitRepository.__init__(self, _format, a_bzrdir, control_files,
-            _revision_store, control_store, text_store, _commit_builder_class,
-            _serializer)
+    def __init__(self, _format, a_bzrdir, control_files, _commit_builder_class,
+        _serializer):
         index_transport = control_files._transport.clone('indices')
         self._pack_collection = RepositoryPackCollection(self, control_files._transport,
             index_transport,
             control_files._transport.clone('upload'),
             control_files._transport.clone('packs'))
-        self._revision_store = KnitPackRevisionStore(self, index_transport, self._revision_store)
-        self.weave_store = KnitPackTextStore(self, index_transport, self.weave_store)
-        self._inv_thunk = InventoryKnitThunk(self, index_transport)
+        KnitRepository.__init__(self, _format, a_bzrdir, control_files,
+            _commit_builder_class, _serializer)
+        self.inventories = KnitVersionedFiles(
+            _KnitGraphIndex(self._pack_collection.inventory_index.combined_index,
+                add_callback=self._pack_collection.inventory_index.add_callback,
+                deltas=True, parents=True, is_locked=self.is_locked),
+            data_access=self._pack_collection.inventory_index.data_access,
+            max_delta_chain=200)
+        self.revisions = KnitVersionedFiles(
+            _KnitGraphIndex(self._pack_collection.revision_index.combined_index,
+                add_callback=self._pack_collection.revision_index.add_callback,
+                deltas=False, parents=True, is_locked=self.is_locked),
+            data_access=self._pack_collection.revision_index.data_access,
+            max_delta_chain=0)
+        self.signatures = KnitVersionedFiles(
+            _KnitGraphIndex(self._pack_collection.signature_index.combined_index,
+                add_callback=self._pack_collection.signature_index.add_callback,
+                deltas=False, parents=False, is_locked=self.is_locked),
+            data_access=self._pack_collection.signature_index.data_access,
+            max_delta_chain=0)
+        self.texts = KnitVersionedFiles(
+            _KnitGraphIndex(self._pack_collection.text_index.combined_index,
+                add_callback=self._pack_collection.text_index.add_callback,
+                deltas=True, parents=True, is_locked=self.is_locked),
+            data_access=self._pack_collection.text_index.data_access,
+            max_delta_chain=200)
         # True when the repository object is 'write locked' (as opposed to the
         # physical lock only taken out around changes to the pack-names list.) 
         # Another way to represent this would be a decorator around the control
@@ -1957,9 +1798,6 @@
     def _commit_write_group(self):
         return self._pack_collection._commit_write_group()
 
-    def get_inventory_weave(self):
-        return self._inv_thunk.get_weave()
-
     def get_transaction(self):
         if self._write_lock_count:
             return self._transaction
@@ -2059,44 +1897,6 @@
     # External references are not supported in pack repositories yet.
     supports_external_lookups = False
 
-    def _get_control_store(self, repo_transport, control_files):
-        """Return the control store for this repository."""
-        return VersionedFileStore(
-            repo_transport,
-            prefixed=False,
-            file_mode=control_files._file_mode,
-            versionedfile_class=knit.make_file_knit,
-            versionedfile_kwargs={'factory': knit.KnitPlainFactory()},
-            )
-
-    def _get_revision_store(self, repo_transport, control_files):
-        """See RepositoryFormat._get_revision_store()."""
-        versioned_file_store = VersionedFileStore(
-            repo_transport,
-            file_mode=control_files._file_mode,
-            prefixed=False,
-            precious=True,
-            versionedfile_class=knit.make_file_knit,
-            versionedfile_kwargs={'delta': False,
-                                  'factory': knit.KnitPlainFactory(),
-                                 },
-            escaped=False,
-            )
-        return KnitRevisionStore(versioned_file_store)
-
-    def _get_text_store(self, transport, control_files):
-        """See RepositoryFormat._get_text_store()."""
-        return self._get_versioned_file_store('knits',
-                                  transport,
-                                  control_files,
-                                  versionedfile_class=knit.make_file_knit,
-                                  versionedfile_kwargs={
-                                      'create_parent_dir': True,
-                                      'delay_create': True,
-                                      'dir_mode': control_files._dir_mode,
-                                  },
-                                  escaped=True)
-
     def initialize(self, a_bzrdir, shared=False):
         """Create a pack based repository.
 
@@ -2129,15 +1929,9 @@
             repo_transport = a_bzrdir.get_repository_transport(None)
         control_files = lockable_files.LockableFiles(repo_transport,
                                 'lock', lockdir.LockDir)
-        text_store = self._get_text_store(repo_transport, control_files)
-        control_store = self._get_control_store(repo_transport, control_files)
-        _revision_store = self._get_revision_store(repo_transport, control_files)
         return self.repository_class(_format=self,
                               a_bzrdir=a_bzrdir,
                               control_files=control_files,
-                              _revision_store=_revision_store,
-                              control_store=control_store,
-                              text_store=text_store,
                               _commit_builder_class=self._commit_builder_class,
                               _serializer=self._serializer)
 

=== modified file 'bzrlib/repofmt/weaverepo.py'
--- bzrlib/repofmt/weaverepo.py	2008-05-12 02:40:40 +0000
+++ bzrlib/repofmt/weaverepo.py	2008-06-11 04:20:16 +0000
@@ -20,7 +20,9 @@
 ghosts.
 """
 
-from StringIO import StringIO
+import os
+from cStringIO import StringIO
+import urllib
 
 from bzrlib import (
     bzrdir,
@@ -30,6 +32,7 @@
     lockdir,
     osutils,
     revision as _mod_revision,
+    versionedfile,
     weave,
     weavefile,
     xml5,
@@ -44,6 +47,12 @@
     )
 from bzrlib.store.text import TextStore
 from bzrlib.trace import mutter
+from bzrlib.tuned_gzip import GzipFile, bytes_to_gzip
+from bzrlib.versionedfile import (
+    AbsentContentFactory,
+    FulltextContentFactory,
+    VersionedFiles,
+    )
 
 
 class AllInOneRepository(Repository):
@@ -51,7 +60,7 @@
 
     _serializer = xml5.serializer_v5
 
-    def __init__(self, _format, a_bzrdir, _revision_store, control_store, text_store):
+    def __init__(self, _format, a_bzrdir):
         # we reuse one control files instance.
         dir_mode = a_bzrdir._control_files._dir_mode
         file_mode = a_bzrdir._control_files._file_mode
@@ -69,23 +78,20 @@
             return store
 
         # not broken out yet because the controlweaves|inventory_store
-        # and text_store | weave_store bits are still different.
+        # and texts bits are still different.
         if isinstance(_format, RepositoryFormat4):
             # cannot remove these - there is still no consistent api 
             # which allows access to this old info.
             self.inventory_store = get_store('inventory-store')
-            text_store = get_store('text-store')
-        super(AllInOneRepository, self).__init__(_format, a_bzrdir, a_bzrdir._control_files, _revision_store, control_store, text_store)
-        if control_store is not None:
-            control_store.get_scope = self.get_transaction
-        text_store.get_scope = self.get_transaction
+            self._text_store = get_store('text-store')
+        super(AllInOneRepository, self).__init__(_format, a_bzrdir, a_bzrdir._control_files)
 
     @needs_read_lock
     def _all_possible_ids(self):
         """Return all the possible revisions that we could find."""
         if 'evil' in debug.debug_flags:
             mutter_callsite(3, "_all_possible_ids scales with size of history.")
-        return self.get_inventory_weave().versions()
+        return [key[-1] for key in self.inventories.keys()]
 
     @needs_read_lock
     def _all_revision_ids(self):
@@ -95,36 +101,31 @@
         present: for weaves ghosts may lead to a lack of correctness until
         the reweave updates the parents list.
         """
-        if self._revision_store.text_store.listable():
-            return self._revision_store.all_revision_ids(self.get_transaction())
-        result = self._all_possible_ids()
-        # TODO: jam 20070210 Ensure that _all_possible_ids returns non-unicode
-        #       ids. (It should, since _revision_store's API should change to
-        #       return utf8 revision_ids)
-        return self._eliminate_revisions_not_present(result)
-
-    def _check_revision_parents(self, revision, inventory):
-        """Private to Repository and Fetch.
-        
-        This checks the parentage of revision in an inventory weave for 
-        consistency and is only applicable to inventory-weave-for-ancestry
-        using repository formats & fetchers.
-        """
-        weave_parents = inventory.get_parent_map(
-            [revision.revision_id])[revision.revision_id]
-        parent_map = inventory.get_parent_map(revision.parent_ids)
-        for parent_id in revision.parent_ids:
-            if parent_id in parent_map:
-                # this parent must not be a ghost.
-                if not parent_id in weave_parents:
-                    # but it is a ghost
-                    raise errors.CorruptRepository(self)
+        return [key[-1] for key in self.revisions.keys()]
+
+    def _activate_new_inventory(self):
+        """Put a replacement inventory.new into use as inventories."""
+        # Copy the content across
+        t = self.bzrdir._control_files._transport
+        t.copy('inventory.new.weave', 'inventory.weave')
+        # delete the temp inventory
+        t.delete('inventory.new.weave')
+        # Check we can parse the new weave properly as a sanity check
+        self.inventories.keys()
+
+    def _backup_inventory(self):
+        t = self.bzrdir._control_files._transport
+        t.copy('inventory.weave', 'inventory.backup.weave')
+
+    def _temp_inventories(self):
+        t = self.bzrdir._control_files._transport
+        return self._format._get_inventories(t, self, 'inventory.new')
 
     def get_commit_builder(self, branch, parents, config, timestamp=None,
                            timezone=None, committer=None, revprops=None,
                            revision_id=None):
         self._check_ascii_revisionid(revision_id, self.get_commit_builder)
-        result = WeaveCommitBuilder(self, parents, config, timestamp, timezone,
+        result = CommitBuilder(self, parents, config, timestamp, timezone,
                               committer, revprops, revision_id)
         self.start_write_group()
         return result
@@ -132,26 +133,18 @@
     @needs_read_lock
     def get_revisions(self, revision_ids):
         revs = self._get_revisions(revision_ids)
-        # weave corruption can lead to absent revision markers that should be
-        # present.
-        # the following test is reasonably cheap (it needs a single weave read)
-        # and the weave is cached in read transactions. In write transactions
-        # it is not cached but typically we only read a small number of
-        # revisions. For knits when they are introduced we will probably want
-        # to ensure that caching write transactions are in use.
-        inv = self.get_inventory_weave()
-        for rev in revs:
-            self._check_revision_parents(rev, inv)
         return revs
 
-    def has_revisions(self, revision_ids):
-        """See Repository.has_revisions()."""
-        result = set()
-        transaction = self.get_transaction()
-        for revision_id in revision_ids:
-            if self._revision_store.has_revision_id(revision_id, transaction):
-                result.add(revision_id)
-        return result
+    def _inventory_add_lines(self, revision_id, parents, lines,
+        check_content=True):
+        """Store lines in inv_vf and return the sha1 of the inventory."""
+        present_parents = self.get_graph().get_parent_map(parents)
+        final_parents = []
+        for parent in parents:
+            if parent in present_parents:
+                final_parents.append((parent,))
+        return self.inventories.add_lines((revision_id,), final_parents, lines,
+            check_content=check_content)[0]
 
     @needs_read_lock
     def is_shared(self):
@@ -190,7 +183,7 @@
         """Return all the possible revisions that we could find."""
         if 'evil' in debug.debug_flags:
             mutter_callsite(3, "_all_possible_ids scales with size of history.")
-        return self.get_inventory_weave().versions()
+        return [key[-1] for key in self.inventories.keys()]
 
     @needs_read_lock
     def _all_revision_ids(self):
@@ -200,36 +193,31 @@
         present: for weaves ghosts may lead to a lack of correctness until
         the reweave updates the parents list.
         """
-        if self._revision_store.text_store.listable():
-            return self._revision_store.all_revision_ids(self.get_transaction())
-        result = self._all_possible_ids()
-        # TODO: jam 20070210 Ensure that _all_possible_ids returns non-unicode
-        #       ids. (It should, since _revision_store's API should change to
-        #       return utf8 revision_ids)
-        return self._eliminate_revisions_not_present(result)
-
-    def _check_revision_parents(self, revision, inventory):
-        """Private to Repository and Fetch.
-        
-        This checks the parentage of revision in an inventory weave for 
-        consistency and is only applicable to inventory-weave-for-ancestry
-        using repository formats & fetchers.
-        """
-        weave_parents = inventory.get_parent_map(
-            [revision.revision_id])[revision.revision_id]
-        parent_map = inventory.get_parent_map(revision.parent_ids)
-        for parent_id in revision.parent_ids:
-            if parent_id in parent_map:
-                # this parent must not be a ghost.
-                if not parent_id in weave_parents:
-                    # but it is a ghost
-                    raise errors.CorruptRepository(self)
+        return [key[-1] for key in self.revisions.keys()]
+
+    def _activate_new_inventory(self):
+        """Put a replacement inventory.new into use as inventories."""
+        # Copy the content across
+        t = self._transport
+        t.copy('inventory.new.weave', 'inventory.weave')
+        # delete the temp inventory
+        t.delete('inventory.new.weave')
+        # Check we can parse the new weave properly as a sanity check
+        self.inventories.keys()
+
+    def _backup_inventory(self):
+        t = self._transport
+        t.copy('inventory.weave', 'inventory.backup.weave')
+
+    def _temp_inventories(self):
+        t = self._transport
+        return self._format._get_inventories(t, self, 'inventory.new')
 
     def get_commit_builder(self, branch, parents, config, timestamp=None,
                            timezone=None, committer=None, revprops=None,
                            revision_id=None):
         self._check_ascii_revisionid(revision_id, self.get_commit_builder)
-        result = WeaveCommitBuilder(self, parents, config, timestamp, timezone,
+        result = CommitBuilder(self, parents, config, timestamp, timezone,
                               committer, revprops, revision_id)
         self.start_write_group()
         return result
@@ -237,31 +225,21 @@
     @needs_read_lock
     def get_revision(self, revision_id):
         """Return the Revision object for a named revision"""
-        # TODO: jam 20070210 get_revision_reconcile should do this for us
         r = self.get_revision_reconcile(revision_id)
-        # weave corruption can lead to absent revision markers that should be
-        # present.
-        # the following test is reasonably cheap (it needs a single weave read)
-        # and the weave is cached in read transactions. In write transactions
-        # it is not cached but typically we only read a small number of
-        # revisions. For knits when they are introduced we will probably want
-        # to ensure that caching write transactions are in use.
-        inv = self.get_inventory_weave()
-        self._check_revision_parents(r, inv)
         return r
 
-    def has_revisions(self, revision_ids):
-        """See Repository.has_revisions()."""
-        result = set()
-        transaction = self.get_transaction()
-        for revision_id in revision_ids:
-            if self._revision_store.has_revision_id(revision_id, transaction):
-                result.add(revision_id)
-        return result
+    def _inventory_add_lines(self, revision_id, parents, lines,
+        check_content=True):
+        """Store lines in inv_vf and return the sha1 of the inventory."""
+        present_parents = self.get_graph().get_parent_map(parents)
+        final_parents = []
+        for parent in parents:
+            if parent in present_parents:
+                final_parents.append((parent,))
+        return self.inventories.add_lines((revision_id,), final_parents, lines,
+            check_content=check_content)[0]
 
     def revision_graph_can_have_wrong_parents(self):
-        # XXX: This is an old format that we don't support full checking on, so
-        # just claim that checking for this inconsistency is not required.
         return False
 
 
@@ -307,17 +285,6 @@
             control_files.unlock()
         return self.open(a_bzrdir, _found=True)
 
-    def _get_control_store(self, repo_transport, control_files):
-        """Return the control store for this repository."""
-        return self._get_versioned_file_store('',
-                                              repo_transport,
-                                              control_files,
-                                              prefixed=False)
-
-    def _get_text_store(self, transport, control_files):
-        """Get a store for file texts for this format."""
-        raise NotImplementedError(self._get_text_store)
-
     def open(self, a_bzrdir, _found=False):
         """See RepositoryFormat.open()."""
         if not _found:
@@ -326,14 +293,12 @@
 
         repo_transport = a_bzrdir.get_repository_transport(None)
         control_files = a_bzrdir._control_files
-        text_store = self._get_text_store(repo_transport, control_files)
-        control_store = self._get_control_store(repo_transport, control_files)
-        _revision_store = self._get_revision_store(repo_transport, control_files)
-        return AllInOneRepository(_format=self,
-                                  a_bzrdir=a_bzrdir,
-                                  _revision_store=_revision_store,
-                                  control_store=control_store,
-                                  text_store=text_store)
+        result = AllInOneRepository(_format=self, a_bzrdir=a_bzrdir)
+        result.revisions = self._get_revisions(repo_transport, result)
+        result.signatures = self._get_signatures(repo_transport, result)
+        result.inventories = self._get_inventories(repo_transport, result)
+        result.texts = self._get_texts(repo_transport, result)
+        return result
 
     def check_conversion_target(self, target_format):
         pass
@@ -373,23 +338,23 @@
         """
         return False
 
-    def _get_control_store(self, repo_transport, control_files):
-        """Format 4 repositories have no formal control store at this point.
-        
-        This will cause any control-file-needing apis to fail - this is desired.
-        """
+    def _get_inventories(self, repo_transport, repo, name='inventory'):
+        # No inventories store written so far.
         return None
-    
-    def _get_revision_store(self, repo_transport, control_files):
-        """See RepositoryFormat._get_revision_store()."""
+
+    def _get_revisions(self, repo_transport, repo):
         from bzrlib.xml4 import serializer_v4
-        return self._get_text_rev_store(repo_transport,
-                                        control_files,
-                                        'revision-store',
-                                        serializer=serializer_v4)
-
-    def _get_text_store(self, transport, control_files):
-        """See RepositoryFormat._get_text_store()."""
+        return RevisionTextStore(repo_transport.clone('revision-store'),
+            serializer_v4, True, versionedfile.PrefixMapper(),
+            repo.is_locked, repo.is_write_locked)
+
+    def _get_signatures(self, repo_transport, repo):
+        return SignatureTextStore(repo_transport.clone('revision-store'),
+            False, versionedfile.PrefixMapper(),
+            repo.is_locked, repo.is_write_locked)
+
+    def _get_texts(self, repo_transport, repo):
+        return None
 
 
 class RepositoryFormat5(PreSplitOutRepositoryFormat):
@@ -411,17 +376,27 @@
         """See RepositoryFormat.get_format_description()."""
         return "Weave repository format 5"
 
-    def _get_revision_store(self, repo_transport, control_files):
-        """See RepositoryFormat._get_revision_store()."""
-        """Return the revision store object for this a_bzrdir."""
-        return self._get_text_rev_store(repo_transport,
-                                        control_files,
-                                        'revision-store',
-                                        compressed=False)
-
-    def _get_text_store(self, transport, control_files):
-        """See RepositoryFormat._get_text_store()."""
-        return self._get_versioned_file_store('weaves', transport, control_files, prefixed=False)
+    def _get_inventories(self, repo_transport, repo, name='inventory'):
+        mapper = versionedfile.ConstantMapper(name)
+        return versionedfile.ThunkedVersionedFiles(repo_transport,
+            weave.WeaveFile, mapper, repo.is_locked)
+
+    def _get_revisions(self, repo_transport, repo):
+        from bzrlib.xml5 import serializer_v5
+        return RevisionTextStore(repo_transport.clone('revision-store'),
+            serializer_v5, False, versionedfile.PrefixMapper(),
+            repo.is_locked, repo.is_write_locked)
+
+    def _get_signatures(self, repo_transport, repo):
+        return SignatureTextStore(repo_transport.clone('revision-store'),
+            False, versionedfile.PrefixMapper(),
+            repo.is_locked, repo.is_write_locked)
+
+    def _get_texts(self, repo_transport, repo):
+        mapper = versionedfile.PrefixMapper()
+        base_transport = repo_transport.clone('weaves')
+        return versionedfile.ThunkedVersionedFiles(base_transport,
+            weave.WeaveFile, mapper, repo.is_locked)
 
 
 class RepositoryFormat6(PreSplitOutRepositoryFormat):
@@ -443,17 +418,28 @@
         """See RepositoryFormat.get_format_description()."""
         return "Weave repository format 6"
 
-    def _get_revision_store(self, repo_transport, control_files):
-        """See RepositoryFormat._get_revision_store()."""
-        return self._get_text_rev_store(repo_transport,
-                                        control_files,
-                                        'revision-store',
-                                        compressed=False,
-                                        prefixed=True)
-
-    def _get_text_store(self, transport, control_files):
-        """See RepositoryFormat._get_text_store()."""
-        return self._get_versioned_file_store('weaves', transport, control_files)
+    def _get_inventories(self, repo_transport, repo, name='inventory'):
+        mapper = versionedfile.ConstantMapper(name)
+        return versionedfile.ThunkedVersionedFiles(repo_transport,
+            weave.WeaveFile, mapper, repo.is_locked)
+
+    def _get_revisions(self, repo_transport, repo):
+        from bzrlib.xml5 import serializer_v5
+        return RevisionTextStore(repo_transport.clone('revision-store'),
+            serializer_v5, False, versionedfile.HashPrefixMapper(),
+            repo.is_locked, repo.is_write_locked)
+
+    def _get_signatures(self, repo_transport, repo):
+        return SignatureTextStore(repo_transport.clone('revision-store'),
+            False, versionedfile.HashPrefixMapper(),
+            repo.is_locked, repo.is_write_locked)
+
+    def _get_texts(self, repo_transport, repo):
+        mapper = versionedfile.HashPrefixMapper()
+        base_transport = repo_transport.clone('weaves')
+        return versionedfile.ThunkedVersionedFiles(base_transport,
+            weave.WeaveFile, mapper, repo.is_locked)
+
 
 class RepositoryFormat7(MetaDirRepositoryFormat):
     """Bzr repository 7.
@@ -470,13 +456,6 @@
     _versionedfile_class = weave.WeaveFile
     supports_ghosts = False
 
-    def _get_control_store(self, repo_transport, control_files):
-        """Return the control store for this repository."""
-        return self._get_versioned_file_store('',
-                                              repo_transport,
-                                              control_files,
-                                              prefixed=False)
-
     def get_format_string(self):
         """See RepositoryFormat.get_format_string()."""
         return "Bazaar-NG Repository format 7"
@@ -488,20 +467,27 @@
     def check_conversion_target(self, target_format):
         pass
 
-    def _get_revision_store(self, repo_transport, control_files):
-        """See RepositoryFormat._get_revision_store()."""
-        return self._get_text_rev_store(repo_transport,
-                                        control_files,
-                                        'revision-store',
-                                        compressed=False,
-                                        prefixed=True,
-                                        )
-
-    def _get_text_store(self, transport, control_files):
-        """See RepositoryFormat._get_text_store()."""
-        return self._get_versioned_file_store('weaves',
-                                              transport,
-                                              control_files)
+    def _get_inventories(self, repo_transport, repo, name='inventory'):
+        mapper = versionedfile.ConstantMapper(name)
+        return versionedfile.ThunkedVersionedFiles(repo_transport,
+            weave.WeaveFile, mapper, repo.is_locked)
+
+    def _get_revisions(self, repo_transport, repo):
+        from bzrlib.xml5 import serializer_v5
+        return RevisionTextStore(repo_transport.clone('revision-store'),
+            serializer_v5, True, versionedfile.HashPrefixMapper(),
+            repo.is_locked, repo.is_write_locked)
+
+    def _get_signatures(self, repo_transport, repo):
+        return SignatureTextStore(repo_transport.clone('revision-store'),
+            True, versionedfile.HashPrefixMapper(),
+            repo.is_locked, repo.is_write_locked)
+
+    def _get_texts(self, repo_transport, repo):
+        mapper = versionedfile.HashPrefixMapper()
+        base_transport = repo_transport.clone('weaves')
+        return versionedfile.ThunkedVersionedFiles(base_transport,
+            weave.WeaveFile, mapper, repo.is_locked)
 
     def initialize(self, a_bzrdir, shared=False):
         """Create a weave repository.
@@ -538,28 +524,183 @@
             repo_transport = a_bzrdir.get_repository_transport(None)
         control_files = lockable_files.LockableFiles(repo_transport,
                                 'lock', lockdir.LockDir)
-        text_store = self._get_text_store(repo_transport, control_files)
-        control_store = self._get_control_store(repo_transport, control_files)
-        _revision_store = self._get_revision_store(repo_transport, control_files)
-        return WeaveMetaDirRepository(_format=self,
-            a_bzrdir=a_bzrdir,
-            control_files=control_files,
-            _revision_store=_revision_store,
-            control_store=control_store,
-            text_store=text_store)
-
-
-class WeaveCommitBuilder(CommitBuilder):
-    """A builder for weave based repos that don't support ghosts."""
-
-    def _add_text_to_weave(self, file_id, new_lines, parents, nostore_sha):
-        versionedfile = self.repository.weave_store.get_weave_or_empty(
-            file_id, self.repository.get_transaction())
-        result = versionedfile.add_lines(
-            self._new_revision_id, parents, new_lines,
-            nostore_sha=nostore_sha)[0:2]
-        return result
-
+        result = WeaveMetaDirRepository(_format=self, a_bzrdir=a_bzrdir,
+            control_files=control_files)
+        result.revisions = self._get_revisions(repo_transport, result)
+        result.signatures = self._get_signatures(repo_transport, result)
+        result.inventories = self._get_inventories(repo_transport, result)
+        result.texts = self._get_texts(repo_transport, result)
+        result._transport = repo_transport
+        return result
+
+
+class TextVersionedFiles(VersionedFiles):
+    """Just-a-bunch-of-files based VersionedFile stores."""
+
+    def __init__(self, transport, compressed, mapper, is_locked, can_write):
+        self._compressed = compressed
+        self._transport = transport
+        self._mapper = mapper
+        if self._compressed:
+            self._ext = '.gz'
+        else:
+            self._ext = ''
+        self._is_locked = is_locked
+        self._can_write = can_write
+
+    def add_lines(self, key, parents, lines):
+        """Add a revision to the store."""
+        if not self._is_locked():
+            raise errors.ObjectNotLocked(self)
+        if not self._can_write():
+            raise errors.ReadOnlyError(self)
+        if '/' in key[-1]:
+            raise ValueError('bad idea to put / in %r' % key)
+        text = ''.join(lines)
+        if self._compressed:
+            text = bytes_to_gzip(text)
+        path = self._map(key)
+        self._transport.put_bytes_non_atomic(path, text, create_parent_dir=True)
+
+    def insert_record_stream(self, stream):
+        adapters = {}
+        for record in stream:
+            # Raise an error when a record is missing.
+            if record.storage_kind == 'absent':
+                raise errors.RevisionNotPresent([record.key[0]], self)
+            # adapt to non-tuple interface
+            if record.storage_kind == 'fulltext':
+                self.add_lines(record.key, None,
+                    osutils.split_lines(record.get_bytes_as('fulltext')))
+            else:
+                adapter_key = record.storage_kind, 'fulltext'
+                try:
+                    adapter = adapters[adapter_key]
+                except KeyError:
+                    adapter_factory = adapter_registry.get(adapter_key)
+                    adapter = adapter_factory(self)
+                    adapters[adapter_key] = adapter
+                lines = osutils.split_lines(adapter.get_bytes(
+                    record, record.get_bytes_as(record.storage_kind)))
+                try:
+                    self.add_lines(record.key, None, lines)
+                except RevisionAlreadyPresent:
+                    pass
+
+    def _load_text(self, key):
+        if not self._is_locked():
+            raise errors.ObjectNotLocked(self)
+        path = self._map(key)
+        try:
+            text = self._transport.get_bytes(path)
+            compressed = self._compressed
+        except errors.NoSuchFile:
+            if self._compressed:
+                # try without the .gz
+                path = path[:-3]
+                try:
+                    text = self._transport.get_bytes(path)
+                    compressed = False
+                except errors.NoSuchFile:
+                    return None
+            else:
+                return None
+        if compressed:
+            text = GzipFile(mode='rb', fileobj=StringIO(text)).read()
+        return text
+
+    def _map(self, key):
+        return self._mapper.map(key) + self._ext
+
+
+class RevisionTextStore(TextVersionedFiles):
+    """Legacy thunk for format 4 repositories."""
+
+    def __init__(self, transport, serializer, compressed, mapper, is_locked,
+        can_write):
+        """Create a RevisionTextStore at transport with serializer."""
+        TextVersionedFiles.__init__(self, transport, compressed, mapper,
+            is_locked, can_write)
+        self._serializer = serializer
+
+    def _load_text_parents(self, key):
+        text = self._load_text(key)
+        if text is None:
+            return None, None
+        parents = self._serializer.read_revision_from_string(text).parent_ids
+        return text, tuple((parent,) for parent in parents)
+
+    def get_parent_map(self, keys):
+        result = {}
+        for key in keys:
+            parents = self._load_text_parents(key)[1]
+            if parents is None:
+                continue
+            result[key] = parents
+        return result
+    
+    def get_record_stream(self, keys, sort_order, include_delta_closure):
+        for key in keys:
+            text, parents = self._load_text_parents(key)
+            if text is None:
+                yield AbsentContentFactory(key)
+            else:
+                yield FulltextContentFactory(key, parents, None, text)
+
+    def keys(self):
+        if not self._is_locked():
+            raise errors.ObjectNotLocked(self)
+        relpaths = set()
+        for quoted_relpath in self._transport.iter_files_recursive():
+            relpath = urllib.unquote(quoted_relpath)
+            path, ext = os.path.splitext(relpath)
+            if ext == '.gz':
+                relpath = path
+            if '.sig' not in relpath:
+                relpaths.add(relpath)
+        paths = list(relpaths)
+        return set([self._mapper.unmap(path) for path in paths])
+
+
+class SignatureTextStore(TextVersionedFiles):
+    """Legacy thunk for format 4-7 repositories."""
+
+    def __init__(self, transport, compressed, mapper, is_locked, can_write):
+        TextVersionedFiles.__init__(self, transport, compressed, mapper,
+            is_locked, can_write)
+        self._ext = '.sig' + self._ext
+
+    def get_parent_map(self, keys):
+        result = {}
+        for key in keys:
+            text = self._load_text(key)
+            if text is None:
+                continue
+            result[key] = None
+        return result
+    
+    def get_record_stream(self, keys, sort_order, include_delta_closure):
+        for key in keys:
+            text = self._load_text(key)
+            if text is None:
+                yield AbsentContentFactory(key)
+            else:
+                yield FulltextContentFactory(key, None, None, text)
+
+    def keys(self):
+        if not self._is_locked():
+            raise errors.ObjectNotLocked(self)
+        relpaths = set()
+        for quoted_relpath in self._transport.iter_files_recursive():
+            relpath = urllib.unquote(quoted_relpath)
+            path, ext = os.path.splitext(relpath)
+            if ext == '.gz':
+                relpath = path
+            if not relpath.endswith('.sig'):
+                continue
+            relpaths.add(relpath[:-4])
+        paths = list(relpaths)
+        return set([self._mapper.unmap(path) for path in paths])
 
 _legacy_formats = [RepositoryFormat4(),
                    RepositoryFormat5(),

=== modified file 'bzrlib/repository.py'
--- bzrlib/repository.py	2008-05-12 02:40:40 +0000
+++ bzrlib/repository.py	2008-06-11 04:20:16 +0000
@@ -408,18 +408,14 @@
         return self._get_delta(ie, basis_inv, path), True
 
     def _add_text_to_weave(self, file_id, new_lines, parents, nostore_sha):
-        versionedfile = self.repository.weave_store.get_weave_or_empty(
-            file_id, self.repository.get_transaction())
-        # Don't change this to add_lines - add_lines_with_ghosts is cheaper
-        # than add_lines, and allows committing when a parent is ghosted for
-        # some reason.
         # Note: as we read the content directly from the tree, we know its not
         # been turned into unicode or badly split - but a broken tree
         # implementation could give us bad output from readlines() so this is
         # not a guarantee of safety. What would be better is always checking
         # the content during test suite execution. RBC 20070912
-        return versionedfile.add_lines_with_ghosts(
-            self._new_revision_id, parents, new_lines,
+        parent_keys = tuple((file_id, parent) for parent in parents)
+        return self.repository.texts.add_lines(
+            (file_id, self._new_revision_id), parent_keys, new_lines,
             nostore_sha=nostore_sha, random_id=self.random_revid,
             check_content=False)[0:2]
 
@@ -453,6 +449,13 @@
     The Repository builds on top of Stores and a Transport, which respectively 
     describe the disk data format and the way of accessing the (possibly 
     remote) disk.
+
+    :ivar revisions: A bzrlib.versionedfile.VersionedFiles instance containing
+        the serialised revisions for the repository. This can be used to obtain
+        revision graph information or to access raw serialised revisions.
+        The result of trying to insert data into the repository via this store
+        is undefined: it should be considered read-only except for implementors
+        of repositories.
     """
 
     # What class to use for a CommitBuilder. Often its simpler to change this
@@ -510,18 +513,14 @@
         if inv.root is None:
             raise AssertionError()
         inv_lines = self._serialise_inventory_to_lines(inv)
-        inv_vf = self.get_inventory_weave()
-        return self._inventory_add_lines(inv_vf, revision_id, parents,
+        return self._inventory_add_lines(revision_id, parents,
             inv_lines, check_content=False)
 
-    def _inventory_add_lines(self, inv_vf, revision_id, parents, lines,
+    def _inventory_add_lines(self, revision_id, parents, lines,
         check_content=True):
         """Store lines in inv_vf and return the sha1 of the inventory."""
-        final_parents = []
-        for parent in parents:
-            if parent in inv_vf:
-                final_parents.append(parent)
-        return inv_vf.add_lines(revision_id, final_parents, lines,
+        parents = [(parent,) for parent in parents]
+        return self.inventories.add_lines((revision_id,), parents, lines,
             check_content=check_content)[0]
 
     def add_revision(self, revision_id, rev, inv=None, config=None):
@@ -544,24 +543,24 @@
             plaintext = Testament(rev, inv).as_short_text()
             self.store_revision_signature(
                 gpg.GPGStrategy(config), plaintext, revision_id)
-        inventory_vf = self.get_inventory_weave()
-        if not revision_id in inventory_vf:
+        # check inventory present
+        if not self.inventories.get_parent_map([(revision_id,)]):
             if inv is None:
                 raise errors.WeaveRevisionNotPresent(revision_id,
-                                                     inventory_vf)
+                                                     self.inventories)
             else:
                 # yes, this is not suitable for adding with ghosts.
                 rev.inventory_sha1 = self.add_inventory(revision_id, inv,
                                                         rev.parent_ids)
         else:
-            rev.inventory_sha1 = inventory_vf.get_sha1s([revision_id])[0]
-        self._revision_store.add_revision(rev, self.get_transaction())
+            rev.inventory_sha1 = self.inventories.get_sha1s([(revision_id,)])[0]
+        self._add_revision(rev)
 
-    def _add_revision_text(self, revision_id, text):
-        revision = self._revision_store._serializer.read_revision_from_string(
-            text)
-        self._revision_store._add_revision(revision, StringIO(text),
-                                           self.get_transaction())
+    def _add_revision(self, revision):
+        text = self._serializer.write_revision_to_string(revision)
+        key = (revision.revision_id,)
+        parents = tuple((parent,) for parent in revision.parent_ids)
+        self.revisions.add_lines(key, parents, osutils.split_lines(text))
 
     def all_revision_ids(self):
         """Returns a list of all the revision ids in the repository. 
@@ -607,11 +606,12 @@
         """Construct the current default format repository in a_bzrdir."""
         return RepositoryFormat.get_default_format().initialize(a_bzrdir)
 
-    def __init__(self, _format, a_bzrdir, control_files, _revision_store, control_store, text_store):
+    def __init__(self, _format, a_bzrdir, control_files):
         """instantiate a Repository.
 
         :param _format: The format of the repository on disk.
         :param a_bzrdir: The BzrDir of the repository.
+        :param revisions: The revisions store for the repository.
 
         In the future we will have a single api for all stores for
         getting file texts, inventories and revisions, then
@@ -622,17 +622,13 @@
         # the following are part of the public API for Repository:
         self.bzrdir = a_bzrdir
         self.control_files = control_files
-        self._revision_store = _revision_store
-        # backwards compatibility
-        self.weave_store = text_store
+        self._transport = control_files._transport
         # for tests
         self._reconcile_does_inventory_gc = True
         self._reconcile_fixes_text_parents = False
         self._reconcile_backsup_inventory = True
         # not right yet - should be more semantically clear ? 
         # 
-        self.control_store = control_store
-        self.control_weaves = control_store
         # TODO: make sure to construct the right store classes, etc, depending
         # on whether escaping is required.
         self._warn_if_deprecated()
@@ -762,10 +758,11 @@
                 last_revision.timezone)
 
         # now gather global repository information
+        # XXX: This is available for many repos regardless of listability.
         if self.bzrdir.root_transport.listable():
-            c, t = self._revision_store.total_size(self.get_transaction())
-            result['revisions'] = c
-            result['size'] = t
+            # XXX: do we want to __define len__() ?
+            result['revisions'] = len(self.revisions.keys())
+            # result['size'] = t
         return result
 
     def find_branches(self, using=False):
@@ -811,57 +808,6 @@
                 branches.extend(repository.find_branches())
         return branches
 
-    def get_data_stream(self, revision_ids):
-        raise NotImplementedError(self.get_data_stream)
-
-    def get_data_stream_for_search(self, search_result):
-        """Get a data stream that can be inserted to a repository.
-
-        :param search_result: A bzrlib.graph.SearchResult selecting the
-            revisions to get.
-        :return: A data stream that can be inserted into a repository using
-            insert_data_stream.
-        """
-        raise NotImplementedError(self.get_data_stream_for_search)
-
-    def insert_data_stream(self, stream):
-        """XXX What does this really do? 
-        
-        Is it a substitute for fetch? 
-        Should it manage its own write group ?
-        """
-        for item_key, bytes in stream:
-            if item_key[0] == 'file':
-                (file_id,) = item_key[1:]
-                knit = self.weave_store.get_weave_or_empty(
-                    file_id, self.get_transaction())
-            elif item_key == ('inventory',):
-                knit = self.get_inventory_weave()
-            elif item_key == ('revisions',):
-                knit = self._revision_store.get_revision_file(
-                    self.get_transaction())
-            elif item_key == ('signatures',):
-                knit = self._revision_store.get_signature_file(
-                    self.get_transaction())
-            else:
-                raise errors.RepositoryDataStreamError(
-                    "Unrecognised data stream key '%s'" % (item_key,))
-            decoded_list = bencode.bdecode(bytes)
-            format = decoded_list.pop(0)
-            data_list = []
-            knit_bytes = ''
-            for version, options, parents, some_bytes in decoded_list:
-                data_list.append((version, options, len(some_bytes), parents))
-                knit_bytes += some_bytes
-            buffer = StringIO(knit_bytes)
-            def reader_func(count):
-                if count is None:
-                    return buffer.read()
-                else:
-                    return buffer.read(count)
-            knit.insert_data_stream(
-                (format, data_list, reader_func))
-
     @needs_read_lock
     def search_missing_revision_ids(self, other, revision_id=None, find_ghosts=True):
         """Return the revision ids that other has that this does not.
@@ -1057,16 +1003,20 @@
         """True if this repository has a copy of the revision."""
         return revision_id in self.has_revisions((revision_id,))
 
+    @needs_read_lock
     def has_revisions(self, revision_ids):
         """Probe to find out the presence of multiple revisions.
 
         :param revision_ids: An iterable of revision_ids.
         :return: A set of the revision_ids that were present.
         """
-        raise NotImplementedError(self.has_revisions)
-
-        return self._revision_store.has_revision_id(revision_id,
-                                                    self.get_transaction())
+        parent_map = self.revisions.get_parent_map(
+            [(rev_id,) for rev_id in revision_ids])
+        result = set()
+        if _mod_revision.NULL_REVISION in revision_ids:
+            result.add(_mod_revision.NULL_REVISION)
+        result.update([key[0] for key in parent_map])
+        return result
 
     @needs_read_lock
     def get_revision(self, revision_id):
@@ -1095,9 +1045,16 @@
         for rev_id in revision_ids:
             if not rev_id or not isinstance(rev_id, basestring):
                 raise errors.InvalidRevisionId(revision_id=rev_id, branch=self)
-        revs = self._revision_store.get_revisions(revision_ids,
-                                                  self.get_transaction())
-        return revs
+        keys = [(key,) for key in revision_ids]
+        stream = self.revisions.get_record_stream(keys, 'unordered', True)
+        revs = {}
+        for record in stream:
+            if record.storage_kind == 'absent':
+                raise errors.NoSuchRevision(self, record.key[0])
+            text = record.get_bytes_as('fulltext')
+            rev = self._serializer.read_revision_from_string(text)
+            revs[record.key[0]] = rev
+        return [revs[revid] for revid in revision_ids]
 
     @needs_read_lock
     def get_revision_xml(self, revision_id):
@@ -1107,7 +1064,7 @@
         rev = self.get_revision(revision_id)
         rev_tmp = StringIO()
         # the current serializer..
-        self._revision_store._serializer.write_revision(rev, rev_tmp)
+        self._serializer.write_revision(rev, rev_tmp)
         rev_tmp.seek(0)
         return rev_tmp.getvalue()
 
@@ -1148,9 +1105,8 @@
 
     @needs_write_lock
     def add_signature_text(self, revision_id, signature):
-        self._revision_store.add_revision_signature_text(revision_id,
-                                                         signature,
-                                                         self.get_transaction())
+        self.signatures.add_lines((revision_id,), (),
+            osutils.split_lines(signature))
 
     def find_text_key_references(self):
         """Find the text key references within the repository.
@@ -1163,12 +1119,12 @@
             revision_id that they contain. The inventory texts from all present
             revision ids are assessed to generate this report.
         """
-        revision_ids = self.all_revision_ids()
-        w = self.get_inventory_weave()
+        revision_keys = self.revisions.keys()
+        w = self.inventories
         pb = ui.ui_factory.nested_progress_bar()
         try:
             return self._find_text_key_references_from_xml_inventory_lines(
-                w.iter_lines_added_or_present_in_versions(revision_ids, pb=pb))
+                w.iter_lines_added_or_present_in_keys(revision_keys, pb=pb))
         finally:
             pb.finished()
 
@@ -1211,7 +1167,7 @@
         search = self._file_ids_altered_regex.search
         unescape = _unescape_xml
         setdefault = result.setdefault
-        for line, version_id in line_iterator:
+        for line, line_key in line_iterator:
             match = search(line)
             if match is None:
                 continue
@@ -1248,7 +1204,7 @@
 
             key = (file_id, revision_id)
             setdefault(key, False)
-            if revision_id == version_id:
+            if revision_id == line_key[-1]:
                 result[key] = True
         return result
 
@@ -1269,13 +1225,13 @@
         """
         result = {}
         setdefault = result.setdefault
-        for file_id, revision_id in \
+        for key in \
             self._find_text_key_references_from_xml_inventory_lines(
                 line_iterator).iterkeys():
             # once data is all ensured-consistent; then this is
             # if revision_id == version_id
-            if revision_id in revision_ids:
-                setdefault(file_id, set()).add(revision_id)
+            if key[-1:] in revision_ids:
+                setdefault(key[0], set()).add(key[-1])
         return result
 
     def fileids_altered_by_revision_ids(self, revision_ids, _inv_weave=None):
@@ -1288,14 +1244,14 @@
         revision_ids. Each altered file-ids has the exact revision_ids that
         altered it listed explicitly.
         """
-        selected_revision_ids = set(revision_ids)
-        w = _inv_weave or self.get_inventory_weave()
+        selected_keys = set((revid,) for revid in revision_ids)
+        w = _inv_weave or self.inventories
         pb = ui.ui_factory.nested_progress_bar()
         try:
             return self._find_file_ids_from_xml_inventory_lines(
-                w.iter_lines_added_or_present_in_versions(
-                    selected_revision_ids, pb=pb),
-                selected_revision_ids)
+                w.iter_lines_added_or_present_in_keys(
+                    selected_keys, pb=pb),
+                selected_keys)
         finally:
             pb.finished()
 
@@ -1312,19 +1268,20 @@
 
         bytes_iterator is an iterable of bytestrings for the file.  The
         kind of iterable and length of the bytestrings are unspecified, but for
-        this implementation, it is a list of lines produced by
-        VersionedFile.get_lines().
+        this implementation, it is a list of bytes produced by
+        VersionedFile.get_record_stream().
 
         :param desired_files: a list of (file_id, revision_id, identifier)
             triples
         """
         transaction = self.get_transaction()
+        text_keys = {}
         for file_id, revision_id, callable_data in desired_files:
-            try:
-                weave = self.weave_store.get_weave(file_id, transaction)
-            except errors.NoSuchFile:
-                raise errors.NoSuchIdInRepository(self, file_id)
-            yield callable_data, weave.get_lines(revision_id)
+            text_keys[(file_id, revision_id)] = callable_data
+        for record in self.texts.get_record_stream(text_keys, 'unordered', True):
+            if record.storage_kind == 'absent':
+                raise errors.RevisionNotPresent(record.key, self)
+            yield text_keys[record.key], record.get_bytes_as('fulltext')
 
     def _generate_text_key_index(self, text_key_references=None,
         ancestors=None):
@@ -1447,7 +1404,7 @@
         # maybe this generator should explicitly have the contract that it
         # should not be iterated until the previously yielded item has been
         # processed?
-        inv_w = self.get_inventory_weave()
+        inv_w = self.inventories
 
         # file ids that changed
         file_ids = self.fileids_altered_by_revision_ids(revision_ids, inv_w)
@@ -1481,11 +1438,6 @@
         yield ("revisions", None, revision_ids)
 
     @needs_read_lock
-    def get_inventory_weave(self):
-        return self.control_weaves.get_weave('inventory',
-            self.get_transaction())
-
-    @needs_read_lock
     def get_inventory(self, revision_id):
         """Get Inventory object by revision id."""
         return self.iter_inventories([revision_id]).next()
@@ -1506,10 +1458,21 @@
 
     def _iter_inventories(self, revision_ids):
         """single-document based inventory iteration."""
-        texts = self.get_inventory_weave().get_texts(revision_ids)
-        for text, revision_id in zip(texts, revision_ids):
+        for text, revision_id in self._iter_inventory_xmls(revision_ids):
             yield self.deserialise_inventory(revision_id, text)
 
+    def _iter_inventory_xmls(self, revision_ids):
+        keys = [(revision_id,) for revision_id in revision_ids]
+        stream = self.inventories.get_record_stream(keys, 'unordered', True)
+        texts = {}
+        for record in stream:
+            if record.storage_kind != 'absent':
+                texts[record.key] = record.get_bytes_as('fulltext')
+            else:
+                raise errors.NoSuchRevision(self, record.key)
+        for key in keys:
+            yield texts[key], key[-1]
+
     def deserialise_inventory(self, revision_id, xml):
         """Transform the xml into an inventory object. 
 
@@ -1534,11 +1497,12 @@
     @needs_read_lock
     def get_inventory_xml(self, revision_id):
         """Get inventory XML as a file object."""
+        texts = self._iter_inventory_xmls([revision_id])
         try:
-            iw = self.get_inventory_weave()
-            return iw.get_text(revision_id)
-        except IndexError:
+            text, revision_id = texts.next()
+        except StopIteration:
             raise errors.HistoryMissing(self, 'inventory', revision_id)
+        return text
 
     @needs_read_lock
     def get_inventory_sha1(self, revision_id):
@@ -1645,9 +1609,21 @@
             return [None]
         if not self.has_revision(revision_id):
             raise errors.NoSuchRevision(self, revision_id)
-        w = self.get_inventory_weave()
-        candidates = w.get_ancestry(revision_id, topo_sorted)
-        return [None] + candidates # self._eliminate_revisions_not_present(candidates)
+        graph = self.get_graph()
+        keys = set()
+        search = graph._make_breadth_first_searcher([revision_id])
+        while True:
+            try:
+                found, ghosts = search.next_with_ghosts()
+            except StopIteration:
+                break
+            keys.update(found)
+        if _mod_revision.NULL_REVISION in keys:
+            keys.remove(_mod_revision.NULL_REVISION)
+        if topo_sorted:
+            parent_map = graph.get_parent_map(keys)
+            keys = tsort.topo_sort(parent_map)
+        return [None] + list(keys)
 
     def pack(self):
         """Compress the data within the repository.
@@ -1683,7 +1659,7 @@
 
     @deprecated_method(symbol_versioning.one_five)
     def revision_parents(self, revision_id):
-        return self.get_inventory_weave().parent_names(revision_id)
+        return self.get_revision(revision_id).parent_ids
 
     @deprecated_method(symbol_versioning.one_one)
     def get_parents(self, revision_ids):
@@ -1763,14 +1739,21 @@
     @needs_read_lock
     def has_signature_for_revision_id(self, revision_id):
         """Query for a revision signature for revision_id in the repository."""
-        return self._revision_store.has_signature(revision_id,
-                                                  self.get_transaction())
+        if not self.has_revision(revision_id):
+            raise errors.NoSuchRevision(self, revision_id)
+        sig_present = (1 == len(
+            self.signatures.get_parent_map([(revision_id,)])))
+        return sig_present
 
     @needs_read_lock
     def get_signature_text(self, revision_id):
         """Return the text for a signature."""
-        return self._revision_store.get_signature_text(revision_id,
-                                                       self.get_transaction())
+        stream = self.signatures.get_record_stream([(revision_id,)],
+            'unordered', True)
+        record = stream.next()
+        if record.storage_kind == 'absent':
+            raise errors.NoSuchRevision(self, revision_id)
+        return record.get_bytes_as('fulltext')
 
     @needs_read_lock
     def check(self, revision_ids=None):
@@ -1903,28 +1886,28 @@
         path, root = entries.next()
         if root.revision != rev.revision_id:
             raise errors.IncompatibleRevision(repr(repository))
+    text_keys = {}
+    for path, ie in entries:
+        text_keys[(ie.file_id, ie.revision)] = ie
+    text_parent_map = repository.texts.get_parent_map(text_keys)
+    missing_texts = set(text_keys) - set(text_parent_map)
     # Add the texts that are not already present
-    for path, ie in entries:
-        w = repository.weave_store.get_weave_or_empty(ie.file_id,
-                repository.get_transaction())
-        if ie.revision not in w:
-            text_parents = []
-            # FIXME: TODO: The following loop *may* be overlapping/duplicate
-            # with InventoryEntry.find_previous_heads(). if it is, then there
-            # is a latent bug here where the parents may have ancestors of each
-            # other. RBC, AB
-            for revision, tree in parent_trees.iteritems():
-                if ie.file_id not in tree:
-                    continue
-                parent_id = tree.inventory[ie.file_id].revision
-                if parent_id in text_parents:
-                    continue
-                text_parents.append(parent_id)
-                    
-            vfile = repository.weave_store.get_weave_or_empty(ie.file_id, 
-                repository.get_transaction())
-            lines = revision_tree.get_file(ie.file_id).readlines()
-            vfile.add_lines(rev.revision_id, text_parents, lines)
+    for text_key in missing_texts:
+        ie = text_keys[text_key]
+        text_parents = []
+        # FIXME: TODO: The following loop overlaps/duplicates that done by
+        # commit to determine parents. There is a latent/real bug here where
+        # the parents inserted are not those commit would do - in particular
+        # they are not filtered by heads(). RBC, AB
+        for revision, tree in parent_trees.iteritems():
+            if ie.file_id not in tree:
+                continue
+            parent_id = tree.inventory[ie.file_id].revision
+            if parent_id in text_parents:
+                continue
+            text_parents.append((ie.file_id, parent_id))
+        lines = revision_tree.get_file(ie.file_id).readlines()
+        repository.texts.add_lines(text_key, text_parents, lines)
     try:
         # install the inventory
         repository.add_inventory(rev.revision_id, inv, present_parents)
@@ -1938,13 +1921,8 @@
 class MetaDirRepository(Repository):
     """Repositories in the new meta-dir layout."""
 
-    def __init__(self, _format, a_bzrdir, control_files, _revision_store, control_store, text_store):
-        super(MetaDirRepository, self).__init__(_format,
-                                                a_bzrdir,
-                                                control_files,
-                                                _revision_store,
-                                                control_store,
-                                                text_store)
+    def __init__(self, _format, a_bzrdir, control_files):
+        super(MetaDirRepository, self).__init__(_format, a_bzrdir, control_files)
         dir_mode = self.control_files._dir_mode
         file_mode = self.control_files._file_mode
 
@@ -1979,12 +1957,9 @@
 class MetaDirVersionedFileRepository(MetaDirRepository):
     """Repositories in a meta-dir, that work via versioned file objects."""
 
-    def __init__(self, _format, a_bzrdir, control_files, _revision_store, control_store, text_store):
+    def __init__(self, _format, a_bzrdir, control_files):
         super(MetaDirVersionedFileRepository, self).__init__(_format, a_bzrdir,
-            control_files, _revision_store, control_store, text_store)
-        _revision_store.get_scope = self.get_transaction
-        control_store.get_scope = self.get_transaction
-        text_store.get_scope = self.get_transaction
+            control_files)
 
 
 class RepositoryFormatRegistry(registry.Registry):
@@ -2086,10 +2061,6 @@
         from bzrlib import bzrdir
         return bzrdir.format_registry.make_bzrdir('default').repository_format
 
-    def _get_control_store(self, repo_transport, control_files):
-        """Return the control store for this repository."""
-        raise NotImplementedError(self._get_control_store)
-
     def get_format_string(self):
         """Return the ASCII format string that identifies this format.
         
@@ -2102,33 +2073,6 @@
         """Return the short description for this format."""
         raise NotImplementedError(self.get_format_description)
 
-    def _get_revision_store(self, repo_transport, control_files):
-        """Return the revision store object for this a_bzrdir."""
-        raise NotImplementedError(self._get_revision_store)
-
-    def _get_text_rev_store(self,
-                            transport,
-                            control_files,
-                            name,
-                            compressed=True,
-                            prefixed=False,
-                            serializer=None):
-        """Common logic for getting a revision store for a repository.
-        
-        see self._get_revision_store for the subclass-overridable method to 
-        get the store for a repository.
-        """
-        from bzrlib.store.revision.text import TextRevisionStore
-        dir_mode = control_files._dir_mode
-        file_mode = control_files._file_mode
-        text_store = TextStore(transport.clone(name),
-                              prefixed=prefixed,
-                              compressed=compressed,
-                              dir_mode=dir_mode,
-                              file_mode=file_mode)
-        _revision_store = TextRevisionStore(text_store, serializer)
-        return _revision_store
-
     # TODO: this shouldn't be in the base class, it's specific to things that
     # use weaves or knits -- mbp 20070207
     def _get_versioned_file_store(self,
@@ -2509,19 +2453,21 @@
         if self.source.control_files._transport.listable():
             pb = ui.ui_factory.nested_progress_bar()
             try:
-                self.target.weave_store.copy_all_ids(
-                    self.source.weave_store,
-                    pb=pb,
-                    from_transaction=self.source.get_transaction(),
-                    to_transaction=self.target.get_transaction())
+                self.target.texts.insert_record_stream(
+                    self.source.texts.get_record_stream(
+                        self.source.texts.keys(), 'topological', False))
                 pb.update('copying inventory', 0, 1)
-                self.target.control_weaves.copy_multi(
-                    self.source.control_weaves, ['inventory'],
-                    from_transaction=self.source.get_transaction(),
-                    to_transaction=self.target.get_transaction())
-                self.target._revision_store.text_store.copy_all_ids(
-                    self.source._revision_store.text_store,
-                    pb=pb)
+                self.target.inventories.insert_record_stream(
+                    self.source.inventories.get_record_stream(
+                        self.source.inventories.keys(), 'topological', False))
+                self.target.signatures.insert_record_stream(
+                    self.source.signatures.get_record_stream(
+                        self.source.signatures.keys(),
+                        'unordered', True))
+                self.target.revisions.insert_record_stream(
+                    self.source.revisions.get_record_stream(
+                        self.source.revisions.keys(),
+                        'topological', True))
             finally:
                 pb.finished()
         else:
@@ -2894,43 +2840,6 @@
         return len(revision_ids), 0
 
 
-class InterRemoteToOther(InterRepository):
-
-    def __init__(self, source, target):
-        InterRepository.__init__(self, source, target)
-        self._real_inter = None
-
-    @staticmethod
-    def is_compatible(source, target):
-        if not isinstance(source, remote.RemoteRepository):
-            return False
-        # Is source's model compatible with target's model?
-        source._ensure_real()
-        real_source = source._real_repository
-        if isinstance(real_source, remote.RemoteRepository):
-            raise NotImplementedError(
-                "We don't support remote repos backed by remote repos yet.")
-        return InterRepository._same_model(real_source, target)
-
-    @needs_write_lock
-    def fetch(self, revision_id=None, pb=None, find_ghosts=False):
-        """See InterRepository.fetch()."""
-        from bzrlib.fetch import RemoteToOtherFetcher
-        mutter("Using fetch logic to copy between %s(remote) and %s(%s)",
-               self.source, self.target, self.target._format)
-        # TODO: jam 20070210 This should be an assert, not a translate
-        revision_id = osutils.safe_revision_id(revision_id)
-        f = RemoteToOtherFetcher(to_repository=self.target,
-                                 from_repository=self.source,
-                                 last_revision=revision_id,
-                                 pb=pb, find_ghosts=find_ghosts)
-        return f.count_copied, f.failed_revisions
-
-    @classmethod
-    def _get_repo_format_to_test(self):
-        return None
-
-
 class InterOtherToRemote(InterRepository):
 
     def __init__(self, source, target):
@@ -2970,7 +2879,6 @@
 InterRepository.register_optimiser(InterModel1and2)
 InterRepository.register_optimiser(InterKnit1and2)
 InterRepository.register_optimiser(InterPackRepo)
-InterRepository.register_optimiser(InterRemoteToOther)
 InterRepository.register_optimiser(InterOtherToRemote)
 
 
@@ -3062,17 +2970,16 @@
         self.repository = repository
         self.text_index = self.repository._generate_text_key_index()
     
-    def calculate_file_version_parents(self, revision_id, file_id):
+    def calculate_file_version_parents(self, text_key):
         """Calculate the correct parents for a file version according to
         the inventories.
         """
-        parent_keys = self.text_index[(file_id, revision_id)]
+        parent_keys = self.text_index[text_key]
         if parent_keys == [_mod_revision.NULL_REVISION]:
             return ()
-        # strip the file_id, for the weave api
-        return tuple([revision_id for file_id, revision_id in parent_keys])
+        return tuple(parent_keys)
 
-    def check_file_version_parents(self, weave, file_id):
+    def check_file_version_parents(self, texts, progress_bar=None):
         """Check the parents stored in a versioned file are correct.
 
         It also detects file versions that are not referenced by their
@@ -3086,24 +2993,30 @@
             file, but not used by the corresponding inventory.
         """
         wrong_parents = {}
-        unused_versions = set()
-        versions = weave.versions()
-        parent_map = weave.get_parent_map(versions)
-        for num, revision_id in enumerate(versions):
+        self.file_ids = set([file_id for file_id, _ in
+            self.text_index.iterkeys()])
+        # text keys is now grouped by file_id
+        n_weaves = len(self.file_ids)
+        files_in_revisions = {}
+        revisions_of_files = {}
+        n_versions = len(self.text_index)
+        progress_bar.update('loading text store', 0, n_versions)
+        parent_map = self.repository.texts.get_parent_map(self.text_index)
+        # On unlistable transports this could well be empty/error...
+        text_keys = self.repository.texts.keys()
+        unused_keys = frozenset(text_keys) - set(self.text_index)
+        for num, key in enumerate(self.text_index.iterkeys()):
+            if progress_bar is not None:
+                progress_bar.update('checking text graph', num, n_versions)
+            correct_parents = self.calculate_file_version_parents(key)
             try:
-                correct_parents = self.calculate_file_version_parents(
-                    revision_id, file_id)
-            except KeyError:
-                # The version is not part of the used keys.
-                unused_versions.add(revision_id)
-            else:
-                try:
-                    knit_parents = tuple(parent_map[revision_id])
-                except errors.RevisionNotPresent:
-                    knit_parents = None
-                if correct_parents != knit_parents:
-                    wrong_parents[revision_id] = (knit_parents, correct_parents)
-        return wrong_parents, unused_versions
+                knit_parents = parent_map[key]
+            except errors.RevisionNotPresent:
+                # Missing text!
+                knit_parents = None
+            if correct_parents != knit_parents:
+                wrong_parents[key] = (knit_parents, correct_parents)
+        return wrong_parents, unused_keys
 
 
 def _old_get_graph(repository, revision_id):

=== modified file 'bzrlib/revisiontree.py'
--- bzrlib/revisiontree.py	2008-05-08 04:33:38 +0000
+++ bzrlib/revisiontree.py	2008-06-11 04:20:16 +0000
@@ -19,6 +19,7 @@
 from cStringIO import StringIO
 
 from bzrlib import (
+    errors,
     osutils,
     revision,
     symbol_versioning,
@@ -39,7 +40,6 @@
         # construction - this will mean that we can change the constructor
         # with much less chance of breaking client code.
         self._repository = branch
-        self._weave_store = branch.weave_store
         self._inventory = inv
         self._revision_id = revision_id
 
@@ -62,34 +62,33 @@
         """Return the revision id associated with this tree."""
         return self._revision_id
 
-    def _get_weave(self, file_id):
-        return self._weave_store.get_weave(file_id,
-                self._repository.get_transaction())
-
     def get_file_lines(self, file_id):
-        ie = self._inventory[file_id]
-        weave = self._get_weave(file_id)
-        return weave.get_lines(ie.revision)
+        return osutils.split_lines(self.get_file_text(file_id))
 
     def get_file_text(self, file_id):
-        return ''.join(self.get_file_lines(file_id))
+        return list(self.iter_files_bytes([(file_id, None)]))[0][1]
 
     def get_file(self, file_id, path=None):
         return StringIO(self.get_file_text(file_id))
 
     def iter_files_bytes(self, desired_files):
-        """See Tree.extract_files_bytes.
+        """See Tree.iter_files_bytes.
 
         This version is implemented on top of Repository.extract_files_bytes"""
         repo_desired_files = [(f, self.inventory[f].revision, i)
                               for f, i in desired_files]
-        return self._repository.iter_files_bytes(repo_desired_files)
+        try:
+            for result in self._repository.iter_files_bytes(repo_desired_files):
+                yield result
+        except errors.RevisionNotPresent, e:
+            raise errors.NoSuchFile(e.revision_id)
 
     def annotate_iter(self, file_id,
                       default_revision=revision.CURRENT_REVISION):
         """See Tree.annotate_iter"""
-        w = self._get_weave(file_id)
-        return w.annotate(self.inventory[file_id].revision)
+        text_key = (file_id, self.inventory[file_id].revision)
+        annotations = self._repository.texts.annotate(text_key)
+        return [(key[-1], line) for key, line in annotations]
 
     def get_file_size(self, file_id):
         """See Tree.get_file_size"""

=== modified file 'bzrlib/smart/repository.py'
--- bzrlib/smart/repository.py	2008-05-08 04:33:38 +0000
+++ bzrlib/smart/repository.py	2008-06-11 04:20:16 +0000
@@ -257,7 +257,6 @@
               firstrev: 1234.230 0
               latestrev: 345.700 3600
               revisions: 2
-              size:45
 
               But containing only fields returned by the gather_stats() call
         """
@@ -394,70 +393,3 @@
             tarball.add(dirname, '.bzr') # recursive by default
         finally:
             tarball.close()
-
-
-class SmartServerRepositoryStreamKnitDataForRevisions(SmartServerRepositoryRequest):
-    """Bzr <= 1.1 streaming pull, buffers all data on server."""
-
-    def do_repository_request(self, repository, *revision_ids):
-        repository.lock_read()
-        try:
-            return self._do_repository_request(repository, revision_ids)
-        finally:
-            repository.unlock()
-
-    def _do_repository_request(self, repository, revision_ids):
-        stream = repository.get_data_stream_for_search(
-            repository.revision_ids_to_search_result(set(revision_ids)))
-        buffer = StringIO()
-        pack = ContainerSerialiser()
-        buffer.write(pack.begin())
-        try:
-            for name_tuple, bytes in stream:
-                buffer.write(pack.bytes_record(bytes, [name_tuple]))
-        except errors.RevisionNotPresent, e:
-            return FailedSmartServerResponse(('NoSuchRevision', e.revision_id))
-        buffer.write(pack.end())
-        return SuccessfulSmartServerResponse(('ok',), buffer.getvalue())
-
-
-class SmartServerRepositoryStreamRevisionsChunked(SmartServerRepositoryRequest):
-    """Bzr 1.1+ streaming pull."""
-
-    def do_body(self, body_bytes):
-        repository = self._repository
-        repository.lock_read()
-        try:
-            search, error = self.recreate_search(repository, body_bytes)
-            if error is not None:
-                repository.unlock()
-                return error
-            stream = repository.get_data_stream_for_search(search.get_result())
-        except Exception:
-            # On non-error, unlocking is done by the body stream handler.
-            repository.unlock()
-            raise
-        return SuccessfulSmartServerResponse(('ok',),
-            body_stream=self.body_stream(stream, repository))
-
-    def body_stream(self, stream, repository):
-        pack = ContainerSerialiser()
-        yield pack.begin()
-        try:
-            try:
-                for name_tuple, bytes in stream:
-                    yield pack.bytes_record(bytes, [name_tuple])
-            except:
-                # Undo the lock_read that that happens once the iterator from
-                # get_data_stream is started.
-                repository.unlock()
-                raise
-        except errors.RevisionNotPresent, e:
-            # This shouldn't be able to happen, but as we don't buffer
-            # everything it can in theory happen.
-            repository.unlock()
-            yield FailedSmartServerResponse(('NoSuchRevision', e.revision_id))
-        else:
-            repository.unlock()
-            pack.end()
-

=== modified file 'bzrlib/store/__init__.py'
--- bzrlib/store/__init__.py	2008-05-12 04:20:02 +0000
+++ bzrlib/store/__init__.py	2008-06-11 04:20:16 +0000
@@ -267,9 +267,7 @@
     def _iter_files_recursive(self):
         """Iterate through the files in the transport."""
         for quoted_relpath in self._transport.iter_files_recursive():
-            # transport iterator always returns quoted paths, regardless of
-            # escaping
-            yield urllib.unquote(quoted_relpath)
+            yield quoted_relpath
 
     def __iter__(self):
         for relpath in self._iter_files_recursive():
@@ -298,8 +296,8 @@
         else:
             suffixes = []
         path = self._mapper.map((fileid,))
-        full_path = u'.'.join([path] + suffixes)
-        return urlutils.escape(full_path)
+        full_path = '.'.join([path] + suffixes)
+        return full_path
 
     def __repr__(self):
         if self._transport is None:

=== removed file 'bzrlib/store/revision/__init__.py'
--- bzrlib/store/revision/__init__.py	2007-09-25 08:14:12 +0000
+++ bzrlib/store/revision/__init__.py	1970-01-01 00:00:00 +0000
@@ -1,119 +0,0 @@
-# Copyright (C) 2006 Canonical Ltd
-#
-# This program is free software; you can redistribute it and/or modify
-# it under the terms of the GNU General Public License as published by
-# the Free Software Foundation; either version 2 of the License, or
-# (at your option) any later version.
-#
-# This program is distributed in the hope that it will be useful,
-# but WITHOUT ANY WARRANTY; without even the implied warranty of
-# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-# GNU General Public License for more details.
-#
-# You should have received a copy of the GNU General Public License
-# along with this program; if not, write to the Free Software
-# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
-
-"""Revision stores.
-
-Revision stores are responsible for storing a group of revisions
-and returning some interesting data about them such as ancestors
-and ghosts information.
-"""
-
-from cStringIO import StringIO
-
-from bzrlib.lazy_import import lazy_import
-lazy_import(globals(), """
-from bzrlib import (
-    errors,
-    osutils,
-    xml5,
-    )
-from bzrlib.trace import mutter
-""")
-
-
-class RevisionStore(object):
-    """A revision store stores revisions."""
-
-    def __init__(self, serializer=None):
-        if serializer is None:
-            serializer = xml5.serializer_v5
-        self._serializer = serializer
-
-    def add_revision(self, revision, transaction):
-        """Add revision to the revision store.
-
-        :param rev: The revision object.
-        """
-        # serialisation : common to all at the moment.
-        rev_tmp = StringIO()
-        self._serializer.write_revision(revision, rev_tmp)
-        rev_tmp.seek(0)
-        self._add_revision(revision, rev_tmp, transaction)
-        mutter('added revision_id {%s}', revision.revision_id)
-
-    def _add_revision(self, revision, revision_as_file, transaction):
-        """Template method helper to store revision in this store."""
-        raise NotImplementedError(self._add_revision)
-
-    def add_revision_signature_text(self, revision_id, signature_text, transaction):
-        """Add signature_text as a signature for revision_id."""
-        raise NotImplementedError(self.add_revision_signature_text)
-
-    def all_revision_ids(self, transaction):
-        """Returns a list of all the revision ids in the revision store. 
-
-        :return: list of revision_ids in topological order.
-        """
-        raise NotImplementedError(self.all_revision_ids)
-
-    def get_revision(self, revision_id, transaction):
-        """Return the Revision object for a named revision."""
-        return self.get_revisions([revision_id], transaction)[0]
-
-    def get_revisions(self, revision_ids, transaction):
-        """Return the Revision objects for a list of named revisions."""
-        raise NotImplementedError(self.get_revision)
-
-    def get_signature_text(self, revision_id, transaction):
-        """Get the signature text for the digital signature of revision_id.
-        
-        :return: a signature text.
-        """
-        self._guard_revision(revision_id, transaction)
-        return self._get_signature_text(revision_id, transaction)
-
-    def _get_signature_text(self, revision_id, transaction):
-        """Helper method for get_signature_text to return the text itself."""
-        raise NotImplementedError(self.get_signature_text)
-
-    def _guard_revision(self, revision_id, transaction):
-        """Guard method for testing the presence of a revision."""
-        if not self.has_revision_id(revision_id, transaction):
-            raise errors.NoSuchRevision(self, revision_id)
-
-    def has_revision_id(self, revision_id, transaction):
-        """True if the store contains revision_id."""
-        raise NotImplementedError(self.has_revision_id)
-
-    def has_signature(self, revision_id, transaction):
-        """True if the store has a signature for revision_id."""
-        self._guard_revision(revision_id, transaction)
-        return self._has_signature(revision_id, transaction)
-
-    def _has_signature(self, revision_id, transaction):
-        """Return the presence of a signature for revision_id.
-
-        A worker memrthod for has_signature, this can assume the
-        revision is present.
-        """
-        return NotImplementedError(self._has_signature)
-        
-    def total_size(self, transaction):
-        """How big is the store?
-
-        :return: (count, bytes) tuple.
-        """
-        raise NotImplementedError(self.total_size)

=== removed file 'bzrlib/store/revision/knit.py'
--- bzrlib/store/revision/knit.py	2008-05-08 04:33:38 +0000
+++ bzrlib/store/revision/knit.py	1970-01-01 00:00:00 +0000
@@ -1,144 +0,0 @@
-# Copyright (C) 2006 Canonical Ltd
-#
-# This program is free software; you can redistribute it and/or modify
-# it under the terms of the GNU General Public License as published by
-# the Free Software Foundation; either version 2 of the License, or
-# (at your option) any later version.
-#
-# This program is distributed in the hope that it will be useful,
-# but WITHOUT ANY WARRANTY; without even the implied warranty of
-# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-# GNU General Public License for more details.
-#
-# You should have received a copy of the GNU General Public License
-# along with this program; if not, write to the Free Software
-# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
-
-"""VersionedFile based revision store.
-
-This stores revisions as individual entries in a knit, and signatures in a 
-parallel knit.
-"""
-
-
-from bzrlib import (
-    errors,
-    osutils,
-    revision as _mod_revision,
-    )
-from bzrlib.knit import make_file_knit, KnitPlainFactory
-from bzrlib.store.revision import RevisionStore
-from bzrlib.store.versioned import VersionedFileStore
-from bzrlib.transport import get_transport
-
-
-class KnitRevisionStoreFactory(object):
-    """Factory to create a KnitRevisionStore for testing."""
-
-    def create(self, url):
-        """Create a revision store at url."""
-        t = get_transport(url)
-        t.mkdir('revision-store')
-        versioned_file_store = VersionedFileStore(
-            t.clone('revision-store'),
-            precious=True,
-            versionedfile_class=make_file_knit,
-            versionedfile_kwargs={'delta':False, 'factory':KnitPlainFactory()})
-        return KnitRevisionStore(versioned_file_store)
-
-    def __str__(self):
-        return "KnitRevisionStore"
-
-
-class KnitRevisionStore(RevisionStore):
-    """A RevisionStore layering on a VersionedFileStore."""
-
-    def __init__(self, versioned_file_store):
-        """Create a KnitRevisionStore object.
-
-        :param versioned_file_store: the text store to use for storing 
-                                     revisions and signatures.
-        """
-        super(KnitRevisionStore, self).__init__()
-        self.versioned_file_store = versioned_file_store
-
-    def __repr__(self):
-        return "%s(%s)" % (self.__class__.__name__,
-                           self.versioned_file_store)
-
-    def _add_revision(self, revision, revision_as_file, transaction):
-        """Template method helper to store revision in this store."""
-        self.get_revision_file(transaction).add_lines_with_ghosts(
-            revision.revision_id,
-            revision.parent_ids,
-            osutils.split_lines(revision_as_file.read()))
-
-    def add_revision_signature_text(self, revision_id, signature_text, transaction):
-        """See RevisionStore.add_revision_signature_text()."""
-        self.get_signature_file(transaction).add_lines(
-            revision_id, [], osutils.split_lines(signature_text))
-
-    def all_revision_ids(self, transaction):
-        """See RevisionStore.all_revision_ids()."""
-        rev_file = self.get_revision_file(transaction)
-        return rev_file.get_ancestry(rev_file.versions())
-
-    def get_revisions(self, revision_ids, transaction):
-        """See RevisionStore.get_revisions()."""
-        texts = self._get_serialized_revisions(revision_ids, transaction)
-        revisions = []
-        try:
-            for text, revision_id in zip(texts, revision_ids):
-                r = self._serializer.read_revision_from_string(text)
-                revisions.append(r)
-        except SyntaxError, e:
-            raise errors.BzrError('failed to unpack revision_xml for %s: %s' %
-                                   (revision_id, str(e)))
-        return revisions
-
-    def _get_serialized_revisions(self, revision_ids, transaction):
-        texts = []
-        vf = self.get_revision_file(transaction)
-        try:
-            return vf.get_texts(revision_ids)
-        except (errors.RevisionNotPresent), e:
-            raise errors.NoSuchRevision(self, e.revision_id)
-
-    def _get_revision_xml(self, revision_id, transaction):
-        try:
-            return self.get_revision_file(transaction).get_text(revision_id)
-        except (errors.RevisionNotPresent):
-            raise errors.NoSuchRevision(self, revision_id)
-
-    def get_revision_file(self, transaction):
-        """Get the revision versioned file object."""
-        vf = self.versioned_file_store.get_weave_or_empty('revisions', transaction)
-        # The revisions knit should always be non-delta, so force delta=False
-        # here.
-        vf.delta = False
-        return vf
-
-    def get_signature_file(self, transaction):
-        """Get the signature text versioned file object."""
-        return self.versioned_file_store.get_weave_or_empty('signatures', transaction)
-
-    def _get_signature_text(self, revision_id, transaction):
-        """See RevisionStore._get_signature_text()."""
-        try:
-            return self.get_signature_file(transaction).get_text(revision_id)
-        except errors.RevisionNotPresent:
-            raise errors.NoSuchRevision(self, revision_id)
-
-    def has_revision_id(self, revision_id, transaction):
-        """True if the store contains revision_id."""
-        return (_mod_revision.is_null(revision_id)
-                or self.get_revision_file(transaction).has_version(revision_id))
-
-    def _has_signature(self, revision_id, transaction):
-        """See RevisionStore._has_signature()."""
-        return self.get_signature_file(transaction).has_version(revision_id)
-
-    def total_size(self, transaction):
-        """ See RevisionStore.total_size()."""
-        return (len(self.all_revision_ids(transaction)),
-            self.versioned_file_store.total_size()[1])

=== removed file 'bzrlib/store/revision/text.py'
--- bzrlib/store/revision/text.py	2008-05-08 04:33:38 +0000
+++ bzrlib/store/revision/text.py	1970-01-01 00:00:00 +0000
@@ -1,131 +0,0 @@
-# Copyright (C) 2006 Canonical Ltd
-#
-# This program is free software; you can redistribute it and/or modify
-# it under the terms of the GNU General Public License as published by
-# the Free Software Foundation; either version 2 of the License, or
-# (at your option) any later version.
-#
-# This program is distributed in the hope that it will be useful,
-# but WITHOUT ANY WARRANTY; without even the implied warranty of
-# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-# GNU General Public License for more details.
-#
-# You should have received a copy of the GNU General Public License
-# along with this program; if not, write to the Free Software
-# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
-
-"""TextStore based revision store.
-
-This stores revisions as individual text entries in a TextStore and 
-requires access to a inventory weave to produce object graphs.
-"""
-
-
-from cStringIO import StringIO
-
-
-from bzrlib import (
-    cache_utf8,
-    errors,
-    osutils,
-    revision as _mod_revision,
-    )
-from bzrlib.store.revision import RevisionStore
-from bzrlib.store.text import TextStore
-from bzrlib.store.versioned import VersionedFileStore
-from bzrlib.transport import get_transport
-from bzrlib.tsort import topo_sort
-
-
-class TextRevisionStoreTestFactory(object):
-    """Factory to create a TextRevisionStore for testing.
-
-    This creates a inventory weave and hooks it into the revision store
-    """
-
-    def create(self, url):
-        """Create a revision store at url."""
-        t = get_transport(url)
-        t.mkdir('revstore')
-        text_store = TextStore(t.clone('revstore'))
-        return TextRevisionStore(text_store)
-
-    def __str__(self):
-        return "TextRevisionStore"
-
-
-class TextRevisionStore(RevisionStore):
-    """A RevisionStore layering on a TextStore and Inventory weave store."""
-
-    def __init__(self, text_store, serializer=None):
-        """Create a TextRevisionStore object.
-
-        :param text_store: the text store to put serialised revisions into.
-        """
-        super(TextRevisionStore, self).__init__(serializer)
-        self.text_store = text_store
-        self.text_store.register_suffix('sig')
-
-    def _add_revision(self, revision, revision_as_file, transaction):
-        """Template method helper to store revision in this store."""
-        self.text_store.add(revision_as_file, revision.revision_id)
-
-    def add_revision_signature_text(self, revision_id, signature_text, transaction):
-        """See RevisionStore.add_revision_signature_text()."""
-        self.text_store.add(StringIO(signature_text), revision_id, "sig")
-
-    def all_revision_ids(self, transaction):
-        """See RevisionStore.all_revision_ids()."""
-        # for TextRevisionStores, this is only functional
-        # on listable transports.
-        result_graph = {}
-        for rev_id in self.text_store:
-            rev = self.get_revision(rev_id, transaction)
-            result_graph[rev_id] = rev.parent_ids
-        # remove ghosts
-        for rev_id, parents in result_graph.items():
-            for parent in list(parents):
-                if not parent in result_graph:
-                    del parents[parents.index(parent)]
-        return topo_sort(result_graph.items())
-
-    def get_revisions(self, revision_ids, transaction):
-        """See RevisionStore.get_revisions()."""
-        revisions = []
-        for revision_id in revision_ids:
-            xml_file = self._get_revision_xml_file(revision_id)
-            try:
-                r = self._serializer.read_revision(xml_file)
-            except SyntaxError, e:
-                raise errors.BzrError('failed to unpack revision_xml %s %s' % (
-                    revision_id, str(e)))
-            xml_file.close()
-            revisions.append(r)
-        return revisions
-
-    def _get_revision_xml_file(self, revision_id):
-        _mod_revision.check_not_reserved_id(revision_id)
-        try:
-            return self.text_store.get(revision_id)
-        except (IndexError, KeyError):
-            raise errors.NoSuchRevision(self, revision_id)
-
-    def _get_signature_text(self, revision_id, transaction):
-        """See RevisionStore._get_signature_text()."""
-        try:
-            return self.text_store.get(revision_id, suffix='sig').read()
-        except (IndexError, KeyError):
-            raise errors.NoSuchRevision(self, revision_id)
-
-    def has_revision_id(self, revision_id, transaction):
-        """True if the store contains revision_id."""
-        return (_mod_revision.is_null(revision_id)
-                or self.text_store.has_id(revision_id))
-
-    def _has_signature(self, revision_id, transaction):
-        """See RevisionStore._has_signature()."""
-        return self.text_store.has_id(revision_id, suffix='sig')
-
-    def total_size(self, transaction):
-        """ See RevisionStore.total_size()."""
-        return self.text_store.total_size()

=== modified file 'bzrlib/store/versioned/__init__.py'
--- bzrlib/store/versioned/__init__.py	2008-05-12 04:20:02 +0000
+++ bzrlib/store/versioned/__init__.py	2008-06-11 04:20:16 +0000
@@ -179,13 +179,10 @@
     def _put_weave(self, file_id, weave, transaction):
         """Preserved here for upgrades-to-weaves to use."""
         myweave = self._make_new_versionedfile(file_id, transaction)
-        myweave.insert_record_stream(weave.get_record_stream(weave.versions(),
+        myweave.insert_record_stream(weave.get_record_stream(
+            [(version,) for version in weave.versions()],
             'topological', False))
 
-    def copy(self, source, result_id, transaction):
-        """Copy the source versioned file to result_id in this store."""
-        source.copy_to(self.filename(result_id), self._transport)
- 
     def copy_all_ids(self, store_from, pb=None, from_transaction=None,
                      to_transaction=None):
         """Copy all the file ids from store_from into self."""
@@ -242,7 +239,8 @@
                 target = self._make_new_versionedfile(f, to_transaction)
                 source = from_store.get_weave(f, from_transaction)
                 target.insert_record_stream(source.get_record_stream(
-                    source.versions(), 'topological', False))
+                    [(version,) for version in source.versions()],
+                    'topological', False))
         finally:
             pb.finished()
 

=== modified file 'bzrlib/tests/__init__.py'
--- bzrlib/tests/__init__.py	2008-05-12 03:04:40 +0000
+++ bzrlib/tests/__init__.py	2008-06-11 04:20:16 +0000
@@ -2630,10 +2630,8 @@
                    'bzrlib.tests.inventory_implementations',
                    'bzrlib.tests.interrepository_implementations',
                    'bzrlib.tests.intertree_implementations',
-                   'bzrlib.tests.interversionedfile_implementations',
                    'bzrlib.tests.per_lock',
                    'bzrlib.tests.repository_implementations',
-                   'bzrlib.tests.revisionstore_implementations',
                    'bzrlib.tests.test__dirstate_helpers',
                    'bzrlib.tests.test_ancestry',
                    'bzrlib.tests.test_annotate',
@@ -2661,7 +2659,6 @@
                    'bzrlib.tests.test_directory_service',
                    'bzrlib.tests.test_email_message',
                    'bzrlib.tests.test_errors',
-                   'bzrlib.tests.test_escaped_store',
                    'bzrlib.tests.test_extract',
                    'bzrlib.tests.test_fetch',
                    'bzrlib.tests.test_ftp_transport',

=== modified file 'bzrlib/tests/blackbox/test_info.py'
--- bzrlib/tests/blackbox/test_info.py	2007-12-15 18:33:10 +0000
+++ bzrlib/tests/blackbox/test_info.py	2008-06-11 04:20:16 +0000
@@ -90,7 +90,6 @@
 
 Repository:
          0 revisions
-         0 KiB
 """, out)
         self.assertEqual('', err)
         tree1.commit('commit one')
@@ -148,12 +147,7 @@
 
 Repository:
          1 revision
-         %d KiB
 """ % (datestring_first, datestring_first,
-       # poking at _revision_store isn't all that clean, but neither is
-       # having the ui test dependent on the exact overhead of a given store.
-       branch2.repository._revision_store.total_size(
-        branch2.repository.get_transaction())[1] / 1024,
        ), out)
         self.assertEqual('', err)
 
@@ -200,15 +194,10 @@
 
 Repository:
          1 revision
-         %d KiB
 """ % (bound_tree._format.get_format_description(),
        branch3._format.get_format_description(),
        branch3.repository._format.get_format_description(),
        datestring_first, datestring_first,
-       # poking at _revision_store isn't all that clean, but neither is
-       # having the ui test dependent on the exact overhead of a given store.
-       branch3.repository._revision_store.total_size(
-        branch3.repository.get_transaction())[1] / 1024,
        ), out)
         self.assertEqual('', err)
 
@@ -249,13 +238,8 @@
 
 Repository:
          1 revision
-         %d KiB
 """ % (branch4.repository._format.get_format_description(),
        datestring_first, datestring_first,
-       # poking at _revision_store isn't all that clean, but neither is
-       # having the ui test dependent on the exact overhead of a given store.
-       branch4.repository._revision_store.total_size(
-        branch4.repository.get_transaction())[1] / 1024,
        ), out)
         self.assertEqual('', err)
 
@@ -295,7 +279,6 @@
 
 Repository:
          1 revision
-         0 KiB
 """ % (datestring_first, datestring_first,), out)
         self.assertEqual('', err)
 
@@ -342,7 +325,6 @@
 
 Repository:
          1 revision
-         0 KiB
 """ % (datestring_first, datestring_first,
        ), out)
         self.assertEqual('', err)
@@ -385,13 +367,8 @@
 
 Repository:
          1 revision
-         %d KiB
 """ % (branch3.repository._format.get_format_description(),
        datestring_first, datestring_first,
-       # poking at _revision_store isn't all that clean, but neither is
-       # having the ui test dependent on the exact overhead of a given store.
-       branch3.repository._revision_store.total_size(
-        branch3.repository.get_transaction())[1] / 1024,
        ), out)
         self.assertEqual('', err)
 
@@ -430,13 +407,8 @@
 
 Repository:
          1 revision
-         %d KiB
 """ % (branch4.repository._format.get_format_description(),
        datestring_first, datestring_first,
-       # poking at _revision_store isn't all that clean, but neither is
-       # having the ui test dependent on the exact overhead of a given store.
-       branch4.repository._revision_store.total_size(
-        branch4.repository.get_transaction())[1] / 1024,
        ), out)
         self.assertEqual('', err)
 
@@ -476,7 +448,6 @@
 
 Repository:
          2 revisions
-         0 KiB
 """ % (datestring_first, datestring_last,), out)
         self.assertEqual('', err)
 
@@ -502,7 +473,6 @@
 
 Repository:
          0 revisions
-         0 KiB
 """ % (info.describe_format(repo.bzrdir, repo, branch, None),
        format.get_branch_format().get_format_description(),
        format.repository_format.get_format_description(),
@@ -528,7 +498,6 @@
 
 Repository:
          0 revisions
-         0 KiB
 """ % ('repo', format.repository_format.get_format_description(),
        ), out)
         self.assertEqual('', err)
@@ -555,7 +524,6 @@
 
 Repository:
          0 revisions
-         0 KiB
 """ % (format.get_branch_format().get_format_description(),
        format.repository_format.get_format_description(),
        ), out)
@@ -615,13 +583,9 @@
 
 Repository:
          1 revision
-         %d KiB
 """ % (format.get_branch_format().get_format_description(),
        format.repository_format.get_format_description(),
        datestring_first, datestring_first,
-       # poking at _revision_store isn't all that clean, but neither is
-       # having the ui test dependent on the exact overhead of a given store.
-       repo._revision_store.total_size(repo.get_transaction())[1] / 1024,
        ), out)
         self.assertEqual('', err)
 
@@ -657,7 +621,6 @@
 
 Repository:
          0 revisions
-         0 KiB
 """ % (format.get_branch_format().get_format_description(),
        format.repository_format.get_format_description(),
        ), out)
@@ -699,13 +662,9 @@
 
 Repository:
          1 revision
-         %d KiB
 """ % (format.get_branch_format().get_format_description(),
        format.repository_format.get_format_description(),
        datestring_first, datestring_first,
-       # poking at _revision_store isn't all that clean, but neither is
-       # having the ui test dependent on the exact overhead of a given store.
-       repo._revision_store.total_size(repo.get_transaction())[1] / 1024,
        ), out)
         self.assertEqual('', err)
         tree3.commit('commit two')
@@ -749,13 +708,9 @@
 
 Repository:
          2 revisions
-         %d KiB
 """ % (format.get_branch_format().get_format_description(),
        format.repository_format.get_format_description(),
        datestring_first, datestring_last,
-       # poking at _revision_store isn't all that clean, but neither is
-       # having the ui test dependent on the exact overhead of a given store.
-       repo._revision_store.total_size(repo.get_transaction())[1] / 1024,
        ), out)
         self.assertEqual('', err)
 
@@ -781,13 +736,9 @@
 
 Repository:
          2 revisions
-         %d KiB
 """ % (format.get_branch_format().get_format_description(),
        format.repository_format.get_format_description(),
        datestring_first, datestring_last,
-       # poking at _revision_store isn't all that clean, but neither is
-       # having the ui test dependent on the exact overhead of a given store.
-       repo._revision_store.total_size(repo.get_transaction())[1] / 1024,
        ), out)
         self.assertEqual('', err)
 
@@ -804,11 +755,7 @@
 
 Repository:
          2 revisions
-         %d KiB
 """ % (format.repository_format.get_format_description(),
-       # poking at _revision_store isn't all that clean, but neither is
-       # having the ui test dependent on the exact overhead of a given store.
-       repo._revision_store.total_size(repo.get_transaction())[1] / 1024,
        ), out)
         self.assertEqual('', err)
 
@@ -833,7 +780,6 @@
 
 Repository:
          0 revisions
-         0 KiB
 """ % (format.repository_format.get_format_description(),
        ), out)
         self.assertEqual('', err)
@@ -874,7 +820,6 @@
 
 Repository:
          0 revisions
-         0 KiB
 """ % (format.get_branch_format().get_format_description(),
        format.repository_format.get_format_description(),
        ), out)
@@ -919,13 +864,9 @@
 
 Repository:
          1 revision
-         %d KiB
 """ % (format.get_branch_format().get_format_description(),
        format.repository_format.get_format_description(),
        datestring_first, datestring_first,
-       # poking at _revision_store isn't all that clean, but neither is
-       # having the ui test dependent on the exact overhead of a given store.
-       repo._revision_store.total_size(repo.get_transaction())[1] / 1024,
        ), out)
         self.assertEqual('', err)
 
@@ -962,12 +903,8 @@
 
 Repository:
          1 revision
-         %d KiB
 """ % (format.get_branch_format().get_format_description(),
        format.repository_format.get_format_description(),
-       # poking at _revision_store isn't all that clean, but neither is
-       # having the ui test dependent on the exact overhead of a given store.
-       repo._revision_store.total_size(repo.get_transaction())[1] / 1024,
        ), out)
         self.assertEqual('', err)
 
@@ -1009,13 +946,9 @@
 
 Repository:
          1 revision
-         %d KiB
 """ % (format.get_branch_format().get_format_description(),
        format.repository_format.get_format_description(),
        datestring_first, datestring_first,
-       # poking at _revision_store isn't all that clean, but neither is
-       # having the ui test dependent on the exact overhead of a given store.
-       repo._revision_store.total_size(repo.get_transaction())[1] / 1024,
        ), out)
         self.assertEqual('', err)
 
@@ -1034,11 +967,7 @@
 
 Repository:
          1 revision
-         %d KiB
 """ % (format.repository_format.get_format_description(),
-       # poking at _revision_store isn't all that clean, but neither is
-       # having the ui test dependent on the exact overhead of a given store.
-       repo._revision_store.total_size(repo.get_transaction())[1] / 1024,
        ),
        out)
         self.assertEqual('', err)
@@ -1064,7 +993,6 @@
 
 Repository:
          0 revisions
-         0 KiB
 """ % (format.repository_format.get_format_description(),
        ), out)
         self.assertEqual('', err)
@@ -1102,7 +1030,6 @@
 
 Repository:
          0 revisions
-         0 KiB
 """ % (format.get_branch_format().get_format_description(),
        format.repository_format.get_format_description(),
        ), out)
@@ -1238,7 +1165,6 @@
 %s
 Repository:
          0 revisions
-         0 KiB
 """ %  (description,
         format,
         tree_data,
@@ -1405,7 +1331,6 @@
 
 Repository:
          0 revisions
-         0 KiB
 """ % ('branch', tree.branch.repository._format.get_format_description(),
        ), out)
         self.assertEqual('', err)
@@ -1439,7 +1364,6 @@
 
 Repository:
          0 revisions
-         0 KiB
 """ % ('branch', tree.branch.repository._format.get_format_description(),
        ), out)
         self.assertEqual('', err)

=== modified file 'bzrlib/tests/branch_implementations/test_branch.py'
--- bzrlib/tests/branch_implementations/test_branch.py	2008-05-02 07:31:24 +0000
+++ bzrlib/tests/branch_implementations/test_branch.py	2008-06-11 04:20:16 +0000
@@ -57,7 +57,8 @@
         tree.lock_read()
         self.addCleanup(tree.unlock)
         graph = tree.branch.repository.get_graph()
-        ancestry_graph = graph.get_parent_map(tree.branch.repository.all_revision_ids())
+        ancestry_graph = graph.get_parent_map(
+            tree.branch.repository.all_revision_ids())
         self.assertEqual({'rev-1':('null:',),
                           'rev-2':('rev-1', ),
                           'rev-1.1.1':('rev-1', ),
@@ -280,6 +281,7 @@
                 branch.repository.commit_write_group()
         finally:
             branch.unlock()
+        # A signature without a revision should not be accessible.
         self.assertRaises(errors.NoSuchRevision,
                           branch.repository.has_signature_for_revision_id,
                           'A')

=== modified file 'bzrlib/tests/bzrdir_implementations/test_bzrdir.py'
--- bzrlib/tests/bzrdir_implementations/test_bzrdir.py	2008-04-10 13:30:05 +0000
+++ bzrlib/tests/bzrdir_implementations/test_bzrdir.py	2008-06-11 04:20:16 +0000
@@ -125,15 +125,16 @@
                     self.assertEqual(left_repo.get_revision(rev_id),
                         right_repo.get_revision(rev_id))
                 # inventories
-                left_inv_weave = left_repo.get_inventory_weave()
-                right_inv_weave = right_repo.get_inventory_weave()
-                self.assertEqual(set(left_inv_weave.versions()),
-                    set(right_inv_weave.versions()))
+                left_inv_weave = left_repo.inventories
+                right_inv_weave = right_repo.inventories
+                self.assertEqual(set(left_inv_weave.keys()),
+                    set(right_inv_weave.keys()))
                 # XXX: currently this does not handle indirectly referenced
                 # inventories (e.g. where the inventory is a delta basis for
                 # one that is fully present but that the revid for that
                 # inventory is not yet present.)
-                self.assertEqual(set(left_inv_weave.versions()), set(all_revs))
+                self.assertEqual(set(left_inv_weave.keys()),
+                    set(left_repo.revisions.keys()))
                 left_trees = left_repo.revision_trees(all_revs)
                 right_trees = right_repo.revision_trees(all_revs)
                 for left_tree, right_tree in izip(left_trees, right_trees):
@@ -142,14 +143,15 @@
                 text_index = left_repo._generate_text_key_index()
                 self.assertEqual(text_index,
                     right_repo._generate_text_key_index())
+                desired_files = []
                 for file_id, revision_id in text_index.iterkeys():
-                    left_weave = left_repo.weave_store.get_weave(
-                        file_id, left_repo.get_transaction())
-                    right_weave = right_repo.weave_store.get_weave(
-                        file_id, right_repo.get_transaction())
-                    self.assertEqual(
-                        left_weave.get_text(revision_id),
-                        right_weave.get_text(revision_id))
+                    desired_files.append(
+                        (file_id, revision_id, (file_id, revision_id)))
+                left_texts = list(left_repo.iter_files_bytes(desired_files))
+                right_texts = list(right_repo.iter_files_bytes(desired_files))
+                left_texts.sort()
+                right_texts.sort()
+                self.assertEqual(left_texts, right_texts)
                 # signatures
                 for rev_id in all_revs:
                     try:
@@ -306,7 +308,8 @@
                                      './.bzr/merge-hashes',
                                      './.bzr/repository',
                                      ])
-        self.assertRepositoryHasSameItems(tree.branch.repository, repo)
+        self.assertRepositoryHasSameItems(tree.branch.repository,
+            target.open_repository())
 
     def test_clone_bzrdir_repository_under_shared(self):
         tree = self.make_branch_and_tree('commit_tree')

=== modified file 'bzrlib/tests/interrepository_implementations/test_fetch.py'
--- bzrlib/tests/interrepository_implementations/test_fetch.py	2008-05-11 23:49:50 +0000
+++ bzrlib/tests/interrepository_implementations/test_fetch.py	2008-06-11 04:20:16 +0000
@@ -205,10 +205,10 @@
         to_repo.fetch(from_tree.branch.repository, tree_rev)
         # to_repo should have a file_graph for from_tree.path2id('subtree') and
         # revid tree_rev.
+        file_id = from_tree.path2id('subtree')
         to_repo.lock_read()
         try:
-            file_vf = to_repo.weave_store.get_weave(
-                from_tree.path2id('subtree'), to_repo.get_transaction())
-            self.assertEqual([tree_rev], file_vf.get_ancestry([tree_rev]))
+            self.assertEqual({(file_id, tree_rev):()},
+                to_repo.texts.get_parent_map([(file_id, tree_rev)]))
         finally:
             to_repo.unlock()

=== modified file 'bzrlib/tests/interrepository_implementations/test_interrepository.py'
--- bzrlib/tests/interrepository_implementations/test_interrepository.py	2008-05-11 23:49:50 +0000
+++ bzrlib/tests/interrepository_implementations/test_interrepository.py	2008-06-11 04:20:16 +0000
@@ -100,8 +100,8 @@
         # add a corrupt inventory 'orphan'
         tree_a.branch.repository.lock_write()
         tree_a.branch.repository.start_write_group()
-        inv_file = tree_a.branch.repository.get_inventory_weave()
-        inv_file.add_lines('orphan', [], [])
+        inv_file = tree_a.branch.repository.inventories
+        inv_file.add_lines(('orphan',), [], [])
         tree_a.branch.repository.commit_write_group()
         tree_a.branch.repository.unlock()
         # add a real revision 'rev1'
@@ -206,9 +206,7 @@
             inv.root.revision = revision_id
             root_id = inv.root.file_id
             sha1 = repo.add_inventory(revision_id, inv, parent_ids)
-            vf = repo.weave_store.get_weave_or_empty(root_id,
-                repo.get_transaction())
-            vf.add_lines(revision_id, [], [])
+            repo.texts.add_lines((root_id, revision_id), [], [])
             rev = bzrlib.revision.Revision(timestamp=0,
                                            timezone=None,
                                            committer="Foo Bar <foo@example.com>",

=== removed directory 'bzrlib/tests/interversionedfile_implementations'
=== removed file 'bzrlib/tests/interversionedfile_implementations/__init__.py'
--- bzrlib/tests/interversionedfile_implementations/__init__.py	2008-04-28 08:57:45 +0000
+++ bzrlib/tests/interversionedfile_implementations/__init__.py	1970-01-01 00:00:00 +0000
@@ -1,108 +0,0 @@
-# Copyright (C) 2006 Canonical Ltd
-# Authors: Robert Collins <robert.collins@canonical.com>
-#
-# This program is free software; you can redistribute it and/or modify
-# it under the terms of the GNU General Public License as published by
-# the Free Software Foundation; either version 2 of the License, or
-# (at your option) any later version.
-#
-# This program is distributed in the hope that it will be useful,
-# but WITHOUT ANY WARRANTY; without even the implied warranty of
-# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-# GNU General Public License for more details.
-#
-# You should have received a copy of the GNU General Public License
-# along with this program; if not, write to the Free Software
-# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
-
-
-"""InterVersioned implementation tests for bzr.
-
-These test the conformance of all the interversionedfile variations to the
-expected API including generally applicable corner cases.
-Specific tests for individual cases are in the tests/test_versionedfile.py file 
-rather than in tests/interversionedfile_implementations/*.py.
-"""
-
-from bzrlib.tests import (
-                          adapt_modules,
-                          default_transport,
-                          TestScenarioApplier,
-                          )
-
-
-class InterVersionedFileTestProviderAdapter(TestScenarioApplier):
-    """A tool to generate a suite testing multiple inter versioned-file classes.
-
-    This is done by copying the test once for each InterVersionedFile provider
-    and injecting the transport_server, transport_readonly_server,
-    versionedfile_factory and versionedfile_factory_to classes into each copy.
-    Each copy is also given a new id() to make it easy to identify.
-    """
-
-    def __init__(self, transport_server, transport_readonly_server, formats):
-        self._transport_server = transport_server
-        self._transport_readonly_server = transport_readonly_server
-        self.scenarios = self.formats_to_scenarios(formats)
-    
-    def formats_to_scenarios(self, formats):
-        """Transform the input formats to a list of scenarios.
-
-        :param formats: A list of tuples:
-            (interversionedfile_class, versionedfile_factory,
-             versionedfile_factory_to).
-        """
-        result = []
-        for (interversionedfile_class,
-             versionedfile_factory,
-             versionedfile_factory_to) in formats:
-            scenario = (interversionedfile_class.__name__, {
-                "transport_server":self._transport_server,
-                "transport_readonly_server":self._transport_readonly_server,
-                "interversionedfile_class":interversionedfile_class,
-                "versionedfile_factory":versionedfile_factory,
-                "versionedfile_factory_to":versionedfile_factory_to,
-                })
-            result.append(scenario)
-        return result
-
-    @staticmethod
-    def default_test_list():
-        """Generate the default list of interversionedfile permutations to test."""
-        from bzrlib.versionedfile import InterVersionedFile
-        from bzrlib.weave import WeaveFile
-        from bzrlib.knit import make_file_knit
-        result = []
-        # test the fallback InterVersionedFile from annotated knits to weave
-        result.append((InterVersionedFile,
-                       make_file_knit,
-                       WeaveFile))
-        for optimiser in InterVersionedFile._optimisers:
-            result.append((optimiser,
-                           optimiser._matching_file_from_factory,
-                           optimiser._matching_file_to_factory
-                           ))
-        # if there are specific combinations we want to use, we can add them 
-        # here.
-        return result
-
-
-def load_tests(basic_tests, module, loader):
-    result = loader.suiteClass()
-    # add the tests for this module
-    result.addTests(basic_tests)
-
-    test_interversionedfile_implementations = [
-        'bzrlib.tests.interversionedfile_implementations.test_join',
-        ]
-    adapter = InterVersionedFileTestProviderAdapter(
-        default_transport,
-        # None here will cause a readonly decorator to be created
-        # by the TestCaseWithTransport.get_readonly_transport method.
-        None,
-        InterVersionedFileTestProviderAdapter.default_test_list()
-        )
-    # add the tests for the sub modules
-    adapt_modules(test_interversionedfile_implementations,
-                  adapter, loader, result)
-    return result

=== removed file 'bzrlib/tests/interversionedfile_implementations/test_join.py'
--- bzrlib/tests/interversionedfile_implementations/test_join.py	2008-04-29 05:27:08 +0000
+++ bzrlib/tests/interversionedfile_implementations/test_join.py	1970-01-01 00:00:00 +0000
@@ -1,279 +0,0 @@
-# Copyright (C) 2006 Canonical Ltd
-#
-# This program is free software; you can redistribute it and/or modify
-# it under the terms of the GNU General Public License as published by
-# the Free Software Foundation; either version 2 of the License, or
-# (at your option) any later version.
-#
-# This program is distributed in the hope that it will be useful,
-# but WITHOUT ANY WARRANTY; without even the implied warranty of
-# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-# GNU General Public License for more details.
-#
-# You should have received a copy of the GNU General Public License
-# along with this program; if not, write to the Free Software
-# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
-
-"""Tests for join between versioned files."""
-
-
-import bzrlib.errors as errors
-from bzrlib.symbol_versioning import one_five
-from bzrlib.tests import TestCaseWithTransport
-from bzrlib.transport import get_transport
-import bzrlib.versionedfile as versionedfile
-
-
-class TestJoin(TestCaseWithTransport):
-    #Tests have self.versionedfile_factory and self.versionedfile_factory_to
-    #available to create source and target versioned files respectively.
-
-    def get_source(self, name='source'):
-        """Get a versioned file we will be joining from."""
-        return self.versionedfile_factory(name,
-                                          get_transport(self.get_url()),
-                                          create=True)
-
-    def get_target(self, name='target', create=True):
-        """"Get an empty versioned file to join into."""
-        return self.versionedfile_factory_to(name,
-                                             get_transport(self.get_url()),
-                                             create=create)
-
-    def test_join(self):
-        f1 = self.get_source()
-        f1.add_lines('r0', [], ['a\n', 'b\n'])
-        f1.add_lines('r1', ['r0'], ['c\n', 'b\n'])
-        f2 = self.get_target()
-        self.applyDeprecated(one_five, f2.join, f1, None)
-        def verify_file(f):
-            self.assertTrue(f.has_version('r0'))
-            self.assertTrue(f.has_version('r1'))
-        verify_file(f2)
-        verify_file(self.get_target())
-
-        self.assertRaises(errors.RevisionNotPresent,
-            self.applyDeprecated, one_five, f2.join, f1, version_ids=['r3'])
-
-    def test_gets_expected_inter_worker(self):
-        source = self.get_source()
-        target = self.get_target()
-        inter = versionedfile.InterVersionedFile.get(source, target)
-        self.assertTrue(isinstance(inter, self.interversionedfile_class))
-
-    def test_join_versions_joins_ancestors_not_siblings(self):
-        # joining with a version list should bring in ancestors of the
-        # named versions but not siblings thereof.
-        target = self.get_target()
-        target.add_lines('base', [], [])
-        source = self.get_source()
-        source.add_lines('base', [], [])
-        source.add_lines('sibling', ['base'], [])
-        source.add_lines('ancestorleft', ['base'], [])
-        source.add_lines('ancestorright', ['base'], [])
-        source.add_lines('namedleft', ['ancestorleft'], [])
-        source.add_lines('namedright', ['ancestorright'], [])
-        self.applyDeprecated(one_five, target.join, source,
-            version_ids=['namedleft', 'namedright'])
-        self.assertFalse(target.has_version('sibling'))
-        self.assertTrue(target.has_version('ancestorleft'))
-        self.assertTrue(target.has_version('ancestorright'))
-        self.assertTrue(target.has_version('namedleft'))
-        self.assertTrue(target.has_version('namedright'))
-
-    def test_join_different_parents_existing_version(self):
-        """This may either ignore or error."""
-        w1 = self.get_target('w1')
-        w2 = self.get_source('w2')
-        w1.add_lines('v-1', [], ['line 1\n'])
-        w2.add_lines('v-2', [], ['line 2\n'])
-        w1.add_lines('v-3', ['v-1'], ['line 1\n'])
-        w2.add_lines('v-3', ['v-2'], ['line 1\n'])
-        try:
-            self.applyDeprecated(one_five, w1.join, w2)
-        except errors.WeaveParentMismatch:
-            # Acceptable behaviour:
-            return
-        self.assertEqual(sorted(w1.versions()),
-                         'v-1 v-2 v-3'.split())
-        self.assertEqualDiff(w1.get_text('v-3'),
-                'line 1\n')
-        self.assertEqual({'v-3':('v-1',)}, w1.get_parent_map(['v-3']))
-        ann = list(w1.annotate('v-3'))
-        self.assertEqual(len(ann), 1)
-        self.assertEqual(ann[0][0], 'v-1')
-        self.assertEqual(ann[0][1], 'line 1\n')
-        
-    def build_weave1(self):
-        weave1 = self.get_source()
-        self.lines1 = ['hello\n']
-        self.lines3 = ['hello\n', 'cruel\n', 'world\n']
-        weave1.add_lines('v1', [], self.lines1)
-        weave1.add_lines('v2', ['v1'], ['hello\n', 'world\n'])
-        weave1.add_lines('v3', ['v2'], self.lines3)
-        return weave1
-        
-    def test_join_with_empty(self):
-        """Reweave adding empty weave"""
-        wb = self.get_target()
-        w1 = self.build_weave1()
-        self.applyDeprecated(one_five, w1.join, wb)
-        self.verify_weave1(w1)
-
-    def verify_weave1(self, w1):
-        self.assertEqual(sorted(w1.versions()), ['v1', 'v2', 'v3'])
-        self.assertEqual(w1.get_lines('v1'), ['hello\n'])
-        self.assertEqual({'v1':()}, w1.get_parent_map(['v1']))
-        self.assertEqual(w1.get_lines('v2'), ['hello\n', 'world\n'])
-        self.assertEqual({'v2':('v1',)}, w1.get_parent_map(['v2']))
-        self.assertEqual(w1.get_lines('v3'), ['hello\n', 'cruel\n', 'world\n'])
-        self.assertEqual({'v3':('v2',)}, w1.get_parent_map(['v3']))
-
-    def test_join_source_has_less_parents_preserves_parents(self):
-        # when the target has a text with more parent info, join 
-        # preserves that.
-        s = self.get_source()
-        s.add_lines('base', [], [])
-        s.add_lines('text', [], [])
-        t = self.get_target()
-        t.add_lines('base', [], [])
-        t.add_lines('text', ['base'], [])
-        self.applyDeprecated(one_five, t.join, s)
-        self.assertEqual({'text':('base',)}, t.get_parent_map(['text']))
-
-    def test_join_with_ghosts(self):
-        """Join that inserts parents of an existing revision.
-
-        This can happen when merging from another branch who knows about
-        revisions the destination does not, and the destinations index is
-        incorrect because it was or is using a ghost-unaware format to
-        represent the index. In this test the second weave knows of an
-        additional parent of v2.
-        
-        However v2 must not be changed because we consider indexes immutable:
-        instead a check or reconcile operation locally should pickup that v2 is
-        wrong and regenerate the index at a later time. So either this errors,
-        or leaves v2 unaltered.
-        """
-        w1 = self.build_weave1()
-        wb = self.get_target()
-        wb.add_lines('x1', [], ['line from x1\n'])
-        wb.add_lines('v1', [], ['hello\n'])
-        wb.add_lines('v2', ['v1', 'x1'], ['hello\n', 'world\n'])
-        try:
-            self.applyDeprecated(one_five, w1.join, wb)
-        except errors.WeaveParentMismatch:
-            # Acceptable behaviour:
-            return
-        self.assertEqual(['v1', 'v2', 'v3', 'x1',], sorted(w1.versions()))
-        self.assertEqual('line from x1\n', w1.get_text('x1'))
-        self.assertEqual(['hello\n', 'world\n'], w1.get_lines('v2'))
-        self.assertEqual({'v2':('v1',)}, w1.get_parent_map(['v2']))
-
-    def test_join_with_ignore_missing_versions(self):
-        # test that ignore_missing=True makes a listed but absent version id
-        # be ignored, and that unlisted version_ids are not integrated.
-        w1 = self.build_weave1()
-        wb = self.get_target()
-        wb.add_lines('x1', [], ['line from x1\n'])
-        wb.add_lines('v1', [], ['hello\n'])
-        wb.add_lines('v2', ['v1', 'x1'], ['hello\n', 'world\n'])
-        self.applyDeprecated(one_five, w1.join, wb, version_ids=['x1', 'z1'],
-            ignore_missing=True)
-        eq = self.assertEquals
-        eq(sorted(w1.versions()), ['v1', 'v2', 'v3', 'x1'])
-        eq(w1.get_text('x1'), 'line from x1\n')
-        eq(w1.get_lines('v2'), ['hello\n', 'world\n'])
-        self.assertEqual({'v2':('v1',)}, w1.get_parent_map(['v2']))
-    
-    def build_source_weave(self, name, *pattern):
-        w = self.get_source(name)
-        for version, parents in pattern:
-            w.add_lines(version, parents, [])
-        return w
-
-    def build_target_weave(self, name, *pattern):
-        w = self.get_target(name)
-        for version, parents in pattern:
-            w.add_lines(version, parents, [])
-        return w
-        
-    def test_joining_ghosts(self):
-        # some versioned file formats allow lines to be added with parent
-        # information that is > than that in the format. Formats that do
-        # not support this need to raise NotImplementedError on the
-        # add_lines_with_ghosts api.
-        # files with ghost information when joined into a file which
-        # supports that must preserve it, when joined into a file which
-        # does not must discard it, and when filling a ghost for a listed
-        # ghost must reconcile it
-        source = self.get_source()
-        target = self.get_target()
-        # try filling target with ghosts and filling in reverse -  
-        try:
-            target.add_lines_with_ghosts('notbase', ['base'], [])
-        except NotImplementedError:
-            # The target does not support ghosts; the test is irrelevant.
-            return
-        try:
-            self.applyDeprecated(one_five, source.join, target)
-        except errors.RevisionNotPresent:
-            return
-        # legacy apis should behave
-        self.assertEqual(['notbase'], source.get_ancestry(['notbase']))
-        self.assertFalse(source.has_version('base'))
-        # ghost data should have been preserved
-        self.assertEqual(['base', 'notbase'], source.get_ancestry_with_ghosts(['notbase']))
-        self.assertEqual(['base'], source.get_parents_with_ghosts('notbase'))
-        self.assertEqual({'notbase':('base',)},
-            source.get_parent_map(source.versions()))
-
-        # if we add something that is fills out what is a ghost, then 
-        # when joining into a ghost aware join it should flesh out the ghosts.
-        source.add_lines('base', [], [])
-        self.applyDeprecated(one_five, target.join, source, version_ids=['base'])
-        self.assertEqual(['base', 'notbase'], target.get_ancestry(['notbase']))
-        self.assertEqual({'base':(),
-                          'notbase':('base', ),
-                          },
-                         target.get_parent_map(target.versions()))
-        self.assertTrue(target.has_version('base'))
-        # we have _with_ghost apis to give us ghost information.
-        self.assertEqual(['base', 'notbase'], target.get_ancestry_with_ghosts(['notbase']))
-        self.assertEqual(['base'], target.get_parents_with_ghosts('notbase'))
-        self.assertEqual({'base':(),
-                          'notbase':('base',),
-                          },
-            source.get_parent_map(source.versions()))
-
-    def test_restricted_join_into_empty(self):
-        # joining into an empty versioned file with a version_ids list
-        # should only grab the selected versions.
-        source = self.get_source()
-        source.add_lines('skip_me', [], ['a\n'])
-        source.add_lines('inherit_me', [], ['b\n'])
-        source.add_lines('select_me', ['inherit_me'], ['b\n'])
-        target = self.get_target()
-        self.applyDeprecated(one_five, target.join, source, version_ids=['select_me'])
-        self.assertEqual(['inherit_me', 'select_me'], target.versions())
-
-    def test_join_odd_records(self):
-        # this tests that joining the 1st, 3rd and 5th records and not the
-        # 2nd and 4th works, particularly after reopening the file.
-        # this test is designed to test versioned files that perform
-        # optimisations on the join. Grabbing partial data and reopening the
-        # file make it likely to trigger a fault.
-        source = self.get_source()
-        source.add_lines('1', [], ['1st\n'])
-        source.add_lines('2', [], ['2nd\n'])
-        source.add_lines('3', ['1'], ['1st\n', '2nd\n'])
-        source.add_lines('4', ['2'], ['1st\n'])
-        source.add_lines('5', ['3'], ['1st\n', '2nd\n', '3rd\n'])
-        target = self.get_target()
-        self.applyDeprecated(one_five, target.join, source, version_ids=['1', '3', '5'])
-        target = self.get_target(create=False)
-        self.assertEqual(set(['1', '3', '5']), set(target.versions()))
-        self.assertEqual(3, len(target.versions()))
-        self.assertEqual(['1st\n'], target.get_lines('1'))
-        self.assertEqual(['1st\n', '2nd\n'], target.get_lines('3'))
-        self.assertEqual(['1st\n', '2nd\n', '3rd\n'], target.get_lines('5'))

=== modified file 'bzrlib/tests/repository_implementations/__init__.py'
--- bzrlib/tests/repository_implementations/__init__.py	2008-04-28 08:57:45 +0000
+++ bzrlib/tests/repository_implementations/__init__.py	2008-06-11 04:20:16 +0000
@@ -369,7 +369,6 @@
         else:
             count = 3
         return [
-            "%d inconsistent parents" % count,
             # will be gc'd
             r"unreferenced version: {rev2} in a-file-id",
             r"unreferenced version: {rev2b} in a-file-id",
@@ -380,6 +379,7 @@
             r"but should have \('rev2c',\)",
             r"a-file-id version rev4 has parents \('rev2',\) "
             r"but should have \('rev1a',\)",
+            "%d inconsistent parents" % count,
             ]
 
     def populate_repository(self, repo):

=== modified file 'bzrlib/tests/repository_implementations/helpers.py'
--- bzrlib/tests/repository_implementations/helpers.py	2007-11-03 13:47:54 +0000
+++ bzrlib/tests/repository_implementations/helpers.py	2008-06-11 04:20:16 +0000
@@ -58,22 +58,17 @@
             inv_sha1 = repo.add_inventory('revision-id', inv, [])
             if repo.supports_rich_root():
                 root_id = inv.root.file_id
-                vf = repo.weave_store.get_weave_or_empty(root_id,
-                    repo.get_transaction())
-                vf.add_lines('revision-id', [], [])
+                repo.texts.add_lines((root_id, 'revision-id'), [], [])
             revision = _mod_revision.Revision('revision-id',
                 committer='jrandom@example.com', timestamp=0,
                 inventory_sha1=inv_sha1, timezone=0, message='message',
                 parent_ids=[])
             # Manually add the revision text using the RevisionStore API, with
             # bad parents.
-            rev_tmp = StringIO()
-            repo._revision_store._serializer.write_revision(revision, rev_tmp)
-            rev_tmp.seek(0)
-            repo._revision_store.get_revision_file(repo.get_transaction()
-                ).add_lines_with_ghosts(revision.revision_id,
-                ['incorrect-parent'],
-                osutils.split_lines(rev_tmp.read()))
+            rev_text = repo._serializer.write_revision_to_string(revision)
+            repo.revisions.add_lines((revision.revision_id,),
+                [('incorrect-parent',)],
+                osutils.split_lines(rev_text))
         except:
             repo.abort_write_group()
             repo.unlock()

=== modified file 'bzrlib/tests/repository_implementations/test_check.py'
--- bzrlib/tests/repository_implementations/test_check.py	2007-11-27 21:58:36 +0000
+++ bzrlib/tests/repository_implementations/test_check.py	2008-06-11 04:20:16 +0000
@@ -20,7 +20,6 @@
 
 from bzrlib import (
     errors,
-    inventory,
     revision as _mod_revision,
     )
 from bzrlib.tests import TestNotApplicable

=== modified file 'bzrlib/tests/repository_implementations/test_check_reconcile.py'
--- bzrlib/tests/repository_implementations/test_check_reconcile.py	2008-04-08 03:39:43 +0000
+++ bzrlib/tests/repository_implementations/test_check_reconcile.py	2008-06-11 04:20:16 +0000
@@ -60,9 +60,7 @@
         inv.root.revision = revision_id
         if repo.supports_rich_root():
             root_id = inv.root.file_id
-            vf = repo.weave_store.get_weave_or_empty(root_id,
-                repo.get_transaction())
-            vf.add_lines(revision_id, [], [])
+            repo.texts.add_lines((root_id, revision_id), [], [])
         repo.add_inventory(revision_id, inv, parent_ids)
         revision = Revision(revision_id, committer='jrandom@example.com',
             timestamp=0, inventory_sha1='', timezone=0, message='foo',
@@ -103,9 +101,8 @@
         entry.text_sha1 = sha.sha(file_contents).hexdigest()
         inv.add(entry)
         if make_file_version:
-            vf = repo.weave_store.get_weave_or_empty(file_id,
-                                                     repo.get_transaction())
-            vf.add_lines(revision, parents, [file_contents])
+            repo.texts.add_lines((file_id, revision),
+                [(file_id, parent) for parent in parents], [file_contents])
         return inv
 
     def require_repo_suffers_text_parent_corruption(self, repo):
@@ -114,13 +111,13 @@
                     "Format does not support text parent reconciliation")
 
     def file_parents(self, repo, revision_id):
-        return repo.weave_store.get_weave('a-file-id',
-            repo.get_transaction()).get_parent_map([revision_id])[revision_id]
+        key = ('a-file-id', revision_id)
+        parent_map = repo.texts.get_parent_map([key])
+        return tuple(parent[-1] for parent in parent_map[key])
 
     def assertFileVersionAbsent(self, repo, revision_id):
-        self.assertFalse(repo.weave_store.get_weave('a-file-id',
-            repo.get_transaction()).has_version(revision_id),
-            'File version %s wrongly present.' % (revision_id,))
+        self.assertEqual({},
+            repo.texts.get_parent_map([('a-file-id', revision_id)]))
 
     def assertParentsMatch(self, expected_parents_for_versions, repo,
                            when_description):
@@ -152,8 +149,8 @@
 
         :returns: A dict of `{version: hash}`.
         """
-        vf = repo.weave_store.get_weave('a-file-id', repo.get_transaction())
-        return dict(zip(versions, vf.get_sha1s(versions)))
+        keys = [('a-file-id', version) for version in versions]
+        return dict(zip(versions, repo.texts.get_sha1s(keys)))
 
     def test_reconcile_behaviour(self):
         """Populate a repository and reconcile it, verifying the state before
@@ -180,12 +177,20 @@
                 self.shas_for_versions_of_file(
                     repo, scenario.all_versions_after_reconcile()))
 
+            # Scenario.corrected_fulltexts contains texts which the test wants
+            # to assert are now fulltexts. However this is an abstraction
+            # violation; really we care that:
+            # - the text is reconstructable
+            # - it has an empty parents list
+            # (we specify it this way because a store can use arbitrary
+            # compression pointers in principle.
             for file_version in scenario.corrected_fulltexts():
-                vf = repo.weave_store.get_weave(
-                    'a-file-id', repo.get_transaction())
-                self.assertEqual('fulltext',
-                    vf._index.get_method(file_version),
-                    '%r should be fulltext' % (file_version,))
+                key = ('a-file-id', file_version)
+                self.assertEqual({key:()}, repo.texts.get_parent_map([key]))
+                self.assertIsInstance(
+                    repo.texts.get_record_stream([key], 'unordered',
+                        True).next().get_bytes_as('fulltext'),
+                    str)
         finally:
             repo.unlock()
 

=== modified file 'bzrlib/tests/repository_implementations/test_commit_builder.py'
--- bzrlib/tests/repository_implementations/test_commit_builder.py	2007-10-26 19:18:48 +0000
+++ bzrlib/tests/repository_implementations/test_commit_builder.py	2008-06-11 04:20:16 +0000
@@ -28,6 +28,7 @@
     repository,
     tests,
     )
+from bzrlib.graph import Graph
 from bzrlib.tests.repository_implementations import test_repository
 
 
@@ -243,7 +244,10 @@
         tree1, tree2 = self._get_revtrees(tree, [rev1, rev2])
         self.assertEqual(rev1, tree1.inventory[name + 'id'].revision)
         self.assertEqual(rev1, tree2.inventory[name + 'id'].revision)
-        self.assertFileAncestry([rev1], tree, name)
+        file_id = name + 'id'
+        expected_graph = {}
+        expected_graph[(file_id, rev1)] = ()
+        self.assertFileGraph(expected_graph, tree, (file_id, rev1))
 
     def test_last_modified_revision_after_commit_dir_unchanged(self):
         # committing without changing a dir does not change the last modified.
@@ -264,7 +268,10 @@
         tree1, tree2 = self._get_revtrees(tree, [rev1, rev2])
         self.assertEqual(rev1, tree1.inventory['dirid'].revision)
         self.assertEqual(rev1, tree2.inventory['dirid'].revision)
-        self.assertFileAncestry([rev1], tree, 'dir')
+        file_id = 'dirid'
+        expected_graph = {}
+        expected_graph[(file_id, rev1)] = ()
+        self.assertFileGraph(expected_graph, tree, (file_id, rev1))
 
     def test_last_modified_revision_after_commit_file_unchanged(self):
         # committing without changing a file does not change the last modified.
@@ -337,7 +344,11 @@
         tree1, tree2 = self._get_revtrees(tree, [rev1, rev2])
         self.assertEqual(rev1, tree1.inventory[name + 'id'].revision)
         self.assertEqual(rev2, tree2.inventory[name + 'id'].revision)
-        self.assertFileAncestry([rev1, rev2], tree, name)
+        file_id = name + 'id'
+        expected_graph = {}
+        expected_graph[(file_id, rev1)] = ()
+        expected_graph[(file_id, rev2)] = ((file_id, rev1),)
+        self.assertFileGraph(expected_graph, tree, (file_id, rev2))
 
     def mini_commit(self, tree, name, new_name, records_version=True,
         delta_against_basis=True):
@@ -407,19 +418,13 @@
             tree.unlock()
         return rev2
 
-    def assertFileAncestry(self, ancestry, tree, name, alt_ancestry=None):
+    def assertFileGraph(self, expected_graph, tree, tip):
         # all the changes that have occured should be in the ancestry
         # (closest to a public per-file graph API we have today)
         tree.lock_read()
         self.addCleanup(tree.unlock)
-        vw = tree.branch.repository.weave_store.get_weave(name + 'id',
-            tree.branch.repository.get_transaction())
-        result = vw.get_ancestry([ancestry[-1]])
-        if alt_ancestry is None:
-            self.assertEqual(ancestry, result)
-        else:
-            self.assertSubset([tuple(result)],
-                [tuple(ancestry), tuple(alt_ancestry)])
+        graph = dict(Graph(tree.branch.repository.texts).iter_ancestry([tip]))
+        self.assertEqual(expected_graph, graph)
 
     def test_last_modified_revision_after_content_file_changes(self):
         # altering a file changes the last modified.
@@ -457,10 +462,13 @@
         rev4 = self.mini_commit(tree1, 'new_' + name, 'new_' + name)
         tree3, = self._get_revtrees(tree1, [rev4])
         self.assertEqual(rev4, tree3.inventory[name + 'id'].revision)
-        # TODO: change this to an assertFileGraph call to check the
-        # parent order of rev4: it should be rev2, rev3
-        self.assertFileAncestry([rev1, rev2, rev3, rev4], tree1, name,
-            [rev1, rev3, rev2, rev4])
+        file_id = name + 'id'
+        expected_graph = {}
+        expected_graph[(file_id, rev1)] = ()
+        expected_graph[(file_id, rev2)] = ((file_id, rev1),)
+        expected_graph[(file_id, rev3)] = ((file_id, rev1),)
+        expected_graph[(file_id, rev4)] = ((file_id, rev2), (file_id, rev3),)
+        self.assertFileGraph(expected_graph, tree1, (file_id, rev4))
 
     def test_last_modified_revision_after_merge_dir_changes(self):
         # merge a dir changes the last modified.
@@ -489,7 +497,11 @@
         rev3 = self.mini_commit(tree1, name, 'new_' + name, False)
         tree3, = self._get_revtrees(tree1, [rev2])
         self.assertEqual(rev2, tree3.inventory[name + 'id'].revision)
-        self.assertFileAncestry([rev1, rev2], tree1, name)
+        file_id = name + 'id'
+        expected_graph = {}
+        expected_graph[(file_id, rev1)] = ()
+        expected_graph[(file_id, rev2)] = ((file_id, rev1),)
+        self.assertFileGraph(expected_graph, tree1, (file_id, rev2))
 
     def test_last_modified_revision_after_converged_merge_dir_changes(self):
         # merge a dir changes the last modified.

=== modified file 'bzrlib/tests/repository_implementations/test_fetch.py'
--- bzrlib/tests/repository_implementations/test_fetch.py	2008-04-25 22:12:34 +0000
+++ bzrlib/tests/repository_implementations/test_fetch.py	2008-06-11 04:20:16 +0000
@@ -69,7 +69,11 @@
         # disk.
         knit3_repo = b_bzrdir.open_repository()
         rev1_tree = knit3_repo.revision_tree('rev1')
-        lines = rev1_tree.get_file_lines(rev1_tree.get_root_id())
+        rev1_tree.lock_read()
+        try:
+            lines = rev1_tree.get_file_lines(rev1_tree.get_root_id())
+        finally:
+            rev1_tree.unlock()
         self.assertEqual([], lines)
         b_branch = b_bzrdir.create_branch()
         b_branch.pull(tree_a.branch)

=== modified file 'bzrlib/tests/repository_implementations/test_reconcile.py'
--- bzrlib/tests/repository_implementations/test_reconcile.py	2008-03-26 21:42:35 +0000
+++ bzrlib/tests/repository_implementations/test_reconcile.py	2008-06-11 04:20:16 +0000
@@ -47,10 +47,11 @@
     def checkNoBackupInventory(self, aBzrDir):
         """Check that there is no backup inventory in aBzrDir."""
         repo = aBzrDir.open_repository()
-        self.assertRaises(errors.NoSuchFile,
-                          repo.control_weaves.get_weave,
-                          'inventory.backup',
-                          repo.get_transaction())
+        # Remote repository, and possibly others, do not have 
+        # _transport.
+        if getattr(repo, '_transport', None) is not None:
+            for path in repo._transport.list_dir('.'):
+                self.assertFalse('inventory.backup' in path)
 
 
 class TestsNeedingReweave(TestReconcile):
@@ -76,9 +77,7 @@
             inv.root.revision = revision_id
             root_id = inv.root.file_id
             sha1 = repo.add_inventory(revision_id, inv, parent_ids)
-            vf = repo.weave_store.get_weave_or_empty(root_id,
-                repo.get_transaction())
-            vf.add_lines(revision_id, [], [])
+            repo.texts.add_lines((root_id, revision_id), [], [])
             rev = bzrlib.revision.Revision(timestamp=0,
                                            timezone=None,
                                            committer="Foo Bar <foo@example.com>",
@@ -195,11 +194,14 @@
 
     def check_missing_was_removed(self, repo):
         if repo._reconcile_backsup_inventory:
-            backup = repo.control_weaves.get_weave('inventory.backup',
-                                                   repo.get_transaction())
-            self.assertTrue('missing' in backup.versions())
-        self.assertRaises(errors.RevisionNotPresent,
-                          repo.get_inventory, 'missing')
+            backed_up = False
+            for path in repo._transport.list_dir('.'):
+                if 'inventory.backup' in path:
+                    backed_up = True
+            self.assertTrue(backed_up)
+            # Not clear how to do this at an interface level:
+            # self.assertTrue('missing' in backup.versions())
+        self.assertRaises(errors.NoSuchRevision, repo.get_inventory, 'missing')
 
     def test_reweave_inventory_without_revision_reconciler(self):
         # smoke test for the all in one Reconciler class,
@@ -325,9 +327,7 @@
         inv.root.revision = 'wrong-secondary-parent'
         if repo.supports_rich_root():
             root_id = inv.root.file_id
-            vf = repo.weave_store.get_weave_or_empty(root_id,
-                repo.get_transaction())
-            vf.add_lines('wrong-secondary-parent', [], [])
+            repo.texts.add_lines((root_id, 'wrong-secondary-parent'), [], [])
         sha1 = repo.add_inventory('wrong-secondary-parent', inv, ['1', '3', '2'])
         rev = Revision(timestamp=0,
                        timezone=None,

=== modified file 'bzrlib/tests/repository_implementations/test_repository.py'
--- bzrlib/tests/repository_implementations/test_repository.py	2008-05-12 02:40:40 +0000
+++ bzrlib/tests/repository_implementations/test_repository.py	2008-06-11 04:20:16 +0000
@@ -64,6 +64,139 @@
 
 class TestRepository(TestCaseWithRepository):
 
+    def test_attribute_inventories_store(self):
+        """Test the existence of the inventories attribute."""
+        tree = self.make_branch_and_tree('tree')
+        repo = tree.branch.repository
+        self.assertIsInstance(repo.inventories,
+            bzrlib.versionedfile.VersionedFiles)
+
+    def test_attribute_inventories_basics(self):
+        """Test basic aspects of the inventories attribute."""
+        tree = self.make_branch_and_tree('tree')
+        repo = tree.branch.repository
+        rev_id = (tree.commit('a'),)
+        tree.lock_read()
+        self.addCleanup(tree.unlock)
+        self.assertEqual(set([rev_id]), set(repo.inventories.keys()))
+
+    def test_attribute_revision_store(self):
+        """Test the existence of the revisions attribute."""
+        tree = self.make_branch_and_tree('tree')
+        repo = tree.branch.repository
+        self.assertIsInstance(repo.revisions,
+            bzrlib.versionedfile.VersionedFiles)
+
+    def test_attribute_revision_store_basics(self):
+        """Test the basic behaviour of the revisions attribute."""
+        tree = self.make_branch_and_tree('tree')
+        repo = tree.branch.repository
+        repo.lock_write()
+        try:
+            self.assertEqual(set(), set(repo.revisions.keys()))
+            revid = (tree.commit("foo"),)
+            self.assertEqual(set([revid]), set(repo.revisions.keys()))
+            self.assertEqual({revid:()},
+                repo.revisions.get_parent_map([revid]))
+        finally:
+            repo.unlock()
+        tree2 = self.make_branch_and_tree('tree2')
+        tree2.pull(tree.branch)
+        left_id = (tree2.commit('left'),)
+        right_id = (tree.commit('right'),)
+        tree.merge_from_branch(tree2.branch)
+        merge_id = (tree.commit('merged'),)
+        repo.lock_read()
+        self.addCleanup(repo.unlock)
+        self.assertEqual(set([revid, left_id, right_id, merge_id]),
+            set(repo.revisions.keys()))
+        self.assertEqual({revid:(), left_id:(revid,), right_id:(revid,),
+             merge_id:(right_id, left_id)},
+            repo.revisions.get_parent_map(repo.revisions.keys()))
+
+    def test_attribute_signature_store(self):
+        """Test the existence of the signatures attribute."""
+        tree = self.make_branch_and_tree('tree')
+        repo = tree.branch.repository
+        self.assertIsInstance(repo.signatures,
+            bzrlib.versionedfile.VersionedFiles)
+
+    def test_attribute_text_store_basics(self):
+        """Test the basic behaviour of the text store."""
+        tree = self.make_branch_and_tree('tree')
+        repo = tree.branch.repository
+        file_id = ("Foo:Bar",)
+        tree.lock_write()
+        try:
+            self.assertEqual(set(), set(repo.texts.keys()))
+            tree.add(['foo'], file_id, ['file'])
+            tree.put_file_bytes_non_atomic(file_id[0], 'content\n')
+            revid = (tree.commit("foo"),)
+            if repo._format.rich_root_data:
+                root_commit = (tree.get_root_id(),) + revid
+                keys = set([root_commit])
+                parents = {root_commit:()}
+            else:
+                keys = set()
+                parents = {}
+            keys.add(file_id + revid)
+            parents[file_id + revid] = ()
+            self.assertEqual(keys, set(repo.texts.keys()))
+            self.assertEqual(parents,
+                repo.texts.get_parent_map(repo.texts.keys()))
+        finally:
+            tree.unlock()
+        tree2 = self.make_branch_and_tree('tree2')
+        tree2.pull(tree.branch)
+        tree2.put_file_bytes_non_atomic('Foo:Bar', 'right\n')
+        right_id = (tree2.commit('right'),)
+        keys.add(file_id + right_id)
+        parents[file_id + right_id] = (file_id + revid,)
+        tree.put_file_bytes_non_atomic('Foo:Bar', 'left\n')
+        left_id = (tree.commit('left'),)
+        keys.add(file_id + left_id)
+        parents[file_id + left_id] = (file_id + revid,)
+        tree.merge_from_branch(tree2.branch)
+        tree.put_file_bytes_non_atomic('Foo:Bar', 'merged\n')
+        try:
+            tree.auto_resolve()
+        except errors.UnsupportedOperation:
+            pass
+        merge_id = (tree.commit('merged'),)
+        keys.add(file_id + merge_id)
+        parents[file_id + merge_id] = (file_id + left_id, file_id + right_id)
+        repo.lock_read()
+        self.addCleanup(repo.unlock)
+        self.assertEqual(keys, set(repo.texts.keys()))
+        self.assertEqual(parents, repo.texts.get_parent_map(repo.texts.keys()))
+
+    def test_attribute_text_store(self):
+        """Test the existence of the texts attribute."""
+        tree = self.make_branch_and_tree('tree')
+        repo = tree.branch.repository
+        self.assertIsInstance(repo.texts,
+            bzrlib.versionedfile.VersionedFiles)
+
+    def test_exposed_versioned_files_are_marked_dirty(self):
+        repo = self.make_repository('.')
+        repo.lock_write()
+        signatures = repo.signatures
+        revisions = repo.revisions
+        inventories = repo.inventories
+        repo.unlock()
+        self.assertRaises(errors.ObjectNotLocked,
+            signatures.keys)
+        self.assertRaises(errors.ObjectNotLocked,
+            revisions.keys)
+        self.assertRaises(errors.ObjectNotLocked,
+            inventories.keys)
+        self.assertRaises(errors.ObjectNotLocked,
+            signatures.add_lines, ('foo',), [], [])
+        self.assertRaises(errors.ObjectNotLocked,
+            revisions.add_lines, ('foo',), [], [])
+        self.assertRaises(errors.ObjectNotLocked,
+            inventories.add_lines, ('foo',), [], [])
+
     def test_clone_to_default_format(self):
         #TODO: Test that cloning a repository preserves all the information
         # such as signatures[not tested yet] etc etc.
@@ -367,126 +500,6 @@
         repo._format.rich_root_data
         repo._format.supports_tree_reference
 
-    def test_get_data_stream_deprecated(self):
-        # If get_data_stream is present it must be deprecated
-        tree = self.make_branch_and_tree('t')
-        self.build_tree(['t/foo'])
-        tree.add('foo', 'file1')
-        tree.commit('message', rev_id='rev_id')
-        repo = tree.branch.repository
-        repo.lock_read()
-        self.addCleanup(repo.unlock)
-        try:
-            stream = self.applyDeprecated(one_two, repo.get_data_stream,
-                ['rev_id'])
-        except NotImplementedError:
-            raise TestNotApplicable("%s doesn't support get_data_stream"
-                % repo._format)
-        except AttributeError:
-            pass
-
-    def test_get_data_stream_for_search(self):
-        # Make a repo with a revision
-        tree = self.make_branch_and_tree('t')
-        self.build_tree(['t/foo'])
-        tree.add('foo', 'file1')
-        tree.commit('message', rev_id='rev_id')
-        repo = tree.branch.repository
-
-        # Get a data stream (a file-like object) for that revision
-        search = graph.SearchResult(set(['rev_id']), set([NULL_REVISION]), 1,
-            set(['rev_id']))
-        repo.lock_read()
-        self.addCleanup(repo.unlock)
-        try:
-            stream = repo.get_data_stream_for_search(search)
-        except NotImplementedError:
-            raise TestNotApplicable("%s doesn't support get_data_stream"
-                % repo._format)
-
-        # The data stream is a iterator that yields (name, versioned_file)
-        # pairs for:
-        #   * the file knit (or knits; if this repo has rich roots there will
-        #     be a file knit for that as well as for 'file1').
-        #   * the inventory knit
-        #   * the revisions knit
-        # in that order.
-        expected_record_names = [
-            ('file', 'file1'),
-            ('inventory',),
-            ('signatures',),
-            ('revisions',)]
-        streamed_names = []
-        for name, bytes in stream:
-            streamed_names.append(name)
-
-        if repo.supports_rich_root():
-            # Check for the root versioned file in the stream, then remove it
-            # from streamed_names so we can compare that with
-            # expected_record_names.
-            # Note that the file knits can be in any order, so this test is
-            # written to allow that.
-            inv = repo.get_inventory('rev_id')
-            expected_record_name = ('file', inv.root.file_id)
-            self.assertTrue(expected_record_name in streamed_names)
-            streamed_names.remove(expected_record_name)
-
-        self.assertEqual(expected_record_names, streamed_names)
-
-    def test_insert_data_stream(self):
-        tree = self.make_branch_and_tree('source')
-        self.build_tree(['source/foo'])
-        tree.add('foo', 'file1')
-        tree.commit('message', rev_id='rev_id')
-        source_repo = tree.branch.repository
-        dest_repo = self.make_repository('dest')
-        search = dest_repo.search_missing_revision_ids(source_repo,
-            revision_id='rev_id')
-        source_repo.lock_read()
-        self.addCleanup(source_repo.unlock)
-        try:
-            stream = source_repo.get_data_stream_for_search(search)
-        except NotImplementedError, e:
-            # Not all repositories support streaming.
-            self.assertContainsRe(str(e), 'get_data_stream_for_search')
-            raise TestSkipped('This format does not support streaming.')
-
-        dest_repo.lock_write()
-        try:
-            dest_repo.start_write_group()
-            try:
-                dest_repo.insert_data_stream(stream)
-            except:
-                dest_repo.abort_write_group()
-                raise
-            else:
-                dest_repo.commit_write_group()
-        finally:
-            dest_repo.unlock()
-        # reopen to be sure it was added.
-        dest_repo = dest_repo.bzrdir.open_repository()
-        self.assertTrue(dest_repo.has_revision('rev_id'))
-
-        # insert the same data stream again, should be no-op
-        stream = source_repo.get_data_stream_for_search(search)
-        dest_repo.lock_write()
-        try:
-            dest_repo.start_write_group()
-            try:
-                dest_repo.insert_data_stream(stream)
-            except:
-                dest_repo.abort_write_group()
-                raise
-            else:
-                dest_repo.commit_write_group()
-        finally:
-            dest_repo.unlock()
-
-        # try to insert data stream with invalid key
-        stream = [[('bogus-key',), '']]
-        self.assertRaises(errors.RepositoryDataStreamError,
-                          dest_repo.insert_data_stream, stream)
-
     def test_get_serializer_format(self):
         repo = self.make_repository('.')
         format = repo.get_serializer_format()
@@ -860,9 +873,9 @@
         tree_a.lock_write()
         try:
             tree_a.branch.repository.start_write_group()
-            inv_file = tree_a.branch.repository.get_inventory_weave()
             try:
-                inv_file.add_lines('orphan', [], [])
+                inv_file = tree_a.branch.repository.inventories
+                inv_file.add_lines(('orphan',), [], [])
             except:
                 tree_a.branch.repository.commit_write_group()
                 raise
@@ -876,7 +889,10 @@
         tree_a.commit('rev2', rev_id='rev2', allow_pointless=True)
         # add a reference to a ghost
         tree_a.add_parent_tree_id('ghost1')
-        tree_a.commit('rev3', rev_id='rev3', allow_pointless=True)
+        try:
+            tree_a.commit('rev3', rev_id='rev3', allow_pointless=True)
+        except errors.RevisionNotPresent:
+            raise TestNotApplicable("Cannot test with ghosts for this format.")
         # add another reference to a ghost, and a second ghost.
         tree_a.add_parent_tree_id('ghost1')
         tree_a.add_parent_tree_id('ghost2')
@@ -906,8 +922,8 @@
 
     def test_all_revision_ids(self):
         # all_revision_ids -> all revisions
-        self.assertEqual(['rev1', 'rev2', 'rev3', 'rev4'],
-                         self.bzrdir.open_repository().all_revision_ids())
+        self.assertEqual(set(['rev1', 'rev2', 'rev3', 'rev4']),
+            set(self.bzrdir.open_repository().all_revision_ids()))
 
     def test_get_ancestry_missing_revision(self):
         # get_ancestry(revision that is in some data but not fully installed
@@ -953,7 +969,10 @@
                                        inventory_sha1=sha1,
                                        revision_id='ghost')
         rev.parent_ids = ['the_ghost']
-        repo.add_revision('ghost', rev)
+        try:
+            repo.add_revision('ghost', rev)
+        except (errors.NoSuchRevision, errors.RevisionNotPresent):
+            raise TestNotApplicable("Cannot test with ghosts for this format.")
          
         inv = Inventory(revision_id = 'the_ghost')
         inv.root.revision = 'the_ghost'
@@ -967,8 +986,10 @@
         rev.parent_ids = []
         repo.add_revision('the_ghost', rev)
         # check its setup usefully
-        inv_weave = repo.get_inventory_weave()
-        self.assertEqual(['ghost'], inv_weave.get_ancestry(['ghost']))
+        inv_weave = repo.inventories
+        possible_parents = (None, (('ghost',),))
+        self.assertSubset(inv_weave.get_parent_map([('ghost',)])[('ghost',)],
+            possible_parents)
         repo.commit_write_group()
         repo.unlock()
 

=== modified file 'bzrlib/tests/repository_implementations/test_statistics.py'
--- bzrlib/tests/repository_implementations/test_statistics.py	2007-02-03 10:56:04 +0000
+++ bzrlib/tests/repository_implementations/test_statistics.py	2008-06-11 04:20:16 +0000
@@ -21,16 +21,6 @@
 
 class TestGatherStats(TestCaseWithRepository):
 
-    def check_stats_has_size(self, stats):
-        """Check that stats has a reasonable size entry."""
-        # actual disk size varies from implementation to implementation,
-        # but they should all provide it on their native transport.
-        self.assertTrue('size' in stats)
-        # and it should be a number
-        self.assertIsInstance(stats['size'], (int, long))
-        # and now remove it to make other assertions work without variation.
-        del stats['size']
-
     def test_gather_stats(self):
         """First smoke test covering the refactoring into the Repository api."""
         tree = self.make_branch_and_memory_tree('.')
@@ -48,7 +38,6 @@
         # now, in the same repository, asking for stats with/without the 
         # committers flag generates the same date information.
         stats = tree.branch.repository.gather_stats(rev2, committers=False)
-        self.check_stats_has_size(stats)
         self.assertEqual({
             'firstrev': (1170491381.0, 0),
             'latestrev': (1171491381.0, 0),
@@ -56,7 +45,6 @@
             },
             stats)
         stats = tree.branch.repository.gather_stats(rev2, committers=True)
-        self.check_stats_has_size(stats)
         self.assertEqual({
             'committers': 2,
             'firstrev': (1170491381.0, 0),
@@ -65,28 +53,11 @@
             },
             stats)
 
-    def test_gather_stats_norevid_gets_size(self):
-        """Without a revid, repository size is still gathered."""
-        tree = self.make_branch_and_memory_tree('.')
-        tree.lock_write()
-        tree.add('')
-        # put something in the repository, because zero-size is borink.
-        rev1 = tree.commit('first post')
-        tree.unlock()
-        # now ask for global repository stats.
-        stats = tree.branch.repository.gather_stats()
-        self.check_stats_has_size(stats)
-        self.assertEqual({
-            'revisions': 1
-            },
-            stats)
-
     def test_gather_stats_empty_repo(self):
-        """An empty repository still has size and revisions."""
+        """An empty repository still has revisions."""
         tree = self.make_branch_and_memory_tree('.')
         # now ask for global repository stats.
         stats = tree.branch.repository.gather_stats()
-        self.check_stats_has_size(stats)
         self.assertEqual({
             'revisions': 0
             },

=== removed directory 'bzrlib/tests/revisionstore_implementations'
=== removed file 'bzrlib/tests/revisionstore_implementations/__init__.py'
--- bzrlib/tests/revisionstore_implementations/__init__.py	2008-04-28 08:57:45 +0000
+++ bzrlib/tests/revisionstore_implementations/__init__.py	1970-01-01 00:00:00 +0000
@@ -1,92 +0,0 @@
-# Copyright (C) 2006 Canonical Ltd
-# Authors: Robert Collins <robert.collins@canonical.com>
-#
-# This program is free software; you can redistribute it and/or modify
-# it under the terms of the GNU General Public License as published by
-# the Free Software Foundation; either version 2 of the License, or
-# (at your option) any later version.
-#
-# This program is distributed in the hope that it will be useful,
-# but WITHOUT ANY WARRANTY; without even the implied warranty of
-# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-# GNU General Public License for more details.
-#
-# You should have received a copy of the GNU General Public License
-# along with this program; if not, write to the Free Software
-# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
-
-
-"""RevisionStore implemnetation tests for bzr.
-
-These test the conformance of all the revision store implementations to the 
-expected API including generally applicable corner cases.
-Specific tests for individual cases are in the tests/test_revisionstore.py file 
-rather than in tests/revisionstore_implementations/*.py.
-"""
-
-from bzrlib.tests import (
-                          adapt_modules,
-                          default_transport,
-                          TestScenarioApplier,
-                          )
-
-
-class RevisionStoreTestProviderAdapter(TestScenarioApplier):
-    """A tool to generate a suite testing multiple repository stores.
-
-    This is done by copying the test once for each repository store
-    and injecting the transport_server, transport_readonly_server,
-    and revision-store-factory into each copy.
-    Each copy is also given a new id() to make it easy to identify.
-    """
-
-    def __init__(self, transport_server, transport_readonly_server, factories):
-        self._transport_server = transport_server
-        self._transport_readonly_server = transport_readonly_server
-        self.scenarios = self.factories_to_scenarios(factories)
-    
-    def factories_to_scenarios(self, factories):
-        """Transform the input factories to a list of scenarios.
-
-        :param factories: A list of factories.
-        """
-        result = []
-        for factory in factories:
-            scenario = (factory, {
-                "transport_server":self._transport_server,
-                "transport_readonly_server":self._transport_readonly_server,
-                "store_factory":factory,
-                })
-            result.append(scenario)
-        return result
-
-    @staticmethod
-    def default_test_list():
-        """Generate the default list of revision store permutations to test."""
-        from bzrlib.store.revision.text import TextRevisionStoreTestFactory
-        from bzrlib.store.revision.knit import KnitRevisionStoreFactory
-        result = []
-        # test the fallback InterVersionedFile from weave to annotated knits
-        result.append(TextRevisionStoreTestFactory())
-        result.append(KnitRevisionStoreFactory())
-        return result
-
-
-def load_tests(basic_tests, module, loader):
-    result = loader.suiteClass()
-    # add the tests for this module
-    result.addTests(basic_tests)
-
-    test_revisionstore_implementations = [
-        'bzrlib.tests.revisionstore_implementations.test_all',
-        ]
-    adapter = RevisionStoreTestProviderAdapter(
-        default_transport,
-        # None here will cause a readonly decorator to be created
-        # by the TestCaseWithTransport.get_readonly_transport method.
-        None,
-        RevisionStoreTestProviderAdapter.default_test_list()
-        )
-    # add the tests for the sub modules
-    adapt_modules(test_revisionstore_implementations, adapter, loader, result)
-    return result

=== removed file 'bzrlib/tests/revisionstore_implementations/test_all.py'
--- bzrlib/tests/revisionstore_implementations/test_all.py	2008-04-04 00:06:58 +0000
+++ bzrlib/tests/revisionstore_implementations/test_all.py	1970-01-01 00:00:00 +0000
@@ -1,131 +0,0 @@
-# Copyright (C) 2006 Canonical Ltd
-#
-# This program is free software; you can redistribute it and/or modify
-# it under the terms of the GNU General Public License as published by
-# the Free Software Foundation; either version 2 of the License, or
-# (at your option) any later version.
-#
-# This program is distributed in the hope that it will be useful,
-# but WITHOUT ANY WARRANTY; without even the implied warranty of
-# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-# GNU General Public License for more details.
-#
-# You should have received a copy of the GNU General Public License
-# along with this program; if not, write to the Free Software
-# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
-
-"""Revision store tests."""
-
-
-import bzrlib.errors as errors
-from bzrlib.revision import Revision
-from bzrlib.store.revision import RevisionStore
-from bzrlib.tests import TestCaseWithTransport
-from bzrlib.transactions import PassThroughTransaction
-
-
-class TestFactory(TestCaseWithTransport):
-
-    def test_factory_keeps_smoke_in(self):
-        s = self.store_factory.create(self.get_url('.'))
-        self.assertTrue(isinstance(s, RevisionStore))
-
-
-class TestAll(TestCaseWithTransport):
-
-    def setUp(self):
-        super(TestAll, self).setUp()
-        self.store = self.store_factory.create(self.get_url('.'))
-        self.transaction = PassThroughTransaction()
-        self.store.get_scope = self.get_transaction
-
-    def get_transaction(self):
-        return self.transaction
-
-    def test_add_has_get(self):
-        rev = self.add_sample_rev()
-        self.assertTrue(self.store.has_revision_id('A', self.transaction))
-        rev2 = self.store.get_revision('A', self.transaction)
-        self.assertEqual(rev, rev2)
-
-    def add_sample_rev(self):
-        rev = Revision(timestamp=0,
-                       timezone=0,
-                       committer="Foo Bar <foo@example.com>",
-                       message="Message",
-                       inventory_sha1='aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa',
-                       revision_id='A')
-        self.store.add_revision(rev, self.transaction)
-        return rev
-
-    def test_has_missing(self):
-        # has of a non present id -> False
-        self.assertFalse(self.store.has_revision_id('missing', self.transaction))
-
-    def test_has_null(self):
-        # has of null -> True
-        self.assertTrue(self.store.has_revision_id('null:', self.transaction))
-
-    def test_get_revision_missing(self):
-        # get_revision('B') -> raises NoSuchRevision
-        self.assertRaises(errors.NoSuchRevision,
-                          self.store.get_revision,
-                          'B',
-                          self.transaction)
-
-    def test_get_revision_null(self):
-        # get_revision(null) -> raises ReservedId
-        self.assertRaises(errors.ReservedId,
-                          self.store.get_revision,
-                          'null:',
-                          self.transaction)
-
-    def test_add_signature_text_missing(self):
-        # add of a text signature for a missing revision must work, to allow
-        # revisions to be added after the signature.
-        self.store.add_revision_signature_text('A', 'foo\nbar', self.transaction)
-        # but must not be visible
-        self.assertRaises(errors.NoSuchRevision,
-                          self.store.has_signature,
-                          'A',
-                          self.transaction)
-        # at all
-        self.assertRaises(errors.NoSuchRevision,
-                          self.store.get_signature_text,
-                          'A',
-                          self.transaction)
-        # until the revision is added
-        self.add_sample_rev()
-        self.assertTrue(self.store.has_signature('A', self.transaction))
-        self.assertEqual('foo\nbar',
-                         self.store.get_signature_text('A', self.transaction))
-    
-    def test_add_signature_text(self):
-        # add a signature to a existing revision works.
-        self.add_sample_rev()
-        self.assertFalse(self.store.has_signature('A', self.transaction))
-        self.assertRaises(errors.NoSuchRevision,
-                          self.store.get_signature_text,
-                          'A',
-                          self.transaction)
-        self.store.add_revision_signature_text('A', 'foo\nbar', self.transaction)
-        self.assertTrue(self.store.has_signature('A', self.transaction))
-        self.assertEqual('foo\nbar',
-                         self.store.get_signature_text('A', self.transaction))
-
-    def test_total_size(self):
-        # we get a revision count and a numeric size figure from total_size().
-        count, bytes = self.store.total_size(self.transaction)
-        self.assertEqual(0, count)
-        # some stores use disk immediately that they are created so we just 
-        # check that its an int.
-        self.assertIsInstance(bytes, (int, long))
-        self.add_sample_rev()
-        count, bytes = self.store.total_size(self.transaction)
-        self.assertEqual(1, count)
-        self.assertNotEqual(0, bytes)
-        
-    def test_all_revision_ids(self):
-        self.assertEqual([], self.store.all_revision_ids(self.transaction))
-        self.add_sample_rev()
-        self.assertEqual(['A'], self.store.all_revision_ids(self.transaction))

=== modified file 'bzrlib/tests/test_bundle.py'
--- bzrlib/tests/test_bundle.py	2008-05-08 04:33:38 +0000
+++ bzrlib/tests/test_bundle.py	2008-06-11 04:20:16 +0000
@@ -469,23 +469,28 @@
         for ancestor in ancestors:
             old = self.b1.repository.revision_tree(ancestor)
             new = tree.branch.repository.revision_tree(ancestor)
-
-            # Check that there aren't any inventory level changes
-            delta = new.changes_from(old)
-            self.assertFalse(delta.has_changed(),
-                             'Revision %s not copied correctly.'
-                             % (ancestor,))
-
-            # Now check that the file contents are all correct
-            for inventory_id in old:
-                try:
-                    old_file = old.get_file(inventory_id)
-                except NoSuchFile:
-                    continue
-                if old_file is None:
-                    continue
-                self.assertEqual(old_file.read(),
-                                 new.get_file(inventory_id).read())
+            old.lock_read()
+            new.lock_read()
+            try:
+                # Check that there aren't any inventory level changes
+                delta = new.changes_from(old)
+                self.assertFalse(delta.has_changed(),
+                                 'Revision %s not copied correctly.'
+                                 % (ancestor,))
+
+                # Now check that the file contents are all correct
+                for inventory_id in old:
+                    try:
+                        old_file = old.get_file(inventory_id)
+                    except NoSuchFile:
+                        continue
+                    if old_file is None:
+                        continue
+                    self.assertEqual(old_file.read(),
+                                     new.get_file(inventory_id).read())
+            finally:
+                new.unlock()
+                old.unlock()
         if not _mod_revision.is_null(rev_id):
             rh = self.b1.revision_history()
             tree.branch.set_revision_history(rh[:rh.index(rev_id)+1])
@@ -1027,9 +1032,12 @@
         repo = self.make_repo_with_installed_revisions()
         inv = repo.get_inventory('rev2')
         self.assertEqual('rev2', inv.root.revision)
-        root_vf = repo.weave_store.get_weave(inv.root.file_id,
-                                             repo.get_transaction())
-        self.assertEqual(root_vf.versions(), ['rev1', 'rev2'])
+        root_id = inv.root.file_id
+        repo.lock_read()
+        self.addCleanup(repo.unlock)
+        self.assertEqual({(root_id, 'rev1'):(),
+            (root_id, 'rev2'):((root_id, 'rev1'),)},
+            repo.texts.get_parent_map([(root_id, 'rev1'), (root_id, 'rev2')]))
 
     def test_inv_hash_across_serializers(self):
         repo = self.make_repo_with_installed_revisions()
@@ -1345,14 +1353,18 @@
         tree2 = self.make_branch_and_tree('target')
         target_repo = tree2.branch.repository
         install_bundle(target_repo, serializer.read(s))
-        vf = target_repo.weave_store.get_weave('fileid-2',
-            target_repo.get_transaction())
-        self.assertEqual('contents1\nstatic\n', vf.get_text('rev1'))
-        self.assertEqual('contents2\nstatic\n', vf.get_text('rev2'))
+        target_repo.lock_read()
+        self.addCleanup(target_repo.unlock)
+        self.assertEqual({'1':'contents1\nstatic\n',
+            '2':'contents2\nstatic\n'},
+            dict(target_repo.iter_files_bytes(
+                [('fileid-2', 'rev1', '1'), ('fileid-2', 'rev2', '2')])))
         rtree = target_repo.revision_tree('rev2')
-        inventory_vf = target_repo.get_inventory_weave()
-        self.assertEqual({'rev2':('rev1',)},
-            inventory_vf.get_parent_map(['rev2']))
+        inventory_vf = target_repo.inventories
+        # If the inventory store has a graph, it must match the revision graph.
+        self.assertSubset(
+            [inventory_vf.get_parent_map([('rev2',)])[('rev2',)]],
+            [None, (('rev1',),)])
         self.assertEqual('changed file',
                          target_repo.get_revision('rev2').message)
 

=== removed file 'bzrlib/tests/test_escaped_store.py'
--- bzrlib/tests/test_escaped_store.py	2008-05-02 03:30:07 +0000
+++ bzrlib/tests/test_escaped_store.py	1970-01-01 00:00:00 +0000
@@ -1,102 +0,0 @@
-# Copyright (C) 2005, 2007 Canonical Ltd
-#
-# This program is free software; you can redistribute it and/or modify
-# it under the terms of the GNU General Public License as published by
-# the Free Software Foundation; either version 2 of the License, or
-# (at your option) any later version.
-#
-# This program is distributed in the hope that it will be useful,
-# but WITHOUT ANY WARRANTY; without even the implied warranty of
-# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-# GNU General Public License for more details.
-#
-# You should have received a copy of the GNU General Public License
-# along with this program; if not, write to the Free Software
-# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
-
-"""Test Escaped Stores."""
-
-from cStringIO import StringIO
-import os
-import gzip
-
-from bzrlib import osutils
-from bzrlib.errors import BzrError, UnlistableStore, NoSuchFile
-from bzrlib.store.text import TextStore
-from bzrlib.tests import TestCaseWithTransport
-import bzrlib.transport
-
-
-class TestEscaped(TestCaseWithTransport):
-    """Mixin template class that provides some common tests for stores"""
-
-    def get_store(self, prefixed=False, escaped=True):
-        t = bzrlib.transport.get_transport(self.get_url())
-        return TextStore(t, prefixed=prefixed, escaped=escaped)
-
-    def test_prefixed(self):
-        # Prefix should be determined by unescaped string
-        text_store = self.get_store(prefixed=True)
-
-        # hash_prefix() is not defined for unicode characters
-        # it is only defined for byte streams.
-        # so hash_prefix() needs to operate on *at most* utf-8
-        # encoded. However urlutils.escape() does both encoding to utf-8
-        # and urllib quoting, so we will use the escaped form
-        # as the path passed to hash_prefix
-
-        self.assertEqual('62/a', text_store._relpath('a'))
-        self.assertEqual('88/%2520', text_store._relpath(' '))
-        self.assertEqual('72/%40%253a%253c%253e',
-                text_store._relpath('@:<>'))
-
-    def test_files(self):
-        text_store = self.get_store(prefixed=True)
-
-        text_store.add(StringIO('a'), 'a')
-        self.failUnlessExists('62/a')
-
-        text_store.add(StringIO('space'), ' ')
-        self.failUnlessExists('88/%20')
-        self.assertEquals('space', text_store.get(' ').read())
-
-        text_store.add(StringIO('surprise'), '@:<>')
-        self.failUnlessExists('72/@%3a%3c%3e')
-        self.assertEquals('surprise', text_store.get('@:<>').read())
-
-        text_store.add(StringIO('utf8'), '\xc2\xb5')
-        self.failUnlessExists('77/%c2%b5')
-        self.assertEquals('utf8', text_store.get('\xc2\xb5').read())
-
-    def test_weave(self):
-        from bzrlib.store.versioned import WeaveStore
-        from bzrlib.transactions import PassThroughTransaction
-
-        trans = PassThroughTransaction()
-
-        t = bzrlib.transport.get_transport(self.get_url())
-        weave_store = WeaveStore(t, prefixed=True, escaped=True)
-        def add_text(file_id, rev_id, contents, parents, transaction):
-            vfile = weave_store.get_weave_or_empty(file_id, transaction)
-            vfile.add_lines(rev_id, parents, contents)
-
-        def check_text(file_id, revision_id, contents):
-            vfile = weave_store.get_weave(file_id, trans)
-            self.assertEqual(contents, vfile.get_lines(revision_id))
-
-        add_text('a', 'r', ['a'], [], trans)
-        self.failUnlessExists('62/a.weave')
-        check_text('a', 'r', ['a'])
-
-        add_text(' ', 'r', ['space'], [], trans)
-        self.failIfExists('21/ .weave')
-        self.failUnlessExists('88/%20.weave')
-        check_text(' ', 'r', ['space'])
-
-        add_text('@:<>', 'r', ['surprise'], [], trans)
-        self.failUnlessExists('72/@%3a%3c%3e.weave')
-        check_text('@:<>', 'r', ['surprise'])
-
-        add_text('\xc2\xb5', 'r', ['utf8'], [], trans)
-        self.failUnlessExists('77/%c2%b5.weave')
-        check_text('\xc2\xb5', 'r', ['utf8'])

=== modified file 'bzrlib/tests/test_fetch.py'
--- bzrlib/tests/test_fetch.py	2008-05-10 03:13:56 +0000
+++ bzrlib/tests/test_fetch.py	2008-06-11 04:20:16 +0000
@@ -146,17 +146,24 @@
         branch = self.make_branch('branch', format=knit2_format)
         branch.pull(tree.branch, stop_revision='rev1')
         repo = branch.repository
-        root_knit = repo.weave_store.get_weave('tree-root',
-                                                repo.get_transaction())
-        # Make sure fetch retrieved only what we requested
-        self.assertTrue('rev1' in root_knit)
-        self.assertTrue('rev2' not in root_knit)
+        repo.lock_read()
+        try:
+            # Make sure fetch retrieved only what we requested
+            self.assertEqual({('tree-root', 'rev1'):()},
+                repo.texts.get_parent_map(
+                    [('tree-root', 'rev1'), ('tree-root', 'rev2')]))
+        finally:
+            repo.unlock()
         branch.pull(tree.branch)
-        root_knit = repo.weave_store.get_weave('tree-root',
-                                                repo.get_transaction())
         # Make sure that the next revision in the root knit was retrieved,
         # even though the text, name, parent_id, etc., were unchanged.
-        self.assertTrue('rev2' in root_knit)
+        repo.lock_read()
+        try:
+            # Make sure fetch retrieved only what we requested
+            self.assertEqual({('tree-root', 'rev2'):(('tree-root', 'rev1'),)},
+                repo.texts.get_parent_map([('tree-root', 'rev2')]))
+        finally:
+            repo.unlock()
 
     def test_fetch_incompatible(self):
         knit_tree = self.make_branch_and_tree('knit', format='knit')
@@ -286,9 +293,7 @@
         # twice?
         self.assertEqual(1, self._count_log_matches('/ce/id.kndx', http_logs))
         self.assertEqual(1, self._count_log_matches('/ce/id.knit', http_logs))
-        # XXX: This *should* be '1', but more intrusive fetch changes are
-        # needed to drop this back to 1.
-        self.assertEqual(2, self._count_log_matches('inventory.kndx', http_logs))
+        self.assertEqual(1, self._count_log_matches('inventory.kndx', http_logs))
         # this r-h check test will prevent regressions, but it currently already 
         # passes, before the patch to cache-rh is applied :[
         self.assertTrue(1 >= self._count_log_matches('revision-history',
@@ -350,9 +355,12 @@
         self.do_fetch_order_test('B', 'A')
 
     def get_parents(self, file_id, revision_id):
-        transaction = self.repo.get_transaction()
-        vf = self.repo.weave_store.get_weave(file_id, transaction)
-        return vf.get_parents_with_ghosts(revision_id)
+        self.repo.lock_read()
+        try:
+            parent_map = self.repo.texts.get_parent_map([(file_id, revision_id)])
+            return parent_map[(file_id, revision_id)]
+        finally:
+            self.repo.unlock()
 
     def test_fetch_ghosts(self):
         self.make_tree_and_repo()
@@ -366,8 +374,10 @@
         self.tree.commit('second commit', rev_id='second-id')
         self.repo.fetch(self.tree.branch.repository, 'second-id')
         root_id = self.tree.get_root_id()
-        self.assertEqual(['left-parent', 'ghost-parent', 'not-ghost-parent'],
-                         self.get_parents(root_id, 'second-id'))
+        self.assertEqual(
+            ((root_id, 'left-parent'), (root_id, 'ghost-parent'),
+             (root_id, 'not-ghost-parent')),
+            self.get_parents(root_id, 'second-id'))
 
     def make_two_commits(self, change_root, fetch_twice):
         self.make_tree_and_repo()
@@ -381,13 +391,13 @@
 
     def test_fetch_changed_root(self):
         self.make_two_commits(change_root=True, fetch_twice=False)
-        self.assertEqual([], self.get_parents('unique-id', 'second-id'))
+        self.assertEqual((), self.get_parents('unique-id', 'second-id'))
 
     def test_two_fetch_changed_root(self):
         self.make_two_commits(change_root=True, fetch_twice=True)
-        self.assertEqual([], self.get_parents('unique-id', 'second-id'))
+        self.assertEqual((), self.get_parents('unique-id', 'second-id'))
 
     def test_two_fetches(self):
         self.make_two_commits(change_root=False, fetch_twice=True)
-        self.assertEqual(['first-id'],
-                         self.get_parents('TREE_ROOT', 'second-id'))
+        self.assertEqual((('TREE_ROOT', 'first-id'),),
+            self.get_parents('TREE_ROOT', 'second-id'))

=== modified file 'bzrlib/tests/test_knit.py'
--- bzrlib/tests/test_knit.py	2008-05-11 23:49:50 +0000
+++ bzrlib/tests/test_knit.py	2008-06-11 04:20:16 +0000
@@ -37,23 +37,15 @@
 from bzrlib.index import *
 from bzrlib.knit import (
     AnnotatedKnitContent,
-    DATA_SUFFIX,
-    INDEX_SUFFIX,
     KnitContent,
-    KnitGraphIndex,
-    KnitVersionedFile,
-    KnitPlainFactory,
-    KnitAnnotateFactory,
-    _KnitAccess,
-    _KnitData,
-    _KnitIndex,
-    make_file_knit,
-    _PackAccess,
+    KnitSequenceMatcher,
+    KnitVersionedFiles,
     PlainKnitContent,
-    _StreamAccess,
-    _StreamIndex,
-    WeaveToKnit,
-    KnitSequenceMatcher,
+    _DirectPackAccess,
+    _KndxIndex,
+    _KnitGraphIndex,
+    _KnitKeyAccess,
+    make_file_factory,
     )
 from bzrlib.osutils import split_lines
 from bzrlib.symbol_versioning import one_four
@@ -66,8 +58,7 @@
 from bzrlib.transport import get_transport
 from bzrlib.transport.memory import MemoryTransport
 from bzrlib.tuned_gzip import GzipFile
-from bzrlib.util import bencode
-from bzrlib.weave import Weave
+from bzrlib.versionedfile import ConstantMapper
 
 
 class _CompiledKnitFeature(Feature):
@@ -148,6 +139,51 @@
         self.assertDerivedBlocksEqual('a\nb\nc', 'a\nb\nc\n', noeol=True)
 
 
+TEXT_1 = """\
+Banana cup cakes:
+
+- bananas
+- eggs
+- broken tea cups
+"""
+
+TEXT_1A = """\
+Banana cup cake recipe
+(serves 6)
+
+- bananas
+- eggs
+- broken tea cups
+- self-raising flour
+"""
+
+TEXT_1B = """\
+Banana cup cake recipe
+
+- bananas (do not use plantains!!!)
+- broken tea cups
+- flour
+"""
+
+delta_1_1a = """\
+0,1,2
+Banana cup cake recipe
+(serves 6)
+5,5,1
+- self-raising flour
+"""
+
+TEXT_2 = """\
+Boeuf bourguignon
+
+- beef
+- red wine
+- small onions
+- carrot
+- mushrooms
+"""
+
+
 class TestPlainKnitContent(TestCase, KnitContentTestsMixin):
 
     def _make_content(self, lines):
@@ -232,20 +268,17 @@
 class KnitRecordAccessTestsMixin(object):
     """Tests for getting and putting knit records."""
 
-    def assertAccessExists(self, access):
-        """Ensure the data area for access has been initialised/exists."""
-        raise NotImplementedError(self.assertAccessExists)
-
     def test_add_raw_records(self):
         """Add_raw_records adds records retrievable later."""
         access = self.get_access()
-        memos = access.add_raw_records([10], '1234567890')
+        memos = access.add_raw_records([('key', 10)], '1234567890')
         self.assertEqual(['1234567890'], list(access.get_raw_records(memos)))
  
     def test_add_several_raw_records(self):
         """add_raw_records with many records and read some back."""
         access = self.get_access()
-        memos = access.add_raw_records([10, 2, 5], '12345678901234567')
+        memos = access.add_raw_records([('key', 10), ('key2', 2), ('key3', 5)],
+            '12345678901234567')
         self.assertEqual(['1234567890', '12', '34567'],
             list(access.get_raw_records(memos)))
         self.assertEqual(['1234567890'],
@@ -257,39 +290,20 @@
         self.assertEqual(['1234567890', '34567'],
             list(access.get_raw_records(memos[0:1] + memos[2:3])))
 
-    def test_create(self):
-        """create() should make a file on disk."""
-        access = self.get_access()
-        access.create()
-        self.assertAccessExists(access)
-
-    def test_open_file(self):
-        """open_file never errors."""
-        access = self.get_access()
-        access.open_file()
-
 
 class TestKnitKnitAccess(TestCaseWithMemoryTransport, KnitRecordAccessTestsMixin):
     """Tests for the .kndx implementation."""
 
-    def assertAccessExists(self, access):
-        self.assertNotEqual(None, access.open_file())
-
     def get_access(self):
         """Get a .knit style access instance."""
-        access = _KnitAccess(self.get_transport(), "foo.knit", None, None,
-            False, False)
+        mapper = ConstantMapper("foo")
+        access = _KnitKeyAccess(self.get_transport(), mapper)
         return access
     
 
 class TestPackKnitAccess(TestCaseWithMemoryTransport, KnitRecordAccessTestsMixin):
     """Tests for the pack based access."""
 
-    def assertAccessExists(self, access):
-        # as pack based access has no backing unless an index maps data, this
-        # is a no-op.
-        pass
-
     def get_access(self):
         return self._get_access()[0]
 
@@ -299,23 +313,23 @@
             transport.append_bytes(packname, bytes)
         writer = pack.ContainerWriter(write_data)
         writer.begin()
-        indices = {index:(transport, packname)}
-        access = _PackAccess(indices, writer=(writer, index))
+        access = _DirectPackAccess({})
+        access.set_writer(writer, index, (transport, packname))
         return access, writer
 
     def test_read_from_several_packs(self):
         access, writer = self._get_access()
         memos = []
-        memos.extend(access.add_raw_records([10], '1234567890'))
+        memos.extend(access.add_raw_records([('key', 10)], '1234567890'))
         writer.end()
         access, writer = self._get_access('pack2', 'FOOBAR')
-        memos.extend(access.add_raw_records([5], '12345'))
+        memos.extend(access.add_raw_records([('key', 5)], '12345'))
         writer.end()
         access, writer = self._get_access('pack3', 'BAZ')
-        memos.extend(access.add_raw_records([5], 'alpha'))
+        memos.extend(access.add_raw_records([('key', 5)], 'alpha'))
         writer.end()
         transport = self.get_transport()
-        access = _PackAccess({"FOO":(transport, 'packfile'),
+        access = _DirectPackAccess({"FOO":(transport, 'packfile'),
             "FOOBAR":(transport, 'pack2'),
             "BAZ":(transport, 'pack3')})
         self.assertEqual(['1234567890', '12345', 'alpha'],
@@ -331,7 +345,7 @@
 
     def test_set_writer(self):
         """The writer should be settable post construction."""
-        access = _PackAccess({})
+        access = _DirectPackAccess({})
         transport = self.get_transport()
         packname = 'packfile'
         index = 'foo'
@@ -340,7 +354,7 @@
         writer = pack.ContainerWriter(write_data)
         writer.begin()
         access.set_writer(writer, index, (transport, packname))
-        memos = access.add_raw_records([10], '1234567890')
+        memos = access.add_raw_records([('key', 10)], '1234567890')
         writer.end()
         self.assertEqual(['1234567890'], list(access.get_raw_records(memos)))
 
@@ -362,15 +376,16 @@
                                         'end rev-id-1\n'
                                         % (sha1sum,))
         transport = MockTransport([gz_txt])
-        access = _KnitAccess(transport, 'filename', None, None, False, False)
-        data = _KnitData(access=access)
-        records = [('rev-id-1', (None, 0, len(gz_txt)))]
-
-        contents = data.read_records(records)
-        self.assertEqual({'rev-id-1':(['foo\n', 'bar\n'], sha1sum)}, contents)
-
-        raw_contents = list(data.read_records_iter_raw(records))
-        self.assertEqual([('rev-id-1', gz_txt, sha1sum)], raw_contents)
+        access = _KnitKeyAccess(transport, ConstantMapper('filename'))
+        knit = KnitVersionedFiles(None, access)
+        records = [(('rev-id-1',), (('rev-id-1',), 0, len(gz_txt)))]
+
+        contents = list(knit._read_records_iter(records))
+        self.assertEqual([(('rev-id-1',), ['foo\n', 'bar\n'],
+            '4e48e2c9a3d2ca8a708cb0cc545700544efb5021')], contents)
+
+        raw_contents = list(knit._read_records_iter_raw(records))
+        self.assertEqual([(('rev-id-1',), gz_txt, sha1sum)], raw_contents)
 
     def test_not_enough_lines(self):
         sha1sum = sha.new('foo\n').hexdigest()
@@ -380,14 +395,15 @@
                                         'end rev-id-1\n'
                                         % (sha1sum,))
         transport = MockTransport([gz_txt])
-        access = _KnitAccess(transport, 'filename', None, None, False, False)
-        data = _KnitData(access=access)
-        records = [('rev-id-1', (None, 0, len(gz_txt)))]
-        self.assertRaises(errors.KnitCorrupt, data.read_records, records)
+        access = _KnitKeyAccess(transport, ConstantMapper('filename'))
+        knit = KnitVersionedFiles(None, access)
+        records = [(('rev-id-1',), (('rev-id-1',), 0, len(gz_txt)))]
+        self.assertRaises(errors.KnitCorrupt, list,
+            knit._read_records_iter(records))
 
         # read_records_iter_raw won't detect that sort of mismatch/corruption
-        raw_contents = list(data.read_records_iter_raw(records))
-        self.assertEqual([('rev-id-1', gz_txt, sha1sum)], raw_contents)
+        raw_contents = list(knit._read_records_iter_raw(records))
+        self.assertEqual([(('rev-id-1',),  gz_txt, sha1sum)], raw_contents)
 
     def test_too_many_lines(self):
         sha1sum = sha.new('foo\nbar\n').hexdigest()
@@ -398,14 +414,15 @@
                                         'end rev-id-1\n'
                                         % (sha1sum,))
         transport = MockTransport([gz_txt])
-        access = _KnitAccess(transport, 'filename', None, None, False, False)
-        data = _KnitData(access=access)
-        records = [('rev-id-1', (None, 0, len(gz_txt)))]
-        self.assertRaises(errors.KnitCorrupt, data.read_records, records)
+        access = _KnitKeyAccess(transport, ConstantMapper('filename'))
+        knit = KnitVersionedFiles(None, access)
+        records = [(('rev-id-1',), (('rev-id-1',), 0, len(gz_txt)))]
+        self.assertRaises(errors.KnitCorrupt, list,
+            knit._read_records_iter(records))
 
         # read_records_iter_raw won't detect that sort of mismatch/corruption
-        raw_contents = list(data.read_records_iter_raw(records))
-        self.assertEqual([('rev-id-1', gz_txt, sha1sum)], raw_contents)
+        raw_contents = list(knit._read_records_iter_raw(records))
+        self.assertEqual([(('rev-id-1',), gz_txt, sha1sum)], raw_contents)
 
     def test_mismatched_version_id(self):
         sha1sum = sha.new('foo\nbar\n').hexdigest()
@@ -415,15 +432,16 @@
                                         'end rev-id-1\n'
                                         % (sha1sum,))
         transport = MockTransport([gz_txt])
-        access = _KnitAccess(transport, 'filename', None, None, False, False)
-        data = _KnitData(access=access)
+        access = _KnitKeyAccess(transport, ConstantMapper('filename'))
+        knit = KnitVersionedFiles(None, access)
         # We are asking for rev-id-2, but the data is rev-id-1
-        records = [('rev-id-2', (None, 0, len(gz_txt)))]
-        self.assertRaises(errors.KnitCorrupt, data.read_records, records)
+        records = [(('rev-id-2',), (('rev-id-2',), 0, len(gz_txt)))]
+        self.assertRaises(errors.KnitCorrupt, list,
+            knit._read_records_iter(records))
 
-        # read_records_iter_raw will notice if we request the wrong version.
+        # read_records_iter_raw detects mismatches in the header
         self.assertRaises(errors.KnitCorrupt, list,
-                          data.read_records_iter_raw(records))
+            knit._read_records_iter_raw(records))
 
     def test_uncompressed_data(self):
         sha1sum = sha.new('foo\nbar\n').hexdigest()
@@ -433,16 +451,17 @@
                'end rev-id-1\n'
                % (sha1sum,))
         transport = MockTransport([txt])
-        access = _KnitAccess(transport, 'filename', None, None, False, False)
-        data = _KnitData(access=access)
-        records = [('rev-id-1', (None, 0, len(txt)))]
+        access = _KnitKeyAccess(transport, ConstantMapper('filename'))
+        knit = KnitVersionedFiles(None, access)
+        records = [(('rev-id-1',), (('rev-id-1',), 0, len(txt)))]
 
         # We don't have valid gzip data ==> corrupt
-        self.assertRaises(errors.KnitCorrupt, data.read_records, records)
+        self.assertRaises(errors.KnitCorrupt, list,
+            knit._read_records_iter(records))
 
         # read_records_iter_raw will notice the bad data
         self.assertRaises(errors.KnitCorrupt, list,
-                          data.read_records_iter_raw(records))
+            knit._read_records_iter_raw(records))
 
     def test_corrupted_data(self):
         sha1sum = sha.new('foo\nbar\n').hexdigest()
@@ -454,373 +473,261 @@
         # Change 2 bytes in the middle to \xff
         gz_txt = gz_txt[:10] + '\xff\xff' + gz_txt[12:]
         transport = MockTransport([gz_txt])
-        access = _KnitAccess(transport, 'filename', None, None, False, False)
-        data = _KnitData(access=access)
-        records = [('rev-id-1', (None, 0, len(gz_txt)))]
-
-        self.assertRaises(errors.KnitCorrupt, data.read_records, records)
-
-        # read_records_iter_raw will notice if we request the wrong version.
-        self.assertRaises(errors.KnitCorrupt, list,
-                          data.read_records_iter_raw(records))
+        access = _KnitKeyAccess(transport, ConstantMapper('filename'))
+        knit = KnitVersionedFiles(None, access)
+        records = [(('rev-id-1',), (('rev-id-1',), 0, len(gz_txt)))]
+        self.assertRaises(errors.KnitCorrupt, list,
+            knit._read_records_iter(records))
+        # read_records_iter_raw will barf on bad gz data
+        self.assertRaises(errors.KnitCorrupt, list,
+            knit._read_records_iter_raw(records))
 
 
 class LowLevelKnitIndexTests(TestCase):
 
-    def get_knit_index(self, *args, **kwargs):
+    def get_knit_index(self, transport, name, mode):
+        mapper = ConstantMapper(name)
         orig = knit._load_data
         def reset():
             knit._load_data = orig
         self.addCleanup(reset)
         from bzrlib._knit_load_data_py import _load_data_py
         knit._load_data = _load_data_py
-        return _KnitIndex(get_scope=lambda:None, *args, **kwargs)
-
-    def test_no_such_file(self):
-        transport = MockTransport()
-
-        self.assertRaises(NoSuchFile, self.get_knit_index,
-                          transport, "filename", "r")
-        self.assertRaises(NoSuchFile, self.get_knit_index,
-                          transport, "filename", "w", create=False)
+        allow_writes = lambda: 'w' in mode
+        return _KndxIndex(transport, mapper, lambda:None, allow_writes, lambda:True)
 
     def test_create_file(self):
         transport = MockTransport()
-
-        index = self.get_knit_index(transport, "filename", "w",
-            file_mode="wb", create=True)
-        self.assertEqual(
-                ("put_bytes_non_atomic",
-                    ("filename", index.HEADER), {"mode": "wb"}),
-                transport.calls.pop(0))
-
-    def test_delay_create_file(self):
-        transport = MockTransport()
-
-        index = self.get_knit_index(transport, "filename", "w",
-            create=True, file_mode="wb", create_parent_dir=True,
-            delay_create=True, dir_mode=0777)
-        self.assertEqual([], transport.calls)
-
-        index.add_versions([])
-        name, (filename, f), kwargs = transport.calls.pop(0)
-        self.assertEqual("put_file_non_atomic", name)
-        self.assertEqual(
-            {"dir_mode": 0777, "create_parent_dir": True, "mode": "wb"},
-            kwargs)
-        self.assertEqual("filename", filename)
-        self.assertEqual(index.HEADER, f.read())
-
-        index.add_versions([])
-        self.assertEqual(("append_bytes", ("filename", ""), {}),
-            transport.calls.pop(0))
+        index = self.get_knit_index(transport, "filename", "w")
+        index.keys()
+        call = transport.calls.pop(0)
+        # call[1][1] is a StringIO - we can't test it by simple equality.
+        self.assertEqual('put_file_non_atomic', call[0])
+        self.assertEqual('filename.kndx', call[1][0])
+        # With no history, _KndxIndex writes a new index:
+        self.assertEqual(_KndxIndex.HEADER,
+            call[1][1].getvalue())
+        self.assertEqual({'create_parent_dir': True}, call[2])
 
     def test_read_utf8_version_id(self):
         unicode_revision_id = u"version-\N{CYRILLIC CAPITAL LETTER A}"
         utf8_revision_id = unicode_revision_id.encode('utf-8')
         transport = MockTransport([
-            _KnitIndex.HEADER,
+            _KndxIndex.HEADER,
             '%s option 0 1 :' % (utf8_revision_id,)
             ])
         index = self.get_knit_index(transport, "filename", "r")
-        # _KnitIndex is a private class, and deals in utf8 revision_ids, not
+        # _KndxIndex is a private class, and deals in utf8 revision_ids, not
         # Unicode revision_ids.
-        self.assertTrue(index.has_version(utf8_revision_id))
-        self.assertFalse(index.has_version(unicode_revision_id))
+        self.assertEqual({(utf8_revision_id,):()},
+            index.get_parent_map(index.keys()))
+        self.assertFalse((unicode_revision_id,) in index.keys())
 
     def test_read_utf8_parents(self):
         unicode_revision_id = u"version-\N{CYRILLIC CAPITAL LETTER A}"
         utf8_revision_id = unicode_revision_id.encode('utf-8')
         transport = MockTransport([
-            _KnitIndex.HEADER,
+            _KndxIndex.HEADER,
             "version option 0 1 .%s :" % (utf8_revision_id,)
             ])
         index = self.get_knit_index(transport, "filename", "r")
-        self.assertEqual((utf8_revision_id,),
-            index.get_parents_with_ghosts("version"))
+        self.assertEqual({("version",):((utf8_revision_id,),)},
+            index.get_parent_map(index.keys()))
 
     def test_read_ignore_corrupted_lines(self):
         transport = MockTransport([
-            _KnitIndex.HEADER,
+            _KndxIndex.HEADER,
             "corrupted",
             "corrupted options 0 1 .b .c ",
             "version options 0 1 :"
             ])
         index = self.get_knit_index(transport, "filename", "r")
-        self.assertEqual(1, index.num_versions())
-        self.assertTrue(index.has_version("version"))
+        self.assertEqual(1, len(index.keys()))
+        self.assertEqual(set([("version",)]), index.keys())
 
     def test_read_corrupted_header(self):
         transport = MockTransport(['not a bzr knit index header\n'])
-        self.assertRaises(KnitHeaderError,
-            self.get_knit_index, transport, "filename", "r")
+        index = self.get_knit_index(transport, "filename", "r")
+        self.assertRaises(KnitHeaderError, index.keys)
 
     def test_read_duplicate_entries(self):
         transport = MockTransport([
-            _KnitIndex.HEADER,
+            _KndxIndex.HEADER,
             "parent options 0 1 :",
             "version options1 0 1 0 :",
             "version options2 1 2 .other :",
             "version options3 3 4 0 .other :"
             ])
         index = self.get_knit_index(transport, "filename", "r")
-        self.assertEqual(2, index.num_versions())
+        self.assertEqual(2, len(index.keys()))
         # check that the index used is the first one written. (Specific
         # to KnitIndex style indices.
-        self.assertEqual("1", index._version_list_to_index(["version"]))
-        self.assertEqual((None, 3, 4), index.get_position("version"))
-        self.assertEqual(["options3"], index.get_options("version"))
-        self.assertEqual(("parent", "other"),
-            index.get_parents_with_ghosts("version"))
+        self.assertEqual("1", index._dictionary_compress([("version",)]))
+        self.assertEqual((("version",), 3, 4), index.get_position(("version",)))
+        self.assertEqual(["options3"], index.get_options(("version",)))
+        self.assertEqual({("version",):(("parent",), ("other",))},
+            index.get_parent_map([("version",)]))
 
     def test_read_compressed_parents(self):
         transport = MockTransport([
-            _KnitIndex.HEADER,
+            _KndxIndex.HEADER,
             "a option 0 1 :",
             "b option 0 1 0 :",
             "c option 0 1 1 0 :",
             ])
         index = self.get_knit_index(transport, "filename", "r")
-        self.assertEqual({"b":("a",), "c":("b", "a")},
-            index.get_parent_map(["b", "c"]))
+        self.assertEqual({("b",):(("a",),), ("c",):(("b",), ("a",))},
+            index.get_parent_map([("b",), ("c",)]))
 
     def test_write_utf8_version_id(self):
         unicode_revision_id = u"version-\N{CYRILLIC CAPITAL LETTER A}"
         utf8_revision_id = unicode_revision_id.encode('utf-8')
         transport = MockTransport([
-            _KnitIndex.HEADER
+            _KndxIndex.HEADER
             ])
         index = self.get_knit_index(transport, "filename", "r")
-        index.add_version(utf8_revision_id, ["option"], (None, 0, 1), [])
-        self.assertEqual(("append_bytes", ("filename",
-            "\n%s option 0 1  :" % (utf8_revision_id,)),
-            {}),
-            transport.calls.pop(0))
+        index.add_records([
+            ((utf8_revision_id,), ["option"], ((utf8_revision_id,), 0, 1), [])])
+        call = transport.calls.pop(0)
+        # call[1][1] is a StringIO - we can't test it by simple equality.
+        self.assertEqual('put_file_non_atomic', call[0])
+        self.assertEqual('filename.kndx', call[1][0])
+        # With no history, _KndxIndex writes a new index:
+        self.assertEqual(_KndxIndex.HEADER +
+            "\n%s option 0 1  :" % (utf8_revision_id,),
+            call[1][1].getvalue())
+        self.assertEqual({'create_parent_dir': True}, call[2])
 
     def test_write_utf8_parents(self):
         unicode_revision_id = u"version-\N{CYRILLIC CAPITAL LETTER A}"
         utf8_revision_id = unicode_revision_id.encode('utf-8')
         transport = MockTransport([
-            _KnitIndex.HEADER
-            ])
-        index = self.get_knit_index(transport, "filename", "r")
-        index.add_version("version", ["option"], (None, 0, 1), [utf8_revision_id])
-        self.assertEqual(("append_bytes", ("filename",
-            "\nversion option 0 1 .%s :" % (utf8_revision_id,)),
-            {}),
-            transport.calls.pop(0))
-
-    def test_get_ancestry(self):
-        transport = MockTransport([
-            _KnitIndex.HEADER,
-            "a option 0 1 :",
-            "b option 0 1 0 .e :",
-            "c option 0 1 1 0 :",
-            "d option 0 1 2 .f :"
-            ])
-        index = self.get_knit_index(transport, "filename", "r")
-
-        self.assertEqual([], index.get_ancestry([]))
-        self.assertEqual(["a"], index.get_ancestry(["a"]))
-        self.assertEqual(["a", "b"], index.get_ancestry(["b"]))
-        self.assertEqual(["a", "b", "c"], index.get_ancestry(["c"]))
-        self.assertEqual(["a", "b", "c", "d"], index.get_ancestry(["d"]))
-        self.assertEqual(["a", "b"], index.get_ancestry(["a", "b"]))
-        self.assertEqual(["a", "b", "c"], index.get_ancestry(["a", "c"]))
-
-        self.assertRaises(RevisionNotPresent, index.get_ancestry, ["e"])
-
-    def test_get_ancestry_with_ghosts(self):
-        transport = MockTransport([
-            _KnitIndex.HEADER,
-            "a option 0 1 :",
-            "b option 0 1 0 .e :",
-            "c option 0 1 0 .f .g :",
-            "d option 0 1 2 .h .j .k :"
-            ])
-        index = self.get_knit_index(transport, "filename", "r")
-
-        self.assertEqual([], index.get_ancestry_with_ghosts([]))
-        self.assertEqual(["a"], index.get_ancestry_with_ghosts(["a"]))
-        self.assertEqual(["a", "e", "b"],
-            index.get_ancestry_with_ghosts(["b"]))
-        self.assertEqual(["a", "g", "f", "c"],
-            index.get_ancestry_with_ghosts(["c"]))
-        self.assertEqual(["a", "g", "f", "c", "k", "j", "h", "d"],
-            index.get_ancestry_with_ghosts(["d"]))
-        self.assertEqual(["a", "e", "b"],
-            index.get_ancestry_with_ghosts(["a", "b"]))
-        self.assertEqual(["a", "g", "f", "c"],
-            index.get_ancestry_with_ghosts(["a", "c"]))
+            _KndxIndex.HEADER
+            ])
+        index = self.get_knit_index(transport, "filename", "r")
+        index.add_records([
+            (("version",), ["option"], (("version",), 0, 1), [(utf8_revision_id,)])])
+        call = transport.calls.pop(0)
+        # call[1][1] is a StringIO - we can't test it by simple equality.
+        self.assertEqual('put_file_non_atomic', call[0])
+        self.assertEqual('filename.kndx', call[1][0])
+        # With no history, _KndxIndex writes a new index:
+        self.assertEqual(_KndxIndex.HEADER +
+            "\nversion option 0 1 .%s :" % (utf8_revision_id,),
+            call[1][1].getvalue())
+        self.assertEqual({'create_parent_dir': True}, call[2])
+
+    def test_keys(self):
+        transport = MockTransport([
+            _KndxIndex.HEADER
+            ])
+        index = self.get_knit_index(transport, "filename", "r")
+
+        self.assertEqual(set(), index.keys())
+
+        index.add_records([(("a",), ["option"], (("a",), 0, 1), [])])
+        self.assertEqual(set([("a",)]), index.keys())
+
+        index.add_records([(("a",), ["option"], (("a",), 0, 1), [])])
+        self.assertEqual(set([("a",)]), index.keys())
+
+        index.add_records([(("b",), ["option"], (("b",), 0, 1), [])])
+        self.assertEqual(set([("a",), ("b",)]), index.keys())
+
+    def add_a_b(self, index, random_id=None):
+        kwargs = {}
+        if random_id is not None:
+            kwargs["random_id"] = random_id
+        index.add_records([
+            (("a",), ["option"], (("a",), 0, 1), [("b",)]),
+            (("a",), ["opt"], (("a",), 1, 2), [("c",)]),
+            (("b",), ["option"], (("b",), 2, 3), [("a",)])
+            ], **kwargs)
+
+    def assertIndexIsAB(self, index):
+        self.assertEqual({
+            ('a',): (('c',),),
+            ('b',): (('a',),),
+            },
+            index.get_parent_map(index.keys()))
+        self.assertEqual((("a",), 1, 2), index.get_position(("a",)))
+        self.assertEqual((("b",), 2, 3), index.get_position(("b",)))
+        self.assertEqual(["opt"], index.get_options(("a",)))
+
+    def test_add_versions(self):
+        transport = MockTransport([
+            _KndxIndex.HEADER
+            ])
+        index = self.get_knit_index(transport, "filename", "r")
+
+        self.add_a_b(index)
+        call = transport.calls.pop(0)
+        # call[1][1] is a StringIO - we can't test it by simple equality.
+        self.assertEqual('put_file_non_atomic', call[0])
+        self.assertEqual('filename.kndx', call[1][0])
+        # With no history, _KndxIndex writes a new index:
         self.assertEqual(
-            ["a", "g", "f", "c", "e", "b", "k", "j", "h", "d"],
-            index.get_ancestry_with_ghosts(["b", "d"]))
-
-        self.assertRaises(RevisionNotPresent,
-            index.get_ancestry_with_ghosts, ["e"])
-
-    def test_num_versions(self):
-        transport = MockTransport([
-            _KnitIndex.HEADER
-            ])
-        index = self.get_knit_index(transport, "filename", "r")
-
-        self.assertEqual(0, index.num_versions())
-        self.assertEqual(0, len(index))
-
-        index.add_version("a", ["option"], (None, 0, 1), [])
-        self.assertEqual(1, index.num_versions())
-        self.assertEqual(1, len(index))
-
-        index.add_version("a", ["option2"], (None, 1, 2), [])
-        self.assertEqual(1, index.num_versions())
-        self.assertEqual(1, len(index))
-
-        index.add_version("b", ["option"], (None, 0, 1), [])
-        self.assertEqual(2, index.num_versions())
-        self.assertEqual(2, len(index))
-
-    def test_get_versions(self):
-        transport = MockTransport([
-            _KnitIndex.HEADER
-            ])
-        index = self.get_knit_index(transport, "filename", "r")
-
-        self.assertEqual([], index.get_versions())
-
-        index.add_version("a", ["option"], (None, 0, 1), [])
-        self.assertEqual(["a"], index.get_versions())
-
-        index.add_version("a", ["option"], (None, 0, 1), [])
-        self.assertEqual(["a"], index.get_versions())
-
-        index.add_version("b", ["option"], (None, 0, 1), [])
-        self.assertEqual(["a", "b"], index.get_versions())
-
-    def test_add_version(self):
-        transport = MockTransport([
-            _KnitIndex.HEADER
-            ])
-        index = self.get_knit_index(transport, "filename", "r")
-
-        index.add_version("a", ["option"], (None, 0, 1), ["b"])
-        self.assertEqual(("append_bytes",
-            ("filename", "\na option 0 1 .b :"),
-            {}), transport.calls.pop(0))
-        self.assertTrue(index.has_version("a"))
-        self.assertEqual(1, index.num_versions())
-        self.assertEqual((None, 0, 1), index.get_position("a"))
-        self.assertEqual(["option"], index.get_options("a"))
-        self.assertEqual(("b",), index.get_parents_with_ghosts("a"))
-
-        index.add_version("a", ["opt"], (None, 1, 2), ["c"])
-        self.assertEqual(("append_bytes",
-            ("filename", "\na opt 1 2 .c :"),
-            {}), transport.calls.pop(0))
-        self.assertTrue(index.has_version("a"))
-        self.assertEqual(1, index.num_versions())
-        self.assertEqual((None, 1, 2), index.get_position("a"))
-        self.assertEqual(["opt"], index.get_options("a"))
-        self.assertEqual(("c",), index.get_parents_with_ghosts("a"))
-
-        index.add_version("b", ["option"], (None, 2, 3), ["a"])
-        self.assertEqual(("append_bytes",
-            ("filename", "\nb option 2 3 0 :"),
-            {}), transport.calls.pop(0))
-        self.assertTrue(index.has_version("b"))
-        self.assertEqual(2, index.num_versions())
-        self.assertEqual((None, 2, 3), index.get_position("b"))
-        self.assertEqual(["option"], index.get_options("b"))
-        self.assertEqual(("a",), index.get_parents_with_ghosts("b"))
-
-    def test_add_versions(self):
-        transport = MockTransport([
-            _KnitIndex.HEADER
-            ])
-        index = self.get_knit_index(transport, "filename", "r")
-
-        index.add_versions([
-            ("a", ["option"], (None, 0, 1), ["b"]),
-            ("a", ["opt"], (None, 1, 2), ["c"]),
-            ("b", ["option"], (None, 2, 3), ["a"])
-            ])
-        self.assertEqual(("append_bytes", ("filename",
+            _KndxIndex.HEADER +
             "\na option 0 1 .b :"
             "\na opt 1 2 .c :"
-            "\nb option 2 3 0 :"
-            ), {}), transport.calls.pop(0))
-        self.assertTrue(index.has_version("a"))
-        self.assertTrue(index.has_version("b"))
-        self.assertEqual(2, index.num_versions())
-        self.assertEqual((None, 1, 2), index.get_position("a"))
-        self.assertEqual((None, 2, 3), index.get_position("b"))
-        self.assertEqual(["opt"], index.get_options("a"))
-        self.assertEqual(["option"], index.get_options("b"))
-        self.assertEqual(("c",), index.get_parents_with_ghosts("a"))
-        self.assertEqual(("a",), index.get_parents_with_ghosts("b"))
+            "\nb option 2 3 0 :",
+            call[1][1].getvalue())
+        self.assertEqual({'create_parent_dir': True}, call[2])
+        self.assertIndexIsAB(index)
 
     def test_add_versions_random_id_is_accepted(self):
         transport = MockTransport([
-            _KnitIndex.HEADER
+            _KndxIndex.HEADER
             ])
         index = self.get_knit_index(transport, "filename", "r")
-
-        index.add_versions([
-            ("a", ["option"], (None, 0, 1), ["b"]),
-            ("a", ["opt"], (None, 1, 2), ["c"]),
-            ("b", ["option"], (None, 2, 3), ["a"])
-            ], random_id=True)
+        self.add_a_b(index, random_id=True)
 
     def test_delay_create_and_add_versions(self):
         transport = MockTransport()
 
-        index = self.get_knit_index(transport, "filename", "w",
-            create=True, file_mode="wb", create_parent_dir=True,
-            delay_create=True, dir_mode=0777)
+        index = self.get_knit_index(transport, "filename", "w")
+        # dir_mode=0777)
         self.assertEqual([], transport.calls)
-
-        index.add_versions([
-            ("a", ["option"], (None, 0, 1), ["b"]),
-            ("a", ["opt"], (None, 1, 2), ["c"]),
-            ("b", ["option"], (None, 2, 3), ["a"])
-            ])
-        name, (filename, f), kwargs = transport.calls.pop(0)
-        self.assertEqual("put_file_non_atomic", name)
-        self.assertEqual(
-            {"dir_mode": 0777, "create_parent_dir": True, "mode": "wb"},
-            kwargs)
-        self.assertEqual("filename", filename)
-        self.assertEqual(
-            index.HEADER +
+        self.add_a_b(index)
+        #self.assertEqual(
+        #[    {"dir_mode": 0777, "create_parent_dir": True, "mode": "wb"},
+        #    kwargs)
+        # Two calls: one during which we load the existing index (and when its
+        # missing create it), then a second where we write the contents out.
+        self.assertEqual(2, len(transport.calls))
+        call = transport.calls.pop(0)
+        self.assertEqual('put_file_non_atomic', call[0])
+        self.assertEqual('filename.kndx', call[1][0])
+        # With no history, _KndxIndex writes a new index:
+        self.assertEqual(_KndxIndex.HEADER, call[1][1].getvalue())
+        self.assertEqual({'create_parent_dir': True}, call[2])
+        call = transport.calls.pop(0)
+        # call[1][1] is a StringIO - we can't test it by simple equality.
+        self.assertEqual('put_file_non_atomic', call[0])
+        self.assertEqual('filename.kndx', call[1][0])
+        # With no history, _KndxIndex writes a new index:
+        self.assertEqual(
+            _KndxIndex.HEADER +
             "\na option 0 1 .b :"
             "\na opt 1 2 .c :"
             "\nb option 2 3 0 :",
-            f.read())
-
-    def test_has_version(self):
-        transport = MockTransport([
-            _KnitIndex.HEADER,
-            "a option 0 1 :"
-            ])
-        index = self.get_knit_index(transport, "filename", "r")
-
-        self.assertTrue(index.has_version("a"))
-        self.assertFalse(index.has_version("b"))
+            call[1][1].getvalue())
+        self.assertEqual({'create_parent_dir': True}, call[2])
 
     def test_get_position(self):
         transport = MockTransport([
-            _KnitIndex.HEADER,
+            _KndxIndex.HEADER,
             "a option 0 1 :",
             "b option 1 2 :"
             ])
         index = self.get_knit_index(transport, "filename", "r")
 
-        self.assertEqual((None, 0, 1), index.get_position("a"))
-        self.assertEqual((None, 1, 2), index.get_position("b"))
+        self.assertEqual((("a",), 0, 1), index.get_position(("a",)))
+        self.assertEqual((("b",), 1, 2), index.get_position(("b",)))
 
     def test_get_method(self):
         transport = MockTransport([
-            _KnitIndex.HEADER,
+            _KndxIndex.HEADER,
             "a fulltext,unknown 0 1 :",
             "b unknown,line-delta 1 2 :",
             "c bad 3 4 :"
@@ -833,7 +740,7 @@
 
     def test_get_options(self):
         transport = MockTransport([
-            _KnitIndex.HEADER,
+            _KndxIndex.HEADER,
             "a opt1 0 1 :",
             "b opt2,opt3 1 2 :"
             ])
@@ -844,7 +751,7 @@
 
     def test_get_parent_map(self):
         transport = MockTransport([
-            _KnitIndex.HEADER,
+            _KndxIndex.HEADER,
             "a option 0 1 :",
             "b option 1 2 0 .c :",
             "c option 1 2 1 0 .e :"
@@ -852,52 +759,21 @@
         index = self.get_knit_index(transport, "filename", "r")
 
         self.assertEqual({
-            "a":(),
-            "b":("a", "c"),
-            "c":("b", "a", "e"),
-            }, index.get_parent_map(["a", "b", "c"]))
-
-    def test_get_parents_with_ghosts(self):
-        transport = MockTransport([
-            _KnitIndex.HEADER,
-            "a option 0 1 :",
-            "b option 1 2 0 .c :",
-            "c option 1 2 1 0 .e :"
-            ])
-        index = self.get_knit_index(transport, "filename", "r")
-
-        self.assertEqual((), index.get_parents_with_ghosts("a"))
-        self.assertEqual(("a", "c"), index.get_parents_with_ghosts("b"))
-        self.assertEqual(("b", "a", "e"),
-            index.get_parents_with_ghosts("c"))
-
-    def test_check_versions_present(self):
-        transport = MockTransport([
-            _KnitIndex.HEADER,
-            "a option 0 1 :",
-            "b option 0 1 :"
-            ])
-        index = self.get_knit_index(transport, "filename", "r")
-
-        check = index.check_versions_present
-
-        check([])
-        check(["a"])
-        check(["b"])
-        check(["a", "b"])
-        self.assertRaises(RevisionNotPresent, check, ["c"])
-        self.assertRaises(RevisionNotPresent, check, ["a", "b", "c"])
+            ("a",):(),
+            ("b",):(("a",), ("c",)),
+            ("c",):(("b",), ("a",), ("e",)),
+            }, index.get_parent_map(index.keys()))
 
     def test_impossible_parent(self):
         """Test we get KnitCorrupt if the parent couldn't possibly exist."""
         transport = MockTransport([
-            _KnitIndex.HEADER,
+            _KndxIndex.HEADER,
             "a option 0 1 :",
             "b option 0 1 4 :"  # We don't have a 4th record
             ])
+        index = self.get_knit_index(transport, 'filename', 'r')
         try:
-            self.assertRaises(errors.KnitCorrupt,
-                              self.get_knit_index, transport, 'filename', 'r')
+            self.assertRaises(errors.KnitCorrupt, index.keys)
         except TypeError, e:
             if (str(e) == ('exceptions must be strings, classes, or instances,'
                            ' not exceptions.IndexError')
@@ -910,14 +786,14 @@
 
     def test_corrupted_parent(self):
         transport = MockTransport([
-            _KnitIndex.HEADER,
+            _KndxIndex.HEADER,
             "a option 0 1 :",
             "b option 0 1 :",
             "c option 0 1 1v :", # Can't have a parent of '1v'
             ])
+        index = self.get_knit_index(transport, 'filename', 'r')
         try:
-            self.assertRaises(errors.KnitCorrupt,
-                              self.get_knit_index, transport, 'filename', 'r')
+            self.assertRaises(errors.KnitCorrupt, index.keys)
         except TypeError, e:
             if (str(e) == ('exceptions must be strings, classes, or instances,'
                            ' not exceptions.ValueError')
@@ -930,14 +806,14 @@
 
     def test_corrupted_parent_in_list(self):
         transport = MockTransport([
-            _KnitIndex.HEADER,
+            _KndxIndex.HEADER,
             "a option 0 1 :",
             "b option 0 1 :",
             "c option 0 1 1 v :", # Can't have a parent of 'v'
             ])
+        index = self.get_knit_index(transport, 'filename', 'r')
         try:
-            self.assertRaises(errors.KnitCorrupt,
-                              self.get_knit_index, transport, 'filename', 'r')
+            self.assertRaises(errors.KnitCorrupt, index.keys)
         except TypeError, e:
             if (str(e) == ('exceptions must be strings, classes, or instances,'
                            ' not exceptions.ValueError')
@@ -950,12 +826,12 @@
 
     def test_invalid_position(self):
         transport = MockTransport([
-            _KnitIndex.HEADER,
+            _KndxIndex.HEADER,
             "a option 1v 1 :",
             ])
+        index = self.get_knit_index(transport, 'filename', 'r')
         try:
-            self.assertRaises(errors.KnitCorrupt,
-                              self.get_knit_index, transport, 'filename', 'r')
+            self.assertRaises(errors.KnitCorrupt, index.keys)
         except TypeError, e:
             if (str(e) == ('exceptions must be strings, classes, or instances,'
                            ' not exceptions.ValueError')
@@ -968,12 +844,12 @@
 
     def test_invalid_size(self):
         transport = MockTransport([
-            _KnitIndex.HEADER,
+            _KndxIndex.HEADER,
             "a option 1 1v :",
             ])
+        index = self.get_knit_index(transport, 'filename', 'r')
         try:
-            self.assertRaises(errors.KnitCorrupt,
-                              self.get_knit_index, transport, 'filename', 'r')
+            self.assertRaises(errors.KnitCorrupt, index.keys)
         except TypeError, e:
             if (str(e) == ('exceptions must be strings, classes, or instances,'
                            ' not exceptions.ValueError')
@@ -986,1049 +862,58 @@
 
     def test_short_line(self):
         transport = MockTransport([
-            _KnitIndex.HEADER,
+            _KndxIndex.HEADER,
             "a option 0 10  :",
             "b option 10 10 0", # This line isn't terminated, ignored
             ])
         index = self.get_knit_index(transport, "filename", "r")
-        self.assertEqual(['a'], index.get_versions())
+        self.assertEqual(set([('a',)]), index.keys())
 
     def test_skip_incomplete_record(self):
         # A line with bogus data should just be skipped
         transport = MockTransport([
-            _KnitIndex.HEADER,
+            _KndxIndex.HEADER,
             "a option 0 10  :",
             "b option 10 10 0", # This line isn't terminated, ignored
             "c option 20 10 0 :", # Properly terminated, and starts with '\n'
             ])
         index = self.get_knit_index(transport, "filename", "r")
-        self.assertEqual(['a', 'c'], index.get_versions())
+        self.assertEqual(set([('a',), ('c',)]), index.keys())
 
     def test_trailing_characters(self):
         # A line with bogus data should just be skipped
         transport = MockTransport([
-            _KnitIndex.HEADER,
+            _KndxIndex.HEADER,
             "a option 0 10  :",
             "b option 10 10 0 :a", # This line has extra trailing characters
             "c option 20 10 0 :", # Properly terminated, and starts with '\n'
             ])
         index = self.get_knit_index(transport, "filename", "r")
-        self.assertEqual(['a', 'c'], index.get_versions())
+        self.assertEqual(set([('a',), ('c',)]), index.keys())
 
 
 class LowLevelKnitIndexTests_c(LowLevelKnitIndexTests):
 
     _test_needs_features = [CompiledKnitFeature]
 
-    def get_knit_index(self, *args, **kwargs):
+    def get_knit_index(self, transport, name, mode):
+        mapper = ConstantMapper(name)
         orig = knit._load_data
         def reset():
             knit._load_data = orig
         self.addCleanup(reset)
         from bzrlib._knit_load_data_c import _load_data_c
         knit._load_data = _load_data_c
-        return _KnitIndex(get_scope=lambda:None, *args, **kwargs)
+        allow_writes = lambda: mode == 'w'
+        return _KndxIndex(transport, mapper, lambda:None, allow_writes, lambda:True)
 
 
 class KnitTests(TestCaseWithTransport):
     """Class containing knit test helper routines."""
 
-    def make_test_knit(self, annotate=False, delay_create=False, index=None,
-                       name='test', delta=True, access_mode='w'):
-        if not annotate:
-            factory = KnitPlainFactory()
-        else:
-            factory = None
-        if index is None:
-            index = _KnitIndex(get_transport('.'), name + INDEX_SUFFIX,
-                access_mode, create=True, file_mode=None,
-                create_parent_dir=False, delay_create=delay_create,
-                dir_mode=None, get_scope=lambda:None)
-        access = _KnitAccess(get_transport('.'), name + DATA_SUFFIX, None,
-            None, delay_create, False)
-        return KnitVersionedFile(name, get_transport('.'), factory=factory,
-            create=True, delay_create=delay_create, index=index,
-            access_method=access, delta=delta)
-
-    def assertRecordContentEqual(self, knit, version_id, candidate_content):
-        """Assert that some raw record content matches the raw record content
-        for a particular version_id in the given knit.
-        """
-        index_memo = knit._index.get_position(version_id)
-        record = (version_id, index_memo)
-        [(_, expected_content, _)] = list(knit._data.read_records_iter_raw([record]))
-        self.assertEqual(expected_content, candidate_content)
-
-
-class BasicKnitTests(KnitTests):
-
-    def add_stock_one_and_one_a(self, k):
-        k.add_lines('text-1', [], split_lines(TEXT_1))
-        k.add_lines('text-1a', ['text-1'], split_lines(TEXT_1A))
-
-    def test_knit_constructor(self):
-        """Construct empty k"""
-        self.make_test_knit()
-
-    def test_make_explicit_index(self):
-        """We can supply an index to use."""
-        knit = KnitVersionedFile('test', get_transport('.'),
-            index='strangelove', access_method="a")
-        self.assertEqual(knit._index, 'strangelove')
-
-    def test_knit_add(self):
-        """Store one text in knit and retrieve"""
-        k = self.make_test_knit()
-        k.add_lines('text-1', [], split_lines(TEXT_1))
-        self.assertTrue(k.has_version('text-1'))
-        self.assertEqualDiff(''.join(k.get_lines('text-1')), TEXT_1)
-
-    def test_newline_empty_lines(self):
-        # ensure that ["\n"] round trips ok.
-        knit = self.make_test_knit()
-        knit.add_lines('a', [], ["\n"])
-        knit.add_lines_with_ghosts('b', [], ["\n"])
-        self.assertEqual(["\n"], knit.get_lines('a'))
-        self.assertEqual(["\n"], knit.get_lines('b'))
-        self.assertEqual(['fulltext'], knit._index.get_options('a'))
-        self.assertEqual(['fulltext'], knit._index.get_options('b'))
-        knit.add_lines('c', ['a'], ["\n"])
-        knit.add_lines_with_ghosts('d', ['b'], ["\n"])
-        self.assertEqual(["\n"], knit.get_lines('c'))
-        self.assertEqual(["\n"], knit.get_lines('d'))
-        self.assertEqual(['line-delta'], knit._index.get_options('c'))
-        self.assertEqual(['line-delta'], knit._index.get_options('d'))
-
-    def test_empty_lines(self):
-        # bizarrely, [] is not listed as having no-eol. 
-        knit = self.make_test_knit()
-        knit.add_lines('a', [], [])
-        knit.add_lines_with_ghosts('b', [], [])
-        self.assertEqual([], knit.get_lines('a'))
-        self.assertEqual([], knit.get_lines('b'))
-        self.assertEqual(['fulltext'], knit._index.get_options('a'))
-        self.assertEqual(['fulltext'], knit._index.get_options('b'))
-        knit.add_lines('c', ['a'], [])
-        knit.add_lines_with_ghosts('d', ['b'], [])
-        self.assertEqual([], knit.get_lines('c'))
-        self.assertEqual([], knit.get_lines('d'))
-        self.assertEqual(['line-delta'], knit._index.get_options('c'))
-        self.assertEqual(['line-delta'], knit._index.get_options('d'))
-
-    def test_knit_reload(self):
-        # test that the content in a reloaded knit is correct
-        k = self.make_test_knit()
-        k.add_lines('text-1', [], split_lines(TEXT_1))
-        del k
-        k2 = make_file_knit('test', get_transport('.'), access_mode='r',
-            factory=KnitPlainFactory(), create=True)
-        self.assertTrue(k2.has_version('text-1'))
-        self.assertEqualDiff(''.join(k2.get_lines('text-1')), TEXT_1)
-
-    def test_knit_several(self):
-        """Store several texts in a knit"""
-        k = self.make_test_knit()
-        k.add_lines('text-1', [], split_lines(TEXT_1))
-        k.add_lines('text-2', [], split_lines(TEXT_2))
-        self.assertEqualDiff(''.join(k.get_lines('text-1')), TEXT_1)
-        self.assertEqualDiff(''.join(k.get_lines('text-2')), TEXT_2)
-        
-    def test_repeated_add(self):
-        """Knit traps attempt to replace existing version"""
-        k = self.make_test_knit()
-        k.add_lines('text-1', [], split_lines(TEXT_1))
-        self.assertRaises(RevisionAlreadyPresent, 
-                k.add_lines,
-                'text-1', [], split_lines(TEXT_1))
-
-    def test_empty(self):
-        k = self.make_test_knit(True)
-        k.add_lines('text-1', [], [])
-        self.assertEquals(k.get_lines('text-1'), [])
-
-    def test_incomplete(self):
-        """Test if texts without a ending line-end can be inserted and
-        extracted."""
-        k = make_file_knit('test', get_transport('.'), delta=False, create=True)
-        k.add_lines('text-1', [], ['a\n',    'b'  ])
-        k.add_lines('text-2', ['text-1'], ['a\rb\n', 'b\n'])
-        # reopening ensures maximum room for confusion
-        k = make_file_knit('test', get_transport('.'), delta=False, create=True)
-        self.assertEquals(k.get_lines('text-1'), ['a\n',    'b'  ])
-        self.assertEquals(k.get_lines('text-2'), ['a\rb\n', 'b\n'])
-
-    def test_delta(self):
-        """Expression of knit delta as lines"""
-        k = self.make_test_knit()
-        td = list(line_delta(TEXT_1.splitlines(True),
-                             TEXT_1A.splitlines(True)))
-        self.assertEqualDiff(''.join(td), delta_1_1a)
-        out = apply_line_delta(TEXT_1.splitlines(True), td)
-        self.assertEqualDiff(''.join(out), TEXT_1A)
-
-    def test_add_with_parents(self):
-        """Store in knit with parents"""
-        k = self.make_test_knit()
-        self.add_stock_one_and_one_a(k)
-        self.assertEqual({'text-1':(), 'text-1a':('text-1',)},
-            k.get_parent_map(['text-1', 'text-1a']))
-
-    def test_ancestry(self):
-        """Store in knit with parents"""
-        k = self.make_test_knit()
-        self.add_stock_one_and_one_a(k)
-        self.assertEquals(set(k.get_ancestry(['text-1a'])), set(['text-1a', 'text-1']))
-
-    def test_add_delta(self):
-        """Store in knit with parents"""
-        k = self.make_test_knit(annotate=False)
-        self.add_stock_one_and_one_a(k)
-        self.assertEqualDiff(''.join(k.get_lines('text-1a')), TEXT_1A)
-
-    def test_add_delta_knit_graph_index(self):
-        """Does adding work with a KnitGraphIndex."""
-        index = InMemoryGraphIndex(2)
-        knit_index = KnitGraphIndex(index, add_callback=index.add_nodes,
-            deltas=True)
-        k = self.make_test_knit(annotate=True, index=knit_index)
-        self.add_stock_one_and_one_a(k)
-        self.assertEqualDiff(''.join(k.get_lines('text-1a')), TEXT_1A)
-        # check the index had the right data added.
-        self.assertEqual(set([
-            (index, ('text-1', ), ' 0 127', ((), ())),
-            (index, ('text-1a', ), ' 127 140', ((('text-1', ),), (('text-1', ),))),
-            ]), set(index.iter_all_entries()))
-        # we should not have a .kndx file
-        self.assertFalse(get_transport('.').has('test.kndx'))
-
-    def test_annotate(self):
-        """Annotations"""
-        k = self.make_test_knit(annotate=True, name='knit')
-        self.insert_and_test_small_annotate(k)
-
-    def insert_and_test_small_annotate(self, k):
-        """test annotation with k works correctly."""
-        k.add_lines('text-1', [], ['a\n', 'b\n'])
-        k.add_lines('text-2', ['text-1'], ['a\n', 'c\n'])
-
-        origins = k.annotate('text-2')
-        self.assertEquals(origins[0], ('text-1', 'a\n'))
-        self.assertEquals(origins[1], ('text-2', 'c\n'))
-
-    def test_annotate_fulltext(self):
-        """Annotations"""
-        k = self.make_test_knit(annotate=True, name='knit', delta=False)
-        self.insert_and_test_small_annotate(k)
-
-    def test_annotate_merge_1(self):
-        k = self.make_test_knit(True)
-        k.add_lines('text-a1', [], ['a\n', 'b\n'])
-        k.add_lines('text-a2', [], ['d\n', 'c\n'])
-        k.add_lines('text-am', ['text-a1', 'text-a2'], ['d\n', 'b\n'])
-        origins = k.annotate('text-am')
-        self.assertEquals(origins[0], ('text-a2', 'd\n'))
-        self.assertEquals(origins[1], ('text-a1', 'b\n'))
-
-    def test_annotate_merge_2(self):
-        k = self.make_test_knit(True)
-        k.add_lines('text-a1', [], ['a\n', 'b\n', 'c\n'])
-        k.add_lines('text-a2', [], ['x\n', 'y\n', 'z\n'])
-        k.add_lines('text-am', ['text-a1', 'text-a2'], ['a\n', 'y\n', 'c\n'])
-        origins = k.annotate('text-am')
-        self.assertEquals(origins[0], ('text-a1', 'a\n'))
-        self.assertEquals(origins[1], ('text-a2', 'y\n'))
-        self.assertEquals(origins[2], ('text-a1', 'c\n'))
-
-    def test_annotate_merge_9(self):
-        k = self.make_test_knit(True)
-        k.add_lines('text-a1', [], ['a\n', 'b\n', 'c\n'])
-        k.add_lines('text-a2', [], ['x\n', 'y\n', 'z\n'])
-        k.add_lines('text-am', ['text-a1', 'text-a2'], ['k\n', 'y\n', 'c\n'])
-        origins = k.annotate('text-am')
-        self.assertEquals(origins[0], ('text-am', 'k\n'))
-        self.assertEquals(origins[1], ('text-a2', 'y\n'))
-        self.assertEquals(origins[2], ('text-a1', 'c\n'))
-
-    def test_annotate_merge_3(self):
-        k = self.make_test_knit(True)
-        k.add_lines('text-a1', [], ['a\n', 'b\n', 'c\n'])
-        k.add_lines('text-a2', [] ,['x\n', 'y\n', 'z\n'])
-        k.add_lines('text-am', ['text-a1', 'text-a2'], ['k\n', 'y\n', 'z\n'])
-        origins = k.annotate('text-am')
-        self.assertEquals(origins[0], ('text-am', 'k\n'))
-        self.assertEquals(origins[1], ('text-a2', 'y\n'))
-        self.assertEquals(origins[2], ('text-a2', 'z\n'))
-
-    def test_annotate_merge_4(self):
-        k = self.make_test_knit(True)
-        k.add_lines('text-a1', [], ['a\n', 'b\n', 'c\n'])
-        k.add_lines('text-a2', [], ['x\n', 'y\n', 'z\n'])
-        k.add_lines('text-a3', ['text-a1'], ['a\n', 'b\n', 'p\n'])
-        k.add_lines('text-am', ['text-a2', 'text-a3'], ['a\n', 'b\n', 'z\n'])
-        origins = k.annotate('text-am')
-        self.assertEquals(origins[0], ('text-a1', 'a\n'))
-        self.assertEquals(origins[1], ('text-a1', 'b\n'))
-        self.assertEquals(origins[2], ('text-a2', 'z\n'))
-
-    def test_annotate_merge_5(self):
-        k = self.make_test_knit(True)
-        k.add_lines('text-a1', [], ['a\n', 'b\n', 'c\n'])
-        k.add_lines('text-a2', [], ['d\n', 'e\n', 'f\n'])
-        k.add_lines('text-a3', [], ['x\n', 'y\n', 'z\n'])
-        k.add_lines('text-am',
-                    ['text-a1', 'text-a2', 'text-a3'],
-                    ['a\n', 'e\n', 'z\n'])
-        origins = k.annotate('text-am')
-        self.assertEquals(origins[0], ('text-a1', 'a\n'))
-        self.assertEquals(origins[1], ('text-a2', 'e\n'))
-        self.assertEquals(origins[2], ('text-a3', 'z\n'))
-
-    def test_annotate_file_cherry_pick(self):
-        k = self.make_test_knit(True)
-        k.add_lines('text-1', [], ['a\n', 'b\n', 'c\n'])
-        k.add_lines('text-2', ['text-1'], ['d\n', 'e\n', 'f\n'])
-        k.add_lines('text-3', ['text-2', 'text-1'], ['a\n', 'b\n', 'c\n'])
-        origins = k.annotate('text-3')
-        self.assertEquals(origins[0], ('text-1', 'a\n'))
-        self.assertEquals(origins[1], ('text-1', 'b\n'))
-        self.assertEquals(origins[2], ('text-1', 'c\n'))
-
-    def test_reannotate(self):
-        k1 = make_file_knit('knit1', get_transport('.'),
-                               factory=KnitAnnotateFactory(), create=True)
-        # 0
-        k1.add_lines('text-a', [], ['a\n', 'b\n'])
-        # 1
-        k1.add_lines('text-b', ['text-a'], ['a\n', 'c\n'])
-
-        k2 = make_file_knit('test2', get_transport('.'),
-                               factory=KnitAnnotateFactory(), create=True)
-        k2.insert_record_stream(k1.get_record_stream(k1.versions(),
-            'unordered', False))
-
-        # 2
-        k1.add_lines('text-X', ['text-b'], ['a\n', 'b\n'])
-        # 2
-        k2.add_lines('text-c', ['text-b'], ['z\n', 'c\n'])
-        # 3
-        k2.add_lines('text-Y', ['text-b'], ['b\n', 'c\n'])
-
-        # test-c will have index 3
-        k1.insert_record_stream(k2.get_record_stream(['text-c'],
-            'unordered', False))
-
-        lines = k1.get_lines('text-c')
-        self.assertEquals(lines, ['z\n', 'c\n'])
-
-        origins = k1.annotate('text-c')
-        self.assertEquals(origins[0], ('text-c', 'z\n'))
-        self.assertEquals(origins[1], ('text-b', 'c\n'))
-
-    def test_get_line_delta_texts(self):
-        """Make sure we can call get_texts on text with reused line deltas"""
-        k1 = make_file_knit('test1', get_transport('.'),
-            factory=KnitPlainFactory(), create=True)
-        for t in range(3):
-            if t == 0:
-                parents = []
-            else:
-                parents = ['%d' % (t-1)]
-            k1.add_lines('%d' % t, parents, ['hello\n'] * t)
-        k1.get_texts(('%d' % t) for t in range(3))
-        
-    def test_iter_lines_reads_in_order(self):
-        instrumented_t = get_transport('trace+memory:///')
-        k1 = make_file_knit('id', instrumented_t, create=True, delta=True)
-        self.assertEqual([('get', 'id.kndx',)], instrumented_t._activity)
-        # add texts with no required ordering
-        k1.add_lines('base', [], ['text\n'])
-        k1.add_lines('base2', [], ['text2\n'])
-        # clear the logged activity, but preserve the list instance in case of
-        # clones pointing at it.
-        del instrumented_t._activity[:]
-        # request a last-first iteration
-        results = list(k1.iter_lines_added_or_present_in_versions(
-            ['base2', 'base']))
-        self.assertEqual(
-            [('readv', 'id.knit', [(0, 87), (87, 89)], False, None)],
-            instrumented_t._activity)
-        self.assertEqual([('text\n', 'base'), ('text2\n', 'base2')], results)
-
-    def test_knit_format(self):
-        # this tests that a new knit index file has the expected content
-        # and that is writes the data we expect as records are added.
-        knit = self.make_test_knit(True)
-        # Now knit files are not created until we first add data to them
-        self.assertFileEqual("# bzr knit index 8\n", 'test.kndx')
-        knit.add_lines_with_ghosts('revid', ['a_ghost'], ['a\n'])
-        self.assertFileEqual(
-            "# bzr knit index 8\n"
-            "\n"
-            "revid fulltext 0 84 .a_ghost :",
-            'test.kndx')
-        knit.add_lines_with_ghosts('revid2', ['revid'], ['a\n'])
-        self.assertFileEqual(
-            "# bzr knit index 8\n"
-            "\nrevid fulltext 0 84 .a_ghost :"
-            "\nrevid2 line-delta 84 82 0 :",
-            'test.kndx')
-        # we should be able to load this file again
-        knit = make_file_knit('test', get_transport('.'), access_mode='r')
-        self.assertEqual(['revid', 'revid2'], knit.versions())
-        # write a short write to the file and ensure that its ignored
-        indexfile = file('test.kndx', 'ab')
-        indexfile.write('\nrevid3 line-delta 166 82 1 2 3 4 5 .phwoar:demo ')
-        indexfile.close()
-        # we should be able to load this file again
-        knit = make_file_knit('test', get_transport('.'), access_mode='w')
-        self.assertEqual(['revid', 'revid2'], knit.versions())
-        # and add a revision with the same id the failed write had
-        knit.add_lines('revid3', ['revid2'], ['a\n'])
-        # and when reading it revid3 should now appear.
-        knit = make_file_knit('test', get_transport('.'), access_mode='r')
-        self.assertEqual(['revid', 'revid2', 'revid3'], knit.versions())
-        self.assertEqual({'revid3':('revid2',)}, knit.get_parent_map(['revid3']))
-
-    def test_delay_create(self):
-        """Test that passing delay_create=True creates files late"""
-        knit = self.make_test_knit(annotate=True, delay_create=True)
-        self.failIfExists('test.knit')
-        self.failIfExists('test.kndx')
-        knit.add_lines_with_ghosts('revid', ['a_ghost'], ['a\n'])
-        self.failUnlessExists('test.knit')
-        self.assertFileEqual(
-            "# bzr knit index 8\n"
-            "\n"
-            "revid fulltext 0 84 .a_ghost :",
-            'test.kndx')
-
-    def test_create_parent_dir(self):
-        """create_parent_dir can create knits in nonexistant dirs"""
-        # Has no effect if we don't set 'delay_create'
-        trans = get_transport('.')
-        self.assertRaises(NoSuchFile, make_file_knit, 'dir/test',
-                          trans, access_mode='w', factory=None,
-                          create=True, create_parent_dir=True)
-        # Nothing should have changed yet
-        knit = make_file_knit('dir/test', trans, access_mode='w',
-                                 factory=None, create=True,
-                                 create_parent_dir=True,
-                                 delay_create=True)
-        self.failIfExists('dir/test.knit')
-        self.failIfExists('dir/test.kndx')
-        self.failIfExists('dir')
-        knit.add_lines('revid', [], ['a\n'])
-        self.failUnlessExists('dir')
-        self.failUnlessExists('dir/test.knit')
-        self.assertFileEqual(
-            "# bzr knit index 8\n"
-            "\n"
-            "revid fulltext 0 84  :",
-            'dir/test.kndx')
-
-    def test_create_mode_700(self):
-        trans = get_transport('.')
-        if not trans._can_roundtrip_unix_modebits():
-            # Can't roundtrip, so no need to run this test
-            return
-        knit = make_file_knit('dir/test', trans, access_mode='w', factory=None,
-            create=True, create_parent_dir=True, delay_create=True,
-            file_mode=0600, dir_mode=0700)
-        knit.add_lines('revid', [], ['a\n'])
-        self.assertTransportMode(trans, 'dir', 0700)
-        self.assertTransportMode(trans, 'dir/test.knit', 0600)
-        self.assertTransportMode(trans, 'dir/test.kndx', 0600)
-
-    def test_create_mode_770(self):
-        trans = get_transport('.')
-        if not trans._can_roundtrip_unix_modebits():
-            # Can't roundtrip, so no need to run this test
-            return
-        knit = make_file_knit('dir/test', trans, access_mode='w', factory=None,
-            create=True, create_parent_dir=True, delay_create=True,
-            file_mode=0660, dir_mode=0770)
-        knit.add_lines('revid', [], ['a\n'])
-        self.assertTransportMode(trans, 'dir', 0770)
-        self.assertTransportMode(trans, 'dir/test.knit', 0660)
-        self.assertTransportMode(trans, 'dir/test.kndx', 0660)
-
-    def test_create_mode_777(self):
-        trans = get_transport('.')
-        if not trans._can_roundtrip_unix_modebits():
-            # Can't roundtrip, so no need to run this test
-            return
-        knit = make_file_knit('dir/test', trans, access_mode='w', factory=None,
-            create=True, create_parent_dir=True, delay_create=True,
-            file_mode=0666, dir_mode=0777)
-        knit.add_lines('revid', [], ['a\n'])
-        self.assertTransportMode(trans, 'dir', 0777)
-        self.assertTransportMode(trans, 'dir/test.knit', 0666)
-        self.assertTransportMode(trans, 'dir/test.kndx', 0666)
-
-    def test_plan_merge(self):
-        my_knit = self.make_test_knit(annotate=True)
-        my_knit.add_lines('text1', [], split_lines(TEXT_1))
-        my_knit.add_lines('text1a', ['text1'], split_lines(TEXT_1A))
-        my_knit.add_lines('text1b', ['text1'], split_lines(TEXT_1B))
-        plan = list(my_knit.plan_merge('text1a', 'text1b'))
-        for plan_line, expected_line in zip(plan, AB_MERGE):
-            self.assertEqual(plan_line, expected_line)
-
-
-class GetDataStreamTests(KnitTests):
-    """Tests for get_data_stream."""
-
-    def test_get_stream_empty(self):
-        """Get a data stream for an empty knit file."""
-        k1 = self.make_test_knit()
-        format, data_list, reader_callable = k1.get_data_stream([])
-        self.assertEqual('knit-plain', format)
-        self.assertEqual([], data_list)
-        content = reader_callable(None)
-        self.assertEqual('', content)
-        self.assertIsInstance(content, str)
-
-    def test_get_stream_one_version(self):
-        """Get a data stream for a single record out of a knit containing just
-        one record.
-        """
-        k1 = self.make_test_knit()
-        test_data = [
-            ('text-a', [], TEXT_1),
-            ]
-        expected_data_list = [
-            # version, options, length, parents
-            ('text-a', ['fulltext'], 122, ()),
-           ]
-        for version_id, parents, lines in test_data:
-            k1.add_lines(version_id, parents, split_lines(lines))
-
-        format, data_list, reader_callable = k1.get_data_stream(['text-a'])
-        self.assertEqual('knit-plain', format)
-        self.assertEqual(expected_data_list, data_list)
-        # There's only one record in the knit, so the content should be the
-        # entire knit data file's contents.
-        self.assertEqual(k1.transport.get_bytes(k1._data._access._filename),
-                         reader_callable(None))
-        
-    def test_get_stream_get_one_version_of_many(self):
-        """Get a data stream for just one version out of a knit containing many
-        versions.
-        """
-        k1 = self.make_test_knit()
-        # Insert the same data as test_knit_join, as they seem to cover a range
-        # of cases (no parents, one parent, multiple parents).
-        test_data = [
-            ('text-a', [], TEXT_1),
-            ('text-b', ['text-a'], TEXT_1),
-            ('text-c', [], TEXT_1),
-            ('text-d', ['text-c'], TEXT_1),
-            ('text-m', ['text-b', 'text-d'], TEXT_1),
-            ]
-        expected_data_list = [
-            # version, options, length, parents
-            ('text-m', ['line-delta'], 84, ('text-b', 'text-d')),
-            ]
-        for version_id, parents, lines in test_data:
-            k1.add_lines(version_id, parents, split_lines(lines))
-
-        format, data_list, reader_callable = k1.get_data_stream(['text-m'])
-        self.assertEqual('knit-plain', format)
-        self.assertEqual(expected_data_list, data_list)
-        self.assertRecordContentEqual(k1, 'text-m', reader_callable(None))
-        
-    def test_get_data_stream_unordered_index(self):
-        """Get a data stream when the knit index reports versions out of order.
-
-        https://bugs.launchpad.net/bzr/+bug/164637
-        """
-        k1 = self.make_test_knit()
-        test_data = [
-            ('text-a', [], TEXT_1),
-            ('text-b', ['text-a'], TEXT_1),
-            ('text-c', [], TEXT_1),
-            ('text-d', ['text-c'], TEXT_1),
-            ('text-m', ['text-b', 'text-d'], TEXT_1),
-            ]
-        for version_id, parents, lines in test_data:
-            k1.add_lines(version_id, parents, split_lines(lines))
-        # monkey-patch versions method to return out of order, as if coming
-        # from multiple independently indexed packs
-        original_versions = k1.versions
-        k1.versions = lambda: reversed(original_versions())
-        expected_data_list = [
-            ('text-a', ['fulltext'], 122, ()),
-            ('text-b', ['line-delta'], 84, ('text-a',))]
-        # now check the fulltext is first and the delta second
-        format, data_list, _ = k1.get_data_stream(['text-a', 'text-b'])
-        self.assertEqual('knit-plain', format)
-        self.assertEqual(expected_data_list, data_list)
-        # and that's true if we ask for them in the opposite order too
-        format, data_list, _ = k1.get_data_stream(['text-b', 'text-a'])
-        self.assertEqual(expected_data_list, data_list)
-        # also try requesting more versions
-        format, data_list, _ = k1.get_data_stream([
-            'text-m', 'text-b', 'text-a'])
-        self.assertEqual([
-            ('text-a', ['fulltext'], 122, ()),
-            ('text-b', ['line-delta'], 84, ('text-a',)),
-            ('text-m', ['line-delta'], 84, ('text-b', 'text-d')),
-            ], data_list)
-
-    def test_get_stream_ghost_parent(self):
-        """Get a data stream for a version with a ghost parent."""
-        k1 = self.make_test_knit()
-        # Test data
-        k1.add_lines('text-a', [], split_lines(TEXT_1))
-        k1.add_lines_with_ghosts('text-b', ['text-a', 'text-ghost'],
-                                 split_lines(TEXT_1))
-        # Expected data
-        expected_data_list = [
-            # version, options, length, parents
-            ('text-b', ['line-delta'], 84, ('text-a', 'text-ghost')),
-            ]
-        
-        format, data_list, reader_callable = k1.get_data_stream(['text-b'])
-        self.assertEqual('knit-plain', format)
-        self.assertEqual(expected_data_list, data_list)
-        self.assertRecordContentEqual(k1, 'text-b', reader_callable(None))
-    
-    def test_get_stream_get_multiple_records(self):
-        """Get a stream for multiple records of a knit."""
-        k1 = self.make_test_knit()
-        # Insert the same data as test_knit_join, as they seem to cover a range
-        # of cases (no parents, one parent, multiple parents).
-        test_data = [
-            ('text-a', [], TEXT_1),
-            ('text-b', ['text-a'], TEXT_1),
-            ('text-c', [], TEXT_1),
-            ('text-d', ['text-c'], TEXT_1),
-            ('text-m', ['text-b', 'text-d'], TEXT_1),
-            ]
-        for version_id, parents, lines in test_data:
-            k1.add_lines(version_id, parents, split_lines(lines))
-
-        # This test is actually a bit strict as the order in which they're
-        # returned is not defined.  This matches the current (deterministic)
-        # behaviour.
-        expected_data_list = [
-            # version, options, length, parents
-            ('text-d', ['line-delta'], 84, ('text-c',)),
-            ('text-b', ['line-delta'], 84, ('text-a',)),
-            ]
-        # Note that even though we request the revision IDs in a particular
-        # order, the data stream may return them in any order it likes.  In this
-        # case, they'll be in the order they were inserted into the knit.
-        format, data_list, reader_callable = k1.get_data_stream(
-            ['text-d', 'text-b'])
-        self.assertEqual('knit-plain', format)
-        self.assertEqual(expected_data_list, data_list)
-        # must match order they're returned
-        self.assertRecordContentEqual(k1, 'text-d', reader_callable(84))
-        self.assertRecordContentEqual(k1, 'text-b', reader_callable(84))
-        self.assertEqual('', reader_callable(None),
-                         "There should be no more bytes left to read.")
-
-    def test_get_stream_all(self):
-        """Get a data stream for all the records in a knit.
-
-        This exercises fulltext records, line-delta records, records with
-        various numbers of parents, and reading multiple records out of the
-        callable.  These cases ought to all be exercised individually by the
-        other test_get_stream_* tests; this test is basically just paranoia.
-        """
-        k1 = self.make_test_knit()
-        # Insert the same data as test_knit_join, as they seem to cover a range
-        # of cases (no parents, one parent, multiple parents).
-        test_data = [
-            ('text-a', [], TEXT_1),
-            ('text-b', ['text-a'], TEXT_1),
-            ('text-c', [], TEXT_1),
-            ('text-d', ['text-c'], TEXT_1),
-            ('text-m', ['text-b', 'text-d'], TEXT_1),
-           ]
-        for version_id, parents, lines in test_data:
-            k1.add_lines(version_id, parents, split_lines(lines))
-
-        # This test is actually a bit strict as the order in which they're
-        # returned is not defined.  This matches the current (deterministic)
-        # behaviour.
-        expected_data_list = [
-            # version, options, length, parents
-            ('text-a', ['fulltext'], 122, ()),
-            ('text-b', ['line-delta'], 84, ('text-a',)),
-            ('text-m', ['line-delta'], 84, ('text-b', 'text-d')),
-            ('text-c', ['fulltext'], 121, ()),
-            ('text-d', ['line-delta'], 84, ('text-c',)),
-            ]
-        format, data_list, reader_callable = k1.get_data_stream(
-            ['text-a', 'text-b', 'text-c', 'text-d', 'text-m'])
-        self.assertEqual('knit-plain', format)
-        self.assertEqual(expected_data_list, data_list)
-        for version_id, options, length, parents in expected_data_list:
-            bytes = reader_callable(length)
-            self.assertRecordContentEqual(k1, version_id, bytes)
-
-
-class InsertDataStreamTests(KnitTests):
-    """Tests for insert_data_stream."""
-
-    def assertKnitFilesEqual(self, knit1, knit2):
-        """Assert that the contents of the index and data files of two knits are
-        equal.
-        """
-        self.assertEqual(
-            knit1.transport.get_bytes(knit1._data._access._filename),
-            knit2.transport.get_bytes(knit2._data._access._filename))
-        self.assertEqual(
-            knit1.transport.get_bytes(knit1._index._filename),
-            knit2.transport.get_bytes(knit2._index._filename))
-
-    def assertKnitValuesEqual(self, left, right):
-        """Assert that the texts, annotations and graph of left and right are
-        the same.
-        """
-        self.assertEqual(set(left.versions()), set(right.versions()))
-        for version in left.versions():
-            self.assertEqual(left.get_parents_with_ghosts(version),
-                right.get_parents_with_ghosts(version))
-            self.assertEqual(left.get_lines(version),
-                right.get_lines(version))
-            self.assertEqual(left.annotate(version),
-                right.annotate(version))
-
-    def test_empty_stream(self):
-        """Inserting a data stream with no records should not put any data into
-        the knit.
-        """
-        k1 = self.make_test_knit()
-        k1.insert_data_stream(
-            (k1.get_format_signature(), [], lambda ignored: ''))
-        self.assertEqual('', k1.transport.get_bytes(k1._data._access._filename),
-                         "The .knit should be completely empty.")
-        self.assertEqual(k1._index.HEADER,
-                         k1.transport.get_bytes(k1._index._filename),
-                         "The .kndx should have nothing apart from the header.")
-
-    def test_one_record(self):
-        """Inserting a data stream with one record from a knit with one record
-        results in byte-identical files.
-        """
-        source = self.make_test_knit(name='source')
-        source.add_lines('text-a', [], split_lines(TEXT_1))
-        data_stream = source.get_data_stream(['text-a'])
-        target = self.make_test_knit(name='target')
-        target.insert_data_stream(data_stream)
-        self.assertKnitFilesEqual(source, target)
-
-    def test_annotated_stream_into_unannotated_knit(self):
-        """Inserting an annotated datastream to an unannotated knit works."""
-        # case one - full texts.
-        source = self.make_test_knit(name='source', annotate=True)
-        target = self.make_test_knit(name='target', annotate=False)
-        source.add_lines('text-a', [], split_lines(TEXT_1))
-        target.insert_data_stream(source.get_data_stream(['text-a']))
-        self.assertKnitValuesEqual(source, target)
-        # case two - deltas.
-        source.add_lines('text-b', ['text-a'], split_lines(TEXT_2))
-        target.insert_data_stream(source.get_data_stream(['text-b']))
-        self.assertKnitValuesEqual(source, target)
-
-    def test_unannotated_stream_into_annotated_knit(self):
-        """Inserting an unannotated datastream to an annotated knit works."""
-        # case one - full texts.
-        source = self.make_test_knit(name='source', annotate=False)
-        target = self.make_test_knit(name='target', annotate=True)
-        source.add_lines('text-a', [], split_lines(TEXT_1))
-        target.insert_data_stream(source.get_data_stream(['text-a']))
-        self.assertKnitValuesEqual(source, target)
-        # case two - deltas.
-        source.add_lines('text-b', ['text-a'], split_lines(TEXT_2))
-        target.insert_data_stream(source.get_data_stream(['text-b']))
-        self.assertKnitValuesEqual(source, target)
-
-    def test_records_already_present(self):
-        """Insert a data stream where some records are alreday present in the
-        target, and some not.  Only the new records are inserted.
-        """
-        source = self.make_test_knit(name='source')
-        target = self.make_test_knit(name='target')
-        # Insert 'text-a' into both source and target
-        source.add_lines('text-a', [], split_lines(TEXT_1))
-        target.insert_data_stream(source.get_data_stream(['text-a']))
-        # Insert 'text-b' into just the source.
-        source.add_lines('text-b', ['text-a'], split_lines(TEXT_1))
-        # Get a data stream of both text-a and text-b, and insert it.
-        data_stream = source.get_data_stream(['text-a', 'text-b'])
-        target.insert_data_stream(data_stream)
-        # The source and target will now be identical.  This means the text-a
-        # record was not added a second time.
-        self.assertKnitFilesEqual(source, target)
-
-    def test_multiple_records(self):
-        """Inserting a data stream of all records from a knit with multiple
-        records results in byte-identical files.
-        """
-        source = self.make_test_knit(name='source')
-        source.add_lines('text-a', [], split_lines(TEXT_1))
-        source.add_lines('text-b', ['text-a'], split_lines(TEXT_1))
-        source.add_lines('text-c', [], split_lines(TEXT_1))
-        data_stream = source.get_data_stream(['text-a', 'text-b', 'text-c'])
-        
-        target = self.make_test_knit(name='target')
-        target.insert_data_stream(data_stream)
-        
-        self.assertKnitFilesEqual(source, target)
-
-    def test_ghost_parent(self):
-        """Insert a data stream with a record that has a ghost parent."""
-        # Make a knit with a record, text-a, that has a ghost parent.
-        source = self.make_test_knit(name='source')
-        source.add_lines_with_ghosts('text-a', ['text-ghost'],
-                                     split_lines(TEXT_1))
-        data_stream = source.get_data_stream(['text-a'])
-
-        target = self.make_test_knit(name='target')
-        target.insert_data_stream(data_stream)
-
-        self.assertKnitFilesEqual(source, target)
-
-        # The target knit object is in a consistent state, i.e. the record we
-        # just added is immediately visible.
-        self.assertTrue(target.has_version('text-a'))
-        self.assertFalse(target.has_version('text-ghost'))
-        self.assertEqual({'text-a':('text-ghost',)},
-            target.get_parent_map(['text-a', 'text-ghost']))
-        self.assertEqual(split_lines(TEXT_1), target.get_lines('text-a'))
-
-    def test_inconsistent_version_lines(self):
-        """Inserting a data stream which has different content for a version_id
-        than already exists in the knit will raise KnitCorrupt.
-        """
-        source = self.make_test_knit(name='source')
-        target = self.make_test_knit(name='target')
-        # Insert a different 'text-a' into both source and target
-        source.add_lines('text-a', [], split_lines(TEXT_1))
-        target.add_lines('text-a', [], split_lines(TEXT_2))
-        # Insert a data stream with conflicting content into the target
-        data_stream = source.get_data_stream(['text-a'])
-        self.assertRaises(
-            errors.KnitCorrupt, target.insert_data_stream, data_stream)
-
-    def test_inconsistent_version_parents(self):
-        """Inserting a data stream which has different parents for a version_id
-        than already exists in the knit will raise KnitCorrupt.
-        """
-        source = self.make_test_knit(name='source')
-        target = self.make_test_knit(name='target')
-        # Insert a different 'text-a' into both source and target.  They differ
-        # only by the parents list, the content is the same.
-        source.add_lines_with_ghosts('text-a', [], split_lines(TEXT_1))
-        target.add_lines_with_ghosts('text-a', ['a-ghost'], split_lines(TEXT_1))
-        # Insert a data stream with conflicting content into the target
-        data_stream = source.get_data_stream(['text-a'])
-        self.assertRaises(
-            errors.KnitCorrupt, target.insert_data_stream, data_stream)
-
-    def test_unknown_stream_format(self):
-        """A data stream in a different format to the target knit cannot be
-        inserted.
-
-        It will raise KnitDataStreamUnknown because the fallback code will fail
-        to make a knit. In future we may need KnitDataStreamIncompatible again,
-        for more exotic cases.
-        """
-        data_stream = ('fake-format-signature', [], lambda _: '')
-        target = self.make_test_knit(name='target')
-        self.assertRaises(
-            errors.KnitDataStreamUnknown,
-            target.insert_data_stream, data_stream)
-
-    def test_bug_208418(self):
-        """You can insert a stream with an incompatible format, even when:
-          * the stream has a line-delta record,
-          * whose parent is in the target, also stored as a line-delta
-
-        See <https://launchpad.net/bugs/208418>.
-        """
-        base_lines = split_lines(TEXT_1)
-        # Make the target
-        target = self.make_test_knit(name='target', annotate=True)
-        target.add_lines('version-1', [], base_lines)
-        target.add_lines('version-2', ['version-1'], base_lines + ['a\n'])
-        # The second record should be a delta.
-        self.assertEqual('line-delta', target._index.get_method('version-2'))
-        
-        # Make a source, with a different format, but the same data
-        source = self.make_test_knit(name='source', annotate=False)
-        source.add_lines('version-1', [], base_lines)
-        source.add_lines('version-2', ['version-1'], base_lines + ['a\n'])
-        # Now add another record, which should be stored as a delta against
-        # version-2.
-        source.add_lines('version-3', ['version-2'], base_lines + ['b\n'])
-        self.assertEqual('line-delta', source._index.get_method('version-3'))
-
-        # Make a stream of the new version
-        data_stream = source.get_data_stream(['version-3'])
-        # And insert into the target
-        target.insert_data_stream(data_stream)
-        # No errors should have been raised.
-
-    def test_line_delta_record_into_non_delta_knit(self):
-        # Make a data stream with a line-delta record
-        source = self.make_test_knit(name='source', delta=True)
-        base_lines = split_lines(TEXT_1)
-        source.add_lines('version-1', [], base_lines)
-        source.add_lines('version-2', ['version-1'], base_lines + ['a\n'])
-        # The second record should be a delta.
-        self.assertEqual('line-delta', source._index.get_method('version-2'))
-        data_stream = source.get_data_stream(['version-1', 'version-2'])
-
-        # Insert the stream into a non-delta knit.
-        target = self.make_test_knit(name='target', delta=False)
-        target.insert_data_stream(data_stream)
-        
-        # Both versions are fulltexts in the target
-        self.assertEqual('fulltext', target._index.get_method('version-1'))
-        self.assertEqual('fulltext', target._index.get_method('version-2'))
-
-
-class DataStreamTests(KnitTests):
-
-    def assertMadeStreamKnit(self, source_knit, versions, target_knit):
-        """Assert that a knit made from a stream is as expected."""
-        a_stream = source_knit.get_data_stream(versions)
-        expected_data = a_stream[2](None)
-        a_stream = source_knit.get_data_stream(versions)
-        a_knit = target_knit._knit_from_datastream(a_stream)
-        self.assertEqual(source_knit.factory.__class__,
-            a_knit.factory.__class__)
-        self.assertIsInstance(a_knit._data._access, _StreamAccess)
-        self.assertIsInstance(a_knit._index, _StreamIndex)
-        self.assertEqual(a_knit._index.data_list, a_stream[1])
-        self.assertEqual(a_knit._data._access.data, expected_data)
-        self.assertEqual(a_knit.filename, target_knit.filename)
-        self.assertEqual(a_knit.transport, target_knit.transport)
-        self.assertEqual(a_knit._index, a_knit._data._access.stream_index)
-        self.assertEqual(target_knit, a_knit._data._access.backing_knit)
-        self.assertIsInstance(a_knit._data._access.orig_factory,
-            source_knit.factory.__class__)
-
-    def test__knit_from_data_stream_empty(self):
-        """Create a knit object from a datastream."""
-        annotated = self.make_test_knit(name='source', annotate=True)
-        plain = self.make_test_knit(name='target', annotate=False)
-        # case 1: annotated source
-        self.assertMadeStreamKnit(annotated, [], annotated)
-        self.assertMadeStreamKnit(annotated, [], plain)
-        # case 2: plain source
-        self.assertMadeStreamKnit(plain, [], annotated)
-        self.assertMadeStreamKnit(plain, [], plain)
-
-    def test__knit_from_data_stream_unknown_format(self):
-        annotated = self.make_test_knit(name='source', annotate=True)
-        self.assertRaises(errors.KnitDataStreamUnknown,
-            annotated._knit_from_datastream, ("unknown", None, None))
-
-
-TEXT_1 = """\
-Banana cup cakes:
-
-- bananas
-- eggs
-- broken tea cups
-"""
-
-TEXT_1A = """\
-Banana cup cake recipe
-(serves 6)
-
-- bananas
-- eggs
-- broken tea cups
-- self-raising flour
-"""
-
-TEXT_1B = """\
-Banana cup cake recipe
-
-- bananas (do not use plantains!!!)
-- broken tea cups
-- flour
-"""
-
-delta_1_1a = """\
-0,1,2
-Banana cup cake recipe
-(serves 6)
-5,5,1
-- self-raising flour
-"""
-
-TEXT_2 = """\
-Boeuf bourguignon
-
-- beef
-- red wine
-- small onions
-- carrot
-- mushrooms
-"""
-
-AB_MERGE_TEXT="""unchanged|Banana cup cake recipe
-new-a|(serves 6)
-unchanged|
-killed-b|- bananas
-killed-b|- eggs
-new-b|- bananas (do not use plantains!!!)
-unchanged|- broken tea cups
-new-a|- self-raising flour
-new-b|- flour
-"""
-AB_MERGE=[tuple(l.split('|')) for l in AB_MERGE_TEXT.splitlines(True)]
-
-
-def line_delta(from_lines, to_lines):
-    """Generate line-based delta from one text to another"""
-    s = difflib.SequenceMatcher(None, from_lines, to_lines)
-    for op in s.get_opcodes():
-        if op[0] == 'equal':
-            continue
-        yield '%d,%d,%d\n' % (op[1], op[2], op[4]-op[3])
-        for i in range(op[3], op[4]):
-            yield to_lines[i]
-
-
-def apply_line_delta(basis_lines, delta_lines):
-    """Apply a line-based perfect diff
-    
-    basis_lines -- text to apply the patch to
-    delta_lines -- diff instructions and content
-    """
-    out = basis_lines[:]
-    i = 0
-    offset = 0
-    while i < len(delta_lines):
-        l = delta_lines[i]
-        a, b, c = map(long, l.split(','))
-        i = i + 1
-        out[offset+a:offset+b] = delta_lines[i:i+c]
-        i = i + c
-        offset = offset + (b - a) + c
-    return out
-
-
-class TestWeaveToKnit(KnitTests):
-
-    def test_weave_to_knit_matches(self):
-        # check that the WeaveToKnit is_compatible function
-        # registers True for a Weave to a Knit.
-        w = Weave(get_scope=lambda:None)
-        k = self.make_test_knit()
-        self.failUnless(WeaveToKnit.is_compatible(w, k))
-        self.failIf(WeaveToKnit.is_compatible(k, w))
-        self.failIf(WeaveToKnit.is_compatible(w, w))
-        self.failIf(WeaveToKnit.is_compatible(k, k))
+    def make_test_knit(self, annotate=False, name='test'):
+        mapper = ConstantMapper(name)
+        return make_file_factory(annotate, mapper)(self.get_transport())
 
 
 class TestKnitIndex(KnitTests):
@@ -2037,15 +922,16 @@
         """Adding versions to the index should update the lookup dict"""
         knit = self.make_test_knit()
         idx = knit._index
-        idx.add_version('a-1', ['fulltext'], (None, 0, 0), [])
+        idx.add_records([(('a-1',), ['fulltext'], (('a-1',), 0, 0), [])])
         self.check_file_contents('test.kndx',
             '# bzr knit index 8\n'
             '\n'
             'a-1 fulltext 0 0  :'
             )
-        idx.add_versions([('a-2', ['fulltext'], (None, 0, 0), ['a-1']),
-                          ('a-3', ['fulltext'], (None, 0, 0), ['a-2']),
-                         ])
+        idx.add_records([
+            (('a-2',), ['fulltext'], (('a-2',), 0, 0), [('a-1',)]),
+            (('a-3',), ['fulltext'], (('a-3',), 0, 0), [('a-2',)]),
+            ])
         self.check_file_contents('test.kndx',
             '# bzr knit index 8\n'
             '\n'
@@ -2053,11 +939,16 @@
             'a-2 fulltext 0 0 0 :\n'
             'a-3 fulltext 0 0 1 :'
             )
-        self.assertEqual(['a-1', 'a-2', 'a-3'], idx._history)
-        self.assertEqual({'a-1':('a-1', ['fulltext'], 0, 0, (), 0),
-                          'a-2':('a-2', ['fulltext'], 0, 0, ('a-1',), 1),
-                          'a-3':('a-3', ['fulltext'], 0, 0, ('a-2',), 2),
-                         }, idx._cache)
+        self.assertEqual(set([('a-3',), ('a-1',), ('a-2',)]), idx.keys())
+        self.assertEqual({
+            ('a-1',): ((('a-1',), 0, 0), None, (), ('fulltext', False)),
+            ('a-2',): ((('a-2',), 0, 0), None, (('a-1',),), ('fulltext', False)),
+            ('a-3',): ((('a-3',), 0, 0), None, (('a-2',),), ('fulltext', False)),
+            }, idx.get_build_details(idx.keys()))
+        self.assertEqual({('a-1',):(),
+            ('a-2',):(('a-1',),),
+            ('a-3',):(('a-2',),),},
+            idx.get_parent_map(idx.keys()))
 
     def test_add_versions_fails_clean(self):
         """If add_versions fails in the middle, it restores a pristine state.
@@ -2073,26 +964,29 @@
 
         knit = self.make_test_knit()
         idx = knit._index
-        idx.add_version('a-1', ['fulltext'], (None, 0, 0), [])
+        idx.add_records([(('a-1',), ['fulltext'], (('a-1',), 0, 0), [])])
 
         class StopEarly(Exception):
             pass
 
         def generate_failure():
             """Add some entries and then raise an exception"""
-            yield ('a-2', ['fulltext'], (None, 0, 0), ('a-1',))
-            yield ('a-3', ['fulltext'], (None, 0, 0), ('a-2',))
+            yield (('a-2',), ['fulltext'], (None, 0, 0), ('a-1',))
+            yield (('a-3',), ['fulltext'], (None, 0, 0), ('a-2',))
             raise StopEarly()
 
         # Assert the pre-condition
-        self.assertEqual(['a-1'], idx._history)
-        self.assertEqual({'a-1':('a-1', ['fulltext'], 0, 0, (), 0)}, idx._cache)
-
-        self.assertRaises(StopEarly, idx.add_versions, generate_failure())
-
+        def assertA1Only():
+            self.assertEqual(set([('a-1',)]), set(idx.keys()))
+            self.assertEqual(
+                {('a-1',): ((('a-1',), 0, 0), None, (), ('fulltext', False))},
+                idx.get_build_details([('a-1',)]))
+            self.assertEqual({('a-1',):()}, idx.get_parent_map(idx.keys()))
+
+        assertA1Only()
+        self.assertRaises(StopEarly, idx.add_records, generate_failure())
         # And it shouldn't be modified
-        self.assertEqual(['a-1'], idx._history)
-        self.assertEqual({'a-1':('a-1', ['fulltext'], 0, 0, (), 0)}, idx._cache)
+        assertA1Only()
 
     def test_knit_index_ignores_empty_files(self):
         # There was a race condition in older bzr, where a ^C at the right time
@@ -2107,8 +1001,8 @@
     def test_knit_index_checks_header(self):
         t = get_transport('.')
         t.put_bytes('test.kndx', '# not really a knit header\n\n')
-
-        self.assertRaises(KnitHeaderError, self.make_test_knit)
+        k = self.make_test_knit()
+        self.assertRaises(KnitHeaderError, k.keys)
 
 
 class TestGraphIndexKnit(KnitTests):
@@ -2153,166 +1047,100 @@
             add_callback = self.catch_add
         else:
             add_callback = None
-        return KnitGraphIndex(combined_index, deltas=deltas,
+        return _KnitGraphIndex(combined_index, lambda:True, deltas=deltas,
             add_callback=add_callback)
 
-    def test_get_ancestry(self):
-        # get_ancestry is defined as eliding ghosts, not erroring.
-        index = self.two_graph_index()
-        self.assertEqual([], index.get_ancestry([]))
-        self.assertEqual(['separate'], index.get_ancestry(['separate']))
-        self.assertEqual(['tail'], index.get_ancestry(['tail']))
-        self.assertEqual(['tail', 'parent'], index.get_ancestry(['parent']))
-        self.assertEqual(['tail', 'parent', 'tip'], index.get_ancestry(['tip']))
-        self.assertTrue(index.get_ancestry(['tip', 'separate']) in
-            (['tail', 'parent', 'tip', 'separate'],
-             ['separate', 'tail', 'parent', 'tip'],
-            ))
-        # and without topo_sort
-        self.assertEqual(set(['separate']),
-            set(index.get_ancestry(['separate'], topo_sorted=False)))
-        self.assertEqual(set(['tail']),
-            set(index.get_ancestry(['tail'], topo_sorted=False)))
-        self.assertEqual(set(['tail', 'parent']),
-            set(index.get_ancestry(['parent'], topo_sorted=False)))
-        self.assertEqual(set(['tail', 'parent', 'tip']),
-            set(index.get_ancestry(['tip'], topo_sorted=False)))
-        self.assertEqual(set(['separate', 'tail', 'parent', 'tip']),
-            set(index.get_ancestry(['tip', 'separate'])))
-        # asking for a ghost makes it go boom.
-        self.assertRaises(errors.RevisionNotPresent, index.get_ancestry, ['ghost'])
-
-    def test_get_ancestry_with_ghosts(self):
-        index = self.two_graph_index()
-        self.assertEqual([], index.get_ancestry_with_ghosts([]))
-        self.assertEqual(['separate'], index.get_ancestry_with_ghosts(['separate']))
-        self.assertEqual(['tail'], index.get_ancestry_with_ghosts(['tail']))
-        self.assertTrue(index.get_ancestry_with_ghosts(['parent']) in
-            (['tail', 'ghost', 'parent'],
-             ['ghost', 'tail', 'parent'],
-            ))
-        self.assertTrue(index.get_ancestry_with_ghosts(['tip']) in
-            (['tail', 'ghost', 'parent', 'tip'],
-             ['ghost', 'tail', 'parent', 'tip'],
-            ))
-        self.assertTrue(index.get_ancestry_with_ghosts(['tip', 'separate']) in
-            (['tail', 'ghost', 'parent', 'tip', 'separate'],
-             ['ghost', 'tail', 'parent', 'tip', 'separate'],
-             ['separate', 'tail', 'ghost', 'parent', 'tip'],
-             ['separate', 'ghost', 'tail', 'parent', 'tip'],
-            ))
-        # asking for a ghost makes it go boom.
-        self.assertRaises(errors.RevisionNotPresent, index.get_ancestry_with_ghosts, ['ghost'])
-
-    def test_num_versions(self):
-        index = self.two_graph_index()
-        self.assertEqual(4, index.num_versions())
-
-    def test_get_versions(self):
-        index = self.two_graph_index()
-        self.assertEqual(set(['tail', 'tip', 'parent', 'separate']),
-            set(index.get_versions()))
-
-    def test_has_version(self):
-        index = self.two_graph_index()
-        self.assertTrue(index.has_version('tail'))
-        self.assertFalse(index.has_version('ghost'))
+    def test_keys(self):
+        index = self.two_graph_index()
+        self.assertEqual(set([('tail',), ('tip',), ('parent',), ('separate',)]),
+            set(index.keys()))
 
     def test_get_position(self):
         index = self.two_graph_index()
-        self.assertEqual((index._graph_index._indices[0], 0, 100), index.get_position('tip'))
-        self.assertEqual((index._graph_index._indices[1], 100, 78), index.get_position('parent'))
+        self.assertEqual((index._graph_index._indices[0], 0, 100), index.get_position(('tip',)))
+        self.assertEqual((index._graph_index._indices[1], 100, 78), index.get_position(('parent',)))
 
     def test_get_method_deltas(self):
         index = self.two_graph_index(deltas=True)
-        self.assertEqual('fulltext', index.get_method('tip'))
-        self.assertEqual('line-delta', index.get_method('parent'))
+        self.assertEqual('fulltext', index.get_method(('tip',)))
+        self.assertEqual('line-delta', index.get_method(('parent',)))
 
     def test_get_method_no_deltas(self):
         # check that the parent-history lookup is ignored with deltas=False.
         index = self.two_graph_index(deltas=False)
-        self.assertEqual('fulltext', index.get_method('tip'))
-        self.assertEqual('fulltext', index.get_method('parent'))
+        self.assertEqual('fulltext', index.get_method(('tip',)))
+        self.assertEqual('fulltext', index.get_method(('parent',)))
 
     def test_get_options_deltas(self):
         index = self.two_graph_index(deltas=True)
-        self.assertEqual(['fulltext', 'no-eol'], index.get_options('tip'))
-        self.assertEqual(['line-delta'], index.get_options('parent'))
+        self.assertEqual(['fulltext', 'no-eol'], index.get_options(('tip',)))
+        self.assertEqual(['line-delta'], index.get_options(('parent',)))
 
     def test_get_options_no_deltas(self):
         # check that the parent-history lookup is ignored with deltas=False.
         index = self.two_graph_index(deltas=False)
-        self.assertEqual(['fulltext', 'no-eol'], index.get_options('tip'))
-        self.assertEqual(['fulltext'], index.get_options('parent'))
-
-    def test_get_parents_with_ghosts(self):
-        index = self.two_graph_index()
-        self.assertEqual(('tail', 'ghost'), index.get_parents_with_ghosts('parent'))
-        # and errors on ghosts.
-        self.assertRaises(errors.RevisionNotPresent,
-            index.get_parents_with_ghosts, 'ghost')
-
-    def test_check_versions_present(self):
-        # ghosts should not be considered present
-        index = self.two_graph_index()
-        self.assertRaises(RevisionNotPresent, index.check_versions_present,
-            ['ghost'])
-        self.assertRaises(RevisionNotPresent, index.check_versions_present,
-            ['tail', 'ghost'])
-        index.check_versions_present(['tail', 'separate'])
+        self.assertEqual(['fulltext', 'no-eol'], index.get_options(('tip',)))
+        self.assertEqual(['fulltext'], index.get_options(('parent',)))
+
+    def test_get_parent_map(self):
+        index = self.two_graph_index()
+        self.assertEqual({('parent',):(('tail',), ('ghost',))},
+            index.get_parent_map([('parent',), ('ghost',)]))
 
     def catch_add(self, entries):
         self.caught_entries.append(entries)
 
     def test_add_no_callback_errors(self):
         index = self.two_graph_index()
-        self.assertRaises(errors.ReadOnlyError, index.add_version,
-            'new', 'fulltext,no-eol', (None, 50, 60), ['separate'])
+        self.assertRaises(errors.ReadOnlyError, index.add_records,
+            [(('new',), 'fulltext,no-eol', (None, 50, 60), ['separate'])])
 
     def test_add_version_smoke(self):
         index = self.two_graph_index(catch_adds=True)
-        index.add_version('new', 'fulltext,no-eol', (None, 50, 60), ['separate'])
+        index.add_records([(('new',), 'fulltext,no-eol', (None, 50, 60),
+            [('separate',)])])
         self.assertEqual([[(('new', ), 'N50 60', ((('separate',),),))]],
             self.caught_entries)
 
     def test_add_version_delta_not_delta_index(self):
         index = self.two_graph_index(catch_adds=True)
-        self.assertRaises(errors.KnitCorrupt, index.add_version,
-            'new', 'no-eol,line-delta', (None, 0, 100), ['parent'])
+        self.assertRaises(errors.KnitCorrupt, index.add_records,
+            [(('new',), 'no-eol,line-delta', (None, 0, 100), [('parent',)])])
         self.assertEqual([], self.caught_entries)
 
     def test_add_version_same_dup(self):
         index = self.two_graph_index(catch_adds=True)
         # options can be spelt two different ways
-        index.add_version('tip', 'fulltext,no-eol', (None, 0, 100), ['parent'])
-        index.add_version('tip', 'no-eol,fulltext', (None, 0, 100), ['parent'])
-        # but neither should have added data.
-        self.assertEqual([[], []], self.caught_entries)
+        index.add_records([(('tip',), 'fulltext,no-eol', (None, 0, 100), [('parent',)])])
+        index.add_records([(('tip',), 'no-eol,fulltext', (None, 0, 100), [('parent',)])])
+        # position/length are ignored (because each pack could have fulltext or
+        # delta, and be at a different position.
+        index.add_records([(('tip',), 'fulltext,no-eol', (None, 50, 100),
+            [('parent',)])])
+        index.add_records([(('tip',), 'fulltext,no-eol', (None, 0, 1000),
+            [('parent',)])])
+        # but neither should have added data:
+        self.assertEqual([[], [], [], []], self.caught_entries)
         
     def test_add_version_different_dup(self):
         index = self.two_graph_index(deltas=True, catch_adds=True)
         # change options
-        self.assertRaises(errors.KnitCorrupt, index.add_version,
-            'tip', 'no-eol,line-delta', (None, 0, 100), ['parent'])
-        self.assertRaises(errors.KnitCorrupt, index.add_version,
-            'tip', 'line-delta,no-eol', (None, 0, 100), ['parent'])
-        self.assertRaises(errors.KnitCorrupt, index.add_version,
-            'tip', 'fulltext', (None, 0, 100), ['parent'])
-        # position/length
-        self.assertRaises(errors.KnitCorrupt, index.add_version,
-            'tip', 'fulltext,no-eol', (None, 50, 100), ['parent'])
-        self.assertRaises(errors.KnitCorrupt, index.add_version,
-            'tip', 'fulltext,no-eol', (None, 0, 1000), ['parent'])
+        self.assertRaises(errors.KnitCorrupt, index.add_records,
+            [(('tip',), 'no-eol,line-delta', (None, 0, 100), [('parent',)])])
+        self.assertRaises(errors.KnitCorrupt, index.add_records,
+            [(('tip',), 'line-delta,no-eol', (None, 0, 100), [('parent',)])])
+        self.assertRaises(errors.KnitCorrupt, index.add_records,
+            [(('tip',), 'fulltext', (None, 0, 100), [('parent',)])])
         # parents
-        self.assertRaises(errors.KnitCorrupt, index.add_version,
-            'tip', 'fulltext,no-eol', (None, 0, 100), [])
+        self.assertRaises(errors.KnitCorrupt, index.add_records,
+            [(('tip',), 'fulltext,no-eol', (None, 0, 100), [])])
         self.assertEqual([], self.caught_entries)
         
     def test_add_versions_nodeltas(self):
         index = self.two_graph_index(catch_adds=True)
-        index.add_versions([
-                ('new', 'fulltext,no-eol', (None, 50, 60), ['separate']),
-                ('new2', 'fulltext', (None, 0, 6), ['new']),
+        index.add_records([
+                (('new',), 'fulltext,no-eol', (None, 50, 60), [('separate',)]),
+                (('new2',), 'fulltext', (None, 0, 6), [('new',)]),
                 ])
         self.assertEqual([(('new', ), 'N50 60', ((('separate',),),)),
             (('new2', ), ' 0 6', ((('new',),),))],
@@ -2321,9 +1149,9 @@
 
     def test_add_versions_deltas(self):
         index = self.two_graph_index(deltas=True, catch_adds=True)
-        index.add_versions([
-                ('new', 'fulltext,no-eol', (None, 50, 60), ['separate']),
-                ('new2', 'line-delta', (None, 0, 6), ['new']),
+        index.add_records([
+                (('new',), 'fulltext,no-eol', (None, 50, 60), [('separate',)]),
+                (('new2',), 'line-delta', (None, 0, 6), [('new',)]),
                 ])
         self.assertEqual([(('new', ), 'N50 60', ((('separate',),), ())),
             (('new2', ), ' 0 6', ((('new',),), (('new',),), ))],
@@ -2332,47 +1160,51 @@
 
     def test_add_versions_delta_not_delta_index(self):
         index = self.two_graph_index(catch_adds=True)
-        self.assertRaises(errors.KnitCorrupt, index.add_versions,
-            [('new', 'no-eol,line-delta', (None, 0, 100), ['parent'])])
+        self.assertRaises(errors.KnitCorrupt, index.add_records,
+            [(('new',), 'no-eol,line-delta', (None, 0, 100), [('parent',)])])
         self.assertEqual([], self.caught_entries)
 
     def test_add_versions_random_id_accepted(self):
         index = self.two_graph_index(catch_adds=True)
-        index.add_versions([], random_id=True)
+        index.add_records([], random_id=True)
 
     def test_add_versions_same_dup(self):
         index = self.two_graph_index(catch_adds=True)
         # options can be spelt two different ways
-        index.add_versions([('tip', 'fulltext,no-eol', (None, 0, 100), ['parent'])])
-        index.add_versions([('tip', 'no-eol,fulltext', (None, 0, 100), ['parent'])])
+        index.add_records([(('tip',), 'fulltext,no-eol', (None, 0, 100),
+            [('parent',)])])
+        index.add_records([(('tip',), 'no-eol,fulltext', (None, 0, 100),
+            [('parent',)])])
+        # position/length are ignored (because each pack could have fulltext or
+        # delta, and be at a different position.
+        index.add_records([(('tip',), 'fulltext,no-eol', (None, 50, 100),
+            [('parent',)])])
+        index.add_records([(('tip',), 'fulltext,no-eol', (None, 0, 1000),
+            [('parent',)])])
         # but neither should have added data.
-        self.assertEqual([[], []], self.caught_entries)
+        self.assertEqual([[], [], [], []], self.caught_entries)
         
     def test_add_versions_different_dup(self):
         index = self.two_graph_index(deltas=True, catch_adds=True)
         # change options
-        self.assertRaises(errors.KnitCorrupt, index.add_versions,
-            [('tip', 'no-eol,line-delta', (None, 0, 100), ['parent'])])
-        self.assertRaises(errors.KnitCorrupt, index.add_versions,
-            [('tip', 'line-delta,no-eol', (None, 0, 100), ['parent'])])
-        self.assertRaises(errors.KnitCorrupt, index.add_versions,
-            [('tip', 'fulltext', (None, 0, 100), ['parent'])])
-        # position/length
-        self.assertRaises(errors.KnitCorrupt, index.add_versions,
-            [('tip', 'fulltext,no-eol', (None, 50, 100), ['parent'])])
-        self.assertRaises(errors.KnitCorrupt, index.add_versions,
-            [('tip', 'fulltext,no-eol', (None, 0, 1000), ['parent'])])
+        self.assertRaises(errors.KnitCorrupt, index.add_records,
+            [(('tip',), 'no-eol,line-delta', (None, 0, 100), [('parent',)])])
+        self.assertRaises(errors.KnitCorrupt, index.add_records,
+            [(('tip',), 'line-delta,no-eol', (None, 0, 100), [('parent',)])])
+        self.assertRaises(errors.KnitCorrupt, index.add_records,
+            [(('tip',), 'fulltext', (None, 0, 100), [('parent',)])])
         # parents
-        self.assertRaises(errors.KnitCorrupt, index.add_versions,
-            [('tip', 'fulltext,no-eol', (None, 0, 100), [])])
+        self.assertRaises(errors.KnitCorrupt, index.add_records,
+            [(('tip',), 'fulltext,no-eol', (None, 0, 100), [])])
         # change options in the second record
-        self.assertRaises(errors.KnitCorrupt, index.add_versions,
-            [('tip', 'fulltext,no-eol', (None, 0, 100), ['parent']),
-             ('tip', 'no-eol,line-delta', (None, 0, 100), ['parent'])])
+        self.assertRaises(errors.KnitCorrupt, index.add_records,
+            [(('tip',), 'fulltext,no-eol', (None, 0, 100), [('parent',)]),
+             (('tip',), 'no-eol,line-delta', (None, 0, 100), [('parent',)])])
         self.assertEqual([], self.caught_entries)
 
+
 class TestNoParentsGraphIndexKnit(KnitTests):
-    """Tests for knits using KnitGraphIndex with no parents."""
+    """Tests for knits using _KnitGraphIndex with no parents."""
 
     def make_g_index(self, name, ref_lists=0, nodes=[]):
         builder = GraphIndexBuilder(ref_lists)
@@ -2385,8 +1217,8 @@
 
     def test_parents_deltas_incompatible(self):
         index = CombinedGraphIndex([])
-        self.assertRaises(errors.KnitError, KnitGraphIndex, index,
-            deltas=True, parents=False)
+        self.assertRaises(errors.KnitError, _KnitGraphIndex, lambda:True,
+            index, deltas=True, parents=False)
 
     def two_graph_index(self, catch_adds=False):
         """Build a two-graph index.
@@ -2408,134 +1240,87 @@
             add_callback = self.catch_add
         else:
             add_callback = None
-        return KnitGraphIndex(combined_index, parents=False,
+        return _KnitGraphIndex(combined_index, lambda:True, parents=False,
             add_callback=add_callback)
 
-    def test_get_ancestry(self):
-        # with no parents, ancestry is always just the key.
-        index = self.two_graph_index()
-        self.assertEqual([], index.get_ancestry([]))
-        self.assertEqual(['separate'], index.get_ancestry(['separate']))
-        self.assertEqual(['tail'], index.get_ancestry(['tail']))
-        self.assertEqual(['parent'], index.get_ancestry(['parent']))
-        self.assertEqual(['tip'], index.get_ancestry(['tip']))
-        self.assertTrue(index.get_ancestry(['tip', 'separate']) in
-            (['tip', 'separate'],
-             ['separate', 'tip'],
-            ))
-        # asking for a ghost makes it go boom.
-        self.assertRaises(errors.RevisionNotPresent, index.get_ancestry, ['ghost'])
-
-    def test_get_ancestry_with_ghosts(self):
-        index = self.two_graph_index()
-        self.assertEqual([], index.get_ancestry_with_ghosts([]))
-        self.assertEqual(['separate'], index.get_ancestry_with_ghosts(['separate']))
-        self.assertEqual(['tail'], index.get_ancestry_with_ghosts(['tail']))
-        self.assertEqual(['parent'], index.get_ancestry_with_ghosts(['parent']))
-        self.assertEqual(['tip'], index.get_ancestry_with_ghosts(['tip']))
-        self.assertTrue(index.get_ancestry_with_ghosts(['tip', 'separate']) in
-            (['tip', 'separate'],
-             ['separate', 'tip'],
-            ))
-        # asking for a ghost makes it go boom.
-        self.assertRaises(errors.RevisionNotPresent, index.get_ancestry_with_ghosts, ['ghost'])
-
-    def test_num_versions(self):
-        index = self.two_graph_index()
-        self.assertEqual(4, index.num_versions())
-
-    def test_get_versions(self):
-        index = self.two_graph_index()
-        self.assertEqual(set(['tail', 'tip', 'parent', 'separate']),
-            set(index.get_versions()))
-
-    def test_has_version(self):
-        index = self.two_graph_index()
-        self.assertTrue(index.has_version('tail'))
-        self.assertFalse(index.has_version('ghost'))
+    def test_keys(self):
+        index = self.two_graph_index()
+        self.assertEqual(set([('tail',), ('tip',), ('parent',), ('separate',)]),
+            set(index.keys()))
 
     def test_get_position(self):
         index = self.two_graph_index()
-        self.assertEqual((index._graph_index._indices[0], 0, 100), index.get_position('tip'))
-        self.assertEqual((index._graph_index._indices[1], 100, 78), index.get_position('parent'))
+        self.assertEqual((index._graph_index._indices[0], 0, 100),
+            index.get_position(('tip',)))
+        self.assertEqual((index._graph_index._indices[1], 100, 78),
+            index.get_position(('parent',)))
 
     def test_get_method(self):
         index = self.two_graph_index()
-        self.assertEqual('fulltext', index.get_method('tip'))
-        self.assertEqual(['fulltext'], index.get_options('parent'))
+        self.assertEqual('fulltext', index.get_method(('tip',)))
+        self.assertEqual(['fulltext'], index.get_options(('parent',)))
 
     def test_get_options(self):
         index = self.two_graph_index()
-        self.assertEqual(['fulltext', 'no-eol'], index.get_options('tip'))
-        self.assertEqual(['fulltext'], index.get_options('parent'))
-
-    def test_get_parents_with_ghosts(self):
-        index = self.two_graph_index()
-        self.assertEqual((), index.get_parents_with_ghosts('parent'))
-        # and errors on ghosts.
-        self.assertRaises(errors.RevisionNotPresent,
-            index.get_parents_with_ghosts, 'ghost')
-
-    def test_check_versions_present(self):
-        index = self.two_graph_index()
-        self.assertRaises(RevisionNotPresent, index.check_versions_present,
-            ['missing'])
-        self.assertRaises(RevisionNotPresent, index.check_versions_present,
-            ['tail', 'missing'])
-        index.check_versions_present(['tail', 'separate'])
+        self.assertEqual(['fulltext', 'no-eol'], index.get_options(('tip',)))
+        self.assertEqual(['fulltext'], index.get_options(('parent',)))
+
+    def test_get_parent_map(self):
+        index = self.two_graph_index()
+        self.assertEqual({('parent',):None},
+            index.get_parent_map([('parent',), ('ghost',)]))
 
     def catch_add(self, entries):
         self.caught_entries.append(entries)
 
     def test_add_no_callback_errors(self):
         index = self.two_graph_index()
-        self.assertRaises(errors.ReadOnlyError, index.add_version,
-            'new', 'fulltext,no-eol', (None, 50, 60), ['separate'])
+        self.assertRaises(errors.ReadOnlyError, index.add_records,
+            [(('new',), 'fulltext,no-eol', (None, 50, 60), [('separate',)])])
 
     def test_add_version_smoke(self):
         index = self.two_graph_index(catch_adds=True)
-        index.add_version('new', 'fulltext,no-eol', (None, 50, 60), [])
+        index.add_records([(('new',), 'fulltext,no-eol', (None, 50, 60), [])])
         self.assertEqual([[(('new', ), 'N50 60')]],
             self.caught_entries)
 
     def test_add_version_delta_not_delta_index(self):
         index = self.two_graph_index(catch_adds=True)
-        self.assertRaises(errors.KnitCorrupt, index.add_version,
-            'new', 'no-eol,line-delta', (None, 0, 100), [])
+        self.assertRaises(errors.KnitCorrupt, index.add_records,
+            [(('new',), 'no-eol,line-delta', (None, 0, 100), [])])
         self.assertEqual([], self.caught_entries)
 
     def test_add_version_same_dup(self):
         index = self.two_graph_index(catch_adds=True)
         # options can be spelt two different ways
-        index.add_version('tip', 'fulltext,no-eol', (None, 0, 100), [])
-        index.add_version('tip', 'no-eol,fulltext', (None, 0, 100), [])
+        index.add_records([(('tip',), 'fulltext,no-eol', (None, 0, 100), [])])
+        index.add_records([(('tip',), 'no-eol,fulltext', (None, 0, 100), [])])
+        # position/length are ignored (because each pack could have fulltext or
+        # delta, and be at a different position.
+        index.add_records([(('tip',), 'fulltext,no-eol', (None, 50, 100), [])])
+        index.add_records([(('tip',), 'fulltext,no-eol', (None, 0, 1000), [])])
         # but neither should have added data.
-        self.assertEqual([[], []], self.caught_entries)
+        self.assertEqual([[], [], [], []], self.caught_entries)
         
     def test_add_version_different_dup(self):
         index = self.two_graph_index(catch_adds=True)
         # change options
-        self.assertRaises(errors.KnitCorrupt, index.add_version,
-            'tip', 'no-eol,line-delta', (None, 0, 100), [])
-        self.assertRaises(errors.KnitCorrupt, index.add_version,
-            'tip', 'line-delta,no-eol', (None, 0, 100), [])
-        self.assertRaises(errors.KnitCorrupt, index.add_version,
-            'tip', 'fulltext', (None, 0, 100), [])
-        # position/length
-        self.assertRaises(errors.KnitCorrupt, index.add_version,
-            'tip', 'fulltext,no-eol', (None, 50, 100), [])
-        self.assertRaises(errors.KnitCorrupt, index.add_version,
-            'tip', 'fulltext,no-eol', (None, 0, 1000), [])
+        self.assertRaises(errors.KnitCorrupt, index.add_records,
+            [(('tip',), 'no-eol,line-delta', (None, 0, 100), [])])
+        self.assertRaises(errors.KnitCorrupt, index.add_records,
+            [(('tip',), 'line-delta,no-eol', (None, 0, 100), [])])
+        self.assertRaises(errors.KnitCorrupt, index.add_records,
+            [(('tip',), 'fulltext', (None, 0, 100), [])])
         # parents
-        self.assertRaises(errors.KnitCorrupt, index.add_version,
-            'tip', 'fulltext,no-eol', (None, 0, 100), ['parent'])
+        self.assertRaises(errors.KnitCorrupt, index.add_records,
+            [(('tip',), 'fulltext,no-eol', (None, 0, 100), [('parent',)])])
         self.assertEqual([], self.caught_entries)
         
     def test_add_versions(self):
         index = self.two_graph_index(catch_adds=True)
-        index.add_versions([
-                ('new', 'fulltext,no-eol', (None, 50, 60), []),
-                ('new2', 'fulltext', (None, 0, 6), []),
+        index.add_records([
+                (('new',), 'fulltext,no-eol', (None, 50, 60), []),
+                (('new2',), 'fulltext', (None, 0, 6), []),
                 ])
         self.assertEqual([(('new', ), 'N50 60'), (('new2', ), ' 0 6')],
             sorted(self.caught_entries[0]))
@@ -2543,274 +1328,46 @@
 
     def test_add_versions_delta_not_delta_index(self):
         index = self.two_graph_index(catch_adds=True)
-        self.assertRaises(errors.KnitCorrupt, index.add_versions,
-            [('new', 'no-eol,line-delta', (None, 0, 100), ['parent'])])
+        self.assertRaises(errors.KnitCorrupt, index.add_records,
+            [(('new',), 'no-eol,line-delta', (None, 0, 100), [('parent',)])])
         self.assertEqual([], self.caught_entries)
 
     def test_add_versions_parents_not_parents_index(self):
         index = self.two_graph_index(catch_adds=True)
-        self.assertRaises(errors.KnitCorrupt, index.add_versions,
-            [('new', 'no-eol,fulltext', (None, 0, 100), ['parent'])])
+        self.assertRaises(errors.KnitCorrupt, index.add_records,
+            [(('new',), 'no-eol,fulltext', (None, 0, 100), [('parent',)])])
         self.assertEqual([], self.caught_entries)
 
     def test_add_versions_random_id_accepted(self):
         index = self.two_graph_index(catch_adds=True)
-        index.add_versions([], random_id=True)
+        index.add_records([], random_id=True)
 
     def test_add_versions_same_dup(self):
         index = self.two_graph_index(catch_adds=True)
         # options can be spelt two different ways
-        index.add_versions([('tip', 'fulltext,no-eol', (None, 0, 100), [])])
-        index.add_versions([('tip', 'no-eol,fulltext', (None, 0, 100), [])])
+        index.add_records([(('tip',), 'fulltext,no-eol', (None, 0, 100), [])])
+        index.add_records([(('tip',), 'no-eol,fulltext', (None, 0, 100), [])])
+        # position/length are ignored (because each pack could have fulltext or
+        # delta, and be at a different position.
+        index.add_records([(('tip',), 'fulltext,no-eol', (None, 50, 100), [])])
+        index.add_records([(('tip',), 'fulltext,no-eol', (None, 0, 1000), [])])
         # but neither should have added data.
-        self.assertEqual([[], []], self.caught_entries)
+        self.assertEqual([[], [], [], []], self.caught_entries)
         
     def test_add_versions_different_dup(self):
         index = self.two_graph_index(catch_adds=True)
         # change options
-        self.assertRaises(errors.KnitCorrupt, index.add_versions,
-            [('tip', 'no-eol,line-delta', (None, 0, 100), [])])
-        self.assertRaises(errors.KnitCorrupt, index.add_versions,
-            [('tip', 'line-delta,no-eol', (None, 0, 100), [])])
-        self.assertRaises(errors.KnitCorrupt, index.add_versions,
-            [('tip', 'fulltext', (None, 0, 100), [])])
-        # position/length
-        self.assertRaises(errors.KnitCorrupt, index.add_versions,
-            [('tip', 'fulltext,no-eol', (None, 50, 100), [])])
-        self.assertRaises(errors.KnitCorrupt, index.add_versions,
-            [('tip', 'fulltext,no-eol', (None, 0, 1000), [])])
+        self.assertRaises(errors.KnitCorrupt, index.add_records,
+            [(('tip',), 'no-eol,line-delta', (None, 0, 100), [])])
+        self.assertRaises(errors.KnitCorrupt, index.add_records,
+            [(('tip',), 'line-delta,no-eol', (None, 0, 100), [])])
+        self.assertRaises(errors.KnitCorrupt, index.add_records,
+            [(('tip',), 'fulltext', (None, 0, 100), [])])
         # parents
-        self.assertRaises(errors.KnitCorrupt, index.add_versions,
-            [('tip', 'fulltext,no-eol', (None, 0, 100), ['parent'])])
+        self.assertRaises(errors.KnitCorrupt, index.add_records,
+            [(('tip',), 'fulltext,no-eol', (None, 0, 100), [('parent',)])])
         # change options in the second record
-        self.assertRaises(errors.KnitCorrupt, index.add_versions,
-            [('tip', 'fulltext,no-eol', (None, 0, 100), []),
-             ('tip', 'no-eol,line-delta', (None, 0, 100), [])])
+        self.assertRaises(errors.KnitCorrupt, index.add_records,
+            [(('tip',), 'fulltext,no-eol', (None, 0, 100), []),
+             (('tip',), 'no-eol,line-delta', (None, 0, 100), [])])
         self.assertEqual([], self.caught_entries)
-
-class TestPackKnits(KnitTests):
-    """Tests that use a _PackAccess and KnitGraphIndex."""
-
-    def test_get_data_stream_packs_ignores_pack_overhead(self):
-        # Packs have an encoding overhead that should not be included in the
-        # 'size' field of a data stream, because it is not returned by the
-        # raw_reading functions - it is why index_memo's are opaque, and
-        # get_data_stream was abusing this.
-        packname = 'test.pack'
-        transport = self.get_transport()
-        def write_data(bytes):
-            transport.append_bytes(packname, bytes)
-        writer = pack.ContainerWriter(write_data)
-        writer.begin()
-        index = InMemoryGraphIndex(2)
-        knit_index = KnitGraphIndex(index, add_callback=index.add_nodes,
-            deltas=True)
-        indices = {index:(transport, packname)}
-        access = _PackAccess(indices, writer=(writer, index))
-        k = KnitVersionedFile('test', get_transport('.'),
-            delta=True, create=True, index=knit_index, access_method=access)
-        # insert something into the knit
-        k.add_lines('text-1', [], ["foo\n"])
-        # get a data stream for it
-        stream = k.get_data_stream(['text-1'])
-        # if the stream has been incorrectly assembled, we will get a short read
-        # reading from the stream (as streams have no trailer)
-        expected_length = stream[1][0][2]
-        # we use -1 to do the read, so that if a trailer is added this test
-        # will fail and we'll adjust it to handle that case correctly, rather
-        # than allowing an over-read that is bogus.
-        self.assertEqual(expected_length, len(stream[2](-1)))
-
-
-class Test_StreamIndex(KnitTests):
-
-    def get_index(self, knit, stream):
-        """Get a _StreamIndex from knit and stream."""
-        return knit._knit_from_datastream(stream)._index
-
-    def assertIndexVersions(self, knit, versions):
-        """Check that the _StreamIndex versions are those of the stream."""
-        index = self.get_index(knit, knit.get_data_stream(versions))
-        self.assertEqual(set(index.get_versions()), set(versions))
-        # check we didn't get duplicates
-        self.assertEqual(len(index.get_versions()), len(versions))
-
-    def assertIndexAncestry(self, knit, ancestry_versions, versions, result):
-        """Check the result of a get_ancestry call on knit."""
-        index = self.get_index(knit, knit.get_data_stream(versions))
-        self.assertEqual(
-            set(result),
-            set(index.get_ancestry(ancestry_versions, False)))
-
-    def assertGetMethod(self, knit, versions, version, result):
-        index = self.get_index(knit, knit.get_data_stream(versions))
-        self.assertEqual(result, index.get_method(version))
-
-    def assertGetOptions(self, knit, version, options):
-        index = self.get_index(knit, knit.get_data_stream(version))
-        self.assertEqual(options, index.get_options(version))
-
-    def assertGetPosition(self, knit, versions, version, result):
-        index = self.get_index(knit, knit.get_data_stream(versions))
-        if result[1] is None:
-            result = (result[0], index, result[2], result[3])
-        self.assertEqual(result, index.get_position(version))
-
-    def assertGetParentsWithGhosts(self, knit, versions, version, parents):
-        index = self.get_index(knit, knit.get_data_stream(versions))
-        self.assertEqual(parents, index.get_parents_with_ghosts(version))
-
-    def make_knit_with_4_versions_2_dags(self):
-        knit = self.make_test_knit()
-        knit.add_lines('a', [], ["foo"])
-        knit.add_lines('b', [], [])
-        knit.add_lines('c', ['b', 'a'], [])
-        knit.add_lines_with_ghosts('d', ['e', 'f'], [])
-        return knit
-
-    def test_versions(self):
-        """The versions of a StreamIndex are those of the datastream."""
-        knit = self.make_knit_with_4_versions_2_dags()
-        # ask for most permutations, which catches bugs like falling back to the
-        # target knit, or showing ghosts, etc.
-        self.assertIndexVersions(knit, [])
-        self.assertIndexVersions(knit, ['a'])
-        self.assertIndexVersions(knit, ['b'])
-        self.assertIndexVersions(knit, ['c'])
-        self.assertIndexVersions(knit, ['d'])
-        self.assertIndexVersions(knit, ['a', 'b'])
-        self.assertIndexVersions(knit, ['b', 'c'])
-        self.assertIndexVersions(knit, ['a', 'c'])
-        self.assertIndexVersions(knit, ['a', 'b', 'c'])
-        self.assertIndexVersions(knit, ['a', 'b', 'c', 'd'])
-
-    def test_construct(self):
-        """Constructing a StreamIndex generates index data."""
-        data_list = [('text-a', ['fulltext'], 127, []),
-            ('text-b', ['option'], 128, ['text-c'])]
-        index = _StreamIndex(data_list, None)
-        self.assertEqual({'text-a':(['fulltext'], (0, 127), []),
-            'text-b':(['option'], (127, 127 + 128), ['text-c'])},
-            index._by_version)
-
-    def test_get_ancestry(self):
-        knit = self.make_knit_with_4_versions_2_dags()
-        self.assertIndexAncestry(knit, ['a'], ['a'], ['a'])
-        self.assertIndexAncestry(knit, ['b'], ['b'], ['b'])
-        self.assertIndexAncestry(knit, ['c'], ['c'], ['c'])
-        self.assertIndexAncestry(knit, ['c'], ['a', 'b', 'c'],
-            set(['a', 'b', 'c']))
-        self.assertIndexAncestry(knit, ['c', 'd'], ['a', 'b', 'c', 'd'],
-            set(['a', 'b', 'c', 'd']))
-
-    def test_get_method(self):
-        knit = self.make_knit_with_4_versions_2_dags()
-        self.assertGetMethod(knit, ['a'], 'a', 'fulltext')
-        self.assertGetMethod(knit, ['c'], 'c', 'line-delta')
-        # get_method on a basis that is not in the datastream (but in the
-        # backing knit) returns 'fulltext', because thats what we'll create as
-        # we thunk across.
-        self.assertGetMethod(knit, ['c'], 'b', 'fulltext')
-
-    def test_get_options(self):
-        knit = self.make_knit_with_4_versions_2_dags()
-        self.assertGetOptions(knit, 'a', ['no-eol', 'fulltext'])
-        self.assertGetOptions(knit, 'c', ['line-delta'])
-
-    def test_get_parents_with_ghosts(self):
-        knit = self.make_knit_with_4_versions_2_dags()
-        self.assertGetParentsWithGhosts(knit, ['a'], 'a', ())
-        self.assertGetParentsWithGhosts(knit, ['c'], 'c', ('b', 'a'))
-        self.assertGetParentsWithGhosts(knit, ['d'], 'd', ('e', 'f'))
-
-    def test_get_position(self):
-        knit = self.make_knit_with_4_versions_2_dags()
-        # get_position returns (thunk_flag, index(can be None), start, end) for
-        # _StreamAccess to use.
-        self.assertGetPosition(knit, ['a'], 'a', (False, None, 0, 78))
-        self.assertGetPosition(knit, ['a', 'c'], 'c', (False, None, 78, 156))
-        # get_position on a text that is not in the datastream (but in the
-        # backing knit) returns (True, 'versionid', None, None) - and then the
-        # access object can construct the relevant data as needed.
-        self.assertGetPosition(knit, ['a', 'c'], 'b', (True, 'b', None, None))
-
-
-class Test_StreamAccess(KnitTests):
-
-    def get_index_access(self, knit, stream):
-        """Get a _StreamAccess from knit and stream."""
-        knit =  knit._knit_from_datastream(stream)
-        return knit._index, knit._data._access
-
-    def assertGetRawRecords(self, knit, versions):
-        index, access = self.get_index_access(knit,
-            knit.get_data_stream(versions))
-        # check that every version asked for can be obtained from the resulting
-        # access object.
-        # batch
-        memos = []
-        for version in versions:
-            memos.append(knit._index.get_position(version))
-        original = {}
-        for version, data in zip(
-            versions, knit._data._access.get_raw_records(memos)):
-            original[version] = data
-        memos = []
-        for version in versions:
-            memos.append(index.get_position(version))
-        streamed = {}
-        for version, data in zip(versions, access.get_raw_records(memos)):
-            streamed[version] = data
-        self.assertEqual(original, streamed)
-        # individually
-        for version in versions:
-            data = list(access.get_raw_records(
-                [index.get_position(version)]))[0]
-            self.assertEqual(original[version], data)
-
-    def make_knit_with_two_versions(self):
-        knit = self.make_test_knit()
-        knit.add_lines('a', [], ["foo"])
-        knit.add_lines('b', [], ["bar"])
-        return knit
-
-    def test_get_raw_records(self):
-        knit = self.make_knit_with_two_versions()
-        self.assertGetRawRecords(knit, ['a', 'b'])
-        self.assertGetRawRecords(knit, ['a'])
-        self.assertGetRawRecords(knit, ['b'])
-    
-    def test_get_raw_record_from_backing_knit(self):
-        # the thunk layer should create an artificial A on-demand when needed.
-        source_knit = self.make_test_knit(name='plain', annotate=False)
-        target_knit = self.make_test_knit(name='annotated', annotate=True)
-        source_knit.add_lines("A", [], ["Foo\n"])
-        # Give the target A, so we can try to thunk across to it.
-        target_knit.insert_record_stream(source_knit.get_record_stream(['A'],
-            'unordered', False))
-        index, access = self.get_index_access(target_knit,
-            source_knit.get_data_stream([]))
-        raw_data = list(access.get_raw_records([(True, "A", None, None)]))[0]
-        df = GzipFile(mode='rb', fileobj=StringIO(raw_data))
-        self.assertEqual(
-            'version A 1 5d36b88bb697a2d778f024048bafabd443d74503\n'
-            'Foo\nend A\n',
-            df.read())
-
-    def test_asking_for_thunk_stream_is_not_plain_errors(self):
-        knit = self.make_test_knit(name='annotated', annotate=True)
-        knit.add_lines("A", [], ["Foo\n"])
-        index, access = self.get_index_access(knit,
-            knit.get_data_stream([]))
-        self.assertRaises(errors.KnitCorrupt,
-            list, access.get_raw_records([(True, "A", None, None)]))
-
-
-class TestFormatSignatures(KnitTests):
-
-    def test_knit_format_signatures(self):
-        """Different formats of knit have different signature strings."""
-        knit = self.make_test_knit(name='a', annotate=True)
-        self.assertEqual('knit-annotated', knit.get_format_signature())
-        knit = self.make_test_knit(name='p', annotate=False)
-        self.assertEqual('knit-plain', knit.get_format_signature())

=== modified file 'bzrlib/tests/test_merge.py'
--- bzrlib/tests/test_merge.py	2008-04-07 04:54:52 +0000
+++ bzrlib/tests/test_merge.py	2008-06-11 04:20:16 +0000
@@ -469,29 +469,30 @@
 
     def setUp(self):
         TestCaseWithMemoryTransport.setUp(self)
-        self.vf = knit.make_file_knit('root', self.get_transport(),
-            create=True)
-        self.plan_merge_vf = versionedfile._PlanMergeVersionedFile('root',
-                                                                   [self.vf])
-
-    def add_version(self, version_id, parents, text):
-        self.vf.add_lines(version_id, parents, [c+'\n' for c in text])
-
-    def add_uncommitted_version(self, version_id, parents, text):
-        self.plan_merge_vf.add_lines(version_id, parents,
+        mapper = versionedfile.PrefixMapper()
+        factory = knit.make_file_factory(True, mapper)
+        self.vf = factory(self.get_transport())
+        self.plan_merge_vf = versionedfile._PlanMergeVersionedFile('root')
+        self.plan_merge_vf.fallback_versionedfiles.append(self.vf)
+
+    def add_version(self, key, parents, text):
+        self.vf.add_lines(key, parents, [c+'\n' for c in text])
+
+    def add_uncommitted_version(self, key, parents, text):
+        self.plan_merge_vf.add_lines(key, parents,
                                      [c+'\n' for c in text])
 
     def setup_plan_merge(self):
-        self.add_version('A', [], 'abc')
-        self.add_version('B', ['A'], 'acehg')
-        self.add_version('C', ['A'], 'fabg')
-        return _PlanMerge('B', 'C', self.plan_merge_vf)
+        self.add_version(('root', 'A'), [], 'abc')
+        self.add_version(('root', 'B'), [('root', 'A')], 'acehg')
+        self.add_version(('root', 'C'), [('root', 'A')], 'fabg')
+        return _PlanMerge('B', 'C', self.plan_merge_vf, ('root',))
 
     def setup_plan_merge_uncommitted(self):
-        self.add_version('A', [], 'abc')
-        self.add_uncommitted_version('B:', ['A'], 'acehg')
-        self.add_uncommitted_version('C:', ['A'], 'fabg')
-        return _PlanMerge('B:', 'C:', self.plan_merge_vf)
+        self.add_version(('root', 'A'), [], 'abc')
+        self.add_uncommitted_version(('root', 'B:'), [('root', 'A')], 'acehg')
+        self.add_uncommitted_version(('root', 'C:'), [('root', 'A')], 'fabg')
+        return _PlanMerge('B:', 'C:', self.plan_merge_vf, ('root',))
 
     def test_unique_lines(self):
         plan = self.setup_plan_merge()
@@ -505,18 +506,19 @@
         self.assertEqual(set([0, 3]), plan._find_new('C'))
 
     def test_find_new2(self):
-        self.add_version('A', [], 'abc')
-        self.add_version('B', ['A'], 'abcde')
-        self.add_version('C', ['A'], 'abcefg')
-        self.add_version('D', ['A', 'B', 'C'], 'abcdegh')
-        my_plan = _PlanMerge('B', 'D', self.plan_merge_vf)
+        self.add_version(('root', 'A'), [], 'abc')
+        self.add_version(('root', 'B'), [('root', 'A')], 'abcde')
+        self.add_version(('root', 'C'), [('root', 'A')], 'abcefg')
+        self.add_version(('root', 'D'),
+            [('root', 'A'), ('root', 'B'), ('root', 'C')], 'abcdegh')
+        my_plan = _PlanMerge('B', 'D', self.plan_merge_vf, ('root',))
         self.assertEqual(set([5, 6]), my_plan._find_new('D'))
         self.assertEqual(set(), my_plan._find_new('A'))
 
     def test_find_new_no_ancestors(self):
-        self.add_version('A', [], 'abc')
-        self.add_version('B', [], 'xyz')
-        my_plan = _PlanMerge('A', 'B', self.vf)
+        self.add_version(('root', 'A'), [], 'abc')
+        self.add_version(('root', 'B'), [], 'xyz')
+        my_plan = _PlanMerge('A', 'B', self.vf, ('root',))
         self.assertEqual(set([0, 1, 2]), my_plan._find_new('A'))
 
     def test_plan_merge(self):
@@ -576,10 +578,10 @@
             list(_PlanMerge._subtract_plans(old_plan, new_plan)))
 
     def setup_merge_with_base(self):
-        self.add_version('COMMON', [], 'abc')
-        self.add_version('THIS', ['COMMON'], 'abcd')
-        self.add_version('BASE', ['COMMON'], 'eabc')
-        self.add_version('OTHER', ['BASE'], 'eafb')
+        self.add_version(('root', 'COMMON'), [], 'abc')
+        self.add_version(('root', 'THIS'), [('root', 'COMMON')], 'abcd')
+        self.add_version(('root', 'BASE'), [('root', 'COMMON')], 'eabc')
+        self.add_version(('root', 'OTHER'), [('root', 'BASE')], 'eafb')
 
     def test_plan_merge_with_base(self):
         self.setup_merge_with_base()
@@ -628,13 +630,15 @@
                          ], list(plan))
 
     def test_plan_lca_merge_with_criss_cross(self):
-        self.add_version('ROOT', [], 'abc')
+        self.add_version(('root', 'ROOT'), [], 'abc')
         # each side makes a change
-        self.add_version('REV1', ['ROOT'], 'abcd')
-        self.add_version('REV2', ['ROOT'], 'abce')
+        self.add_version(('root', 'REV1'), [('root', 'ROOT')], 'abcd')
+        self.add_version(('root', 'REV2'), [('root', 'ROOT')], 'abce')
         # both sides merge, discarding others' changes
-        self.add_version('LCA1', ['REV1', 'REV2'], 'abcd')
-        self.add_version('LCA2', ['REV1', 'REV2'], 'fabce')
+        self.add_version(('root', 'LCA1'),
+            [('root', 'REV1'), ('root', 'REV2')], 'abcd')
+        self.add_version(('root', 'LCA2'),
+            [('root', 'REV1'), ('root', 'REV2')], 'fabce')
         plan = self.plan_merge_vf.plan_lca_merge('LCA1', 'LCA2')
         self.assertEqual([('new-b', 'f\n'),
                           ('unchanged', 'a\n'),

=== modified file 'bzrlib/tests/test_merge_core.py'
--- bzrlib/tests/test_merge_core.py	2008-04-24 07:22:53 +0000
+++ bzrlib/tests/test_merge_core.py	2008-06-11 04:20:16 +0000
@@ -724,7 +724,7 @@
 
     def test_from_revision_ids(self):
         this, other = self.set_up_trees()
-        self.assertRaises(errors.RevisionNotPresent, Merger.from_revision_ids,
+        self.assertRaises(errors.NoSuchRevision, Merger.from_revision_ids,
                           progress.DummyProgress(), this, 'rev2b')
         this.lock_write()
         self.addCleanup(this.unlock)

=== modified file 'bzrlib/tests/test_remote.py'
--- bzrlib/tests/test_remote.py	2008-05-07 06:03:09 +0000
+++ bzrlib/tests/test_remote.py	2008-06-11 04:20:16 +0000
@@ -1107,68 +1107,3 @@
         self.assertFalse(isinstance(dest_repo, RemoteRepository))
         self.assertTrue(isinstance(src_repo, RemoteRepository))
         src_repo.copy_content_into(dest_repo)
-
-
-class TestRepositoryStreamKnitData(TestRemoteRepository):
-
-    def make_pack_file(self, records):
-        pack_file = StringIO()
-        pack_writer = pack.ContainerWriter(pack_file.write)
-        pack_writer.begin()
-        for bytes, names in records:
-            pack_writer.add_bytes_record(bytes, names)
-        pack_writer.end()
-        pack_file.seek(0)
-        return pack_file
-
-    def make_pack_stream(self, records):
-        pack_serialiser = pack.ContainerSerialiser()
-        yield pack_serialiser.begin()
-        for bytes, names in records:
-            yield pack_serialiser.bytes_record(bytes, names)
-        yield pack_serialiser.end()
-
-    def test_bad_pack_from_server(self):
-        """A response with invalid data (e.g. it has a record with multiple
-        names) triggers an exception.
-        
-        Not all possible errors will be caught at this stage, but obviously
-        malformed data should be.
-        """
-        record = ('bytes', [('name1',), ('name2',)])
-        pack_stream = self.make_pack_stream([record])
-        responses = [(('ok',), pack_stream), ]
-        transport_path = 'quack'
-        repo, client = self.setup_fake_client_and_repository(
-            responses, transport_path)
-        search = graph.SearchResult(set(['revid']), set(), 1, set(['revid']))
-        stream = repo.get_data_stream_for_search(search)
-        self.assertRaises(errors.SmartProtocolError, list, stream)
-    
-    def test_backwards_compatibility(self):
-        """If the server doesn't recognise this request, fallback to VFS."""
-        responses = [
-            (('unknown verb', 'Repository.stream_revisions_chunked'), '')]
-        repo, client = self.setup_fake_client_and_repository(
-            responses, 'path')
-        self.mock_called = False
-        repo._real_repository = MockRealRepository(self)
-        search = graph.SearchResult(set(['revid']), set(), 1, set(['revid']))
-        repo.get_data_stream_for_search(search)
-        self.assertTrue(self.mock_called)
-        self.failIf(client.expecting_body,
-            "The protocol has been left in an unclean state that will cause "
-            "TooManyConcurrentRequests errors.")
-
-
-class MockRealRepository(object):
-    """Helper class for TestRepositoryStreamKnitData.test_unknown_method."""
-
-    def __init__(self, test):
-        self.test = test
-
-    def get_data_stream_for_search(self, search):
-        self.test.assertEqual(set(['revid']), search.get_keys())
-        self.test.mock_called = True
-
-

=== modified file 'bzrlib/tests/test_repository.py'
--- bzrlib/tests/test_repository.py	2008-04-26 16:34:47 +0000
+++ bzrlib/tests/test_repository.py	2008-06-11 04:20:16 +0000
@@ -167,15 +167,6 @@
                           control.transport.get,
                           'ancestry.weave')
 
-    def test_exposed_versioned_files_are_marked_dirty(self):
-        control = bzrdir.BzrDirFormat6().initialize(self.get_url())
-        repo = weaverepo.RepositoryFormat6().initialize(control)
-        repo.lock_write()
-        inv = repo.get_inventory_weave()
-        repo.unlock()
-        self.assertRaises(errors.OutSideTransaction,
-            inv.add_lines, 'foo', [], [])
-
     def test_supports_external_lookups(self):
         control = bzrdir.BzrDirFormat6().initialize(self.get_url())
         repo = weaverepo.RepositoryFormat6().initialize(control)
@@ -205,6 +196,25 @@
                              'w\n'
                              'W\n',
                              t.get('inventory.weave').read())
+        # Creating a file with id Foo:Bar results in a non-escaped file name on
+        # disk.
+        control.create_branch()
+        tree = control.create_workingtree()
+        tree.add(['foo'], ['Foo:Bar'], ['file'])
+        tree.put_file_bytes_non_atomic('Foo:Bar', 'content\n')
+        tree.commit('first post', rev_id='first')
+        self.assertEqualDiff(
+            '# bzr weave file v5\n'
+            'i\n'
+            '1 7fe70820e08a1aac0ef224d9c66ab66831cc4ab1\n'
+            'n first\n'
+            '\n'
+            'w\n'
+            '{ 0\n'
+            '. content\n'
+            '}\n'
+            'W\n',
+            t.get('weaves/74/Foo%3ABar.weave').read())
 
     def test_shared_disk_layout(self):
         control = bzrdir.BzrDirMetaFormat1().initialize(self.get_url())
@@ -285,15 +295,6 @@
                              'W\n',
                              t.get('inventory.weave').read())
 
-    def test_exposed_versioned_files_are_marked_dirty(self):
-        control = bzrdir.BzrDirMetaFormat1().initialize(self.get_url())
-        repo = weaverepo.RepositoryFormat7().initialize(control)
-        repo.lock_write()
-        inv = repo.get_inventory_weave()
-        repo.unlock()
-        self.assertRaises(errors.OutSideTransaction,
-            inv.add_lines, 'foo', [], [])
-
     def test_supports_external_lookups(self):
         control = bzrdir.BzrDirMetaFormat1().initialize(self.get_url())
         repo = weaverepo.RepositoryFormat7().initialize(control)
@@ -321,13 +322,19 @@
         # self.assertEqualDiff('', t.get('lock').read())
         self.assertTrue(S_ISDIR(t.stat('knits').st_mode))
         self.check_knits(t)
+        # Check per-file knits.
+        branch = control.create_branch()
+        tree = control.create_workingtree()
+        tree.add(['foo'], ['Nasty-IdC:'], ['file'])
+        tree.put_file_bytes_non_atomic('Nasty-IdC:', '')
+        tree.commit('1st post', rev_id='foo')
+        self.assertHasKnit(t, 'knits/e8/%254easty-%2549d%2543%253a',
+            '\nfoo fulltext 0 81  :')
 
-    def assertHasKnit(self, t, knit_name):
+    def assertHasKnit(self, t, knit_name, extra_content=''):
         """Assert that knit_name exists on t."""
-        self.assertEqualDiff('# bzr knit index 8\n',
+        self.assertEqualDiff('# bzr knit index 8\n' + extra_content,
                              t.get(knit_name + '.kndx').read())
-        # no default content
-        self.assertTrue(t.has(knit_name + '.knit'))
 
     def check_knits(self, t):
         """check knit content for a repository."""
@@ -377,16 +384,6 @@
         self.assertTrue(S_ISDIR(t.stat('knits').st_mode))
         self.check_knits(t)
 
-    def test_exposed_versioned_files_are_marked_dirty(self):
-        format = bzrdir.BzrDirMetaFormat1()
-        format.repository_format = knitrepo.RepositoryFormatKnit1()
-        repo = self.make_repository('.', format=format)
-        repo.lock_write()
-        inv = repo.get_inventory_weave()
-        repo.unlock()
-        self.assertRaises(errors.OutSideTransaction,
-            inv.add_lines, 'foo', [], [])
-
     def test_deserialise_sets_root_revision(self):
         """We must have a inventory.root.revision
 
@@ -420,69 +417,6 @@
         self.assertFalse(repo._format.supports_external_lookups)
 
 
-class KnitRepositoryStreamTests(test_knit.KnitTests):
-    """Tests for knitrepo._get_stream_as_bytes."""
-
-    def test_get_stream_as_bytes(self):
-        # Make a simple knit
-        k1 = self.make_test_knit()
-        k1.add_lines('text-a', [], test_knit.split_lines(test_knit.TEXT_1))
-        
-        # Serialise it, check the output.
-        bytes = knitrepo._get_stream_as_bytes(k1, ['text-a'])
-        data = bencode.bdecode(bytes)
-        format, record = data
-        self.assertEqual('knit-plain', format)
-        self.assertEqual(['text-a', ['fulltext'], []], record[:3])
-        self.assertRecordContentEqual(k1, 'text-a', record[3])
-
-    def test_get_stream_as_bytes_all(self):
-        """Get a serialised data stream for all the records in a knit.
-
-        Much like test_get_stream_all, except for get_stream_as_bytes.
-        """
-        k1 = self.make_test_knit()
-        # Insert the same data as BasicKnitTests.test_knit_join, as they seem
-        # to cover a range of cases (no parents, one parent, multiple parents).
-        test_data = [
-            ('text-a', [], test_knit.TEXT_1),
-            ('text-b', ['text-a'], test_knit.TEXT_1),
-            ('text-c', [], test_knit.TEXT_1),
-            ('text-d', ['text-c'], test_knit.TEXT_1),
-            ('text-m', ['text-b', 'text-d'], test_knit.TEXT_1),
-           ]
-        # This test is actually a bit strict as the order in which they're
-        # returned is not defined.  This matches the current (deterministic)
-        # behaviour.
-        expected_data_list = [
-            # version, options, parents
-            ('text-a', ['fulltext'], []),
-            ('text-b', ['line-delta'], ['text-a']),
-            ('text-m', ['line-delta'], ['text-b', 'text-d']),
-            ('text-c', ['fulltext'], []),
-            ('text-d', ['line-delta'], ['text-c']),
-            ]
-        for version_id, parents, lines in test_data:
-            k1.add_lines(version_id, parents, test_knit.split_lines(lines))
-
-        bytes = knitrepo._get_stream_as_bytes(
-            k1, ['text-a', 'text-b', 'text-m', 'text-c', 'text-d', ])
-
-        data = bencode.bdecode(bytes)
-        format = data.pop(0)
-        self.assertEqual('knit-plain', format)
-
-        for expected, actual in zip(expected_data_list, data):
-            expected_version = expected[0]
-            expected_options = expected[1]
-            expected_parents = expected[2]
-            version, options, parents, bytes = actual
-            self.assertEqual(expected_version, version)
-            self.assertEqual(expected_options, options)
-            self.assertEqual(expected_parents, parents)
-            self.assertRecordContentEqual(k1, version, bytes)
-
-
 class DummyRepository(object):
     """A dummy repository for testing."""
 
@@ -593,49 +527,6 @@
                                                         repo_b).__class__)
 
 
-class TestInterRemoteToOther(TestCaseWithTransport):
-
-    def make_remote_repository(self, path, backing_format=None):
-        """Make a RemoteRepository object backed by a real repository that will
-        be created at the given path."""
-        self.make_repository(path, format=backing_format)
-        smart_server = server.SmartTCPServer_for_testing()
-        smart_server.setUp()
-        remote_transport = get_transport(smart_server.get_url()).clone(path)
-        self.addCleanup(smart_server.tearDown)
-        remote_bzrdir = bzrdir.BzrDir.open_from_transport(remote_transport)
-        remote_repo = remote_bzrdir.open_repository()
-        return remote_repo
-
-    def test_is_compatible_same_format(self):
-        """InterRemoteToOther is compatible with a remote repository and a
-        second repository that have the same format."""
-        local_repo = self.make_repository('local')
-        remote_repo = self.make_remote_repository('remote')
-        is_compatible = repository.InterRemoteToOther.is_compatible
-        self.assertTrue(
-            is_compatible(remote_repo, local_repo),
-            "InterRemoteToOther(%r, %r) is false" % (remote_repo, local_repo))
-          
-    def test_is_incompatible_different_format(self):
-        local_repo = self.make_repository('local', 'dirstate')
-        remote_repo = self.make_remote_repository('a', 'dirstate-with-subtree')
-        is_compatible = repository.InterRemoteToOther.is_compatible
-        self.assertFalse(
-            is_compatible(remote_repo, local_repo),
-            "InterRemoteToOther(%r, %r) is true" % (local_repo, remote_repo))
-
-    def test_is_incompatible_different_format_both_remote(self):
-        remote_repo_a = self.make_remote_repository(
-            'a', 'dirstate-with-subtree')
-        remote_repo_b = self.make_remote_repository('b', 'dirstate')
-        is_compatible = repository.InterRemoteToOther.is_compatible
-        self.assertFalse(
-            is_compatible(remote_repo_a, remote_repo_b),
-            "InterRemoteToOther(%r, %r) is true"
-            % (remote_repo_a, remote_repo_b))
-
-
 class TestRepositoryConverter(TestCaseWithTransport):
 
     def test_convert_empty(self):
@@ -670,28 +561,28 @@
         tree = self.make_branch_and_tree('.', format)
         tree.commit("Dull commit", rev_id="dull")
         revision_tree = tree.branch.repository.revision_tree('dull')
-        self.assertRaises(errors.NoSuchFile, revision_tree.get_file_lines,
-            revision_tree.inventory.root.file_id)
+        revision_tree.lock_read()
+        try:
+            self.assertRaises(errors.NoSuchFile, revision_tree.get_file_lines,
+                revision_tree.inventory.root.file_id)
+        finally:
+            revision_tree.unlock()
         format = bzrdir.BzrDirMetaFormat1()
         format.repository_format = knitrepo.RepositoryFormatKnit3()
         upgrade.Convert('.', format)
         tree = workingtree.WorkingTree.open('.')
         revision_tree = tree.branch.repository.revision_tree('dull')
-        revision_tree.get_file_lines(revision_tree.inventory.root.file_id)
+        revision_tree.lock_read()
+        try:
+            revision_tree.get_file_lines(revision_tree.inventory.root.file_id)
+        finally:
+            revision_tree.unlock()
         tree.commit("Another dull commit", rev_id='dull2')
         revision_tree = tree.branch.repository.revision_tree('dull2')
+        revision_tree.lock_read()
+        self.addCleanup(revision_tree.unlock)
         self.assertEqual('dull', revision_tree.inventory.root.revision)
 
-    def test_exposed_versioned_files_are_marked_dirty(self):
-        format = bzrdir.BzrDirMetaFormat1()
-        format.repository_format = knitrepo.RepositoryFormatKnit3()
-        repo = self.make_repository('.', format=format)
-        repo.lock_write()
-        inv = repo.get_inventory_weave()
-        repo.unlock()
-        self.assertRaises(errors.OutSideTransaction,
-            inv.add_lines, 'foo', [], [])
-
     def test_supports_external_lookups(self):
         format = bzrdir.BzrDirMetaFormat1()
         format.repository_format = knitrepo.RepositoryFormatKnit3()
@@ -769,9 +660,9 @@
         entry.revision = revision
         entry.text_size = 0
         inv.add(entry)
-        vf = repo.weave_store.get_weave_or_empty(file_id,
-                                                 repo.get_transaction())
-        vf.add_lines(revision, parents, ['line\n'])
+        text_key = (file_id, revision)
+        parent_keys = [(file_id, parent) for parent in parents]
+        repo.texts.add_lines(text_key, parent_keys, ['line\n'])
 
     def test_insert_from_broken_repo(self):
         """Inserting a data stream from a broken repository won't silently
@@ -779,19 +670,7 @@
         """
         broken_repo = self.make_broken_repository()
         empty_repo = self.make_repository('empty-repo')
-        search = graph.SearchResult(set(['rev1a', 'rev2', 'rev3']),
-            set(), 3, ['rev1a', 'rev2', 'rev3'])
-        broken_repo.lock_read()
-        self.addCleanup(broken_repo.unlock)
-        stream = broken_repo.get_data_stream_for_search(search)
-        empty_repo.lock_write()
-        self.addCleanup(empty_repo.unlock)
-        empty_repo.start_write_group()
-        try:
-            self.assertRaises(
-                errors.KnitCorrupt, empty_repo.insert_data_stream, stream)
-        finally:
-            empty_repo.abort_write_group()
+        self.assertRaises(errors.RevisionNotPresent, empty_repo.fetch, broken_repo)
 
 
 class TestKnitPackNoSubtrees(TestCaseWithTransport):
@@ -816,11 +695,6 @@
             "Bazaar pack repository format 1 (needs bzr 0.92)\n",
                              t.get('format').read())
 
-    def assertHasKndx(self, t, knit_name):
-        """Assert that knit_name exists on t."""
-        self.assertEqualDiff('# bzr knit index 8\n',
-                             t.get(knit_name + '.kndx').read())
-
     def assertHasNoKndx(self, t, knit_name):
         """Assert that knit_name has no index on t."""
         self.assertFalse(t.has(knit_name + '.kndx'))
@@ -990,8 +864,7 @@
 
     def _add_text(self, repo, fileid):
         """Add a text to the repository within a write group."""
-        vf =repo.weave_store.get_weave(fileid, repo.get_transaction())
-        vf.add_lines('samplerev+' + fileid, [], [])
+        repo.texts.add_lines((fileid, 'samplerev+'+fileid), [], [])
 
     def test_concurrent_writers_merge_new_packs(self):
         format = self.get_format()
@@ -1168,9 +1041,7 @@
             inv.root.revision = revision_id
             root_id = inv.root.file_id
             sha1 = repo.add_inventory(revision_id, inv, [])
-            vf = repo.weave_store.get_weave_or_empty(root_id,
-                repo.get_transaction())
-            vf.add_lines(revision_id, [], [])
+            repo.texts.add_lines((root_id, revision_id), [], [])
             rev = bzrlib.revision.Revision(timestamp=0,
                                            timezone=None,
                                            committer="Foo Bar <foo@example.com>",
@@ -1191,7 +1062,7 @@
         inv = missing_ghost.get_inventory('tip')
         self.assertRaises(errors.NoSuchRevision,
             missing_ghost.get_revision, 'ghost')
-        self.assertRaises(errors.RevisionNotPresent,
+        self.assertRaises(errors.NoSuchRevision,
             missing_ghost.get_inventory, 'ghost')
 
     def test_supports_external_lookups(self):

=== modified file 'bzrlib/tests/test_selftest.py'
--- bzrlib/tests/test_selftest.py	2008-05-12 03:04:40 +0000
+++ bzrlib/tests/test_selftest.py	2008-06-11 04:20:16 +0000
@@ -396,61 +396,6 @@
             adapter.formats_to_scenarios(formats))
 
 
-class TestInterVersionedFileProviderAdapter(TestCase):
-    """A group of tests that test the InterVersionedFile test adapter."""
-
-    def test_scenarios(self):
-        # check that constructor parameters are passed through to the adapted
-        # test.
-        from bzrlib.tests.interversionedfile_implementations \
-            import InterVersionedFileTestProviderAdapter
-        server1 = "a"
-        server2 = "b"
-        formats = [(str, "C1", "C2"), (int, "D1", "D2")]
-        adapter = InterVersionedFileTestProviderAdapter(server1, server2, formats)
-        self.assertEqual([
-            ('str',
-             {'interversionedfile_class':str,
-              'transport_readonly_server': 'b',
-              'transport_server': 'a',
-              'versionedfile_factory': 'C1',
-              'versionedfile_factory_to': 'C2'}),
-            ('int',
-             {'interversionedfile_class': int,
-              'transport_readonly_server': 'b',
-              'transport_server': 'a',
-              'versionedfile_factory': 'D1',
-              'versionedfile_factory_to': 'D2'})],
-            adapter.scenarios)
-
-
-class TestRevisionStoreProviderAdapter(TestCase):
-    """A group of tests that test the RevisionStore test adapter."""
-
-    def test_scenarios(self):
-        # check that constructor parameters are passed through to the adapted
-        # test.
-        from bzrlib.tests.revisionstore_implementations \
-            import RevisionStoreTestProviderAdapter
-        # revision stores need a store factory - i.e. RevisionKnit
-        #, a readonly and rw transport 
-        # transport servers:
-        server1 = "a"
-        server2 = "b"
-        store_factories = ["c", "d"]
-        adapter = RevisionStoreTestProviderAdapter(server1, server2, store_factories)
-        self.assertEqual([
-            ('c',
-             {'store_factory': 'c',
-              'transport_readonly_server': 'b',
-              'transport_server': 'a'}),
-            ('d',
-             {'store_factory': 'd',
-              'transport_readonly_server': 'b',
-              'transport_server': 'a'})],
-            adapter.scenarios)
-
-
 class TestWorkingTreeProviderAdapter(TestCase):
     """A group of tests that test the workingtree implementation test adapter."""
 

=== modified file 'bzrlib/tests/test_smart.py'
--- bzrlib/tests/test_smart.py	2008-04-29 13:43:40 +0000
+++ bzrlib/tests/test_smart.py	2008-06-11 04:20:16 +0000
@@ -796,8 +796,7 @@
         request = smart.repository.SmartServerRepositoryGatherStats(backing)
         repository = self.make_repository('.')
         stats = repository.gather_stats()
-        size = stats['size']
-        expected_body = 'revisions: 0\nsize: %d\n' % size
+        expected_body = 'revisions: 0\n'
         self.assertEqual(SmartServerResponse(('ok', ), expected_body),
                          request.execute('', '', 'no'))
 
@@ -816,11 +815,9 @@
         tree.unlock()
 
         stats = tree.branch.repository.gather_stats()
-        size = stats['size']
         expected_body = ('firstrev: 123456.200 3600\n'
                          'latestrev: 654321.400 0\n'
-                         'revisions: 2\n'
-                         'size: %d\n' % size)
+                         'revisions: 2\n')
         self.assertEqual(SmartServerResponse(('ok', ), expected_body),
                          request.execute('',
                                          rev_id_utf8, 'no'))
@@ -841,12 +838,10 @@
         tree.unlock()
         stats = tree.branch.repository.gather_stats()
 
-        size = stats['size']
         expected_body = ('committers: 2\n'
                          'firstrev: 123456.200 3600\n'
                          'latestrev: 654321.400 0\n'
-                         'revisions: 2\n'
-                         'size: %d\n' % size)
+                         'revisions: 2\n')
         self.assertEqual(SmartServerResponse(('ok', ), expected_body),
                          request.execute('',
                                          rev_id_utf8, 'yes'))
@@ -938,111 +933,6 @@
             SmartServerResponse(('TokenMismatch',)), response)
 
 
-class TestSmartServerRepositoryTarball(tests.TestCaseWithTransport):
-
-    def test_repository_tarball(self):
-        backing = self.get_transport()
-        request = smart.repository.SmartServerRepositoryTarball(backing)
-        repository = self.make_repository('.')
-        # make some extraneous junk in the repository directory which should
-        # not be copied
-        self.build_tree(['.bzr/repository/extra-junk'])
-        response = request.execute('', 'bz2')
-        self.assertEqual(('ok',), response.args)
-        # body should be a tbz2
-        body_file = StringIO(response.body)
-        body_tar = tarfile.open('body_tar.tbz2', fileobj=body_file,
-            mode='r|bz2')
-        # let's make sure there are some key repository components inside it.
-        # the tarfile returns directories with trailing slashes...
-        names = set([n.rstrip('/') for n in body_tar.getnames()])
-        self.assertTrue('.bzr/repository/lock' in names)
-        self.assertTrue('.bzr/repository/format' in names)
-        self.assertTrue('.bzr/repository/extra-junk' not in names,
-            "extraneous file present in tar file")
-
-
-class TestSmartServerRepositoryStreamKnitData(tests.TestCaseWithMemoryTransport):
-
-    def test_fetch_revisions(self):
-        backing = self.get_transport()
-        request = smart.repository.SmartServerRepositoryStreamKnitDataForRevisions(backing)
-        tree = self.make_branch_and_memory_tree('.')
-        tree.lock_write()
-        tree.add('')
-        rev_id1_utf8 = u'\xc8'.encode('utf-8')
-        rev_id2_utf8 = u'\xc9'.encode('utf-8')
-        r1 = tree.commit('1st commit', rev_id=rev_id1_utf8)
-        r1 = tree.commit('2nd commit', rev_id=rev_id2_utf8)
-        tree.unlock()
-
-        response = request.execute('', rev_id2_utf8)
-        self.assertEqual(('ok',), response.args)
-        unpacker = pack.ContainerReader(StringIO(response.body))
-        names = []
-        for [name], read_bytes in unpacker.iter_records():
-            names.append(name)
-            bytes = read_bytes(None)
-            # The bytes should be a valid bencoded string.
-            bencode.bdecode(bytes)
-            # XXX: assert that the bencoded knit records have the right
-            # contents?
-        
-    def test_no_such_revision_error(self):
-        backing = self.get_transport()
-        request = smart.repository.SmartServerRepositoryStreamKnitDataForRevisions(backing)
-        repo = self.make_repository('.')
-        rev_id1_utf8 = u'\xc8'.encode('utf-8')
-        response = request.execute('', rev_id1_utf8)
-        self.assertEqual(
-            SmartServerResponse(('NoSuchRevision', rev_id1_utf8)),
-            response)
-
-
-class TestSmartServerRepositoryStreamRevisionsChunked(tests.TestCaseWithMemoryTransport):
-
-    def test_fetch_revisions(self):
-        backing = self.get_transport()
-        request = smart.repository.SmartServerRepositoryStreamRevisionsChunked(
-            backing)
-        tree = self.make_branch_and_memory_tree('.')
-        tree.lock_write()
-        tree.add('')
-        rev_id1_utf8 = u'\xc8'.encode('utf-8')
-        rev_id2_utf8 = u'\xc9'.encode('utf-8')
-        tree.commit('1st commit', rev_id=rev_id1_utf8)
-        tree.commit('2nd commit', rev_id=rev_id2_utf8)
-        tree.unlock()
-
-        response = request.execute('')
-        self.assertEqual(None, response)
-        response = request.do_body("%s\n%s\n1" % (rev_id2_utf8, rev_id1_utf8))
-        self.assertEqual(('ok',), response.args)
-        parser = pack.ContainerPushParser()
-        names = []
-        for stream_bytes in response.body_stream:
-            parser.accept_bytes(stream_bytes)
-            for [name], record_bytes in parser.read_pending_records():
-                names.append(name)
-                # The bytes should be a valid bencoded string.
-                bencode.bdecode(record_bytes)
-                # XXX: assert that the bencoded knit records have the right
-                # contents?
-        
-    def test_no_such_revision_error(self):
-        backing = self.get_transport()
-        request = smart.repository.SmartServerRepositoryStreamRevisionsChunked(
-            backing)
-        repo = self.make_repository('.')
-        rev_id1_utf8 = u'\xc8'.encode('utf-8')
-        response = request.execute('')
-        self.assertEqual(None, response)
-        response = request.do_body("%s\n\n1" % (rev_id1_utf8,))
-        self.assertEqual(
-            FailedSmartServerResponse(('NoSuchRevision', )),
-            response)
-
-
 class TestSmartServerIsReadonly(tests.TestCaseWithMemoryTransport):
 
     def test_is_readonly_no(self):

=== modified file 'bzrlib/tests/test_upgrade.py'
--- bzrlib/tests/test_upgrade.py	2008-04-24 07:22:53 +0000
+++ bzrlib/tests/test_upgrade.py	2008-06-11 04:20:16 +0000
@@ -135,11 +135,10 @@
         # 'dir-20051005095101-da1441ea3fa6917a'
         repo.lock_read()
         self.addCleanup(repo.unlock)
-        self.assertNotEqual(
-            [],
-            repo.weave_store.get_weave(
-                'dir-20051005095101-da1441ea3fa6917a',
-                repo.get_transaction()).versions())
+        text_keys = repo.texts.keys()
+        dir_keys = [key for key in text_keys if key[0] ==
+                'dir-20051005095101-da1441ea3fa6917a']
+        self.assertNotEqual([], dir_keys)
 
     def test_upgrade_to_meta_sets_workingtree_last_revision(self):
         self.build_tree_contents(_upgrade_dir_template)

=== modified file 'bzrlib/tests/test_versionedfile.py'
--- bzrlib/tests/test_versionedfile.py	2008-05-12 04:37:12 +0000
+++ bzrlib/tests/test_versionedfile.py	2008-06-11 04:20:16 +0000
@@ -37,7 +37,9 @@
                            )
 from bzrlib import knit as _mod_knit
 from bzrlib.knit import (
-    make_file_knit,
+    cleanup_pack_knit,
+    make_file_factory,
+    make_pack_factory,
     KnitAnnotateFactory,
     KnitPlainFactory,
     )
@@ -61,7 +63,6 @@
     ConstantMapper,
     HashEscapedPrefixMapper,
     PrefixMapper,
-    make_pack_factory,
     make_versioned_files_factory,
     )
 from bzrlib.weave import WeaveFile
@@ -86,42 +87,56 @@
     # plain text knits in packs (texts)
     len_one_adapter.scenarios = [
         ('weave-named', {
+            'cleanup':None,
             'factory':make_versioned_files_factory(WeaveFile,
                 ConstantMapper('inventory')),
-            'len':1,
+            'graph':True,
+            'key_length':1,
             }),
         ('named-knit', {
-            'factory':make_versioned_files_factory(make_file_knit,
-                ConstantMapper('revisions')),
-            'len':1,
+            'cleanup':None,
+            'factory':make_file_factory(False, ConstantMapper('revisions')),
+            'graph':True,
+            'key_length':1,
             }),
         ('named-nograph-knit-pack', {
+            'cleanup':cleanup_pack_knit,
             'factory':make_pack_factory(False, False, 1),
-            'len':1,
+            'graph':False,
+            'key_length':1,
             }),
         ('named-graph-knit-pack', {
+            'cleanup':cleanup_pack_knit,
             'factory':make_pack_factory(True, True, 1),
-            'len':1,
+            'graph':True,
+            'key_length':1,
             }),
         ('named-graph-nodelta-knit-pack', {
+            'cleanup':cleanup_pack_knit,
             'factory':make_pack_factory(True, False, 1),
-            'len':1,
+            'graph':True,
+            'key_length':1,
             }),
         ]
     len_two_adapter.scenarios = [
         ('weave-prefix', {
+            'cleanup':None,
             'factory':make_versioned_files_factory(WeaveFile,
                 PrefixMapper()),
-            'len':2,
+            'graph':True,
+            'key_length':2,
             }),
         ('annotated-knit-escape', {
-            'factory':make_versioned_files_factory(make_file_knit,
-                HashEscapedPrefixMapper()),
-            'len':2,
+            'cleanup':None,
+            'factory':make_file_factory(True, HashEscapedPrefixMapper()),
+            'graph':True,
+            'key_length':2,
             }),
         ('plain-knit-pack', {
+            'cleanup':cleanup_pack_knit,
             'factory':make_pack_factory(True, True, 2),
-            'len':2,
+            'graph':True,
+            'key_length':2,
             }),
         ]
     for test in iter_suite_tests(to_adapt):
@@ -158,6 +173,62 @@
     return f, parents
 
 
+def get_diamond_files(files, key_length, trailing_eol=True, left_only=False,
+    nograph=False):
+    """Get a diamond graph to exercise deltas and merges.
+
+    This creates a 5-node graph in files. If files supports 2-length keys two
+    graphs are made to exercise the support for multiple ids.
+    
+    :param trailing_eol: If True end the last line with \n.
+    :param key_length: The length of keys in files. Currently supports length 1
+        and 2 keys.
+    :param left_only: If True do not add the right and merged nodes.
+    :param nograph: If True, do not provide parents to the add_lines calls;
+        this is useful for tests that need inserted data but have graphless
+        stores.
+    :return: The results of the add_lines calls.
+    """
+    if key_length == 1:
+        prefixes = [()]
+    else:
+        prefixes = [('FileA',), ('FileB',)]
+    # insert a diamond graph to exercise deltas and merges.
+    if trailing_eol:
+        last_char = '\n'
+    else:
+        last_char = ''
+    result = []
+    def get_parents(suffix_list):
+        if nograph:
+            return ()
+        else:
+            result = [prefix + suffix for suffix in suffix_list]
+            return result
+    # we loop over each key because that spreads the inserts across prefixes,
+    # which is how commit operates.
+    for prefix in prefixes:
+        result.append(files.add_lines(prefix + ('origin',), (),
+            ['origin' + last_char]))
+    for prefix in prefixes:
+        result.append(files.add_lines(prefix + ('base',),
+            get_parents([('origin',)]), ['base' + last_char]))
+    for prefix in prefixes:
+        result.append(files.add_lines(prefix + ('left',),
+            get_parents([('base',)]),
+            ['base\n', 'left' + last_char]))
+    if not left_only:
+        for prefix in prefixes:
+            result.append(files.add_lines(prefix + ('right',),
+                get_parents([('base',)]),
+                ['base\n', 'right' + last_char]))
+        for prefix in prefixes:
+            result.append(files.add_lines(prefix + ('merged',),
+                get_parents([('left',), ('right',)]),
+                ['base\n', 'left\n', 'right\n', 'merged' + last_char]))
+    return result
+
+
 class VersionedFileTestMixIn(object):
     """A mixin test class for testing VersionedFiles.
 
@@ -194,266 +265,6 @@
         f = self.reopen_file(create=True)
         verify_file(f)
 
-    def test_get_record_stream_empty(self):
-        """get_record_stream is a replacement for get_data_stream."""
-        f = self.get_file()
-        entries = f.get_record_stream([], 'unordered', False)
-        self.assertEqual([], list(entries))
-
-    def assertValidStorageKind(self, storage_kind):
-        """Assert that storage_kind is a valid storage_kind."""
-        self.assertSubset([storage_kind],
-            ['mpdiff', 'knit-annotated-ft', 'knit-annotated-delta',
-             'knit-ft', 'knit-delta', 'fulltext', 'knit-annotated-ft-gz',
-             'knit-annotated-delta-gz', 'knit-ft-gz', 'knit-delta-gz'])
-
-    def capture_stream(self, f, entries, on_seen, parents):
-        """Capture a stream for testing."""
-        for factory in entries:
-            on_seen(factory.key)
-            self.assertValidStorageKind(factory.storage_kind)
-            self.assertEqual(f.get_sha1s([factory.key[0]])[0], factory.sha1)
-            self.assertEqual(parents[factory.key[0]], factory.parents)
-            self.assertIsInstance(factory.get_bytes_as(factory.storage_kind),
-                str)
-
-    def test_get_record_stream_interface(self):
-        """Each item in a stream has to provide a regular interface."""
-        f, parents = get_diamond_vf(self.get_file())
-        entries = f.get_record_stream(['merged', 'left', 'right', 'base'],
-            'unordered', False)
-        seen = set()
-        self.capture_stream(f, entries, seen.add, parents)
-        self.assertEqual(set([('base',), ('left',), ('right',), ('merged',)]),
-            seen)
-
-    def test_get_record_stream_interface_ordered(self):
-        """Each item in a stream has to provide a regular interface."""
-        f, parents = get_diamond_vf(self.get_file())
-        entries = f.get_record_stream(['merged', 'left', 'right', 'base'],
-            'topological', False)
-        seen = []
-        self.capture_stream(f, entries, seen.append, parents)
-        self.assertSubset([tuple(seen)],
-            (
-             (('base',), ('left',), ('right',), ('merged',)),
-             (('base',), ('right',), ('left',), ('merged',)),
-            ))
-
-    def test_get_record_stream_interface_ordered_with_delta_closure(self):
-        """Each item in a stream has to provide a regular interface."""
-        f, parents = get_diamond_vf(self.get_file())
-        entries = f.get_record_stream(['merged', 'left', 'right', 'base'],
-            'topological', True)
-        seen = []
-        for factory in entries:
-            seen.append(factory.key)
-            self.assertValidStorageKind(factory.storage_kind)
-            self.assertEqual(f.get_sha1s([factory.key[0]])[0], factory.sha1)
-            self.assertEqual(parents[factory.key[0]], factory.parents)
-            self.assertEqual(f.get_text(factory.key[0]),
-                factory.get_bytes_as('fulltext'))
-            self.assertIsInstance(factory.get_bytes_as(factory.storage_kind),
-                str)
-        self.assertSubset([tuple(seen)],
-            (
-             (('base',), ('left',), ('right',), ('merged',)),
-             (('base',), ('right',), ('left',), ('merged',)),
-            ))
-
-    def test_get_record_stream_unknown_storage_kind_raises(self):
-        """Asking for a storage kind that the stream cannot supply raises."""
-        f, parents = get_diamond_vf(self.get_file())
-        entries = f.get_record_stream(['merged', 'left', 'right', 'base'],
-            'unordered', False)
-        # We track the contents because we should be able to try, fail a
-        # particular kind and then ask for one that works and continue.
-        seen = set()
-        for factory in entries:
-            seen.add(factory.key)
-            self.assertValidStorageKind(factory.storage_kind)
-            self.assertEqual(f.get_sha1s([factory.key[0]])[0], factory.sha1)
-            self.assertEqual(parents[factory.key[0]], factory.parents)
-            # currently no stream emits mpdiff
-            self.assertRaises(errors.UnavailableRepresentation,
-                factory.get_bytes_as, 'mpdiff')
-            self.assertIsInstance(factory.get_bytes_as(factory.storage_kind),
-                str)
-        self.assertEqual(set([('base',), ('left',), ('right',), ('merged',)]),
-            seen)
-
-    def test_get_record_stream_missing_records_are_absent(self):
-        f, parents = get_diamond_vf(self.get_file())
-        entries = f.get_record_stream(['merged', 'left', 'right', 'or', 'base'],
-            'unordered', False)
-        self.assertAbsentRecord(f, parents, entries)
-        entries = f.get_record_stream(['merged', 'left', 'right', 'or', 'base'],
-            'topological', False)
-        self.assertAbsentRecord(f, parents, entries)
-
-    def assertAbsentRecord(self, f, parents, entries):
-        """Helper for test_get_record_stream_missing_records_are_absent."""
-        seen = set()
-        for factory in entries:
-            seen.add(factory.key)
-            if factory.key == ('or',):
-                self.assertEqual('absent', factory.storage_kind)
-                self.assertEqual(None, factory.sha1)
-                self.assertEqual(None, factory.parents)
-            else:
-                self.assertValidStorageKind(factory.storage_kind)
-                self.assertEqual(f.get_sha1s([factory.key[0]])[0], factory.sha1)
-                self.assertEqual(parents[factory.key[0]], factory.parents)
-                self.assertIsInstance(factory.get_bytes_as(factory.storage_kind),
-                    str)
-        self.assertEqual(
-            set([('base',), ('left',), ('right',), ('merged',), ('or',)]),
-            seen)
-
-    def test_filter_absent_records(self):
-        """Requested missing records can be filter trivially."""
-        f, parents = get_diamond_vf(self.get_file())
-        entries = f.get_record_stream(['merged', 'left', 'right', 'extra', 'base'],
-            'unordered', False)
-        seen = set()
-        self.capture_stream(f, versionedfile.filter_absent(entries), seen.add,
-            parents)
-        self.assertEqual(set([('base',), ('left',), ('right',), ('merged',)]),
-            seen)
-
-    def test_insert_record_stream_empty(self):
-        """Inserting an empty record stream should work."""
-        f = self.get_file()
-        stream = []
-        f.insert_record_stream([])
-
-    def assertIdenticalVersionedFile(self, left, right):
-        """Assert that left and right have the same contents."""
-        self.assertEqual(set(left.versions()), set(right.versions()))
-        self.assertEqual(left.get_parent_map(left.versions()),
-            right.get_parent_map(right.versions()))
-        for v in left.versions():
-            self.assertEqual(left.get_text(v), right.get_text(v))
-
-    def test_insert_record_stream_fulltexts(self):
-        """Any file should accept a stream of fulltexts."""
-        f = self.get_file()
-        weave_vf = WeaveFile('source', get_transport(self.get_url('.')),
-            create=True, get_scope=self.get_transaction)
-        source, _ = get_diamond_vf(weave_vf)
-        stream = source.get_record_stream(source.versions(), 'topological',
-            False)
-        f.insert_record_stream(stream)
-        self.assertIdenticalVersionedFile(f, source)
-
-    def test_insert_record_stream_fulltexts_noeol(self):
-        """Any file should accept a stream of fulltexts."""
-        f = self.get_file()
-        weave_vf = WeaveFile('source', get_transport(self.get_url('.')),
-            create=True, get_scope=self.get_transaction)
-        source, _ = get_diamond_vf(weave_vf, trailing_eol=False)
-        stream = source.get_record_stream(source.versions(), 'topological',
-            False)
-        f.insert_record_stream(stream)
-        self.assertIdenticalVersionedFile(f, source)
-
-    def test_insert_record_stream_annotated_knits(self):
-        """Any file should accept a stream from plain knits."""
-        f = self.get_file()
-        source = make_file_knit('source', get_transport(self.get_url('.')),
-            create=True)
-        get_diamond_vf(source)
-        stream = source.get_record_stream(source.versions(), 'topological',
-            False)
-        f.insert_record_stream(stream)
-        self.assertIdenticalVersionedFile(f, source)
-
-    def test_insert_record_stream_annotated_knits_noeol(self):
-        """Any file should accept a stream from plain knits."""
-        f = self.get_file()
-        source = make_file_knit('source', get_transport(self.get_url('.')),
-            create=True)
-        get_diamond_vf(source, trailing_eol=False)
-        stream = source.get_record_stream(source.versions(), 'topological',
-            False)
-        f.insert_record_stream(stream)
-        self.assertIdenticalVersionedFile(f, source)
-
-    def test_insert_record_stream_plain_knits(self):
-        """Any file should accept a stream from plain knits."""
-        f = self.get_file()
-        source = make_file_knit('source', get_transport(self.get_url('.')),
-            create=True, factory=KnitPlainFactory())
-        get_diamond_vf(source)
-        stream = source.get_record_stream(source.versions(), 'topological',
-            False)
-        f.insert_record_stream(stream)
-        self.assertIdenticalVersionedFile(f, source)
-
-    def test_insert_record_stream_plain_knits_noeol(self):
-        """Any file should accept a stream from plain knits."""
-        f = self.get_file()
-        source = make_file_knit('source', get_transport(self.get_url('.')),
-            create=True, factory=KnitPlainFactory())
-        get_diamond_vf(source, trailing_eol=False)
-        stream = source.get_record_stream(source.versions(), 'topological',
-            False)
-        f.insert_record_stream(stream)
-        self.assertIdenticalVersionedFile(f, source)
-
-    def test_insert_record_stream_existing_keys(self):
-        """Inserting keys already in a file should not error."""
-        f = self.get_file()
-        source = make_file_knit('source', get_transport(self.get_url('.')),
-            create=True, factory=KnitPlainFactory())
-        get_diamond_vf(source)
-        # insert some keys into f.
-        get_diamond_vf(f, left_only=True)
-        stream = source.get_record_stream(source.versions(), 'topological',
-            False)
-        f.insert_record_stream(stream)
-        self.assertIdenticalVersionedFile(f, source)
-
-    def test_insert_record_stream_missing_keys(self):
-        """Inserting a stream with absent keys should raise an error."""
-        f = self.get_file()
-        source = make_file_knit('source', get_transport(self.get_url('.')),
-            create=True, factory=KnitPlainFactory())
-        stream = source.get_record_stream(['missing'], 'topological',
-            False)
-        self.assertRaises(errors.RevisionNotPresent, f.insert_record_stream,
-            stream)
-
-    def test_insert_record_stream_out_of_order(self):
-        """An out of order stream can either error or work."""
-        f, parents = get_diamond_vf(self.get_file())
-        origin_entries = f.get_record_stream(['origin'], 'unordered', False)
-        end_entries = f.get_record_stream(['merged', 'left'],
-            'topological', False)
-        start_entries = f.get_record_stream(['right', 'base'],
-            'topological', False)
-        entries = chain(origin_entries, end_entries, start_entries)
-        target = self.get_file('target')
-        try:
-            target.insert_record_stream(entries)
-        except RevisionNotPresent:
-            # Must not have corrupted the file.
-            target.check()
-        else:
-            self.assertIdenticalVersionedFile(f, target)
-
-    def test_insert_record_stream_delta_missing_basis_no_corruption(self):
-        """Insertion where a needed basis is not included aborts safely."""
-        # Annotated source - deltas can be used in any knit.
-        source = make_file_knit('source', get_transport(self.get_url('.')),
-            create=True)
-        get_diamond_vf(source)
-        entries = source.get_record_stream(['origin', 'merged'], 'unordered', False)
-        f = self.get_file()
-        self.assertRaises(RevisionNotPresent, f.insert_record_stream, entries)
-        f.check()
-        self.assertFalse(f.has_version('merged'))
-
     def test_adds_with_parent_texts(self):
         f = self.get_file()
         parent_texts = {}
@@ -725,8 +536,6 @@
         self._transaction = 'after'
         self.assertRaises(errors.OutSideTransaction, f.add_lines, '', [], [])
         self.assertRaises(errors.OutSideTransaction, f.add_lines_with_ghosts, '', [], [])
-        self.assertRaises(errors.OutSideTransaction, self.applyDeprecated,
-            one_five, f.join, '')
         
     def test_copy_to(self):
         f = self.get_file()
@@ -936,8 +745,6 @@
                           'base',
                           [],
                           [])
-        self.assertRaises(errors.ReadOnlyError, self.applyDeprecated, one_five,
-            vf.join, 'base')
     
     def test_get_sha1s(self):
         # check the sha1 data is available
@@ -1010,171 +817,61 @@
         return WeaveFile
 
 
-class TestKnit(TestCaseWithMemoryTransport, VersionedFileTestMixIn):
-
-    def get_file(self, name='foo', create=True):
-        return make_file_knit(name, get_transport(self.get_url('.')),
-            delta=True, create=True, get_scope=self.get_transaction)
-
-    def get_factory(self):
-        return make_file_knit
-
-    def get_file_corrupted_text(self):
-        knit = self.get_file()
-        knit.add_lines('v1', [], ['hello\n'])
-        knit.add_lines('v2', ['v1'], ['hello\n', 'there\n'])
-        return knit
-
-    def reopen_file(self, name='foo', create=False):
-        return self.get_file(name, create)
-
-    def test_detection(self):
-        knit = self.get_file()
-        knit.check()
-
-    def test_no_implicit_create(self):
-        self.assertRaises(errors.NoSuchFile, self.get_factory(), 'foo',
-            get_transport(self.get_url('.')))
-
-
-class TestPlaintextKnit(TestKnit):
-    """Test a knit with no cached annotations"""
-
-    def get_file(self, name='foo', create=True):
-        return make_file_knit(name, get_transport(self.get_url('.')),
-            delta=True, create=create, get_scope=self.get_transaction,
-            factory=_mod_knit.KnitPlainFactory())
-
-
 class TestPlanMergeVersionedFile(TestCaseWithMemoryTransport):
 
     def setUp(self):
         TestCaseWithMemoryTransport.setUp(self)
-        self.vf1 = make_file_knit('root', self.get_transport(), create=True)
-        self.vf2 = make_file_knit('root', self.get_transport(), create=True)
-        self.plan_merge_vf = versionedfile._PlanMergeVersionedFile('root',
-            [self.vf1, self.vf2])
+        mapper = PrefixMapper()
+        factory = make_file_factory(True, mapper)
+        self.vf1 = factory(self.get_transport('root-1'))
+        self.vf2 = factory(self.get_transport('root-2'))
+        self.plan_merge_vf = versionedfile._PlanMergeVersionedFile('root')
+        self.plan_merge_vf.fallback_versionedfiles.extend([self.vf1, self.vf2])
 
     def test_add_lines(self):
-        self.plan_merge_vf.add_lines('a:', [], [])
-        self.assertRaises(ValueError, self.plan_merge_vf.add_lines, 'a', [],
-                          [])
-        self.assertRaises(ValueError, self.plan_merge_vf.add_lines, 'a:', None,
-                          [])
-        self.assertRaises(ValueError, self.plan_merge_vf.add_lines, 'a:', [],
-                          None)
-
-    def test_ancestry(self):
-        self.vf1.add_lines('A', [], [])
-        self.vf1.add_lines('B', ['A'], [])
-        self.plan_merge_vf.add_lines('C:', ['B'], [])
-        self.plan_merge_vf.add_lines('D:', ['C:'], [])
-        self.assertEqual(set(['A', 'B', 'C:', 'D:']),
-            self.plan_merge_vf.get_ancestry('D:', topo_sorted=False))
+        self.plan_merge_vf.add_lines(('root', 'a:'), [], [])
+        self.assertRaises(ValueError, self.plan_merge_vf.add_lines,
+            ('root', 'a'), [], [])
+        self.assertRaises(ValueError, self.plan_merge_vf.add_lines,
+            ('root', 'a:'), None, [])
+        self.assertRaises(ValueError, self.plan_merge_vf.add_lines,
+            ('root', 'a:'), [], None)
 
     def setup_abcde(self):
-        self.vf1.add_lines('A', [], ['a'])
-        self.vf1.add_lines('B', ['A'], ['b'])
-        self.vf2.add_lines('C', [], ['c'])
-        self.vf2.add_lines('D', ['C'], ['d'])
-        self.plan_merge_vf.add_lines('E:', ['B', 'D'], ['e'])
-
-    def test_ancestry_uses_all_versionedfiles(self):
-        self.setup_abcde()
-        self.assertEqual(set(['A', 'B', 'C', 'D', 'E:']),
-            self.plan_merge_vf.get_ancestry('E:', topo_sorted=False))
-
-    def test_ancestry_raises_revision_not_present(self):
-        error = self.assertRaises(errors.RevisionNotPresent,
-                                  self.plan_merge_vf.get_ancestry, 'E:', False)
-        self.assertContainsRe(str(error), '{E:} not present in "root"')
+        self.vf1.add_lines(('root', 'A'), [], ['a'])
+        self.vf1.add_lines(('root', 'B'), [('root', 'A')], ['b'])
+        self.vf2.add_lines(('root', 'C'), [], ['c'])
+        self.vf2.add_lines(('root', 'D'), [('root', 'C')], ['d'])
+        self.plan_merge_vf.add_lines(('root', 'E:'),
+            [('root', 'B'), ('root', 'D')], ['e'])
 
     def test_get_parents(self):
         self.setup_abcde()
-        self.assertEqual({'B':('A',)}, self.plan_merge_vf.get_parent_map(['B']))
-        self.assertEqual({'D':('C',)}, self.plan_merge_vf.get_parent_map(['D']))
-        self.assertEqual({'E:':('B', 'D')},
-            self.plan_merge_vf.get_parent_map(['E:']))
-        self.assertEqual({}, self.plan_merge_vf.get_parent_map(['F']))
+        self.assertEqual({('root', 'B'):(('root', 'A'),)},
+            self.plan_merge_vf.get_parent_map([('root', 'B')]))
+        self.assertEqual({('root', 'D'):(('root', 'C'),)},
+            self.plan_merge_vf.get_parent_map([('root', 'D')]))
+        self.assertEqual({('root', 'E:'):(('root', 'B'),('root', 'D'))},
+            self.plan_merge_vf.get_parent_map([('root', 'E:')]))
+        self.assertEqual({},
+            self.plan_merge_vf.get_parent_map([('root', 'F')]))
         self.assertEqual({
-                'B':('A',),
-                'D':('C',),
-                'E:':('B', 'D'),
-                }, self.plan_merge_vf.get_parent_map(['B', 'D', 'E:', 'F']))
+                ('root', 'B'):(('root', 'A'),),
+                ('root', 'D'):(('root', 'C'),),
+                ('root', 'E:'):(('root', 'B'),('root', 'D')),
+                },
+            self.plan_merge_vf.get_parent_map(
+                [('root', 'B'), ('root', 'D'), ('root', 'E:'), ('root', 'F')]))
 
-    def test_get_lines(self):
+    def test_get_record_stream(self):
         self.setup_abcde()
-        self.assertEqual(['a'], self.plan_merge_vf.get_lines('A'))
-        self.assertEqual(['c'], self.plan_merge_vf.get_lines('C'))
-        self.assertEqual(['e'], self.plan_merge_vf.get_lines('E:'))
-        error = self.assertRaises(errors.RevisionNotPresent,
-                                  self.plan_merge_vf.get_lines, 'F')
-        self.assertContainsRe(str(error), '{F} not present in "root"')
-
-
-class InterString(versionedfile.InterVersionedFile):
-    """An inter-versionedfile optimised code path for strings.
-
-    This is for use during testing where we use strings as versionedfiles
-    so that none of the default regsitered interversionedfile classes will
-    match - which lets us test the match logic.
-    """
-
-    @staticmethod
-    def is_compatible(source, target):
-        """InterString is compatible with strings-as-versionedfiles."""
-        return isinstance(source, str) and isinstance(target, str)
-
-
-# TODO this and the InterRepository core logic should be consolidatable
-# if we make the registry a separate class though we still need to 
-# test the behaviour in the active registry to catch failure-to-handle-
-# stange-objects
-class TestInterVersionedFile(TestCaseWithMemoryTransport):
-
-    def test_get_default_inter_versionedfile(self):
-        # test that the InterVersionedFile.get(a, b) probes
-        # for a class where is_compatible(a, b) returns
-        # true and returns a default interversionedfile otherwise.
-        # This also tests that the default registered optimised interversionedfile
-        # classes do not barf inappropriately when a surprising versionedfile type
-        # is handed to them.
-        dummy_a = "VersionedFile 1."
-        dummy_b = "VersionedFile 2."
-        self.assertGetsDefaultInterVersionedFile(dummy_a, dummy_b)
-
-    def assertGetsDefaultInterVersionedFile(self, a, b):
-        """Asserts that InterVersionedFile.get(a, b) -> the default."""
-        inter = versionedfile.InterVersionedFile.get(a, b)
-        self.assertEqual(versionedfile.InterVersionedFile,
-                         inter.__class__)
-        self.assertEqual(a, inter.source)
-        self.assertEqual(b, inter.target)
-
-    def test_register_inter_versionedfile_class(self):
-        # test that a optimised code path provider - a
-        # InterVersionedFile subclass can be registered and unregistered
-        # and that it is correctly selected when given a versionedfile
-        # pair that it returns true on for the is_compatible static method
-        # check
-        dummy_a = "VersionedFile 1."
-        dummy_b = "VersionedFile 2."
-        versionedfile.InterVersionedFile.register_optimiser(InterString)
-        try:
-            # we should get the default for something InterString returns False
-            # to
-            self.assertFalse(InterString.is_compatible(dummy_a, None))
-            self.assertGetsDefaultInterVersionedFile(dummy_a, None)
-            # and we should get an InterString for a pair it 'likes'
-            self.assertTrue(InterString.is_compatible(dummy_a, dummy_b))
-            inter = versionedfile.InterVersionedFile.get(dummy_a, dummy_b)
-            self.assertEqual(InterString, inter.__class__)
-            self.assertEqual(dummy_a, inter.source)
-            self.assertEqual(dummy_b, inter.target)
-        finally:
-            versionedfile.InterVersionedFile.unregister_optimiser(InterString)
-        # now we should get the default InterVersionedFile object again.
-        self.assertGetsDefaultInterVersionedFile(dummy_a, dummy_b)
+        def get_record(suffix):
+            return self.plan_merge_vf.get_record_stream(
+                [('root', suffix)], 'unordered', True).next()
+        self.assertEqual('a', get_record('A').get_bytes_as('fulltext'))
+        self.assertEqual('c', get_record('C').get_bytes_as('fulltext'))
+        self.assertEqual('e', get_record('E:').get_bytes_as('fulltext'))
+        self.assertEqual('absent', get_record('F').storage_kind)
 
 
 class TestReadonlyHttpMixin(object):
@@ -1207,16 +904,6 @@
         return WeaveFile
 
 
-class TestKnitHTTP(TestCaseWithWebserver, TestReadonlyHttpMixin):
-
-    def get_file(self):
-        return make_file_knit('foo', get_transport(self.get_url('.')),
-            delta=True, create=True, get_scope=self.get_transaction)
-
-    def get_factory(self):
-        return make_file_knit
-
-
 class MergeCasesMixin(object):
 
     def doMerge(self, base, a, b, mp):
@@ -1454,16 +1141,6 @@
         self._test_merge_from_strings(base, a, b, result)
 
 
-class TestKnitMerge(TestCaseWithMemoryTransport, MergeCasesMixin):
-
-    def get_file(self, name='foo'):
-        return make_file_knit(name, get_transport(self.get_url('.')),
-                                 delta=True, create=True)
-
-    def log_contents(self, w):
-        pass
-
-
 class TestWeaveMerge(TestCaseWithMemoryTransport, MergeCasesMixin):
 
     def get_file(self, name='foo'):
@@ -1504,21 +1181,18 @@
             self.assertIsInstance(adapter, klass)
 
     def get_knit(self, annotated=True):
-        if annotated:
-            factory = KnitAnnotateFactory()
-        else:
-            factory = KnitPlainFactory()
-        return make_file_knit('knit', self.get_transport('.'), delta=True,
-            create=True, factory=factory)
+        mapper = ConstantMapper('knit')
+        transport = self.get_transport()
+        return make_file_factory(annotated, mapper)(transport)
 
     def helpGetBytes(self, f, ft_adapter, delta_adapter):
         """Grab the interested adapted texts for tests."""
         # origin is a fulltext
-        entries = f.get_record_stream(['origin'], 'unordered', False)
+        entries = f.get_record_stream([('origin',)], 'unordered', False)
         base = entries.next()
         ft_data = ft_adapter.get_bytes(base, base.get_bytes_as(base.storage_kind))
         # merged is both a delta and multiple parents.
-        entries = f.get_record_stream(['merged'], 'unordered', False)
+        entries = f.get_record_stream([('merged',)], 'unordered', False)
         merged = entries.next()
         delta_data = delta_adapter.get_bytes(merged,
             merged.get_bytes_as(merged.storage_kind))
@@ -1527,7 +1201,8 @@
     def test_deannotation_noeol(self):
         """Test converting annotated knits to unannotated knits."""
         # we need a full text, and a delta
-        f, parents = get_diamond_vf(self.get_knit(), trailing_eol=False)
+        f = self.get_knit()
+        get_diamond_files(f, 1, trailing_eol=False)
         ft_data, delta_data = self.helpGetBytes(f,
             _mod_knit.FTAnnotatedToUnannotated(None),
             _mod_knit.DeltaAnnotatedToUnannotated(None))
@@ -1544,7 +1219,8 @@
     def test_deannotation(self):
         """Test converting annotated knits to unannotated knits."""
         # we need a full text, and a delta
-        f, parents = get_diamond_vf(self.get_knit())
+        f = self.get_knit()
+        get_diamond_files(f, 1)
         ft_data, delta_data = self.helpGetBytes(f,
             _mod_knit.FTAnnotatedToUnannotated(None),
             _mod_knit.DeltaAnnotatedToUnannotated(None))
@@ -1561,30 +1237,34 @@
     def test_annotated_to_fulltext_no_eol(self):
         """Test adapting annotated knits to full texts (for -> weaves)."""
         # we need a full text, and a delta
-        f, parents = get_diamond_vf(self.get_knit(), trailing_eol=False)
+        f = self.get_knit()
+        get_diamond_files(f, 1, trailing_eol=False)
         # Reconstructing a full text requires a backing versioned file, and it
         # must have the base lines requested from it.
-        logged_vf = versionedfile.RecordingVersionedFileDecorator(f)
+        logged_vf = versionedfile.RecordingVersionedFilesDecorator(f)
         ft_data, delta_data = self.helpGetBytes(f,
             _mod_knit.FTAnnotatedToFullText(None),
             _mod_knit.DeltaAnnotatedToFullText(logged_vf))
         self.assertEqual('origin', ft_data)
         self.assertEqual('base\nleft\nright\nmerged', delta_data)
-        self.assertEqual([('get_lines', 'left')], logged_vf.calls)
+        self.assertEqual([('get_record_stream', [('left',)], 'unordered',
+            True)], logged_vf.calls)
 
     def test_annotated_to_fulltext(self):
         """Test adapting annotated knits to full texts (for -> weaves)."""
         # we need a full text, and a delta
-        f, parents = get_diamond_vf(self.get_knit())
+        f = self.get_knit()
+        get_diamond_files(f, 1)
         # Reconstructing a full text requires a backing versioned file, and it
         # must have the base lines requested from it.
-        logged_vf = versionedfile.RecordingVersionedFileDecorator(f)
+        logged_vf = versionedfile.RecordingVersionedFilesDecorator(f)
         ft_data, delta_data = self.helpGetBytes(f,
             _mod_knit.FTAnnotatedToFullText(None),
             _mod_knit.DeltaAnnotatedToFullText(logged_vf))
         self.assertEqual('origin\n', ft_data)
         self.assertEqual('base\nleft\nright\nmerged\n', delta_data)
-        self.assertEqual([('get_lines', 'left')], logged_vf.calls)
+        self.assertEqual([('get_record_stream', [('left',)], 'unordered',
+            True)], logged_vf.calls)
 
     def test_unannotated_to_fulltext(self):
         """Test adapting unannotated knits to full texts.
@@ -1592,16 +1272,18 @@
         This is used for -> weaves, and for -> annotated knits.
         """
         # we need a full text, and a delta
-        f, parents = get_diamond_vf(self.get_knit(annotated=False))
+        f = self.get_knit(annotated=False)
+        get_diamond_files(f, 1)
         # Reconstructing a full text requires a backing versioned file, and it
         # must have the base lines requested from it.
-        logged_vf = versionedfile.RecordingVersionedFileDecorator(f)
+        logged_vf = versionedfile.RecordingVersionedFilesDecorator(f)
         ft_data, delta_data = self.helpGetBytes(f,
             _mod_knit.FTPlainToFullText(None),
             _mod_knit.DeltaPlainToFullText(logged_vf))
         self.assertEqual('origin\n', ft_data)
         self.assertEqual('base\nleft\nright\nmerged\n', delta_data)
-        self.assertEqual([('get_lines', 'left')], logged_vf.calls)
+        self.assertEqual([('get_record_stream', [('left',)], 'unordered',
+            True)], logged_vf.calls)
 
     def test_unannotated_to_fulltext_no_eol(self):
         """Test adapting unannotated knits to full texts.
@@ -1609,17 +1291,18 @@
         This is used for -> weaves, and for -> annotated knits.
         """
         # we need a full text, and a delta
-        f, parents = get_diamond_vf(self.get_knit(annotated=False),
-            trailing_eol=False)
+        f = self.get_knit(annotated=False)
+        get_diamond_files(f, 1, trailing_eol=False)
         # Reconstructing a full text requires a backing versioned file, and it
         # must have the base lines requested from it.
-        logged_vf = versionedfile.RecordingVersionedFileDecorator(f)
+        logged_vf = versionedfile.RecordingVersionedFilesDecorator(f)
         ft_data, delta_data = self.helpGetBytes(f,
             _mod_knit.FTPlainToFullText(None),
             _mod_knit.DeltaPlainToFullText(logged_vf))
         self.assertEqual('origin', ft_data)
         self.assertEqual('base\nleft\nright\nmerged', delta_data)
-        self.assertEqual([('get_lines', 'left')], logged_vf.calls)
+        self.assertEqual([('get_record_stream', [('left',)], 'unordered',
+            True)], logged_vf.calls)
 
 
 class TestKeyMapper(TestCaseWithMemoryTransport):
@@ -1649,18 +1332,737 @@
     def test_hash_escaped_mapper(self):
         #knit1: hash + escaped
         mapper = versionedfile.HashEscapedPrefixMapper()
-        self.assertEqual("88/%20", mapper.map((" ", "revision-id")))
-        self.assertEqual("ed/fil%45-%49d", mapper.map(("filE-Id",
-            "revision-id")))
-        self.assertEqual("88/ne%57-%49d", mapper.map(("neW-Id",
-            "revision-id")))
-        self.assertEqual(('filE-Id',), mapper.unmap("ed/fil%45-%49d"))
-        self.assertEqual(('neW-Id',), mapper.unmap("88/ne%57-%49d"))
+        self.assertEqual("88/%2520", mapper.map((" ", "revision-id")))
+        self.assertEqual("ed/fil%2545-%2549d", mapper.map(("filE-Id",
+            "revision-id")))
+        self.assertEqual("88/ne%2557-%2549d", mapper.map(("neW-Id",
+            "revision-id")))
+        self.assertEqual(('filE-Id',), mapper.unmap("ed/fil%2545-%2549d"))
+        self.assertEqual(('neW-Id',), mapper.unmap("88/ne%2557-%2549d"))
 
 
 class TestVersionedFiles(TestCaseWithMemoryTransport):
     """Tests for the multiple-file variant of VersionedFile."""
 
+    def get_versionedfiles(self, relpath='files'):
+        transport = self.get_transport(relpath)
+        if relpath != '.':
+            transport.mkdir('.')
+        files = self.factory(transport)
+        if self.cleanup is not None:
+            self.addCleanup(lambda:self.cleanup(files))
+        return files
+
+    def test_annotate(self):
+        files = self.get_versionedfiles()
+        self.get_diamond_files(files)
+        if self.key_length == 1:
+            prefix = ()
+        else:
+            prefix = ('FileA',)
+        # introduced full text
+        origins = files.annotate(prefix + ('origin',))
+        self.assertEqual([
+            (prefix + ('origin',), 'origin\n')],
+            origins)
+        # a delta
+        origins = files.annotate(prefix + ('base',))
+        self.assertEqual([
+            (prefix + ('base',), 'base\n')],
+            origins)
+        # a merge
+        origins = files.annotate(prefix + ('merged',))
+        if self.graph:
+            self.assertEqual([
+                (prefix + ('base',), 'base\n'),
+                (prefix + ('left',), 'left\n'),
+                (prefix + ('right',), 'right\n'),
+                (prefix + ('merged',), 'merged\n')
+                ],
+                origins)
+        else:
+            # Without a graph everything is new.
+            self.assertEqual([
+                (prefix + ('merged',), 'base\n'),
+                (prefix + ('merged',), 'left\n'),
+                (prefix + ('merged',), 'right\n'),
+                (prefix + ('merged',), 'merged\n')
+                ],
+                origins)
+        self.assertRaises(RevisionNotPresent,
+            files.annotate, prefix + ('missing-key',))
+
     def test_construct(self):
         """Each parameterised test can be constructed on a transport."""
-        files = self.factory(self.get_transport())
+        files = self.get_versionedfiles()
+
+    def get_diamond_files(self, files, trailing_eol=True, left_only=False):
+        return get_diamond_files(files, self.key_length,
+            trailing_eol=trailing_eol, nograph=not self.graph,
+            left_only=left_only)
+
+    def test_add_lines_return(self):
+        files = self.get_versionedfiles()
+        # save code by using the stock data insertion helper.
+        adds = self.get_diamond_files(files)
+        results = []
+        # We can only validate the first 2 elements returned from add_lines.
+        for add in adds:
+            self.assertEqual(3, len(add))
+            results.append(add[:2])
+        if self.key_length == 1:
+            self.assertEqual([
+                ('00e364d235126be43292ab09cb4686cf703ddc17', 7),
+                ('51c64a6f4fc375daf0d24aafbabe4d91b6f4bb44', 5),
+                ('a8478686da38e370e32e42e8a0c220e33ee9132f', 10),
+                ('9ef09dfa9d86780bdec9219a22560c6ece8e0ef1', 11),
+                ('ed8bce375198ea62444dc71952b22cfc2b09226d', 23)],
+                results)
+        elif self.key_length == 2:
+            self.assertEqual([
+                ('00e364d235126be43292ab09cb4686cf703ddc17', 7),
+                ('00e364d235126be43292ab09cb4686cf703ddc17', 7),
+                ('51c64a6f4fc375daf0d24aafbabe4d91b6f4bb44', 5),
+                ('51c64a6f4fc375daf0d24aafbabe4d91b6f4bb44', 5),
+                ('a8478686da38e370e32e42e8a0c220e33ee9132f', 10),
+                ('a8478686da38e370e32e42e8a0c220e33ee9132f', 10),
+                ('9ef09dfa9d86780bdec9219a22560c6ece8e0ef1', 11),
+                ('9ef09dfa9d86780bdec9219a22560c6ece8e0ef1', 11),
+                ('ed8bce375198ea62444dc71952b22cfc2b09226d', 23),
+                ('ed8bce375198ea62444dc71952b22cfc2b09226d', 23)],
+                results)
+
+    def test_empty_lines(self):
+        """Empty files can be stored."""
+        f = self.get_versionedfiles()
+        key_a = self.get_simple_key('a')
+        f.add_lines(key_a, [], [])
+        self.assertEqual('',
+            f.get_record_stream([key_a], 'unordered', True
+                ).next().get_bytes_as('fulltext'))
+        key_b = self.get_simple_key('b')
+        f.add_lines(key_b, self.get_parents([key_a]), [])
+        self.assertEqual('',
+            f.get_record_stream([key_b], 'unordered', True
+                ).next().get_bytes_as('fulltext'))
+
+    def test_newline_only(self):
+        f = self.get_versionedfiles()
+        key_a = self.get_simple_key('a')
+        f.add_lines(key_a, [], ['\n'])
+        self.assertEqual('\n',
+            f.get_record_stream([key_a], 'unordered', True
+                ).next().get_bytes_as('fulltext'))
+        key_b = self.get_simple_key('b')
+        f.add_lines(key_b, self.get_parents([key_a]), ['\n'])
+        self.assertEqual('\n',
+            f.get_record_stream([key_b], 'unordered', True
+                ).next().get_bytes_as('fulltext'))
+
+    def test_get_record_stream_empty(self):
+        """An empty stream can be requested without error."""
+        f = self.get_versionedfiles()
+        entries = f.get_record_stream([], 'unordered', False)
+        self.assertEqual([], list(entries))
+
+    def assertValidStorageKind(self, storage_kind):
+        """Assert that storage_kind is a valid storage_kind."""
+        self.assertSubset([storage_kind],
+            ['mpdiff', 'knit-annotated-ft', 'knit-annotated-delta',
+             'knit-ft', 'knit-delta', 'fulltext', 'knit-annotated-ft-gz',
+             'knit-annotated-delta-gz', 'knit-ft-gz', 'knit-delta-gz'])
+
+    def capture_stream(self, f, entries, on_seen, parents):
+        """Capture a stream for testing."""
+        for factory in entries:
+            on_seen(factory.key)
+            self.assertValidStorageKind(factory.storage_kind)
+            self.assertEqual(f.get_sha1s([factory.key])[0], factory.sha1)
+            self.assertEqual(parents[factory.key], factory.parents)
+            self.assertIsInstance(factory.get_bytes_as(factory.storage_kind),
+                str)
+
+    def test_get_record_stream_interface(self):
+        """each item in a stream has to provide a regular interface."""
+        files = self.get_versionedfiles()
+        self.get_diamond_files(files)
+        keys, _ = self.get_keys_and_sort_order()
+        parent_map = files.get_parent_map(keys)
+        entries = files.get_record_stream(keys, 'unordered', False)
+        seen = set()
+        self.capture_stream(files, entries, seen.add, parent_map)
+        self.assertEqual(set(keys), seen)
+
+    def get_simple_key(self, suffix):
+        """Return a key for the object under test."""
+        if self.key_length == 1:
+            return (suffix,)
+        else:
+            return ('FileA',) + (suffix,)
+
+    def get_keys_and_sort_order(self):
+        """Get diamond test keys list, and their sort ordering."""
+        if self.key_length == 1:
+            keys = [('merged',), ('left',), ('right',), ('base',)]
+            sort_order = {('merged',):2, ('left',):1, ('right',):1, ('base',):0}
+        else:
+            keys = [
+                ('FileA', 'merged'), ('FileA', 'left'), ('FileA', 'right'),
+                ('FileA', 'base'),
+                ('FileB', 'merged'), ('FileB', 'left'), ('FileB', 'right'),
+                ('FileB', 'base'),
+                ]
+            sort_order = {
+                ('FileA', 'merged'):2, ('FileA', 'left'):1, ('FileA', 'right'):1,
+                ('FileA', 'base'):0,
+                ('FileB', 'merged'):2, ('FileB', 'left'):1, ('FileB', 'right'):1,
+                ('FileB', 'base'):0,
+                }
+        return keys, sort_order
+
+    def test_get_record_stream_interface_ordered(self):
+        """each item in a stream has to provide a regular interface."""
+        files = self.get_versionedfiles()
+        self.get_diamond_files(files)
+        keys, sort_order = self.get_keys_and_sort_order()
+        parent_map = files.get_parent_map(keys)
+        entries = files.get_record_stream(keys, 'topological', False)
+        seen = []
+        self.capture_stream(files, entries, seen.append, parent_map)
+        self.assertStreamOrder(sort_order, seen, keys)
+
+    def test_get_record_stream_interface_ordered_with_delta_closure(self):
+        """each item must be accessible as a fulltext."""
+        files = self.get_versionedfiles()
+        self.get_diamond_files(files)
+        keys, sort_order = self.get_keys_and_sort_order()
+        parent_map = files.get_parent_map(keys)
+        entries = files.get_record_stream(keys, 'topological', True)
+        seen = []
+        for factory in entries:
+            seen.append(factory.key)
+            self.assertValidStorageKind(factory.storage_kind)
+            self.assertSubset([factory.sha1], [None, files.get_sha1s([factory.key])[0]])
+            self.assertEqual(parent_map[factory.key], factory.parents)
+            # self.assertEqual(files.get_text(factory.key),
+            self.assertIsInstance(factory.get_bytes_as('fulltext'), str)
+            self.assertIsInstance(factory.get_bytes_as(factory.storage_kind),
+                str)
+        self.assertStreamOrder(sort_order, seen, keys)
+
+    def assertStreamOrder(self, sort_order, seen, keys):
+        self.assertEqual(len(set(seen)), len(keys))
+        if self.key_length == 1:
+            lows = {():0}
+        else:
+            lows = {('FileA',):0, ('FileB',):0}
+        if not self.graph:
+            self.assertEqual(set(keys), set(seen))
+        else:
+            for key in seen:
+                sort_pos = sort_order[key]
+                self.assertTrue(sort_pos >= lows[key[:-1]],
+                    "Out of order in sorted stream: %r, %r" % (key, seen))
+                lows[key[:-1]] = sort_pos
+
+    def test_get_record_stream_unknown_storage_kind_raises(self):
+        """Asking for a storage kind that the stream cannot supply raises."""
+        files = self.get_versionedfiles()
+        self.get_diamond_files(files)
+        if self.key_length == 1:
+            keys = [('merged',), ('left',), ('right',), ('base',)]
+        else:
+            keys = [
+                ('FileA', 'merged'), ('FileA', 'left'), ('FileA', 'right'),
+                ('FileA', 'base'),
+                ('FileB', 'merged'), ('FileB', 'left'), ('FileB', 'right'),
+                ('FileB', 'base'),
+                ]
+        parent_map = files.get_parent_map(keys)
+        entries = files.get_record_stream(keys, 'unordered', False)
+        # We track the contents because we should be able to try, fail a
+        # particular kind and then ask for one that works and continue.
+        seen = set()
+        for factory in entries:
+            seen.add(factory.key)
+            self.assertValidStorageKind(factory.storage_kind)
+            self.assertEqual(files.get_sha1s([factory.key])[0], factory.sha1)
+            self.assertEqual(parent_map[factory.key], factory.parents)
+            # currently no stream emits mpdiff
+            self.assertRaises(errors.UnavailableRepresentation,
+                factory.get_bytes_as, 'mpdiff')
+            self.assertIsInstance(factory.get_bytes_as(factory.storage_kind),
+                str)
+        self.assertEqual(set(keys), seen)
+
+    def test_get_record_stream_missing_records_are_absent(self):
+        files = self.get_versionedfiles()
+        self.get_diamond_files(files)
+        if self.key_length == 1:
+            keys = [('merged',), ('left',), ('right',), ('absent',), ('base',)]
+        else:
+            keys = [
+                ('FileA', 'merged'), ('FileA', 'left'), ('FileA', 'right'),
+                ('FileA', 'absent'), ('FileA', 'base'),
+                ('FileB', 'merged'), ('FileB', 'left'), ('FileB', 'right'),
+                ('FileB', 'absent'), ('FileB', 'base'),
+                ('absent', 'absent'),
+                ]
+        parent_map = files.get_parent_map(keys)
+        entries = files.get_record_stream(keys, 'unordered', False)
+        self.assertAbsentRecord(files, keys, parent_map, entries)
+        entries = files.get_record_stream(keys, 'topological', False)
+        self.assertAbsentRecord(files, keys, parent_map, entries)
+
+    def assertAbsentRecord(self, files, keys, parents, entries):
+        """Helper for test_get_record_stream_missing_records_are_absent."""
+        seen = set()
+        for factory in entries:
+            seen.add(factory.key)
+            if factory.key[-1] == 'absent':
+                self.assertEqual('absent', factory.storage_kind)
+                self.assertEqual(None, factory.sha1)
+                self.assertEqual(None, factory.parents)
+            else:
+                self.assertValidStorageKind(factory.storage_kind)
+                self.assertEqual(files.get_sha1s([factory.key])[0], factory.sha1)
+                self.assertEqual(parents[factory.key], factory.parents)
+                self.assertIsInstance(factory.get_bytes_as(factory.storage_kind),
+                    str)
+        self.assertEqual(set(keys), seen)
+
+    def test_filter_absent_records(self):
+        """Requested missing records can be filter trivially."""
+        files = self.get_versionedfiles()
+        self.get_diamond_files(files)
+        keys, _ = self.get_keys_and_sort_order()
+        parent_map = files.get_parent_map(keys)
+        # Add an absent record in the middle of the present keys. (We don't ask
+        # for just absent keys to ensure that content before and after the
+        # absent keys is still delivered).
+        present_keys = list(keys)
+        if self.key_length == 1:
+            keys.insert(2, ('extra',))
+        else:
+            keys.insert(2, ('extra', 'extra'))
+        entries = files.get_record_stream(keys, 'unordered', False)
+        seen = set()
+        self.capture_stream(files, versionedfile.filter_absent(entries), seen.add,
+            parent_map)
+        self.assertEqual(set(present_keys), seen)
+
+    def get_mapper(self):
+        """Get a mapper suitable for the key length of the test interface."""
+        if self.key_length == 1:
+            return ConstantMapper('source')
+        else:
+            return HashEscapedPrefixMapper()
+
+    def get_parents(self, parents):
+        """Get parents, taking self.graph into consideration."""
+        if self.graph:
+            return parents
+        else:
+            return None
+
+    def test_get_parent_map(self):
+        files = self.get_versionedfiles()
+        if self.key_length == 1:
+            parent_details = [
+                (('r0',), self.get_parents(())),
+                (('r1',), self.get_parents((('r0',),))),
+                (('r2',), self.get_parents(())),
+                (('r3',), self.get_parents(())),
+                (('m',), self.get_parents((('r0',),('r1',),('r2',),('r3',)))),
+                ]
+        else:
+            parent_details = [
+                (('FileA', 'r0'), self.get_parents(())),
+                (('FileA', 'r1'), self.get_parents((('FileA', 'r0'),))),
+                (('FileA', 'r2'), self.get_parents(())),
+                (('FileA', 'r3'), self.get_parents(())),
+                (('FileA', 'm'), self.get_parents((('FileA', 'r0'),
+                    ('FileA', 'r1'), ('FileA', 'r2'), ('FileA', 'r3')))),
+                ]
+        for key, parents in parent_details:
+            files.add_lines(key, parents, [])
+            # immediately after adding it should be queryable.
+            self.assertEqual({key:parents}, files.get_parent_map([key]))
+        # We can ask for an empty set
+        self.assertEqual({}, files.get_parent_map([]))
+        # We can ask for many keys
+        all_parents = dict(parent_details)
+        self.assertEqual(all_parents, files.get_parent_map(all_parents.keys()))
+        # Absent keys are just not included in the result.
+        keys = all_parents.keys()
+        if self.key_length == 1:
+            keys.insert(1, ('missing',))
+        else:
+            keys.insert(1, ('missing', 'missing'))
+        # Absent keys are just ignored
+        self.assertEqual(all_parents, files.get_parent_map(keys))
+
+    def test_get_sha1s(self):
+        files = self.get_versionedfiles()
+        self.get_diamond_files(files)
+        if self.key_length == 1:
+            keys = [('base',), ('origin',), ('left',), ('merged',), ('right',)]
+        else:
+            # ask for shas from different prefixes.
+            keys = [
+                ('FileA', 'base'), ('FileB', 'origin'), ('FileA', 'left'),
+                ('FileA', 'merged'), ('FileB', 'right'),
+                ]
+        self.assertEqual([
+            '51c64a6f4fc375daf0d24aafbabe4d91b6f4bb44',
+            '00e364d235126be43292ab09cb4686cf703ddc17',
+            'a8478686da38e370e32e42e8a0c220e33ee9132f',
+            'ed8bce375198ea62444dc71952b22cfc2b09226d',
+            '9ef09dfa9d86780bdec9219a22560c6ece8e0ef1',
+            ],
+            files.get_sha1s(keys))
+        
+    def test_insert_record_stream_empty(self):
+        """Inserting an empty record stream should work."""
+        files = self.get_versionedfiles()
+        files.insert_record_stream([])
+
+    def assertIdenticalVersionedFile(self, expected, actual):
+        """Assert that left and right have the same contents."""
+        self.assertEqual(set(actual.keys()), set(expected.keys()))
+        actual_parents = actual.get_parent_map(actual.keys())
+        if self.graph:
+            self.assertEqual(actual_parents, expected.get_parent_map(expected.keys()))
+        else:
+            for key, parents in actual_parents.items():
+                self.assertEqual(None, parents)
+        for key in actual.keys():
+            actual_text = actual.get_record_stream(
+                [key], 'unordered', True).next().get_bytes_as('fulltext')
+            expected_text = expected.get_record_stream(
+                [key], 'unordered', True).next().get_bytes_as('fulltext')
+            self.assertEqual(actual_text, expected_text)
+
+    def test_insert_record_stream_fulltexts(self):
+        """Any file should accept a stream of fulltexts."""
+        files = self.get_versionedfiles()
+        mapper = self.get_mapper()
+        source_transport = self.get_transport('source')
+        source_transport.mkdir('.')
+        # weaves always output fulltexts.
+        source = make_versioned_files_factory(WeaveFile, mapper)(
+            source_transport)
+        self.get_diamond_files(source, trailing_eol=False)
+        stream = source.get_record_stream(source.keys(), 'topological',
+            False)
+        files.insert_record_stream(stream)
+        self.assertIdenticalVersionedFile(source, files)
+
+    def test_insert_record_stream_fulltexts_noeol(self):
+        """Any file should accept a stream of fulltexts."""
+        files = self.get_versionedfiles()
+        mapper = self.get_mapper()
+        source_transport = self.get_transport('source')
+        source_transport.mkdir('.')
+        # weaves always output fulltexts.
+        source = make_versioned_files_factory(WeaveFile, mapper)(
+            source_transport)
+        self.get_diamond_files(source, trailing_eol=False)
+        stream = source.get_record_stream(source.keys(), 'topological',
+            False)
+        files.insert_record_stream(stream)
+        self.assertIdenticalVersionedFile(source, files)
+
+    def test_insert_record_stream_annotated_knits(self):
+        """Any file should accept a stream from plain knits."""
+        files = self.get_versionedfiles()
+        mapper = self.get_mapper()
+        source_transport = self.get_transport('source')
+        source_transport.mkdir('.')
+        source = make_file_factory(True, mapper)(source_transport)
+        self.get_diamond_files(source)
+        stream = source.get_record_stream(source.keys(), 'topological',
+            False)
+        files.insert_record_stream(stream)
+        self.assertIdenticalVersionedFile(source, files)
+
+    def test_insert_record_stream_annotated_knits_noeol(self):
+        """Any file should accept a stream from plain knits."""
+        files = self.get_versionedfiles()
+        mapper = self.get_mapper()
+        source_transport = self.get_transport('source')
+        source_transport.mkdir('.')
+        source = make_file_factory(True, mapper)(source_transport)
+        self.get_diamond_files(source, trailing_eol=False)
+        stream = source.get_record_stream(source.keys(), 'topological',
+            False)
+        files.insert_record_stream(stream)
+        self.assertIdenticalVersionedFile(source, files)
+
+    def test_insert_record_stream_plain_knits(self):
+        """Any file should accept a stream from plain knits."""
+        files = self.get_versionedfiles()
+        mapper = self.get_mapper()
+        source_transport = self.get_transport('source')
+        source_transport.mkdir('.')
+        source = make_file_factory(False, mapper)(source_transport)
+        self.get_diamond_files(source)
+        stream = source.get_record_stream(source.keys(), 'topological',
+            False)
+        files.insert_record_stream(stream)
+        self.assertIdenticalVersionedFile(source, files)
+
+    def test_insert_record_stream_plain_knits_noeol(self):
+        """Any file should accept a stream from plain knits."""
+        files = self.get_versionedfiles()
+        mapper = self.get_mapper()
+        source_transport = self.get_transport('source')
+        source_transport.mkdir('.')
+        source = make_file_factory(False, mapper)(source_transport)
+        self.get_diamond_files(source, trailing_eol=False)
+        stream = source.get_record_stream(source.keys(), 'topological',
+            False)
+        files.insert_record_stream(stream)
+        self.assertIdenticalVersionedFile(source, files)
+
+    def test_insert_record_stream_existing_keys(self):
+        """Inserting keys already in a file should not error."""
+        files = self.get_versionedfiles()
+        source = self.get_versionedfiles('source')
+        self.get_diamond_files(source)
+        # insert some keys into f.
+        self.get_diamond_files(files, left_only=True)
+        stream = source.get_record_stream(source.keys(), 'topological',
+            False)
+        files.insert_record_stream(stream)
+        self.assertIdenticalVersionedFile(source, files)
+
+    def test_insert_record_stream_missing_keys(self):
+        """Inserting a stream with absent keys should raise an error."""
+        files = self.get_versionedfiles()
+        source = self.get_versionedfiles('source')
+        stream = source.get_record_stream([('missing',) * self.key_length],
+            'topological', False)
+        self.assertRaises(errors.RevisionNotPresent, files.insert_record_stream,
+            stream)
+
+    def test_insert_record_stream_out_of_order(self):
+        """An out of order stream can either error or work."""
+        files = self.get_versionedfiles()
+        source = self.get_versionedfiles('source')
+        self.get_diamond_files(source)
+        if self.key_length == 1:
+            origin_keys = [('origin',)]
+            end_keys = [('merged',), ('left',)]
+            start_keys = [('right',), ('base',)]
+        else:
+            origin_keys = [('FileA', 'origin'), ('FileB', 'origin')]
+            end_keys = [('FileA', 'merged',), ('FileA', 'left',),
+                ('FileB', 'merged',), ('FileB', 'left',)]
+            start_keys = [('FileA', 'right',), ('FileA', 'base',),
+                ('FileB', 'right',), ('FileB', 'base',)]
+        origin_entries = source.get_record_stream(origin_keys, 'unordered', False)
+        end_entries = source.get_record_stream(end_keys, 'topological', False)
+        start_entries = source.get_record_stream(start_keys, 'topological', False)
+        entries = chain(origin_entries, end_entries, start_entries)
+        try:
+            files.insert_record_stream(entries)
+        except RevisionNotPresent:
+            # Must not have corrupted the file.
+            files.check()
+        else:
+            self.assertIdenticalVersionedFile(source, files)
+
+    def test_insert_record_stream_delta_missing_basis_no_corruption(self):
+        """Insertion where a needed basis is not included aborts safely."""
+        # We use a knit always here to be sure we are getting a binary delta.
+        mapper = self.get_mapper()
+        source_transport = self.get_transport('source')
+        source_transport.mkdir('.')
+        source = make_file_factory(False, mapper)(source_transport)
+        self.get_diamond_files(source)
+        entries = source.get_record_stream(['origin', 'merged'], 'unordered', False)
+        files = self.get_versionedfiles()
+        self.assertRaises(RevisionNotPresent, files.insert_record_stream,
+            entries)
+        files.check()
+        self.assertEqual({}, files.get_parent_map([]))
+
+    def test_iter_lines_added_or_present_in_keys(self):
+        # test that we get at least an equalset of the lines added by
+        # versions in the store.
+        # the ordering here is to make a tree so that dumb searches have
+        # more changes to muck up.
+
+        class InstrumentedProgress(progress.DummyProgress):
+
+            def __init__(self):
+
+                progress.DummyProgress.__init__(self)
+                self.updates = []
+
+            def update(self, msg=None, current=None, total=None):
+                self.updates.append((msg, current, total))
+
+        files = self.get_versionedfiles()
+        # add a base to get included
+        files.add_lines(self.get_simple_key('base'), (), ['base\n'])
+        # add a ancestor to be included on one side
+        files.add_lines(self.get_simple_key('lancestor'), (), ['lancestor\n'])
+        # add a ancestor to be included on the other side
+        files.add_lines(self.get_simple_key('rancestor'),
+            self.get_parents([self.get_simple_key('base')]), ['rancestor\n'])
+        # add a child of rancestor with no eofile-nl
+        files.add_lines(self.get_simple_key('child'),
+            self.get_parents([self.get_simple_key('rancestor')]),
+            ['base\n', 'child\n'])
+        # add a child of lancestor and base to join the two roots
+        files.add_lines(self.get_simple_key('otherchild'),
+            self.get_parents([self.get_simple_key('lancestor'),
+                self.get_simple_key('base')]),
+            ['base\n', 'lancestor\n', 'otherchild\n'])
+        def iter_with_keys(keys, expected):
+            # now we need to see what lines are returned, and how often.
+            lines = {}
+            progress = InstrumentedProgress()
+            # iterate over the lines
+            for line in files.iter_lines_added_or_present_in_keys(keys,
+                pb=progress):
+                lines.setdefault(line, 0)
+                lines[line] += 1
+            if []!= progress.updates:
+                self.assertEqual(expected, progress.updates)
+            return lines
+        lines = iter_with_keys(
+            [self.get_simple_key('child'), self.get_simple_key('otherchild')],
+            [('Walking content.', 0, 2),
+             ('Walking content.', 1, 2),
+             ('Walking content.', 2, 2)])
+        # we must see child and otherchild
+        self.assertTrue(lines[('child\n', self.get_simple_key('child'))] > 0)
+        self.assertTrue(
+            lines[('otherchild\n', self.get_simple_key('otherchild'))] > 0)
+        # we dont care if we got more than that.
+        
+        # test all lines
+        lines = iter_with_keys(files.keys(),
+            [('Walking content.', 0, 5),
+             ('Walking content.', 1, 5),
+             ('Walking content.', 2, 5),
+             ('Walking content.', 3, 5),
+             ('Walking content.', 4, 5),
+             ('Walking content.', 5, 5)])
+        # all lines must be seen at least once
+        self.assertTrue(lines[('base\n', self.get_simple_key('base'))] > 0)
+        self.assertTrue(
+            lines[('lancestor\n', self.get_simple_key('lancestor'))] > 0)
+        self.assertTrue(
+            lines[('rancestor\n', self.get_simple_key('rancestor'))] > 0)
+        self.assertTrue(lines[('child\n', self.get_simple_key('child'))] > 0)
+        self.assertTrue(
+            lines[('otherchild\n', self.get_simple_key('otherchild'))] > 0)
+
+    def test_make_mpdiffs(self):
+        from bzrlib import multiparent
+        files = self.get_versionedfiles('source')
+        # add texts that should trip the knit maximum delta chain threshold
+        # as well as doing parallel chains of data in knits.
+        # this is done by two chains of 25 insertions
+        files.add_lines(self.get_simple_key('base'), [], ['line\n'])
+        files.add_lines(self.get_simple_key('noeol'),
+            self.get_parents([self.get_simple_key('base')]), ['line'])
+        # detailed eol tests:
+        # shared last line with parent no-eol
+        files.add_lines(self.get_simple_key('noeolsecond'),
+            self.get_parents([self.get_simple_key('noeol')]),
+                ['line\n', 'line'])
+        # differing last line with parent, both no-eol
+        files.add_lines(self.get_simple_key('noeolnotshared'),
+            self.get_parents([self.get_simple_key('noeolsecond')]),
+                ['line\n', 'phone'])
+        # add eol following a noneol parent, change content
+        files.add_lines(self.get_simple_key('eol'),
+            self.get_parents([self.get_simple_key('noeol')]), ['phone\n'])
+        # add eol following a noneol parent, no change content
+        files.add_lines(self.get_simple_key('eolline'),
+            self.get_parents([self.get_simple_key('noeol')]), ['line\n'])
+        # noeol with no parents:
+        files.add_lines(self.get_simple_key('noeolbase'), [], ['line'])
+        # noeol preceeding its leftmost parent in the output:
+        # this is done by making it a merge of two parents with no common
+        # anestry: noeolbase and noeol with the 
+        # later-inserted parent the leftmost.
+        files.add_lines(self.get_simple_key('eolbeforefirstparent'),
+            self.get_parents([self.get_simple_key('noeolbase'),
+                self.get_simple_key('noeol')]),
+            ['line'])
+        # two identical eol texts
+        files.add_lines(self.get_simple_key('noeoldup'),
+            self.get_parents([self.get_simple_key('noeol')]), ['line'])
+        next_parent = self.get_simple_key('base')
+        text_name = 'chain1-'
+        text = ['line\n']
+        sha1s = {0 :'da6d3141cb4a5e6f464bf6e0518042ddc7bfd079',
+                 1 :'45e21ea146a81ea44a821737acdb4f9791c8abe7',
+                 2 :'e1f11570edf3e2a070052366c582837a4fe4e9fa',
+                 3 :'26b4b8626da827088c514b8f9bbe4ebf181edda1',
+                 4 :'e28a5510be25ba84d31121cff00956f9970ae6f6',
+                 5 :'d63ec0ce22e11dcf65a931b69255d3ac747a318d',
+                 6 :'2c2888d288cb5e1d98009d822fedfe6019c6a4ea',
+                 7 :'95c14da9cafbf828e3e74a6f016d87926ba234ab',
+                 8 :'779e9a0b28f9f832528d4b21e17e168c67697272',
+                 9 :'1f8ff4e5c6ff78ac106fcfe6b1e8cb8740ff9a8f',
+                 10:'131a2ae712cf51ed62f143e3fbac3d4206c25a05',
+                 11:'c5a9d6f520d2515e1ec401a8f8a67e6c3c89f199',
+                 12:'31a2286267f24d8bedaa43355f8ad7129509ea85',
+                 13:'dc2a7fe80e8ec5cae920973973a8ee28b2da5e0a',
+                 14:'2c4b1736566b8ca6051e668de68650686a3922f2',
+                 15:'5912e4ecd9b0c07be4d013e7e2bdcf9323276cde',
+                 16:'b0d2e18d3559a00580f6b49804c23fea500feab3',
+                 17:'8e1d43ad72f7562d7cb8f57ee584e20eb1a69fc7',
+                 18:'5cf64a3459ae28efa60239e44b20312d25b253f3',
+                 19:'1ebed371807ba5935958ad0884595126e8c4e823',
+                 20:'2aa62a8b06fb3b3b892a3292a068ade69d5ee0d3',
+                 21:'01edc447978004f6e4e962b417a4ae1955b6fe5d',
+                 22:'d8d8dc49c4bf0bab401e0298bb5ad827768618bb',
+                 23:'c21f62b1c482862983a8ffb2b0c64b3451876e3f',
+                 24:'c0593fe795e00dff6b3c0fe857a074364d5f04fc',
+                 25:'dd1a1cf2ba9cc225c3aff729953e6364bf1d1855',
+                 }
+        for depth in range(26):
+            new_version = self.get_simple_key(text_name + '%s' % depth)
+            text = text + ['line\n']
+            files.add_lines(new_version, self.get_parents([next_parent]), text)
+            next_parent = new_version
+        next_parent = self.get_simple_key('base')
+        text_name = 'chain2-'
+        text = ['line\n']
+        for depth in range(26):
+            new_version = self.get_simple_key(text_name + '%s' % depth)
+            text = text + ['line\n']
+            files.add_lines(new_version, self.get_parents([next_parent]), text)
+            next_parent = new_version
+        target = self.get_versionedfiles('target')
+        for key in multiparent.topo_iter_keys(files, files.keys()):
+            mpdiff = files.make_mpdiffs([key])[0]
+            parents = files.get_parent_map([key])[key] or []
+            target.add_mpdiffs(
+                [(key, parents, files.get_sha1s([key])[0], mpdiff)])
+            self.assertEqualDiff(
+                files.get_record_stream([key], 'unordered',
+                    True).next().get_bytes_as('fulltext'),
+                target.get_record_stream([key], 'unordered',
+                    True).next().get_bytes_as('fulltext')
+                )
+
+    def test_keys(self):
+        # While use is discouraged, versions() is still needed by aspects of
+        # bzr.
+        files = self.get_versionedfiles()
+        self.assertEqual(set(), set(files.keys()))
+        if self.key_length == 1:
+            key = ('foo',)
+        else:
+            key = ('foo', 'bar',)
+        files.add_lines(key, (), [])
+        self.assertEqual(set([key]), set(files.keys()))

=== modified file 'bzrlib/tests/tree_implementations/test_inv.py'
--- bzrlib/tests/tree_implementations/test_inv.py	2008-05-08 04:33:38 +0000
+++ bzrlib/tests/tree_implementations/test_inv.py	2008-06-11 04:20:16 +0000
@@ -64,8 +64,6 @@
         self.tree.lock_read()
         self.addCleanup(self.tree.unlock)
         self.file_active = get_entry(self.tree, 'fileid')
-        self.weave = self.branch.repository.weave_store.get_weave('fileid',
-            self.branch.repository.get_transaction())
 
     # TODO: test two inventories with the same file revision
 

=== modified file 'bzrlib/tree.py'
--- bzrlib/tree.py	2008-05-08 04:33:38 +0000
+++ bzrlib/tree.py	2008-06-11 04:20:16 +0000
@@ -358,6 +358,7 @@
                                  last_revision_base)
 
     def _get_file_revision(self, file_id, vf, tree_revision):
+        """Ensure that file_id, tree_revision is in vf to plan the merge."""
         def file_revision(revision_tree):
             revision_tree.lock_read()
             try:
@@ -372,18 +373,19 @@
                 except:
                     yield self.repository.revision_tree(revision_id)
 
-        if getattr(self, '_get_weave', None) is None:
+        if getattr(self, '_repository', None) is None:
             last_revision = tree_revision
-            parent_revisions = [file_revision(t) for t in iter_parent_trees()]
-            vf.add_lines(last_revision, parent_revisions,
+            parent_keys = [(file_id, file_revision(t)) for t in
+                iter_parent_trees()]
+            vf.add_lines((file_id, last_revision), parent_keys,
                          self.get_file(file_id).readlines())
             repo = self.branch.repository
-            transaction = repo.get_transaction()
-            base_vf = repo.weave_store.get_weave(file_id, transaction)
+            base_vf = repo.texts
         else:
             last_revision = file_revision(self)
-            base_vf = self._get_weave(file_id)
-        vf.fallback_versionedfiles.append(base_vf)
+            base_vf = self._repository.texts
+        if base_vf not in vf.fallback_versionedfiles:
+            vf.fallback_versionedfiles.append(base_vf)
         return last_revision
 
     inventory = property(_get_inventory,

=== modified file 'bzrlib/versionedfile.py'
--- bzrlib/versionedfile.py	2008-05-12 04:37:12 +0000
+++ bzrlib/versionedfile.py	2008-06-11 04:20:16 +0000
@@ -20,6 +20,7 @@
 """Versioned text file storage api."""
 
 from cStringIO import StringIO
+import os
 import urllib
 from zlib import adler32
 
@@ -34,7 +35,7 @@
     revision,
     ui,
     )
-from bzrlib.graph import Graph
+from bzrlib.graph import DictParentsProvider, Graph, _StackedParentsProvider
 from bzrlib.transport.memory import MemoryTransport
 """)
 from bzrlib.inter import InterObject
@@ -71,7 +72,7 @@
     :ivar parents: A tuple of parent keys for self.key. If the object has
         no parent information, None (as opposed to () for an empty list of
         parents).
-        """
+    """
 
     def __init__(self):
         """Create a ContentFactory."""
@@ -81,7 +82,38 @@
         self.parents = None
 
 
-class AbsentContentFactory(object):
+class FulltextContentFactory(ContentFactory):
+    """Static data content factory.
+
+    This takes a fulltext when created and just returns that during
+    get_bytes_as('fulltext').
+    
+    :ivar sha1: None, or the sha1 of the content fulltext.
+    :ivar storage_kind: The native storage kind of this factory. Always
+        'fulltext'.
+    :ivar key: The key of this content. Each key is a tuple with a single
+        string in it.
+    :ivar parents: A tuple of parent keys for self.key. If the object has
+        no parent information, None (as opposed to () for an empty list of
+        parents).
+     """
+
+    def __init__(self, key, parents, sha1, text):
+        """Create a ContentFactory."""
+        self.sha1 = sha1
+        self.storage_kind = 'fulltext'
+        self.key = key
+        self.parents = parents
+        self._text = text
+
+    def get_bytes_as(self, storage_kind):
+        if storage_kind == self.storage_kind:
+            return self._text
+        raise errors.UnavailableRepresentation(self.key, storage_kind,
+            self.storage_kind)
+
+
+class AbsentContentFactory(ContentFactory):
     """A placeholder content factory for unavailable texts.
     
     :ivar sha1: None.
@@ -99,6 +131,23 @@
         self.parents = None
 
 
+class AdapterFactory(ContentFactory):
+    """A content factory to adapt between key prefix's."""
+
+    def __init__(self, key, parents, adapted):
+        """Create an adapter factory instance."""
+        self.key = key
+        self.parents = parents
+        self._adapted = adapted
+
+    def __getattr__(self, attr):
+        """Return a member from the adapted object."""
+        if attr in ('key', 'parents'):
+            return self.__dict__[attr]
+        else:
+            return getattr(self._adapted, attr)
+
+
 def filter_absent(record_stream):
     """Adapt a record stream to remove absent records."""
     for record in record_stream:
@@ -415,25 +464,6 @@
         """
         raise NotImplementedError(self.annotate)
 
-    @deprecated_method(one_five)
-    def join(self, other, pb=None, msg=None, version_ids=None,
-             ignore_missing=False):
-        """Integrate versions from other into this versioned file.
-
-        If version_ids is None all versions from other should be
-        incorporated into this versioned file.
-
-        Must raise RevisionNotPresent if any of the specified versions
-        are not present in the other file's history unless ignore_missing
-        is supplied in which case they are silently skipped.
-        """
-        self._check_write_ok()
-        return InterVersionedFile.get(other, self).join(
-            pb,
-            msg,
-            version_ids,
-            ignore_missing)
-
     def iter_lines_added_or_present_in_versions(self, version_ids=None,
                                                 pb=None):
         """Iterate over the lines in the versioned file from version_ids.
@@ -483,8 +513,8 @@
         return PlanWeaveMerge(plan, a_marker, b_marker).merge_lines()[0]
 
 
-class RecordingVersionedFileDecorator(object):
-    """A minimal versioned file that records calls made on it.
+class RecordingVersionedFilesDecorator(object):
+    """A minimal versioned files that records calls made on it.
     
     Only enough methods have been added to support tests using it to date.
 
@@ -493,145 +523,678 @@
     """
 
     def __init__(self, backing_vf):
-        """Create a RecordingVersionedFileDecorator decorating backing_vf.
+        """Create a RecordingVersionedFileDsecorator decorating backing_vf.
         
         :param backing_vf: The versioned file to answer all methods.
         """
         self._backing_vf = backing_vf
         self.calls = []
 
-    def get_lines(self, version_ids):
-        self.calls.append(("get_lines", version_ids))
-        return self._backing_vf.get_lines(version_ids)
-
-
-class _PlanMergeVersionedFile(object):
+    def get_record_stream(self, keys, sort_order, include_delta_closure):
+        self.calls.append(("get_record_stream", keys, sort_order,
+            include_delta_closure))
+        return self._backing_vf.get_record_stream(keys, sort_order,
+            include_delta_closure)
+
+
+class KeyMapper(object):
+    """KeyMappers map between keys and underlying paritioned storage."""
+
+    def map(self, key):
+        """Map key to an underlying storage identifier.
+
+        :param key: A key tuple e.g. ('file-id', 'revision-id').
+        :return: An underlying storage identifier, specific to the partitioning
+            mechanism.
+        """
+        raise NotImplementedError(self.map)
+
+    def unmap(self, partition_id):
+        """Map a partitioned storage id back to a key prefix.
+        
+        :param partition_id: The underlying partition id.
+        :return: As much of a key (or prefix) as is derivable from the parition
+            id.
+        """
+        raise NotImplementedError(self.unmap)
+
+
+class ConstantMapper(KeyMapper):
+    """A key mapper that maps to a constant result."""
+
+    def __init__(self, result):
+        """Create a ConstantMapper which will return result for all maps."""
+        self._result = result
+
+    def map(self, key):
+        """See KeyMapper.map()."""
+        return self._result
+
+
+class URLEscapeMapper(KeyMapper):
+    """Base class for use with transport backed storage.
+
+    This provides a map and unmap wrapper that respectively url escape and
+    unescape their outputs and inputs.
+    """
+
+    def map(self, key):
+        """See KeyMapper.map()."""
+        return urllib.quote(self._map(key))
+
+    def unmap(self, partition_id):
+        """See KeyMapper.unmap()."""
+        return self._unmap(urllib.unquote(partition_id))
+
+
+class PrefixMapper(URLEscapeMapper):
+    """A key mapper that extracts the first component of a key.
+    
+    This mapper is for use with a transport based backend.
+    """
+
+    def _map(self, key):
+        """See KeyMapper.map()."""
+        return key[0]
+
+    def _unmap(self, partition_id):
+        """See KeyMapper.unmap()."""
+        return (partition_id,)
+
+
+class HashPrefixMapper(URLEscapeMapper):
+    """A key mapper that combines the first component of a key with a hash.
+
+    This mapper is for use with a transport based backend.
+    """
+
+    def _map(self, key):
+        """See KeyMapper.map()."""
+        prefix = self._escape(key[0])
+        return "%02x/%s" % (adler32(prefix) & 0xff, prefix)
+
+    def _escape(self, prefix):
+        """No escaping needed here."""
+        return prefix
+
+    def _unmap(self, partition_id):
+        """See KeyMapper.unmap()."""
+        return (self._unescape(osutils.basename(partition_id)),)
+
+    def _unescape(self, basename):
+        """No unescaping needed for HashPrefixMapper."""
+        return basename
+
+
+class HashEscapedPrefixMapper(HashPrefixMapper):
+    """Combines the escaped first component of a key with a hash.
+    
+    This mapper is for use with a transport based backend.
+    """
+
+    _safe = "abcdefghijklmnopqrstuvwxyz0123456789-_@,."
+
+    def _escape(self, prefix):
+        """Turn a key element into a filesystem safe string.
+
+        This is similar to a plain urllib.quote, except
+        it uses specific safe characters, so that it doesn't
+        have to translate a lot of valid file ids.
+        """
+        # @ does not get escaped. This is because it is a valid
+        # filesystem character we use all the time, and it looks
+        # a lot better than seeing %40 all the time.
+        r = [((c in self._safe) and c or ('%%%02x' % ord(c)))
+             for c in prefix]
+        return ''.join(r)
+
+    def _unescape(self, basename):
+        """Escaped names are easily unescaped by urlutils."""
+        return urllib.unquote(basename)
+
+
+def make_versioned_files_factory(versioned_file_factory, mapper):
+    """Create a ThunkedVersionedFiles factory.
+
+    This will create a callable which when called creates a
+    ThunkedVersionedFiles on a transport, using mapper to access individual
+    versioned files, and versioned_file_factory to create each individual file.
+    """
+    def factory(transport):
+        return ThunkedVersionedFiles(transport, versioned_file_factory, mapper,
+            lambda:True)
+    return factory
+
+
+class VersionedFiles(object):
+    """Storage for many versioned files.
+
+    This object allows a single keyspace for accessing the history graph and
+    contents of named bytestrings.
+
+    Currently no implementation allows the graph of different key prefixes to
+    intersect, but the API does allow such implementations in the future.
+    """
+
+    def add_lines(self, key, parents, lines, parent_texts=None,
+        left_matching_blocks=None, nostore_sha=None, random_id=False,
+        check_content=True):
+        """Add a text to the store.
+
+        :param key: The key tuple of the text to add.
+        :param parents: The parents key tuples of the text to add.
+        :param lines: A list of lines. Each line must be a bytestring. And all
+            of them except the last must be terminated with \n and contain no
+            other \n's. The last line may either contain no \n's or a single
+            terminating \n. If the lines list does meet this constraint the add
+            routine may error or may succeed - but you will be unable to read
+            the data back accurately. (Checking the lines have been split
+            correctly is expensive and extremely unlikely to catch bugs so it
+            is not done at runtime unless check_content is True.)
+        :param parent_texts: An optional dictionary containing the opaque 
+            representations of some or all of the parents of version_id to
+            allow delta optimisations.  VERY IMPORTANT: the texts must be those
+            returned by add_lines or data corruption can be caused.
+        :param left_matching_blocks: a hint about which areas are common
+            between the text and its left-hand-parent.  The format is
+            the SequenceMatcher.get_matching_blocks format.
+        :param nostore_sha: Raise ExistingContent and do not add the lines to
+            the versioned file if the digest of the lines matches this.
+        :param random_id: If True a random id has been selected rather than
+            an id determined by some deterministic process such as a converter
+            from a foreign VCS. When True the backend may choose not to check
+            for uniqueness of the resulting key within the versioned file, so
+            this should only be done when the result is expected to be unique
+            anyway.
+        :param check_content: If True, the lines supplied are verified to be
+            bytestrings that are correctly formed lines.
+        :return: The text sha1, the number of bytes in the text, and an opaque
+                 representation of the inserted version which can be provided
+                 back to future add_lines calls in the parent_texts dictionary.
+        """
+        raise NotImplementedError(self.add_lines)
+
+    def add_mpdiffs(self, records):
+        """Add mpdiffs to this VersionedFile.
+
+        Records should be iterables of version, parents, expected_sha1,
+        mpdiff. mpdiff should be a MultiParent instance.
+        """
+        vf_parents = {}
+        mpvf = multiparent.MultiMemoryVersionedFile()
+        versions = []
+        for version, parent_ids, expected_sha1, mpdiff in records:
+            versions.append(version)
+            mpvf.add_diff(mpdiff, version, parent_ids)
+        needed_parents = set()
+        for version, parent_ids, expected_sha1, mpdiff in records:
+            needed_parents.update(p for p in parent_ids
+                                  if not mpvf.has_version(p))
+        # It seems likely that adding all the present parents as fulltexts can
+        # easily exhaust memory.
+        present_parents = set(self.get_parent_map(needed_parents).keys())
+        split_lines = osutils.split_lines
+        for record in self.get_record_stream(present_parents, 'unordered',
+            True):
+            mpvf.add_version(split_lines(record.get_bytes_as('fulltext')),
+                record.key, [])
+        for (key, parent_keys, expected_sha1, mpdiff), lines in\
+            zip(records, mpvf.get_line_list(versions)):
+            if len(parent_keys) == 1:
+                left_matching_blocks = list(mpdiff.get_matching_blocks(0,
+                    mpvf.get_diff(parent_keys[0]).num_lines()))
+            else:
+                left_matching_blocks = None
+            version_sha1, _, version_text = self.add_lines(key,
+                parent_keys, lines, vf_parents,
+                left_matching_blocks=left_matching_blocks)
+            if version_sha1 != expected_sha1:
+                raise errors.VersionedFileInvalidChecksum(version)
+            vf_parents[key] = version_text
+
+    def annotate(self, key):
+        """Return a list of (version-key, line) tuples for the text of key.
+
+        :raise RevisionNotPresent: If the key is not present.
+        """
+        raise NotImplementedError(self.annotate)
+
+    def check(self, progress_bar=None):
+        """Check this object for integrity."""
+        raise NotImplementedError(self.check)
+
+    @staticmethod
+    def check_not_reserved_id(version_id):
+        revision.check_not_reserved_id(version_id)
+
+    def _check_lines_not_unicode(self, lines):
+        """Check that lines being added to a versioned file are not unicode."""
+        for line in lines:
+            if line.__class__ is not str:
+                raise errors.BzrBadParameterUnicode("lines")
+
+    def _check_lines_are_lines(self, lines):
+        """Check that the lines really are full lines without inline EOL."""
+        for line in lines:
+            if '\n' in line[:-1]:
+                raise errors.BzrBadParameterContainsNewline("lines")
+
+    def get_parent_map(self, keys):
+        """Get a map of the parents of keys.
+
+        :param keys: The keys to look up parents for.
+        :return: A mapping from keys to parents. Absent keys are absent from
+            the mapping.
+        """
+        raise NotImplementedError(self.get_parent_map)
+
+    def get_record_stream(self, keys, ordering, include_delta_closure):
+        """Get a stream of records for keys.
+
+        :param keys: The keys to include.
+        :param ordering: Either 'unordered' or 'topological'. A topologically
+            sorted stream has compression parents strictly before their
+            children.
+        :param include_delta_closure: If True then the closure across any
+            compression parents will be included (in the opaque data).
+        :return: An iterator of ContentFactory objects, each of which is only
+            valid until the iterator is advanced.
+        """
+        raise NotImplementedError(self.get_record_stream)
+
+    def get_sha1s(self, keys):
+        """Get the sha1's of the texts for the given keys.
+
+        :param keys: The names of the keys to lookup
+        :return: a list of sha1s matching keys.
+        """
+        raise NotImplementedError(self.get_sha1s)
+
+    def insert_record_stream(self, stream):
+        """Insert a record stream into this container.
+
+        :param stream: A stream of records to insert. 
+        :return: None
+        :seealso VersionedFile.get_record_stream:
+        """
+        raise NotImplementedError
+
+    def iter_lines_added_or_present_in_keys(self, keys, pb=None):
+        """Iterate over the lines in the versioned files from keys.
+
+        This may return lines from other keys. Each item the returned
+        iterator yields is a tuple of a line and a text version that that line
+        is present in (not introduced in).
+
+        Ordering of results is in whatever order is most suitable for the
+        underlying storage format.
+
+        If a progress bar is supplied, it may be used to indicate progress.
+        The caller is responsible for cleaning up progress bars (because this
+        is an iterator).
+
+        NOTES:
+         * Lines are normalised by the underlying store: they will all have \n
+           terminators.
+         * Lines are returned in arbitrary order.
+
+        :return: An iterator over (line, key).
+        """
+        raise NotImplementedError(self.iter_lines_added_or_present_in_keys)
+
+    def keys(self):
+        """Return a iterable of the keys for all the contained texts."""
+        raise NotImplementedError(self.keys)
+
+    def make_mpdiffs(self, keys):
+        """Create multiparent diffs for specified keys."""
+        keys_order = tuple(keys)
+        keys = frozenset(keys)
+        knit_keys = set(keys)
+        parent_map = self.get_parent_map(keys)
+        for parent_keys in parent_map.itervalues():
+            if parent_keys:
+                knit_keys.update(parent_keys)
+        missing_keys = keys - set(parent_map)
+        if missing_keys:
+            raise errors.RevisionNotPresent(missing_keys.pop(), self)
+        # We need to filter out ghosts, because we can't diff against them.
+        maybe_ghosts = knit_keys - keys
+        ghosts = maybe_ghosts - set(self.get_parent_map(maybe_ghosts))
+        knit_keys.difference_update(ghosts)
+        lines = {}
+        split_lines = osutils.split_lines
+        for record in self.get_record_stream(knit_keys, 'topological', True):
+            lines[record.key] = split_lines(record.get_bytes_as('fulltext'))
+            # line_block_dict = {}
+            # for parent, blocks in record.extract_line_blocks():
+            #   line_blocks[parent] = blocks
+            # line_blocks[record.key] = line_block_dict
+        diffs = []
+        for key in keys_order:
+            target = lines[key]
+            parents = parent_map[key] or []
+            # Note that filtering knit_keys can lead to a parent difference
+            # between the creation and the application of the mpdiff.
+            parent_lines = [lines[p] for p in parents if p in knit_keys]
+            if len(parent_lines) > 0:
+                left_parent_blocks = self._extract_blocks(key, parent_lines[0],
+                    target)
+            else:
+                left_parent_blocks = None
+            diffs.append(multiparent.MultiParent.from_lines(target,
+                parent_lines, left_parent_blocks))
+        return diffs
+
+    def _extract_blocks(self, version_id, source, target):
+        return None
+
+
+class ThunkedVersionedFiles(VersionedFiles):
+    """Storage for many versioned files thunked onto a 'VersionedFile' class.
+
+    This object allows a single keyspace for accessing the history graph and
+    contents of named bytestrings.
+
+    Currently no implementation allows the graph of different key prefixes to
+    intersect, but the API does allow such implementations in the future.
+    """
+
+    def __init__(self, transport, file_factory, mapper, is_locked):
+        """Create a ThunkedVersionedFiles."""
+        self._transport = transport
+        self._file_factory = file_factory
+        self._mapper = mapper
+        self._is_locked = is_locked
+
+    def add_lines(self, key, parents, lines, parent_texts=None,
+        left_matching_blocks=None, nostore_sha=None, random_id=False,
+        check_content=True):
+        """See VersionedFiles.add_lines()."""
+        path = self._mapper.map(key)
+        version_id = key[-1]
+        parents = [parent[-1] for parent in parents]
+        vf = self._get_vf(path)
+        try:
+            try:
+                return vf.add_lines_with_ghosts(version_id, parents, lines,
+                    parent_texts=parent_texts,
+                    left_matching_blocks=left_matching_blocks,
+                    nostore_sha=nostore_sha, random_id=random_id,
+                    check_content=check_content)
+            except NotImplementedError:
+                return vf.add_lines(version_id, parents, lines,
+                    parent_texts=parent_texts,
+                    left_matching_blocks=left_matching_blocks,
+                    nostore_sha=nostore_sha, random_id=random_id,
+                    check_content=check_content)
+        except errors.NoSuchFile:
+            # parent directory may be missing, try again.
+            self._transport.mkdir(osutils.dirname(path))
+            try:
+                return vf.add_lines_with_ghosts(version_id, parents, lines,
+                    parent_texts=parent_texts,
+                    left_matching_blocks=left_matching_blocks,
+                    nostore_sha=nostore_sha, random_id=random_id,
+                    check_content=check_content)
+            except NotImplementedError:
+                return vf.add_lines(version_id, parents, lines,
+                    parent_texts=parent_texts,
+                    left_matching_blocks=left_matching_blocks,
+                    nostore_sha=nostore_sha, random_id=random_id,
+                    check_content=check_content)
+
+    def annotate(self, key):
+        """Return a list of (version-key, line) tuples for the text of key.
+
+        :raise RevisionNotPresent: If the key is not present.
+        """
+        prefix = key[:-1]
+        path = self._mapper.map(prefix)
+        vf = self._get_vf(path)
+        origins = vf.annotate(key[-1])
+        result = []
+        for origin, line in origins:
+            result.append((prefix + (origin,), line))
+        return result
+
+    def check(self, progress_bar=None):
+        """See VersionedFiles.check()."""
+        for prefix, vf in self._iter_all_components():
+            vf.check()
+
+    def get_parent_map(self, keys):
+        """Get a map of the parents of keys.
+
+        :param keys: The keys to look up parents for.
+        :return: A mapping from keys to parents. Absent keys are absent from
+            the mapping.
+        """
+        prefixes = self._partition_keys(keys)
+        result = {}
+        for prefix, suffixes in prefixes.items():
+            path = self._mapper.map(prefix)
+            vf = self._get_vf(path)
+            parent_map = vf.get_parent_map(suffixes)
+            for key, parents in parent_map.items():
+                result[prefix + (key,)] = tuple(
+                    prefix + (parent,) for parent in parents)
+        return result
+
+    def _get_vf(self, path):
+        if not self._is_locked():
+            raise errors.ObjectNotLocked(self)
+        return self._file_factory(path, self._transport, create=True,
+            get_scope=lambda:None)
+
+    def _partition_keys(self, keys):
+        """Turn keys into a dict of prefix:suffix_list."""
+        result = {}
+        for key in keys:
+            prefix_keys = result.setdefault(key[:-1], [])
+            prefix_keys.append(key[-1])
+        return result
+
+    def _get_all_prefixes(self):
+        # Identify all key prefixes.
+        # XXX: A bit hacky, needs polish.
+        if type(self._mapper) == ConstantMapper:
+            paths = [self._mapper.map(())]
+            prefixes = [()]
+        else:
+            relpaths = set()
+            for quoted_relpath in self._transport.iter_files_recursive():
+                path, ext = os.path.splitext(quoted_relpath)
+                relpaths.add(path)
+            paths = list(relpaths)
+            prefixes = [self._mapper.unmap(path) for path in paths]
+        return zip(paths, prefixes)
+
+    def get_record_stream(self, keys, ordering, include_delta_closure):
+        """See VersionedFiles.get_record_stream()."""
+        # Ordering will be taken care of by each partitioned store; group keys
+        # by partition.
+        keys = sorted(keys)
+        for prefix, suffixes, vf in self._iter_keys_vf(keys):
+            suffixes = [(suffix,) for suffix in suffixes]
+            for record in vf.get_record_stream(suffixes, ordering,
+                include_delta_closure):
+                if record.parents is not None:
+                    record.parents = tuple(
+                        prefix + parent for parent in record.parents)
+                record.key = prefix + record.key
+                yield record
+
+    def _iter_keys_vf(self, keys):
+        prefixes = self._partition_keys(keys)
+        sha1s = {}
+        for prefix, suffixes in prefixes.items():
+            path = self._mapper.map(prefix)
+            vf = self._get_vf(path)
+            yield prefix, suffixes, vf
+
+    def get_sha1s(self, keys):
+        """See VersionedFiles.get_sha1s()."""
+        sha1s = {}
+        for prefix,suffixes, vf in self._iter_keys_vf(keys):
+            vf_sha1s = vf.get_sha1s(suffixes)
+            for suffix, sha1 in zip(suffixes, vf.get_sha1s(suffixes)):
+                sha1s[prefix + (suffix,)] = sha1
+        return [sha1s[key] for key in keys]
+
+    def insert_record_stream(self, stream):
+        """Insert a record stream into this container.
+
+        :param stream: A stream of records to insert. 
+        :return: None
+        :seealso VersionedFile.get_record_stream:
+        """
+        for record in stream:
+            prefix = record.key[:-1]
+            key = record.key[-1:]
+            if record.parents is not None:
+                parents = [parent[-1:] for parent in record.parents]
+            else:
+                parents = None
+            thunk_record = AdapterFactory(key, parents, record)
+            path = self._mapper.map(prefix)
+            # Note that this parses the file many times; we can do better but
+            # as this only impacts weaves in terms of performance, it is
+            # tolerable.
+            vf = self._get_vf(path)
+            vf.insert_record_stream([thunk_record])
+
+    def iter_lines_added_or_present_in_keys(self, keys, pb=None):
+        """Iterate over the lines in the versioned files from keys.
+
+        This may return lines from other keys. Each item the returned
+        iterator yields is a tuple of a line and a text version that that line
+        is present in (not introduced in).
+
+        Ordering of results is in whatever order is most suitable for the
+        underlying storage format.
+
+        If a progress bar is supplied, it may be used to indicate progress.
+        The caller is responsible for cleaning up progress bars (because this
+        is an iterator).
+
+        NOTES:
+         * Lines are normalised by the underlying store: they will all have \n
+           terminators.
+         * Lines are returned in arbitrary order.
+
+        :return: An iterator over (line, key).
+        """
+        for prefix, suffixes, vf in self._iter_keys_vf(keys):
+            for line, version in vf.iter_lines_added_or_present_in_versions(suffixes):
+                yield line, prefix + (version,)
+
+    def _iter_all_components(self):
+        for path, prefix in self._get_all_prefixes():
+            yield prefix, self._get_vf(path)
+
+    def keys(self):
+        """See VersionedFiles.keys()."""
+        result = set()
+        for prefix, vf in self._iter_all_components():
+            for suffix in vf.versions():
+                result.add(prefix + (suffix,))
+        return result
+
+
+class _PlanMergeVersionedFile(VersionedFiles):
     """A VersionedFile for uncommitted and committed texts.
 
     It is intended to allow merges to be planned with working tree texts.
-    It implements only the small part of the VersionedFile interface used by
+    It implements only the small part of the VersionedFiles interface used by
     PlanMerge.  It falls back to multiple versionedfiles for data not stored in
     _PlanMergeVersionedFile itself.
+
+    :ivar: fallback_versionedfiles a list of VersionedFiles objects that can be
+        queried for missing texts.
     """
 
-    def __init__(self, file_id, fallback_versionedfiles=None):
-        """Constuctor
+    def __init__(self, file_id):
+        """Create a _PlanMergeVersionedFile.
 
-        :param file_id: Used when raising exceptions.
-        :param fallback_versionedfiles: If supplied, the set of fallbacks to
-            use.  Otherwise, _PlanMergeVersionedFile.fallback_versionedfiles
-            can be appended to later.
+        :param file_id: Used with _PlanMerge code which is not yet fully
+            tuple-keyspace aware.
         """
         self._file_id = file_id
-        if fallback_versionedfiles is None:
-            self.fallback_versionedfiles = []
-        else:
-            self.fallback_versionedfiles = fallback_versionedfiles
+        # fallback locations
+        self.fallback_versionedfiles = []
+        # Parents for locally held keys.
         self._parents = {}
+        # line data for locally held keys.
         self._lines = {}
+        # key lookup providers
+        self._providers = [DictParentsProvider(self._parents)]
 
     def plan_merge(self, ver_a, ver_b, base=None):
         """See VersionedFile.plan_merge"""
         from bzrlib.merge import _PlanMerge
         if base is None:
-            return _PlanMerge(ver_a, ver_b, self).plan_merge()
-        old_plan = list(_PlanMerge(ver_a, base, self).plan_merge())
-        new_plan = list(_PlanMerge(ver_a, ver_b, self).plan_merge())
+            return _PlanMerge(ver_a, ver_b, self, (self._file_id,)).plan_merge()
+        old_plan = list(_PlanMerge(ver_a, base, self, (self._file_id,)).plan_merge())
+        new_plan = list(_PlanMerge(ver_a, ver_b, self, (self._file_id,)).plan_merge())
         return _PlanMerge._subtract_plans(old_plan, new_plan)
 
     def plan_lca_merge(self, ver_a, ver_b, base=None):
         from bzrlib.merge import _PlanLCAMerge
-        graph = self._get_graph()
-        new_plan = _PlanLCAMerge(ver_a, ver_b, self, graph).plan_merge()
+        graph = Graph(self)
+        new_plan = _PlanLCAMerge(ver_a, ver_b, self, (self._file_id,), graph).plan_merge()
         if base is None:
             return new_plan
-        old_plan = _PlanLCAMerge(ver_a, base, self, graph).plan_merge()
+        old_plan = _PlanLCAMerge(ver_a, base, self, (self._file_id,), graph).plan_merge()
         return _PlanLCAMerge._subtract_plans(list(old_plan), list(new_plan))
 
-    def add_lines(self, version_id, parents, lines):
-        """See VersionedFile.add_lines
+    def add_lines(self, key, parents, lines):
+        """See VersionedFiles.add_lines
 
-        Lines are added locally, not fallback versionedfiles.  Also, ghosts are
-        permitted.  Only reserved ids are permitted.
+        Lines are added locally, not to fallback versionedfiles.  Also, ghosts
+        are permitted.  Only reserved ids are permitted.
         """
-        if not revision.is_reserved_id(version_id):
+        if type(key) != tuple:
+            import pdb;pdb.set_trace()
+        if not revision.is_reserved_id(key[-1]):
             raise ValueError('Only reserved ids may be used')
         if parents is None:
             raise ValueError('Parents may not be None')
         if lines is None:
             raise ValueError('Lines may not be None')
-        self._parents[version_id] = tuple(parents)
-        self._lines[version_id] = lines
+        self._parents[key] = tuple(parents)
+        self._lines[key] = lines
 
-    def get_lines(self, version_id):
-        """See VersionedFile.get_ancestry"""
-        lines = self._lines.get(version_id)
-        if lines is not None:
-            return lines
+    def get_record_stream(self, keys, ordering, include_delta_closure):
+        pending = set(keys)
+        for key in keys:
+            if key in self._lines:
+                lines = self._lines[key]
+                parents = self._parents[key]
+                pending.remove(key)
+                yield FulltextContentFactory(key, parents, None,
+                    ''.join(lines))
         for versionedfile in self.fallback_versionedfiles:
-            try:
-                return versionedfile.get_lines(version_id)
-            except errors.RevisionNotPresent:
-                continue
-        else:
-            raise errors.RevisionNotPresent(version_id, self._file_id)
-
-    def get_ancestry(self, version_id, topo_sorted=False):
-        """See VersionedFile.get_ancestry.
-
-        Note that this implementation assumes that if a VersionedFile can
-        answer get_ancestry at all, it can give an authoritative answer.  In
-        fact, ghosts can invalidate this assumption.  But it's good enough
-        99% of the time, and far cheaper/simpler.
-
-        Also note that the results of this version are never topologically
-        sorted, and are a set.
-        """
-        if topo_sorted:
-            raise ValueError('This implementation does not provide sorting')
-        parents = self._parents.get(version_id)
-        if parents is None:
-            for vf in self.fallback_versionedfiles:
-                try:
-                    return vf.get_ancestry(version_id, topo_sorted=False)
-                except errors.RevisionNotPresent:
+            for record in versionedfile.get_record_stream(
+                pending, 'unordered', True):
+                if record.storage_kind == 'absent':
                     continue
-            else:
-                raise errors.RevisionNotPresent(version_id, self._file_id)
-        ancestry = set([version_id])
-        for parent in parents:
-            ancestry.update(self.get_ancestry(parent, topo_sorted=False))
-        return ancestry
-
-    def get_parent_map(self, version_ids):
-        """See VersionedFile.get_parent_map"""
-        result = {}
-        pending = set(version_ids)
-        for key in version_ids:
-            try:
-                result[key] = self._parents[key]
-            except KeyError:
-                pass
-        pending = pending - set(result.keys())
-        for versionedfile in self.fallback_versionedfiles:
-            parents = versionedfile.get_parent_map(pending)
-            result.update(parents)
-            pending = pending - set(parents.keys())
+                else:
+                    pending.remove(record.key)
+                    yield record
             if not pending:
-                return result
-        return result
+                return
+        # report absent entries
+        for key in pending:
+            yield AbsentContentFactory(key)
 
-    def _get_graph(self):
-        from bzrlib.graph import (
-            DictParentsProvider,
-            Graph,
-            _StackedParentsProvider,
-            )
-        from bzrlib.repofmt.knitrepo import _KnitParentsProvider
-        parent_providers = [DictParentsProvider(self._parents)]
-        for vf in self.fallback_versionedfiles:
-            parent_providers.append(_KnitParentsProvider(vf))
-        return Graph(_StackedParentsProvider(parent_providers))
+    def get_parent_map(self, keys):
+        """See VersionedFiles.get_parent_map"""
+        # We create a new provider because a fallback may have been added.
+        # If we make fallbacks private we can update a stack list and avoid
+        # object creation thrashing.
+        self._providers = self._providers[:1] + self.fallback_versionedfiles
+        return _StackedParentsProvider(self._providers).get_parent_map(keys)
 
 
 class PlanWeaveMerge(TextMerge):
@@ -714,212 +1277,3 @@
         PlanWeaveMerge.__init__(self, plan, a_marker, b_marker)
 
 
-class InterVersionedFile(InterObject):
-    """This class represents operations taking place between two VersionedFiles.
-
-    Its instances have methods like join, and contain
-    references to the source and target versionedfiles these operations can be 
-    carried out on.
-
-    Often we will provide convenience methods on 'versionedfile' which carry out
-    operations with another versionedfile - they will always forward to
-    InterVersionedFile.get(other).method_name(parameters).
-    """
-
-    _optimisers = []
-    """The available optimised InterVersionedFile types."""
-
-    def join(self, pb=None, msg=None, version_ids=None, ignore_missing=False):
-        """Integrate versions from self.source into self.target.
-
-        If version_ids is None all versions from source should be
-        incorporated into this versioned file.
-
-        Must raise RevisionNotPresent if any of the specified versions
-        are not present in the other file's history unless ignore_missing is 
-        supplied in which case they are silently skipped.
-        """
-        target = self.target
-        version_ids = self._get_source_version_ids(version_ids, ignore_missing)
-        graph = Graph(self.source)
-        search = graph._make_breadth_first_searcher(version_ids)
-        transitive_ids = set()
-        map(transitive_ids.update, list(search))
-        parent_map = self.source.get_parent_map(transitive_ids)
-        order = tsort.topo_sort(parent_map.items())
-        pb = ui.ui_factory.nested_progress_bar()
-        parent_texts = {}
-        try:
-            # TODO for incremental cross-format work:
-            # make a versioned file with the following content:
-            # all revisions we have been asked to join
-            # all their ancestors that are *not* in target already.
-            # the immediate parents of the above two sets, with 
-            # empty parent lists - these versions are in target already
-            # and the incorrect version data will be ignored.
-            # TODO: for all ancestors that are present in target already,
-            # check them for consistent data, this requires moving sha1 from
-            # 
-            # TODO: remove parent texts when they are not relevant any more for 
-            # memory pressure reduction. RBC 20060313
-            # pb.update('Converting versioned data', 0, len(order))
-            total = len(order)
-            for index, version in enumerate(order):
-                pb.update('Converting versioned data', index, total)
-                if version in target:
-                    continue
-                _, _, parent_text = target.add_lines(version,
-                                               parent_map[version],
-                                               self.source.get_lines(version),
-                                               parent_texts=parent_texts)
-                parent_texts[version] = parent_text
-            return total
-        finally:
-            pb.finished()
-
-    def _get_source_version_ids(self, version_ids, ignore_missing):
-        """Determine the version ids to be used from self.source.
-
-        :param version_ids: The caller-supplied version ids to check. (None 
-                            for all). If None is in version_ids, it is stripped.
-        :param ignore_missing: if True, remove missing ids from the version 
-                               list. If False, raise RevisionNotPresent on
-                               a missing version id.
-        :return: A set of version ids.
-        """
-        if version_ids is None:
-            # None cannot be in source.versions
-            return set(self.source.versions())
-        else:
-            if ignore_missing:
-                return set(self.source.versions()).intersection(set(version_ids))
-            else:
-                new_version_ids = set()
-                for version in version_ids:
-                    if version is None:
-                        continue
-                    if not self.source.has_version(version):
-                        raise errors.RevisionNotPresent(version, str(self.source))
-                    else:
-                        new_version_ids.add(version)
-                return new_version_ids
-
-
-class KeyMapper(object):
-    """KeyMappers map between keys and underlying paritioned storage."""
-
-    def map(self, key):
-        """Map key to an underlying storage identifier.
-
-        :param key: A key tuple e.g. ('file-id', 'revision-id').
-        :return: An underlying storage identifier, specific to the partitioning
-            mechanism.
-        """
-
-    def unmap(self, partition_id):
-        """Map a partitioned storage id back to a key prefix.
-        
-        :param partition_id: The underlying partition id.
-        :return: As much of a key (or prefix) as is derivable from the parition
-            id.
-        """
-
-
-class ConstantMapper(KeyMapper):
-    """A key mapper that maps to a constant result."""
-
-    def __init__(self, result):
-        """Create a ConstantMapper which will return result for all maps."""
-        self._result = result
-
-    def map(self, key):
-        """See KeyMapper.map()."""
-        return self._result
-
-
-class PrefixMapper(KeyMapper):
-    """A key mapper that extracts the first component of a key."""
-
-    def map(self, key):
-        """See KeyMapper.map()."""
-        return key[0]
-
-    def unmap(self, partition_id):
-        """See KeyMapper.unmap()."""
-        return (partition_id,)
-
-
-class HashPrefixMapper(KeyMapper):
-    """A key mapper that combines the first component of a key with a hash."""
-
-    def map(self, key):
-        """See KeyMapper.map()."""
-        prefix = self._escape(key[0])
-        return "%02x/%s" % (adler32(prefix) & 0xff, prefix)
-
-    def _escape(self, prefix):
-        """No escaping needed here."""
-        return prefix
-
-    def unmap(self, partition_id):
-        """See KeyMapper.unmap()."""
-        return (self._unescape(osutils.basename(partition_id)),)
-
-    def _unescape(self, basename):
-        """No unescaping needed for HashPrefixMapper."""
-        return basename
-
-
-class HashEscapedPrefixMapper(HashPrefixMapper):
-    """Combines the escaped first component of a key with a hash."""
-
-    _safe = "abcdefghijklmnopqrstuvwxyz0123456789-_@,."
-
-    def _escape(self, prefix):
-        """Turn a key element into a filesystem safe string.
-
-        This is similar to a plain urllib.quote, except
-        it uses specific safe characters, so that it doesn't
-        have to translate a lot of valid file ids.
-        """
-        # @ does not get escaped. This is because it is a valid
-        # filesystem character we use all the time, and it looks
-        # a lot better than seeing %40 all the time.
-        r = [((c in self._safe) and c or ('%%%02x' % ord(c)))
-             for c in prefix]
-        return ''.join(r)
-
-    def _unescape(self, basename):
-        """Escaped names are unescaped by urlutils."""
-        return urllib.unquote(basename)
-
-
-def make_pack_factory(graph, delta, keylength):
-    """Create a factory for creating a pack based VersionedFiles.
-    
-    :param graph: Store a graph.
-    :param delta: Delta compress contents.
-    :param keylength: How long should keys be.
-    """
-    return lambda x:None
-
-
-def make_versioned_files_factory(versioned_file_factory, mapper):
-    """Create a ThunkedVersionedFiles factory.
-
-    This will create a callable which when called creates a
-    ThunkedVersionedFiles on a transport, using mapper to access individual
-    versioned files, and versioned_file_factory to create each individual file.
-    """
-    return lambda x:None
-
-
-class VersionedFiles(object):
-    """Storage for many versioned files.
-
-    This object allows a single keyspace for accessing the history graph and
-    contents of named bytestrings.
-
-    Currently no implementation allows the graph of different key prefixes to
-    intersect, but the API does allow such implementations in the future.
-    """

=== modified file 'bzrlib/weave.py'
--- bzrlib/weave.py	2008-05-11 23:49:50 +0000
+++ bzrlib/weave.py	2008-06-11 04:20:16 +0000
@@ -86,8 +86,9 @@
         WeaveRevisionNotPresent,
         )
 import bzrlib.errors as errors
-from bzrlib.osutils import sha_strings, split_lines
+from bzrlib.osutils import dirname, sha_strings, split_lines
 import bzrlib.patiencediff
+from bzrlib.revision import NULL_REVISION
 from bzrlib.symbol_versioning import *
 from bzrlib.trace import mutter
 from bzrlib.tsort import topo_sort
@@ -95,7 +96,6 @@
     AbsentContentFactory,
     adapter_registry,
     ContentFactory,
-    InterVersionedFile,
     VersionedFile,
     )
 from bzrlib.weavefile import _read_weave_v5, write_weave_v5
@@ -119,7 +119,7 @@
 
     def get_bytes_as(self, storage_kind):
         if storage_kind == 'fulltext':
-            return self._weave.get_text(self.key[0])
+            return self._weave.get_text(self.key[-1])
         else:
             raise UnavailableRepresentation(self.key, storage_kind, 'fulltext')
 
@@ -307,6 +307,7 @@
         :return: An iterator of ContentFactory objects, each of which is only
             valid until the iterator is advanced.
         """
+        versions = [version[-1] for version in versions]
         if ordering == 'topological':
             parents = self.get_parent_map(versions)
             new_versions = topo_sort(parents)
@@ -322,11 +323,16 @@
         """See VersionedFile.get_parent_map."""
         result = {}
         for version_id in version_ids:
-            try:
-                result[version_id] = tuple(
-                    map(self._idx_to_name, self._parents[self._lookup(version_id)]))
-            except RevisionNotPresent:
-                pass
+            if version_id == NULL_REVISION:
+                parents = ()
+            else:
+                try:
+                    parents = tuple(
+                        map(self._idx_to_name,
+                            self._parents[self._lookup(version_id)]))
+                except RevisionNotPresent:
+                    continue
+            result[version_id] = parents
         return result
 
     def get_parents_with_ghosts(self, version_id):
@@ -834,59 +840,6 @@
         # no lines outside of insertion blocks, that deletions are
         # properly paired, etc.
 
-    def _join(self, other, pb, msg, version_ids, ignore_missing):
-        """Worker routine for join()."""
-        if not other.versions():
-            return          # nothing to update, easy
-
-        if not version_ids:
-            # versions is never none, InterWeave checks this.
-            return 0
-
-        # two loops so that we do not change ourselves before verifying it
-        # will be ok
-        # work through in index order to make sure we get all dependencies
-        names_to_join = []
-        processed = 0
-        # get the selected versions only that are in other.versions.
-        version_ids = set(other.versions()).intersection(set(version_ids))
-        # pull in the referenced graph.
-        version_ids = other.get_ancestry(version_ids)
-        pending_parents = other.get_parent_map(version_ids)
-        pending_graph = pending_parents.items()
-        if len(pending_graph) != len(version_ids):
-            raise RevisionNotPresent(
-                set(version_ids) - set(pending_parents.keys()), self)
-        for name in topo_sort(pending_graph):
-            other_idx = other._name_map[name]
-            # returns True if we have it, False if we need it.
-            if not self._check_version_consistent(other, other_idx, name):
-                names_to_join.append((other_idx, name))
-            processed += 1
-
-        if pb and not msg:
-            msg = 'weave join'
-
-        merged = 0
-        time0 = time.time()
-        for other_idx, name in names_to_join:
-            # TODO: If all the parents of the other version are already
-            # present then we can avoid some work by just taking the delta
-            # and adjusting the offsets.
-            new_parents = self._imported_parents(other, other_idx)
-            sha1 = other._sha1s[other_idx]
-
-            merged += 1
-
-            if pb:
-                pb.update(msg, merged, len(names_to_join))
-           
-            lines = other.get_lines(other_idx)
-            self._add(name, lines, new_parents, sha1)
-
-        mutter("merged = %d, processed = %d, file_id=%s; deltat=%d"%(
-                merged, processed, self._weave_name, time.time()-time0))
- 
     def _imported_parents(self, other, other_idx):
         """Return list of parents in self corresponding to indexes in other."""
         new_parents = []
@@ -989,9 +942,13 @@
         sio = StringIO()
         write_weave_v5(self, sio)
         sio.seek(0)
-        self._transport.put_file(self._weave_name + WeaveFile.WEAVE_SUFFIX,
-                                 sio,
-                                 self._filemode)
+        bytes = sio.getvalue()
+        path = self._weave_name + WeaveFile.WEAVE_SUFFIX
+        try:
+            self._transport.put_bytes(path, bytes, self._filemode)
+        except errors.NoSuchFile:
+            self._transport.mkdir(dirname(path))
+            self._transport.put_bytes(path, bytes, self._filemode)
 
     @staticmethod
     def get_suffixes():
@@ -1272,30 +1229,3 @@
 if __name__ == '__main__':
     import sys
     sys.exit(main(sys.argv))
-
-
-class InterWeave(InterVersionedFile):
-    """Optimised code paths for weave to weave operations."""
-    
-    _matching_file_from_factory = staticmethod(WeaveFile)
-    _matching_file_to_factory = staticmethod(WeaveFile)
-    
-    @staticmethod
-    def is_compatible(source, target):
-        """Be compatible with weaves."""
-        try:
-            return (isinstance(source, Weave) and
-                    isinstance(target, Weave))
-        except AttributeError:
-            return False
-
-    def join(self, pb=None, msg=None, version_ids=None, ignore_missing=False):
-        """See InterVersionedFile.join."""
-        version_ids = self._get_source_version_ids(version_ids, ignore_missing)
-        if self.target.versions() == [] and version_ids is None:
-            self.target._copy_weave_content(self.source)
-            return
-        self.target._join(self.source, pb, msg, version_ids, ignore_missing)
-
-
-InterVersionedFile.register_optimiser(InterWeave)

=== modified file 'bzrlib/weave_commands.py'
--- bzrlib/weave_commands.py	2007-06-17 22:39:37 +0000
+++ bzrlib/weave_commands.py	2008-06-11 04:20:16 +0000
@@ -36,37 +36,12 @@
         from bzrlib.weavefile import read_weave
         from bzrlib.transport import get_transport
         from bzrlib import osutils
-        from bzrlib.knit import KnitVersionedFile
-        if filename.endswith('.knit'):
-            transport = get_transport(osutils.dirname(filename))
-            relpath = osutils.basename(filename)[:-len('.knit')]
-            vf = KnitVersionedFile(relpath, transport)
-        else:
-            vf = read_weave(file(filename, 'rb'))
+        vf = read_weave(file(filename, 'rb'))
         names = vf.versions()
         names.sort()
         print '\n'.join(names)
 
 
-class cmd_weave_join(Command):
-    """Join the contents of two weave files.
-
-    The resulting weave is sent to stdout.
-
-    This command is only intended for bzr developer use.
-    """
-
-    hidden = True
-    takes_args = ['weave1', 'weave2']
-
-    def run(self, weave1, weave2):
-        from bzrlib.weavefile import read_weave, write_weave
-        w1 = read_weave(file(weave1, 'rb'))
-        w2 = read_weave(file(weave2, 'rb'))
-        w1.join(w2)
-        write_weave(w1, sys.stdout)
-
-
 class cmd_weave_plan_merge(Command):
     """Show the plan for merging two versions within a weave"""
     hidden = True

=== modified file 'bzrlib/workingtree.py'
--- bzrlib/workingtree.py	2008-05-08 04:33:38 +0000
+++ bzrlib/workingtree.py	2008-06-11 04:20:16 +0000
@@ -395,7 +395,7 @@
         # at this point ?
         try:
             return self.branch.repository.revision_tree(revision_id)
-        except errors.RevisionNotPresent:
+        except (errors.RevisionNotPresent, errors.NoSuchRevision):
             # the basis tree *may* be a ghost or a low level error may have
             # occured. If the revision is present, its a problem, if its not
             # its a ghost.

=== modified file 'bzrlib/workingtree_4.py'
--- bzrlib/workingtree_4.py	2008-05-08 04:33:38 +0000
+++ bzrlib/workingtree_4.py	2008-06-11 04:20:16 +0000
@@ -1405,8 +1405,9 @@
     def annotate_iter(self, file_id,
                       default_revision=_mod_revision.CURRENT_REVISION):
         """See Tree.annotate_iter"""
-        w = self._get_weave(file_id)
-        return w.annotate(self.inventory[file_id].revision)
+        text_key = (file_id, self.inventory[file_id].revision)
+        annotations = self._repository.texts.annotate(text_key)
+        return [(key[-1], line) for key, line in annotations]
 
     def _get_ancestors(self, default_revision):
         return set(self._repository.get_ancestry(self._revision_id,
@@ -1564,25 +1565,18 @@
             return parent_details[1]
         return None
 
-    def _get_weave(self, file_id):
-        return self._repository.weave_store.get_weave(file_id,
-                self._repository.get_transaction())
-
     def get_file(self, file_id, path=None):
         return StringIO(self.get_file_text(file_id))
 
     def get_file_lines(self, file_id):
-        entry = self._get_entry(file_id=file_id)[1]
-        if entry is None:
-            raise errors.NoSuchId(tree=self, file_id=file_id)
-        return self._get_weave(file_id).get_lines(entry[1][4])
+        return osutils.split_lines(self.get_file_text(file_id))
 
     def get_file_size(self, file_id):
         """See Tree.get_file_size"""
         return self.inventory[file_id].text_size
 
     def get_file_text(self, file_id):
-        return ''.join(self.get_file_lines(file_id))
+        return list(self.iter_files_bytes([(file_id, None)]))[0][1]
 
     def get_reference_revision(self, file_id, path=None):
         return self.inventory[file_id].reference_revision

